;----------------------------------------------------------------------------
;    PSATOR - Library of routines for amateur radio AMTOR/PACTOR
;    Copyright (C) 1993-1995 Johan B. Forrer, KC7WW
;
;    This is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
;----------------------------------------------------------------------------
;                           PSABIOS                                       
;  ##############SPECIAL VERSION FOR PSA SOUNDCARD########################
;        Copyright (C) J.B.Forrer 1991,1994 - All rights reserved                       
; 
;----------------------------------------------------------------------------
;                      Assembly instructions:
;
;> tasm   /ml psabios.asm
;----------------------------------------------------------------------------
;  History:
;	V0.01 - Derived from TBIOS V2.08 May, 1994
;----------------------------------------------------------------------------
; Hardware implementation
;       RTS - (request to send)     PTT
;       TD  - (Data terminal ready) TXDATA
;----------------------------------------------------------------------------
; A note on the meaning of "MASTER/SLAVE" and "CALLER ORIGINATOR".
; In ARQ mode, you may be the caller or the callee. If you initiated
; the ARQ call, you are MASTER (caller) and the station called the SLAVE.
; When resynch has to happen, it is the duty of the master to initiate
; such resynching.
; During the ARQ qso, control would often be passed over between the
; master and slave. This program uses different bit phasing when
; expecting "single character CONTROL" data and expecting "3 character 
; BLOCK" data.  For purposes of bit phasing, the mode when sending 
; BLOCKS is called "MASTER", and when it receiving CONTROL'S - "SLAVE".
; For purposes of resynching etc, however, it remembers who initiated 
; the contact as being the "ORIG".
; This is slightly different in the old AMT-type systems where
; MASTER/SLAVE has only the meaning of originator. These systems
; also uses much simpler bit phasing algorithms.
;----------------------------------------------------------------------------
; A note on the meanings of CTRX,BKRX, CTTX and BKTX
; CT means "Control" character timing, BK means "3 character block" timing,
; TX for transmit timings, RX for receive timings.
; These constants have meaning only for ARQ. 
;
; When in ARQ slave:=
; CTRX is the time till the expected control character is 
;      ready (390ms: 210 for received block + 120 till mid dead window),
; BKRX is the time till the expected block of three chars is 
;      ready (250ms: 240 dead window).
;
; When in ARQ master:=
; CTTX is the master timing for sending controls: (not adjusted in
; the bit phasing algorithm) (380ms=1*7*10ms for a ctrl + 240ms dead time),
; BKTX is the master timing for sending blocks: (not adjusted in
; the bit phasing algorithm) (240ms = 3*7*10ms + 240ms dead time).
; CTRX(BKRX) hold the predicted time AFTER a control(block) is sent for
; reception of the next block(control). NOTE: This must not add up to
; more than 240ms, nor be less than about 15ms).
;---------------ASSEMBLY DEFINE SWITCHES-------------------------------------

RXBUFSIZ	= 256		; AMTOR RX buffer size
PBUFSIZ		= 4		; ITA2 Buffer size
STKSIZE 	= 1024		; AMTOR stacksize

LIMS		= 128		; for bit tracking threshold   (slave)
ADJS		= 256		; for bit tracking adjustments (slave)
LIM		= 128           ; master/slave FEC
ADJ		= 256           ; 

MAX_TIMEOUT	= 32		; 32 cycles (15 seconds)

GOOD_CNT	= 5		; number of good blocks to go to listen mode
ERR_CNT		= 4		; number of bad blocks till resynch on mode L
SPEC_SYM	= 0FFh		; special symbol used as placeholder for mode
				; Listen error characters
DECAY		= 50		; Decay factor for tuning display
;----------------------------------------------------------------------------
; Access variables off the stack 
; For "Large models", i.e.    mov	bx,ARG1		; get first arg
; C calling sequence func(ARG1, ARG2, ARG3, ARG4)
;
CARG1	EQU	<[BP+6]>
CARG2	EQU	<[BP+8]>
CARG3	EQU	<[BP+10]>
CARG4	EQU	<[BP+12]>
CARG5	EQU	<[BP+14]>

;----------------------------------------------------------------------------


; EVM modem op-codes:
; These are the codes sent out the serial port to the 56002 EVM
; DSP modem.

EVM_NOP		=	08h
EVM_TX_PREV	=	04h
EVM_TX_0	=	05h
EVM_TX_1	=	06h
EVM_RX		=	00h
EVM_100		=	10h
EVM_200		=	11h
EVM_RESET	=	20h



;----------------------------------------------------------------------------
cseg    segment para public 'CODE'
        org     100H
        assume cs:cseg,ds:cseg

installmsg      db      'TORBIOS V2.08 (c) J.B.Forrer 1991-1994',0  ; signature
DISPLAY "PSABIOS V0.01 (c) J.B.Forrer 1991-1994"

;----------------------------------------------------------------------------
;
; TABLE OF ITA2 (BAUDOT) CODES IN CCIR 476 (AMTOR ORDER). 
;                0=INVALID, MSB0=NONPR
;----------------------------------------------------------------------------
	EVEN
RTAB    DB      00h,00h,00h,00h,00h,00h,00h,00h		; 0
        DB      00h,00h,00h,00h,00h,00h,00h,0Fh		
        DB      00h,00h,00h,00h,00h,00h,00h,8Bh		; 1
        DB      00h,00h,00h,8Dh,00h,8Eh,8Fh,00h
        DB      00h,00h,00h,00h,00h,00h,00h,93h		; 2
        DB      00h,00h,00h,95h,00h,96h,97h,00h
        DB      00h,00h,00h,33h,00h,9Ah,9Bh,00h		; 3
        DB      00h,9Ch,9Dh,00h,9Eh,00h,00h,00h
        DB      00h,00h,00h,00h,00h,00h,00h,83h		; 4
        DB      00h,00h,00h,85h,00h,86h,87h,00h
        DB      00h,00h,00h,89h,00h,8Ah,81h,00h		; 5
        DB      00h,8Ch,9Fh,00h,84h,00h,00h,00h
        DB      00h,00h,00h,91h,00h,92h,66h,00h		; 6
        DB      00h,94h,80h,00h,82h,00h,00h,00h		; note blnk=>80h
        DB      00h,98h,99h,00h,90h,00h,00h,00h		; 7
        DB      88h,00h,00h,00h,00h,00h,00h,00h
; TABLE OF CCIR 476 (AMTOR) CODES IN ITA2 (BAUDOT) ORDER          
	EVEN
TTAB    DB      6Ah,56h,6Ch,47h,5Ch,4Bh,4Dh,4Eh
        DB      78h,53h,55h,17h,59h,1Bh,1Dh,1Eh
        DB      74h,63h,65h,27h,69h,2Bh,2Dh,2Eh
        DB      71h,72h,35h,36h,39h,3Ah,3Ch,5Ah
;----------------------------------------------------------------------------
; TABLE OF ASCII CHARACTERS IN ORDER OF ITA2 (BAUDOT) CODE
; NOTE THAT ASCII $ TRANSLATES TO (WRU - figs D) 
;----------------------------------------------------------------------------
	EVEN
;                              LTRS
EXTAB   DB      80h,45h,0Ah,41h,20h,53h,49h,55h
        DB      0Dh,44h,52h,4Ah,4Eh,46h,43h,4Bh
        DB      54h,5Ah,4Ch,57h,48h,59h,50h,51h
        DB      4Fh,42h,47h,00h,4Dh,58h,56h,00h
;                              FIGS
        DB      80h,33h,0Ah,2Dh,20h,27h,38h,37h
        DB      0Dh,24h,34h,2Ah,2Ch,25h,3Ah,28h
        DB      35h,2Bh,29h,32h,23h,36h,30h,31h
        DB      39h,3Fh,40h,00h,2Eh,2Fh,3Dh,00h
;----------------------------------------------------------------------------
; Table of ITA2 characters in order of ASCII code
; 00 are non-translateables, these are just ingored.
; Null gets converted to 40h ITA2, that gets returned as 80h ITA2, that can
; be converted to 64h AMTOR code. Similarly 64h AMTOR gets converted
; to 80h ITA2, and since its printable and only actually 5 bits wide,
; is returned as ASCII null 00.
; The upper two bits have special interpretations:
;
; 	bit 6 = 1  case doesnt matter (null/blnk, cr, lf, space)
; 	bit 7 = 0 LTRS case
;             = 1 FIGS case
;----------------------------------------------------------------------------
	EVEN
ITASC	DB 	 40h,00h,00h,00h,00h,00h,00h,00h	; 0  	
	DB	 00h,00h,42h,00h,00h,48h,00h,00h	
	DB	 00h,00h,00h,00h,00h,00h,00h,00h	; 1 
	DB	 00h,00h,00h,00h,00h,00h,00h,00h
	DB	 44h,00h,00h,94h,89h,8Dh,00h,85h	; 2
	DB	 8Fh,92h,8Bh,91h,8Ch,83h,9Ch,9Dh
;
	DB	 96h,97h,93h,81h,8Ah,90h,95h,87h    	; 3
	DB	 86h,98h,8Eh,00h,00h,9Eh,00h,99h
	DB	 9Ah,03h,19h,0Eh,09h,01h,0Dh,1Ah     	; 4
	DB	 14h,06h,0Bh,0Fh,12h,1Ch,0Ch,18h
	DB	 16h,17h,0Ah,05h,10h,07h,1Eh,13h     	; 5
	DB	 9Dh,95h,91h,00h,00h,00h,00h,00h
	DB	 00h,03h,19h,0Eh,09h,01h,0Dh,1Ah     	; 6
	DB	 14h,06h,0Bh,0Fh,12h,1Ch,0Ch,18h
	DB	 16h,17h,0Ah,05h,10h,07h,1Eh,13h     	; 7
	DB	 9Dh,95h,91h,00h,00h,00h,00h,00h
;----------------------------------------------------------------------------
BKRQ    DB      66h,66h,66h             ; 3 RQS IN A BLOCK
IDLE    DB      66h,0Fh,66h             ; MODE B IDLE PATTERN RECOGNITION
BKBAB   DB      33h,0Fh,33h             ; BETA-ALPHA-BETA RECOGNITION BLOCK
BKALF   DB      0Fh,0Fh,0Fh             ; BLOCK OF ALPHAS FOR QRT
AMT_1   DB      'V:07A I:???? T:30 B:45 S:20 L:1 N:1'
	DB	0Ah,0Dh,00h		; AMT-1 V07A signature
;----------------------------------------------------------------------------
;    VARIABLES CRITICAL TO AMTOR OPERATION
;
CNTRL   DB      1 dup (?)        ; COMMAND INPUT
STSTS   DB      1 dup (?)        ; STATUS INPUT
ECHO    DB      1 dup (?)        ; ECHO SWITCH, 00=NO ECHO
DELAY   DB      1 dup (?)        ; TX SWITCH-ON DELAY
IDENT   DB      4 dup (?)        ; OWN SELCAL CODE
SELCAL  DB      4 dup (?)        ; OUTGOING SELCAL CODE
RXDTA   DB      1 dup (?)        ; RCVD ITA2 CODE TO USER
TXDTA   DB      1 dup (?)        ; ITA2 CODE FROM USER FOR TX
BA_dat  DB	1 dup (?)	 ; Ascii/ITA2 character to be echoed
	EVEN
CD_DELAY DW	1 dup (?)	 ; Control delay
PIPTR   DW      1 dup (?)        ; RX BUFFER INPUT POINTER
POPTR   DW      1 dup (?)        ; RX BUFFER OUTPUT "
KIPTR   DW      1 dup (?)        ; TX BUFFER INPUT POINTER
KOPTR   DW      1 dup (?)        ; TX BUFFER OUTPUT "
CLOCK   DW      1 dup (?)        ; MAIN SYSTEM CLOCK
				; These items used in conjunction with
 				; bit phasing (see INTRPT for more)
PHASE   DW      1 dup (?)        ; BIT PHASE COUNTER
ACPH    DW      1 dup (?)        ; ACCUMULATED BITPHASE ERROR
RAFG    DB      1 dup (?)        ; RX ACTIVE FLAG (HI=ACTIVE)
BFLG    DB      1 dup (?)        ; HIGH WHEN B-MODE ACTIVE
	EVEN
LINP    DB      1 dup (?)        ; LAST RCVD INPUT LOGIC LEVEL
HIST	DB	1 dup (?)	 ; history bit used for improved reliability
HIST_1	DB	1 dup (?)	 ; 
STFG    DB      1 dup (?)        ; STANDBY FLAG
NOCY    DB      1 dup (?)        ; MODE A/L CYCLE TIMEOUT COUNT
SLCAL   DB      6 dup (?)        ; MODE-A SELCAL BLOCKS
TRQFG   DB      1 dup (?)        ; SET HI TO REQUEST C/O TO TX IN MODE A
TFLG    DB      1 dup (?)        ; SET HI IN ARQ TX
QFLG    DB      1 dup (?)        ; SET HIGH TO CLOSE DOWN MODE A OR B 
	EVEN			 ;     (Baudot and ascii as well)
TEMP1   DW      1 dup (?)        ; RQ&ERROR FLAGS,0=NOT ACTIVE
TEMP2   DW      1 dup (?)              
RXBUF   DB      3 dup (?)        ; 3 CHARACTERS RECEIVED 
TXBUF   DB      3 dup (?)        ; 3 CHARACTERS TO BE TRANSMITTED
	EVEN
MASTER  DB      1 dup (?)        ; MASTER=0, SLAVE!=0
ORIG    DB      1 dup (?)        ; IF WE ARE THE CALLER=0
MASKOUT DB      1 dup (?)        ; setting output bit levels
	EVEN
MASKIN  DB      1 dup (?)        ; setting input  bit levels
	EVEN
CTRX    DW      1 dup (?)        ; TIME OF START OF CONTROL RECEIVE
BKRX    DW      1 dup (?)        ; DITTO BLOCK
CTTX    DW      1 dup (?)        ; DITTO TRANSMIT CONTROL
BKTX    DW      1 dup (?)        ; DITTO BLOCK TRANSMIT
RPTR    DW      1 dup (?)        ; SEARCH S-REG POINTER
RBFG    DB      1 dup (?)        ; LAST CONTROL CODE SENT IN MODE A
LCTR    DB      1 dup (?)        ; LAST CONTROL RCVD IN MODE A
COUNT   DB      1 dup (?)        ; MODE B COUNTER
SBFG    DB      1 dup (?)        ; MODE A CALL BLOCK TOGGLE
ERRT    DW      1 dup (?)        ; MODE B RCV ERROR RATE
	EVEN
RXCP    DB      3 dup (?)        ; BLOCK-COMPARE VARIABLES
SHIFT   DB      1 dup (?)        ; LTRS=0, FIGS=$40
RSHFT   DB      1 dup (?)        ; Receive  shift state
TSHFT   DB      1 dup (?)        ; Transmit shift state
LSTCH   DB      1 dup (?)        ; LAST CHARACTER RECEIVED
AMTAX   DW      1 dup (?)        ; TEMP ax-REG IN AMTOR PROG
AMTBX   DW      1 dup (?)        ; TEMP bx-REG "
AMTCX	DW	1 dup (?)	 ; TEMP cx-REG "
	EVEN
REGS    DB      30 dup (?)       ; SEARCH SHIFTREGISTERS (see description in
                                 ;                        routine SMPLE)
KBUF    DB      4 dup (?)        ; TX BUFFER
PBUF    DB      PBUFSIZ dup (?)	 ; RX BUFFER
	EVEN
AMTSP   DW      1 dup (?)        ; AMTOR SP
USRSP_OFF DW    1 dup (?)        ; USER SP
USRSP_SEG DW    1 dup (?)       
;
; Data I/O buffer space
;
rx_store_at_ptr	 DW	1 dup (?)	; Store at pointer
rx_take_from_ptr DW	1 dup (?)	; take from pointer	
RXBUFR           DB    RXBUFSIZ dup (?) ; Data received from AMTOR
rx_ascii_case	 DB	1 dup (?)	; 0=upper,1=lower

tx_store_at_ptr	 DW	1 dup (?)	; Store at pointer
tx_take_from_ptr DW	1 dup (?)	; take from pointer	
TXBUFR           DB   960 dup (?)       ; Data to be transmitted to AMTOR
tx_ascii_case	 DB	1 dup (?)	; 0=upper, 1=lower

case_flag	 db	1 dup (?)	; enable upper/lower when 1
RELY		 db	1 dup (?)	; enable higher reliability when !=0
RAW_CODE	 db	1 dup (?)	; enable reception of raw ITA2 codes
TUN_IND		 db 	1 dup (?)	; enable software tuning indicator
MARQ		 DB 	1 dup(?)	; memory ARQ enable/disable
Mem_flg		 DB	1 dup(?)	; memory ARQ in progress flag
Blk2		 DB	3 dup(?)	; Block2 used during memory ARQ

temp_selcal      DB     4 dup (?)       ; Temp buffer for assembling SELCAL

	EVEN
data_bit	db  1 dup(?)		; copy of last data input

;----------------------------------------------------------------------------
; AMTOR stack, used when AMTOR module gains control

	DB	STKSIZE dup (?)	 ; AMTOR STACK
AM_STACK db	1	dup (?) 
        DB      128     dup (?)  ; dummy padding
;----------------------------------------------------------------------------
; Misc. variables. 

processor_type  dw 1 dup(?)	 ; Processor class	
tst_pat DB      1 dup (?)        ; used for testing pattern

	EVEN
param           dw  1 dup(?)     ; Save calling parameters
param2          dw  1 dup(?)
irq		dw  1 dup(?)	 ; IRQ used

evm_in		db  1 dup(?)
evm_out		db  1 dup(?)

systate         db  1 dup(?)     ; System input state 
status_byte     db  1 dup(?)     ; status byte
	EVEN
RQ_COUNT  	DW  1 dup (?)	 ; Cumlative count of RQ's
ER1_COUNT 	DW  1 dup (?)	 ; Cumulative count of ERROR's
ER2_COUNT 	DW  1 dup (?)	 ; Cumulative count of successful MARQ's
Bits		db  1 dup(?)	 ; Byte width of UART
	EVEN
AD_mark_data	dw  1 dup(?)	;  Peak mark  A/D value
AD_space_data	dw  1 dup(?)	;  Peak space A/D value
	EVEN
INT_BUSY	db  1 dup(?)	 ; re-entrancy flag
TX_STATE	db  1 dup(?)	;  TX_STATE=0 receive, =1 transmit

; Handle dynamic allocation of COM port
	EVEN
COM     	= 03F8h         ; COM1 Port address default
com_port	dw  COM
com_port1	dw  COM+1
com_port2	dw  COM+2
com_port3	dw  COM+3
com_port4	dw  COM+4
com_port5	dw  COM+5
com_port6	dw  COM+6



AMTOR	proc	near
;----------------------------------------------------------------------------
;                         AMTOR STATUS BYTE
;      
;  7 6 5 4 3 2 1 0 (bits)
;            = = = -----------0 = ERROR
;                             1 = RQ
;                             2 = TRAFFIC                         
;                             3 = IDLE                         
;                             4 = OVER                         
;                             5 = PHASE                         
;                             6 = STBY                         
;          3 -----------------1 = TX,  0 = RX
;        4 -------------------1 = ARQ, 0 = FEC     
;      5 ---------------------?
;    6 -----------------------?
;  7 -------------------------? 
;----------------------------------------------------------------------------
;            STANDBY MODE
;----------------------------------------------------------------------------
	EVEN
STBY:	lea     bx,AM_STACK
        mov     sp,bx           ; SP
        mov     ax,cs
        mov     ss,ax           ; SS
	
        lea     bx,REGS         ; set register pointer
        mov     RPTR,bx
        
        lea     bx,PBUF         ; set rxbuf pointer
        mov     PIPTR,bx
        mov     POPTR,bx
        
        lea     bx,KBUF         ; set txbuf pointer
        mov     KIPTR,bx

        call    NXTK
        mov     KOPTR,bx        ; set TXBUF "pseudo-full"
 
        lea     bx,IDENT        ; set up SELCAL code
        call    MVSEL

        mov     al,065h         ; CONTROL 1
        mov     STFG,al
        mov     RBFG,al         ; last control sent
        mov     LCTR,al         ; set rubbish in last CTRL rcvd

        mov     al,080h
        mov     TXDTA,al        ; busy TXDATA to user program

        xor     al,al           ; clear al                             
        mov     TFLG,al         ; clear transmit request
        mov     TRQFG,al
        mov     QFLG,al         ;  .. and QRT request
        mov     RAFG,al         ; turn phasing off
        mov     RXDTA,al        ; clear rubbish RXDATA
	mov	cs:Mem_flg,al	; turn memory ARQ-in-progress off

	mov	rx_ascii_case,al ; Set both rx and tx case for upper
	mov	tx_ascii_case,al
	mov 	TX_STATE,al	; in receive state here

        mov     al,066h         ; standby status code
        mov     STSTS,al        ; put in status byte for user's use

        ; TX off, set to mark
	mov	al,EVM_RX
	mov	cs:evm_out,al


SRCH:   mov     ah,210          ; search for ARQ selcal or mode B signals
SRC0:   call    SMPLE           ; get next ms sample into shiftregisters
        dec     ah
        jnz     SRC0            ; loop until shiftreg filled with 210 bits

SRC1:   call    SMPLE           ; get another RX sample
        mov     al,[bx]         ; form shiftreg into 7-bit characters
        mov     ah,[bx+10]
        shr     al,1
        mov     RXCP+2,al
        mov     al,[bx+20]
        rcr     ah,1
        rcr     al,1
        shr     ah,1
        mov     RXCP+1,ah
        rcr     al,1
        shr     al,1
        mov     RXCP,al

        lea     bx,SLCAL        ; point to our ARQ selcal code
        call    CPBK            ; check RXCP against our ARQ selcal
        je      ASY1L           ; and skip out if match
        lea     bx,SLCAL+3      ; check other half of selcal
        call    CPBK
        je      ASY2L           ; and skip if match

        lea     bx,IDLE         ; ckeck if mode B signal
	call	CPBK
        je      BSYL            ; and skip if yes

        mov     ah,QFLG         ; test if qrt is set (during RESYNCH)
	or	ah,ah		; SET UP FLAGS
        jnz     STBYL           ; exit to STBY if yes
        jmp     SHORT SRC1      ; else keep searching

BSYL:   jmp     BSY
STBYL:  jmp     STBY        
ASY1L:  jmp     ASY1
ASY2L:  jmp     ASY2

; COMPARE 3 CODES IN RXCP WITH 3 AT (bx)
	EVEN
CPBK:   mov     al,RXCP        ; get 1st code
        cmp     al,[bx]        ; does it match ?
        jne     CPB1           ; exit Z-bit clear if not
        mov     al,RXCP+1      ; ditto 2nd
        cmp     al,[bx+1]
        jne     CPB1
        mov     al,RXCP+2      ; ditto 3rd
        cmp     al,[bx+2]
CPB1:   ret                    ; return with Z-bit set if match

;----------------------------------------------------------------------------
; SAMPLES RXFSK AT NEXT MS AND SHIFTS INTO SEARCH REGISTERS
;
; The scanning for ARQ selcal, or a valid mode B signal is done
; using ten shift registers called REGS. RPTR points to one
; of ten of these registers. On return from "DROP", i.e. after
; the next 1ms period, the data input line is sampled, and the bit
; rotated into the shiftregister being pointed to. Note that the
; shift register is actually made up of three linked bytes. For
; example the register at RPTR=&REG consists of (REG)(REG+10)(REG+20).
; After 210 cycles through SMPLE, we end up with ten shift registers
; full of data, each shift register a 3 character CCIR.476 block, where
; the shiftregisters are 1ms apart.
; After filling the shiftregisters initially, each call to SMPLE thus
; gives a 3 character block - 1ms later. This way the search for
; synch may proceed at 1ms resolution.
; 
;----------------------------------------------------------------------------
	EVEN
SMPLE:  call    DROP            ; go to user till next IRQ
        mov     bx,RPTR        	; get current shiftreg pointer
        inc     bx              ; next section of register
        cmp     bx,offset REGS+10 ; test gone too far ?
        jne     NXR1

        lea     bx,REGS        	; if so start round again
NXR1:   mov     RPTR,bx

	mov	al,data_bit

        shl     al,1            ; shift into carry 
        rcr     byte ptr [bx],1 ; and into shiftregister
        rcr     byte ptr [bx+10],1
        rcr     byte ptr [bx+20],1
        ret

; HERE IF RECOGNISED HALF A SELCAL

ASY1:   lea     bx,SLCAL+3      ; point to 2nd half
        jmp     SHORT ASY3

ASY2:   lea     bx,SLCAL        ; point to 1st half
ASY3:   mov     TEMP2,bx

        mov     bx,-3       	; set clock 3ms early in waiting for
        mov     CLOCK,bx	; the rest of the last bit to pass by

        mov     byte ptr BFLG,0 ; set to count 450ms cycles (mode A)
        mov     bx,cs:CD_DELAY	; to help with local QSO's or master's
				; rx recovery problems
        mov     CTTX,bx         ; set transmit timing constants
        mov     BKTX,bx         

        mov     bx,390          ; set RX timing constants  (0186h)
        mov     CTRX,bx
        mov     bx,250
        mov     BKRX,bx
        mov     byte ptr MASTER,1        ; set for slave
	mov	byte ptr TX_STATE,1
	mov	byte ptr ORIG,1
        call    RBLK            ; receive the next block

        mov     bx, TEMP2       ; recover which selcal half
        call    CPBK            ; see if it matches
        je      ASY5            ; skip if yes
        jmp     SRCH            ; else go back to searching

ASY5:   mov     byte ptr NOCY,0 ; clearing rephasing timeout counter
        mov     ah,STFG         ; test if in standby mode (or rephase)
	or	ah,ah		; SET UP FLAGS
        je      ASY6            ; skip if rephase

        mov     bx,KIPTR        ; if STANDBY, clear TXBUFFER to enable
        call    ENBLE           ; enable keyboard

        mov     al,15h          ; set mode A slave phasing status
        mov     STSTS,al
				; bit phasing will be on, start off with
ASY6:   mov	word ptr ACPH, 0 ; no accum. error
	jmp     ARX6            ; go to mode A program

; HERE FOR PROVISIONAL RECOGNITION OF MODE B SIGNAL
BSY:    mov     bx,57           ; set clock 3ms early of end of character
        mov     CLOCK,bx
        mov     MASTER,al       ; set hi for slave phasing
	mov	byte ptr ORIG,1
 
        mov     BFLG,al         ; set clock counting by 70ms
	mov     word ptr ACPH,0	; note that bit phasing will be active
                                ; starting off with no accumulated error
        call    RXB             ; receive a mode B character
        cmp     al,0Fh          ; test if it is still a mode B idle
        je      BSY2            ; skip if it is
        jmp     SRCH            ; else go back searching

BSY2:   mov     byte ptr ERRT+1,0  ; hi part of ERRT=0
        mov     ah,023h         ; mode B idle status
        mov     STSTS,ah
        jmp     BRX             ; go to mode B rx program

; MODE A RECEIVE PROGRAM LOOP
	EVEN
ARX:	cmp	cs:MARQ,0
	jnz	MARX
	jmp	ARX_normal   

;-------Memory ARQ---------------------------------------------------------
; Here are some details on how memory ARQ is supposed to work in PCTOR.
;
; From Barry McLarnon's article in Jan. 1988 QEX: New directions
; in Amateur HF data Transmission systems and Metzner and Chang's
; May 1985 article, IEEE Trans on Communications. VOL. COM-33(5):
; Efficient selective repeat ARQ strategies for very noisy and fluctuating 
; channels.
;
; (Assume we digitize the data into values (+0  to +15) and (0 to -15).
;
; Level 1 (use only hard bits): 
;	Assume the first block's hard bits and soft bit values is 
; saved. From two packets, attempt reconstruction using complete characters
; form both blocks. Where there is a choice between blocks, prefer the first
; one. Move the reconstructed block into the slot of the first block.
; 
; Level 2 (use hard and soft bits):
; 	Assume the level 1 test failed, so now we resort to soft bit values.
; From the two blocks, consider only bits from acceptable characters, i.e.
; all ones, or all zeroes is bad, regardless of the soft bit values. We
; know we are dealing with a 4/3 ratio code, thus a threshold of 
; (4*15/2=30) for the sum of positive soft bits, and (3* -15/2)=-23 for 
; negative soft bits will indicate that a character is reasonable.
; Now add soft bit values on a bit-by-bit basis, remember to limit at
; +/- 15. Then construct the binary block by threshholding the soft bits
; using a threshold of the halfway mark, i.e. 15/2=7 for positive values, 
; -15/2=-7 for negative	values. Perform the 4/3 test. Always replace the 
; old block with the latest reconstructed one. This scheme should converge 
; quickly.
;
;-------Memory ARQ---------------------------------------------------------

MARX:	call    MRBLK            ; receive a block
        mov     al, byte ptr TEMP1   ; test if this block contained an RQ
	or	al,al		; SET UP FLAGS
        jnz     MARRQ0          ; skip if it did

        mov     al,byte ptr TEMP1+1  ; test if this block contained an error
	or	al,al		; SET UP FLAGS
        jnz     MARER           ; skip if it did

	mov	byte ptr cs:Mem_flg,0	; good block now, turn memory arq off again
        mov     al,RXBUF        ; get 1st character in block
        cmp     al,0Fh          ; test if this block is all alphas
        jne     MARX1
        mov     bx, word ptr RXBUF+1  ; RXBUF+1, RXBUF+2 order does'nt matter
        cmp     bx,0F0Fh
        je      MARCL           ; skip to QRT if it is

MARX1:  mov     bx,PIPTR        ; see if RX buffer is empty       
        cmp     bx,POPTR       
        jne     MARRQ           ; RQ cycle if not empty
        
        call    OUTC            ; for testing 1st char for traffic/idle status
        call    UPDTE           ; and update idle/traffic status

        mov     al,RXBUF+1      ; get 2nd character
        call    OUTC            ; and output it

        mov     al,RXBUF+2      ; 3rd character
        call    OUTC

MAROK:  mov     al,RBFG         ; get last control sent
        xor     al,0Fh          ; change to the other side
        mov     RBFG,al         ; and save it

        mov     al,TRQFG
        mov     TFLG,al         ; copy TX request across
        mov     byte ptr NOCY,0   ; clear timeout

MARX6:  mov     al,TFLG         ; test for transmit
	or	al,al		; SET UP FLAGS
        je      MARX7
        jmp     ARTC           	; ..yes, go changeover

MARRQ0:	inc	word ptr RQ_COUNT ; update statistics $$$$$$$$$$$$
MARRQ:  mov	byte ptr cs:Mem_flg,0	; clear memory arq flag
	mov     ah,01           ; here for RQ cycly, set RQ status
        jmp     SHORT MARCT

MARER:  mov	byte ptr cs:Mem_flg,1 ; initiate memory arq
      	mov	ah,07h		; Indicate MARQ in progress
	inc	word ptr ER1_COUNT ; update statistics $$$$$$$$$$$$
				; here for error cycle, set error status
MARCT:  call    UPDTE           ; update ERROR/RQ status
MARX7:  mov     al,RBFG         ; get control code to send
        call    TCTL            ; send it
        jmp     ARX		; and go round for next block

MARCL:  xor     al,RBFG         ; here for QRT, form NEXT control code
        call    TCTL            ; send it
        jmp     STBY            ; and go to STANDBY

; Normal (non MAQR) ----------
ARX_normal:
	call    RBLK            ; receive a block
        mov     al, byte ptr TEMP1   ; test if this block contained an RQ
	or	al,al		; SET UP FLAGS
        jnz     ARRQ0           ; skip if it did

        mov     al,byte ptr TEMP1+1  ; test if this block contained an error
	or	al,al		; SET UP FLAGS
        jnz     ARER            ; skip if it did

        mov     al,RXBUF        ; get 1st character in block
        cmp     al,0Fh          ; test if this block is all alphas
        jne     ARX1
        mov     bx, word ptr RXBUF+1  ; RXBUF+1, RXBUF+2 order does'nt matter
        cmp     bx,0F0Fh
        je      ARCL            ; skip to QRT if it is

ARX1:   mov     bx,PIPTR        ; see if RX buffer is empty       
        cmp     bx,POPTR       
        jne     ARRQ            ; RQ cycle if not empty
        
        call    OUTC            ; for testing 1st char for traffic/idle status
        call    UPDTE           ; and update idle/traffic status

        mov     al,RXBUF+1      ; get 2nd character
        call    OUTC            ; and output it

        mov     al,RXBUF+2      ; 3rd character
        call    OUTC

AROK:   mov     al,RBFG         ; get last control sent
        xor     al,0Fh          ; change to the other side
        mov     RBFG,al         ; and save it

        mov     al,TRQFG
        mov     TFLG,al         ; copy TX request across
        mov     byte ptr NOCY,0   ; clear timeout

; This is the entry point for mode A reading blocks -----
ARX6:   mov     al,TFLG         ; test for transmit
	or	al,al		; SET UP FLAGS
        je      ARX7
        jmp     ARTC            ; ..yes, go changeover

ARRQ0:	inc	word ptr RQ_COUNT ; update statistics
ARRQ:   mov     ah,01           ; here for RQ cycly, set RQ status
        jmp     SHORT ARCT

ARER:   xor     ah,ah           ; here for error cycle, set error status
	inc	word ptr ER1_COUNT ; update statistics
ARCT:   call    UPDTE           ; update ERROR/RQ status
ARX7:   mov     al,RBFG         ; get control code to send
        call    TCTL            ; send it
        jmp     ARX		; and go round for next block

ARCL:   xor     al,RBFG         ; here for QRT, form NEXT control code
        call    TCTL            ; send it
        jmp     STBY            ; and go to STANDBY

; OUTPUT CHARACTER FROM A-REG INTO RX BUFFER
	EVEN
OUTC:   cmp	al,SPEC_SYM	; special character for mode L
	je	OUT4		; place token, skip all case setting

	or	al,al		; SET UP FLAGS
	js      OUT1            ; test if printable character
        mov     ah,03h          ; if not, set idle status
        ret                     ; and exit

OUT1:   mov     ah,040h         ; preset FIGS
        cmp     al,09Bh         ; test to see if character is "FIGS"
        je      OUT2            ; skip if it is

        cmp     al,09Fh         ; test if character is "LTRS"
        jne     OUT3            ; skip if not
        
        xor     ah,ah           ; reclear to LTRS
OUT2:   mov     SHIFT,ah        ; update shift if required
        jmp     SHORT OUT4

OUT3:   or      al,SHIFT        ; mark bit 6 if figs
        mov     ah,LSTCH
        mov     LSTCH,al        ; swap last character
        cmp     al,0D9h         ; is it "?"
        jne     OUT4            ; skip if not
        cmp     ah,0D1h         ; was last character "+"
        jne     OUT4            ; skip if not
        mov     TRQFG,ah        ; if yes, set TX RQST flag

OUT4:   mov     ah,02           ; traffic status code
OUT5:   mov     bx,PIPTR        ; get RX buffer pointer
        call    NXTP            ; move next free location in RXBUF
        mov     byte ptr [bx],al ; put character in it
        mov     PIPTR,bx        ; save new pointer
        ret

; RECEIVE A 7-BIT CODE STARTING 10MS FROM NOW.
;----------------------------------------------------------------------------
;  data bit 0         1         2
;	        10ms      10ms     10ms
;     phase |---- ++++|---- ++++|---- ++++
;     data  -----\_________/---------\_____
;  samples  ^         ^         ^          ^
;----------------------------------------------------------------------------
; This modified procedure takes two samples of the data bits,
; one sample is taken in the interrupt routine, prior to returning
; control to the data sampling routine (a couple of us later).
; If both bits are zero, result is zero, if both bits is one, result
; is one, else result is an error
;----------------------------------------------------------------------------
	EVEN
RCV:    call    WTEN            ; wait 10ms
RCV1:   mov     al,040h         ; preload shiftregister in AL
        mov     RAFG,al         ; enable phasing operation

RCV2:   mov     word ptr PHASE,-5 ; set phasing counter to -5
                                ; get RX data bit
	xchg	al,ah		; save bytes

	mov	al,data_bit

	xchg    al,ah		; swap back
	and     ah,080h         ; mask to get to data 0/1 bit
	cmp	byte ptr RELY,0 ; is high reliability on ?
	jz	RCV2C		; skip out if yes
	rol	ah,1		; otherwise use bit averaging
	rol	byte ptr HIST_1,1 ; combine with previous bit sample
	add	ah,byte ptr HIST_1
	cmp	ah,2 		; 1,1
	je	RCV2A
	cmp	ah,0		; 0,0
	je	RCV2B
	inc 	byte ptr TEMP1+1 	; possible error
	jmp 	SHORT RCV2C
RCV2A:	mov	ah,080h
	jmp	SHORT RCV2C
RCV2B:	mov	ah,0
RCV2C:	shl     ah,1            ; shift into carry 
        rcr     al,1
        jc      RCV3            ; skip if preloaded bit now in carry
        call    WTEN            ; else wait for next bit
        jmp     RCV2            ; and go back and get it
RCV3:   mov     byte ptr RAFG,0   ; when finished switch phasing off
        shr     al,1            ; justify code to lsb-end of byte
        mov     ah,al           ; save code in ah

RCV4:   push    ax
        xor     ah,ah           ; look up ITA2 code from table
        lea     bx,RTAB
        add     bx,ax
        pop     ax
        mov     al, byte ptr[bx] ; get AMTOR character

	or	al,al		; SET UP FLAGS
        jnz     RCV5
        inc     byte ptr TEMP1+1    ; if invalid, set error flag
	
RCV5:   cmp     al,066h         ; if it was an RQ..
        jne     RCV6
        mov     byte ptr TEMP1,al   ; .. then set TEMP1 to mark RQ

RCV6:   ret

;------Memory ARQ------------------------------------------------------------
; Compose a character using memory ARQ 
	EVEN
MRCV:   call    WTEN            ; wait 10ms
MRCV1:  mov     al,040h         ; preload shiftregister in AL
        mov     RAFG,al         ; enable phasing operation

MRCV2:  mov     word ptr PHASE,-5 ; set phasing counter to -5
                                ; get RX data bit
	xchg	al,ah		; save bytes

	mov	al,data_bit

	xchg    al,ah		; swap back
	and     ah,080h         ; mask to get to data 0/1 bit
	cmp	byte ptr RELY,0 ; is high reliability on ?
	jz	MRCV2C		; skip out if yes
	rol	ah,1		; otherwise use bit averaging
	rol	byte ptr HIST_1,1 ; combine with previous bit sample
	add	ah,byte ptr HIST_1
	cmp	ah,2 		; 1,1
	je	MRCV2A
	cmp	ah,0		; 0,0
	je	MRCV2B
	inc 	byte ptr TEMP1+1	; possible error
	jmp 	SHORT MRCV2C
MRCV2A:	mov	ah,080h
	jmp	SHORT MRCV2C
MRCV2B:	mov	ah,0
MRCV2C:	shl     ah,1            ; shift into carry 
        rcr     al,1
        jc      MRCV3           ; skip if preloaded bit now in carry
        call    WTEN            ; else wait for next bit
        jmp     MRCV2           ; and go back and get it
MRCV3:  mov     byte ptr RAFG,0   ; when finished switch phasing off
        shr     al,1            ; justify code to lsb-end of byte
        mov     ah,al           ; save code in ah

MRCV4:  push    ax
        xor     ah,ah           ; look up ITA2 code from table
        lea     bx,RTAB
        add     bx,ax
        pop     ax
        mov     al, byte ptr[bx] ; get AMTOR character

	or	al,al		; SET UP FLAGS
        jnz     MRCV5
        inc     byte ptr TEMP1+1    ; if invalid, set error flag

MRCV5:  cmp     al,066h         ; if it was an RQ..
        jne     MRCV6
        mov     byte ptr TEMP1,al   ; .. then set TEMP1 to mark RQ

MRCV6:  ret
;---------------------------------------------------------------------------
; WAIT TILL CLOCK IS EQUAL TO bx-REGISTER
	EVEN
HOLD1:	call	DROP		; go to user while waiting for next IRQ
HOLD:	cmp	bx,word ptr CLOCK	
	jne	HOLD1		; loop till bx = CLOCK
	ret

; WAIT TEN MILLISECS
	EVEN
WTEN:	mov	ah,10	        ; set count of ten
WAITE:	call	DROP		; go to user till next IRQ
	dec	ah	        ; count down
	jne	WAITE		; loop back till done
	ret		

; RECEIVE A MODE A BLOCK (Non memory ARQ)
	EVEN
RBLK:	mov	word ptr TEMP1,0  ; clear ERROR and RQ flags

	mov	bx,BKRX		; get block RX time constant
	call	HOLD		; wait till this time
	call	RCV1		; receive a character starting now
	mov	RXBUF,al	; put the ITA2 code away
	mov	RXCP,ah		; put the raw AMTOR code away

	call	RCV		; get the next character 10ms from now
	mov	RXBUF+1,al
	mov	RXCP+1,ah

	call	RCV		; and the 3rd
	mov	RXBUF+2,al
	mov	RXCP+2,ah
	ret

;----Memory ARQ--------------------------------------------------------------
; Receive a mode A block, using memory when needed ARQ
	EVEN
MRBLK:	mov	word ptr TEMP1,0  ; clear ERROR and RQ flags

	mov	bx,BKRX		; get block RX time constant
	call	HOLD		; wait till this time

	cmp	cs:Mem_flg,0	; memory ARQ in progress ?
	jnz	MRQ
	jmp	MRB0
     	       			; YES, memory ARQ in progress
MRQ:	call	MRCV1		; receive a character starting now
	mov	Blk2,al		; put the ITA2 code away
	mov	RXCP,ah		; put the raw AMTOR code away

	call	MRCV		; get the next character 10ms from now
	mov	Blk2+1,al
	mov	RXCP+1,ah

	call	MRCV		; and the 3rd
	mov	Blk2+2,al
	mov	RXCP+2,ah

; does Block2 contain RQ ?
        mov     al, byte ptr TEMP1   
	or	al,al		; SET UP FLAGS
        jz      MRA             ; skip if it did
	jmp	MRBx

; does Block2 contain Errors ?
MRA:    mov     al, byte ptr TEMP1+1   
	or	al,al		; SET UP FLAGS
        jnz	MRB0a

	mov	al,Blk2		; NO, then RXBUF <= Blk2
	mov	RXBUF,al
	mov	al,Blk2+1
	mov	RXBUF+1,al
	mov	al,Blk2+2
	mov	RXBUF+2,al
	jmp	SHORT MRBx  	; and exit

; Try reconstruct a good block from the two bad ones
; If there is a choice between the two blocks, arbitrarily use
; Block1 (less data to shuffle).

MRB0a:	mov	byte ptr TEMP1+1,0 ; assume no errors

	mov	al,byte ptr cs:RXBUF
	or	al,al		; is Block1 character OK?
	jnz	MRB0b		
				; NO, see about Block2 character
	mov	al,byte ptr cs:Blk2
	mov	byte ptr cs:RXBUF,al
	or	al,al		; is Block2 character OK
	jnz	MRB0b
	inc	byte ptr cs:TEMP1+1 ; Error

MRB0b:	mov	al,byte ptr cs:RXBUF+1
	or	al,al		; is Block1 character OK?
	jnz	MRB0c		
				; NO, see about Block2 character
	mov	al,byte ptr cs:Blk2+1
	mov	byte ptr cs:RXBUF+1,al
	or	al,al		; is Block2 character OK
	jnz	MRB0c
	inc	byte ptr cs:TEMP1+1 ; Error

MRB0c:	mov	al,byte ptr cs:RXBUF+2
	or	al,al		; is Block1 character OK?
	jnz	MRB0d		
				; NO, see about Block2 character
	mov	al,byte ptr cs:Blk2+2
	mov	byte ptr cs:RXBUF+2,al
	or	al,al		; is Block2 character OK
	jnz	MRB0d
	inc	byte ptr cs:TEMP1+1 ; Error

MRB0d:	cmp	byte ptr cs:TEMP1+1,0
	jnz	MRB0d0
	inc	word ptr cs:ER2_COUNT	; update statistics

MRB0d0:	jmp	SHORT MRBx  	; and exit

				; No, Memory ARQ not in progress	
MRB0:	call	MRCV1		; receive a character starting now
	mov	RXBUF,al		; put the ITA2 code away
	mov	RXCP,ah		; put the raw AMTOR code away

	call	MRCV		; get the next character 10ms from now
	mov	RXBUF+1,al
	mov	RXCP+1,ah

	call	MRCV		; and the 3rd
	mov	RXBUF+2,al
	mov	RXCP+2,ah

MRBx:	ret
;---------------------------------------------------------------------------
	
; POINT THE bx-REG AT THE NEXT LOCATION IN THE RX BUFFER
	EVEN
NXTP:	inc	bx		; next location
	cmp	bx, offset PBUF+PBUFSIZ ; test if off end of the buffer space
	jne	NXP1
	lea	bx,PBUF		; if so, start at the beginning again
NXP1:	ret

; TRANSMIT A CONTROL CODE FROM THE al-REG
	EVEN
TCTL:	mov	bx,CTTX		; get control transmit time constant
	call	TXON		; go wait till time and turn TX on
	call	TXC2		; then go send 7 bit code

; turn TX off 
TCT2:   push    ax

	; TX off
	mov	al,EVM_RX
	mov	cs:evm_out,al

        pop     ax                
	ret
	

;----------------------------------------------------------------
; Transmit the bit on the carry flag, preserving AX
;----------------------------------------------------------------
	EVEN
TXBIT:	push    ax  
        jc      TXCS1           	; is bit 1 or 0
        mov     al,EVM_TX_0
        jmp     SHORT TXCS2     

TXCS1:  mov     al,EVM_TX_1
TXCS2:	xor     al,MASKOUT
	mov	cs:evm_out,al
        pop     ax
	ret



; MODE A CHANGE-TO-TX ROUTINE
	EVEN
ARTC:	mov	ah,0Ch		; change to transmit status
	call	UPDTE
	mov	al,059h		; set CONTROL3 code
	call	TCTL		; send it
	call	RBLK		; receive the next block
	lea	bx,BKBAB	; point to beta-alpha-beta string
	call	CPBK		; is this block beta-alpha-beta ?
	jne	ARTC		; if not loop back

; After change-over, reset the receive case
	call	_force_ltrs
	mov	byte ptr TX_STATE,1	; we are sending now
	xor	al,al
	mov	byte ptr NOCY,al   ; clear timeout
	mov	byte ptr TRQFG,al  ; clear TX RQST flag
	mov	ah,ORIG         ; test if master or slave
	or	ah,ah		; SET UP FLAGS
	jne	ATXBL		; if slave, go send block of RQ's

	mov	al,066h		; if master, just send a single RQ
	call	TCTL		; .. as a control code
        xor	byte ptr MASTER,0FFh  ; change master/slave status
	jmp	ATX		; and go to transmit routine
ATXBL:	jmp     ATXB

; TEST FOR TIMEOUT, TURN TX ON AT RIGHT TIME.

TXON:	mov	ah,NOCY		; get timeout counter
	cmp	ah,MAX_TIMEOUT	; test for timeout
	jns	REPHA		; go rephase if exceeded
	call	HOLD		; else wait for the time constant

TXNO:	push    ax
        ; Turn on TX and send mark data
	mov	al,EVM_TX_1
	mov	cs:evm_out,al
        pop     ax               

	mov	ah,DELAY	; get TX delay constant
	jmp	WAITE		; and wait that long, then return

; HERE IF TIMEOUT IN MODE A

REPHA:  lea     bx, AM_STACK 	; reset AMTOR stack pointer
        mov     sp,bx           ; SP
        mov     ax,cs           ; SS
        mov     ss,ax

	mov	ah,STSTS	; get SEND/RECEIVE status bit
	and	ah,08h	
	add	ah,05h		; add phasing status code
	call	UPDTE

	mov	ah,ORIG         ; test if we originated the call (we=0)
	or	ah,ah		; SET UP FLAGS
	jnz	RPH2		; skip if slave
	jmp	CLA1		; go make master call

RPH2:	jmp	SRCH		; go search for call if slave

; TRANSMIT A CHARACTER FROM (bx) STARTING NOW
	EVEN
TXCH:	mov	al,byte ptr [bx]  ; get character to send
TXC2:	or	al,080h		; set bit 7
TXC3:	shr	al,1		; shift end bit into carry
	jz	TXC4		; skip if no bits left to send

        call	TXBIT

        call	WTEN		; and wait for next bit
	jmp	SHORT TXC3	; and go back for it

TXC4:	inc	bx		; point to next  character to send
	ret

; TRANSMIT A BLOCK OF 3 FROM (bx)
	EVEN
TBLK:	mov	TEMP1,bx	; save location of block
	mov	bx,BKTX		; get block transmit time constant
	call	TXON		; go turn on that time
	mov	bx,TEMP1	; recover block pointer
	call	TXCH		; send three characters
	call	TXCH  
	call	TXCH  
	jmp	TCT2		; and turn TX off and return

; UPDATE STATUS BITS 0-3 FROM ah
	EVEN
UPDTE:	xor	ah,STSTS
	and	ah,0Fh 
	xor	ah,STSTS
	mov	STSTS,ah
	ret

; RECEIVE CONTROL CODE INTO A-REG
	EVEN
RCTL:	mov	bx,CTRX		; get control receive time constant
	call	HOLD		; wait till correct time
	jmp	RCV1		; then go get code and return

; MODE A TRANSMIT LOOP

ATX:	call	RCTL		; receive a control code
ATX2:	cmp	al,LCTR		; same as last one received ?
	je	ATX5		; skip if yes (repeat requested)

	cmp	al,092h		; is it CONTROL ?
	je	ATX3		; skip if yes (next block requested)

	cmp	al,080h		; CONTROL 2 ?
	je	ATX3		; skip if yes (next block requested)

	cmp	al,08Ch		; CONTROL 3?
	jne	ATXA
	jmp	ATRC		; if C3, then changeover requested

ATXA:	mov	ah,08h		; if none, then error, set ERROR status
	call	UPDTE
ATXB:   lea     bx,BKRQ         ; point to a block of RQ's
        jmp     SHORT ATX7      ; and go to request rpt control

ATX3:   mov     LCTR,al         ; save (good) control received
        call    GETC            ; get a charater to send
        mov     TXBUF,al        ; and put it for sending
        cmp     al,033h         ; is it an idle ?
        jne     ATX4

        mov     al,QFLG         ; yes, the test for QRT request
	or	al,al		; SET UP FLAGS
        jnz     AEOC            ; if QRT set, then go QRT

ATX4:   call    UPDTE           ; else update idle/traffic status
        call    GETC            ; get remaining two characters for tx
        mov     TXBUF+1,al      ;
        call    GETC
        mov     TXBUF+2,al
        mov     byte ptr NOCY,0 ; clear timeout
        jmp     SHORT ATX6      ; go send new block
  
ATX5:   mov     ah,09h          ; here if repeat requested, set RQ status
        call    UPDTE

ATX6:   lea     bx,TXBUF        ; point to the block to be sent
ATX7:   call    TBLK            ; and send whichever block
        jmp     SHORT ATX       ; and go round for next cycle


; SEND QRT GROUP IN MODE A AND CLOSEDOWN
AEOC:   mov     ah,0Ch          ; transmit changeover status
        call    UPDTE
        mov     ah,4
        mov     TFLG,ah         ; (temp counter only)

        mov     byte ptr NOCY,0 ; clear timeout to prevent rephasing on
AEC1:   lea     bx,BKALF        ; point to block of alphas
        call    TBLK            ; send them
        dec     byte ptr TFLG   ; count down to zero
        jz      AEC2            ; skip if done 4
        call    RCTL            ; receive control code
        cmp     al,LCTR         ; if repeat requested
        je      AEC1            ; then send another block
        cmp     al,092h         ; if C1, then block was received ok
        je      AEC2            ; ..so close down
        cmp     al,080h         ; ..and same if C2 was received
        jne     AEC1            ; else go back and send another block

AEC2:   jmp     STBY

; GET A CHARACTER FROM THE TX BUFFER INTO A-REG (OR IDLE)
	EVEN
GETC:   mov     bx,KOPTR        ; test if buffer empty
        cmp     bx,KIPTR
        jne     GET1
        mov     al,033h         ; if so, set idle code
        mov     ah,0Bh          ; set idle status
        ret
	EVEN
GET1:   call    NXTK            ; else point to next in buffer
        mov     KOPTR,bx        ; and save its location
        mov     al,byte ptr[bx] ; and get next character
        or      al,080h         ; add transmit lookup table offset
                                ; this actually sets index to point
                                ; to TTAB
        mov     ah,ECHO         ; test local copy echo flag
        or      ah,ah           ; SET FLAGS
        jz      GET2            ; no colcal copy if ECHO = 00
        call    OUT5            ; put character out for RX buffer

GET2:   xor     ah,ah           ; look up CCIR 476 code
        lea     bx,RTAB
        add     bx,ax
        mov     al, byte ptr[bx]     

        mov     ah,0Ah          ; traffic status
        ret                     ; return with character and status

; POINT TO NEXT LOCATION IN TX BUFFER
	EVEN
NXTK:   inc     bx
        cmp     bx, offset KBUF+4  ; test if off end of bufferspace
        jne     NXK1
        lea     bx,KBUF         ; if so, start at beginning again
NXK1:   ret


; MODE A CHANGE TO TO RX ROUTINE
ATRC:   mov     ah,04h          ; change to receive status
        mov     LCTR,ah         ; clear transmit block sequence
        call    UPDTE   
        lea     bx,BKBAB        ; point to BETA-ALPHA-BETA BLOCK
        call    TBLK            ; send it
        call    RCTL            ; receive control code
        cmp     al,66h          ; is it an RQ ?
        jne     ATRC            ; if not, go back and try again

	mov	byte ptr TX_STATE,0 ; we are receiving now 
        mov     byte ptr NOCY,0 ; if OK, then clear timeout
        mov     byte ptr TFLG,0 ; clear transmit request flag
        xor     byte ptr MASTER,0FFh ; change master/slave status
        jmp     ARX7            ; and go to receive loop

; MODE B RECEIVE PROGRAM, ENTER WHEN IN SYNC
	EVEN
BRX:    call    RXB             ; receive a 7-bit mode B character
        jnc     BRX4            ; exit if (carry set) errors too high

        mov     ah,TXBUF        ; move stored characters along by one
        mov     TXBUF,al
        mov     al,TXBUF+1
        mov     TXBUF+1,ah
        mov     TXBUF+2,al
        call    RXB             ; receive another 7 bit character
        jnc     BRX4            ; exit if errors too high

        mov     ah,TXBUF+2      ; recover earlier version of this char
        mov     bx, word ptr TXBUF   ; get last 2 "1st repeat" characters
        cmp     bx,0F0Fh        ; test for closedown code
        je      BRX4            ; and skip out if closedown
        or      al,al           ; SET FLAGS to see if 2nd repeat char is OK
        jne     BRX2            ; skip if it is oK

        mov     al,ah           ; else use 1st repeat character instead
        or      al,al           ; SET FLAGS
        jne     BRX2            ; and skip if that OK too

        xor     ah,ah           ; else set error status
	mov	al,SPEC_SYM	; if repeats are errors, show placeholder
	call    UPDTE	   	; update the error status
	call    OUTC            ; go output the placeholder character
        jmp     SHORT BRX       ; go for next character

BRX2:   call    OUTC            ; go output the character
	call    UPDTE
        jmp     SHORT BRX       ; go for next character

BRX4:   jmp     STBY            ; if bad errors or closedown, go standby

; RECEIVE A MODE B CHARACTER AND UPDATE ERROR RATE
	EVEN
RXB:    xor     bx,bx           ; clear bx
        call    HOLD            ; wait till clock comes round
        call    RCV1            ; tehn receive 7 bits starting now
        mov     cl,al           ; save ITA2 code
        or      al,al           ; was it a valid code ?
        jne     RXB1            ; skip if yes
        dec     al              ; else set al to FF (error high)
        jmp     SHORT RXB2
	EVEN
RXB1:   xor     al,al           ; if valid set error low
RXB2:   xor     ah,ah
        sub     al,byte ptr ERRT+1 ; form new-old error value (hi part)
        sbb     ah,0            ; and carry thu to msb
        shl     al,1            ; multiply by 4
        rcl     ah,1
        shl     al,1  
        rcl     ah,1
        add     al,byte ptr ERRT ; add to (OLD)*256  (lo part)
        mov     byte ptr ERRT,al
        adc     ah,byte ptr ERRT+1 ; carry thru
        mov     byte ptr ERRT+1,ah
        cmp     ah,0A0h         ; is smoothed ERRT now too high ?
        mov     al,cl           ; recover code       
        ret                     ; return with it and carry set if ERROR

; TRANSMIT MODE A SELCAL BLOCK
	EVEN
TSBK:   xor     SBFG,0FFh       ; toggle the selcal block flag
        jnz     TSB1
        lea     bx,SLCAL        ; if SBFG=0 then point to 1st half of
        jmp     TSB2            ;  selcal and go send it
TSB1:   lea     bx,SLCAL+3      
TSB2:   mov     NOCY,0          ; clear timeout to prevent rephasing
        jmp     TBLK            ; and go send the block


; MODE A MASTER CALLING ROUTINE, ENTERED WITH CONTRL CODE 03
CALLA:  lea     bx,AM_STACK  	; reset AMTOR stack pointer
        mov     sp,bx           ; SP
        mov     ax,cs           ; SS
        mov     ss,ax

        mov     ah,03Dh         ; mode A master/slave
        mov     STSTS,ah
        xor     al,al           ; clear al                             
        mov     MASTER,al       ; clear for master phasing
        mov     ORIG,al
        mov     BFLG,al         ; clear to count 450ms cycle

	mov	rx_ascii_case,al ; Set both rx and tx case for upper
	mov	tx_ascii_case,al

        inc     al
        mov     LCTR,al         ; set rubbish in last control RCVD
        lea     bx,SELCAL
        call    MVSEL           ; setup selcal for sending
        mov     bx,240          ; set up transmit time constants
        mov     BKTX,bx
        mov     bx,380
        mov     CTTX,bx

CLA1:   mov     byte ptr SBFG,0 ; clear selcal block toggle flag
CLA2:   mov     ah,QFLG         ; test if qrt set (in resynch)
        or      ah,ah           ; SET FLAGS
        jz      CLA21
        jmp     STBY            ; if so, go standby directly

CLA21:  call    TSBK            ; send a selcal block

        mov     ah,62           ; set shiftcounter
CLA3:   call    SMPLE           ; get RX sample at next ms
        dec     ah              ; count down samples
        jnz     CLA3            ; loop till shiftreg filled

CLA4:   xor     ah,ah           ; clear good code count
CLA5:   call    SMPLE           ; get another sample

        mov     al,byte ptr CLOCK
        cmp     al,240          ; test for end of receive gap
        je      CLA2            ; if endgap, go for another block

        mov     al, byte ptr[bx] ; get shiftregister contents
        shr     al,1            ; form into 7 bits
        cmp     al,065h         ; it it CONTROL1 ?
        je      CLA7            ; skip out if yes
        cmp     al,6Ah          ; is it CONTROL2 ?
        je      CLA7            ; skip if yes
        cmp     al,59h          
        jne     CLA4            ; if no control recognized, shift some

CLA7:   inc     ah              ; if good control, count it
        cmp     ah,5            ; 5 in a row ?
        jne     CLA5            ; if not, count some more
        call    RCV4            ; if good, convert to ITA2
        mov     cl,al           ; save temporary 

        mov     ax,CLOCK        ; get clock at this point
        sub     ax,62           ; estimate control start time (clock-70ms)
				; including approx 10ms delay
        mov     CTRX,ax         ; set control receive time constant
        mov     byte ptr MASTER,0        

        call    TSBK            ; send another selcal block
        call    RCTL            ; receive the next control at this time
        cmp     al,cl           ; is it the same as last one ? 
        jne     CLA2            ; if not, false alarm, keep calling
        mov     bx,CTRX         ; now sure we have a reply
        mov     BKRX,bx         ; set block time constant same as CTL
        mov     ah,STFG
        or      ah,ah           ; SET FLAGS
        jz      CLA8            ; skip if already in QSO
	mov	byte ptr TX_STATE,1	; transmitting now
        call    ENBLE           ; else enable keyboard

CLA8:   jmp     ATX2            ; go to main TX loop

ENBLE:  mov     bx,KIPTR
        mov     KOPTR,bx        ; set keyboard buffer empty
        mov     byte ptr TXDTA,0 ; clear dummy TX character
        mov     byte ptr STFG,0  ; clear standby flag
        ret

; MODE B TRANSMIT PROGRAM, ENTER WITH CONTROL CODE 04
CALLB:  lea     bx,AM_STACK 	; reset AMTOR stack pointer
        mov     sp,bx           ; SP
        mov     ax,cs           ; SS
        mov     ss,ax

	xor	al,al
	mov	tx_ascii_case,al  ; Set and tx case for upper
	mov	byte ptr TX_STATE,1	; transmitting now

        mov     ah,4Dh          ; mode B tx/phasing
        mov     STSTS,ah
        mov     bx,KIPTR
        call    ENBLE
        xor     bx,bx           ; clear bx
        mov     CLOCK,bx        ; set clock within 0-70
        call    TXNO            ; turn on now
        
        mov     al,10           ; set clock counting 0-70
        mov     BFLG,al         ; by setting BFLG=10

CLB1:   mov     COUNT,al        ; set count of 10
CLB2:   call    SIDL            ; send an idle
        dec     byte ptr COUNT           ; count down
        jnz     CLB2            ; loop back till sent 10 idles

        mov     al,28           ; set 28 character count
        mov     COUNT,al
CLB6:   dec     byte ptr COUNT  ; count down
        jz      CLB7            ; if done 28 characters, send 5 idles
        call    GETC            ; else get character to send
        call    UPDTE           ; update traffic/idle status
                                ; ah=0B idle, 0A= traffic
        cmp     al,33h          ; idle ?
        jne     CLBA            
        mov     ah,QFLG         ; if idle, test closedown request
        or      ah,ah           ; SET FLAGS
        jnz     CLBC            ; and if set, go closedown

CLB7:   mov     al,5           	; set count of 5
        jmp     SHORT CLB1      ; go send 5 idles
CLBA:   call    SCHR            ; here for traffic character, send it
        jmp     SHORT CLB6      ; then go back and send it.
CLBC:   mov     cx,15          	; here for closedown, sent idles for abt 2sec
CLBD:   mov     al,0Fh          ; set alpha code
        call    SCHR            ; send it (1st repeat get sent only)
        loop    CLBD            ; loop back till all sent
        jmp     STBY            ; then go to standby

; SEND IDLE/TRAFFIC CHARACTER IN MODE B
	EVEN
SIDL:   mov     al,66h          ; set 1st repeat channel idle code
        mov     ah,0Fh          ; set 2nd repeat channel idle code
        jmp     SHORT SCR1      ; go send them

SCHR:   mov     ah,al           ; set traffic character in both channel
SCR1:   mov     TXBUF,ah        ; save 2nd repeat temp.
        call    TCHR            ; send 1st repeat character
        mov     al,TXBUF        ; shift 2nd repeat thru delay store
        mov     ah,TXBUF+1
        mov     TXBUF+1,al
        mov     al,TXBUF+2
        mov     TXBUF+2,ah
TCHR:   mov     bx,7            ; send character, set bx counter
        call    HOLD
TCR1:   shr     al,1            ; shift into carry

        call	TXBIT

        dec     bx              ; count down
        jz      TCR2            ; skip out if finished
        call    WTEN            ; else wait till next bit
        jmp     SHORT TCR1      ; and go back and do it
        
TCR2:   ret

;----------------------------------------------------------------------------
; LISTENING MODE PROGRAM, ENTER WITH CONTROL CODE 05
; The shift register is used first to locate a posible
; synch by finding at least "5" good characters  in the
; bit stream. These are most probably the same character at
; different millisecond intervals. Then at that time a
; block (of three) characters is read at the given synch.
; If the block is OK, then we have achieved synch.
;---------------------------------------------------------------------------
	EVEN
LISN:   lea     bx,AM_STACK 	; reset AMTOR stack pointer
        mov     sp,bx           ; SP
        mov     ax,cs           ; SS
        mov     ss,ax

        xor     al,al           ; clear al                             
        mov     BFLG,al         ; set up for 450ms counting
        mov     TRQFG,al        ; clear restart flag

        inc     al
        mov     MASTER,al       ; set MASTER to trigger slave phasing
	mov	ORIG,al		;

        call    TCT2            ; turn transmitter off        

        mov     ah,55h          ; mode L phasing status
        mov     STSTS,ah
        mov     bx,250          ; set block receive time as slave
        mov     BKRX,bx
	mov 	word ptr ACPH,0	; bit phasing will be active, start off with
                                ; no accum. error
LSN0:   mov     ah,210          ; set shiftcount
LSN1:   call    SMPLE           ; get RX sample at next ms
        dec     ah
        jnz     LSN1            ; loop till shiftreg full of new data

LS00:   mov     byte ptr COUNT,0         ; clear goodblock count
        mov     byte ptr TEMP1+1,0       ; clear error flag

LS01:   cmp	byte ptr cs:QFLG,0	; ready to quit?
	je	LS01a
	jmp	STBY		; yes, return to standby
LS01a:	call    SMPLE           ; get another sample
        mov     al,[bx]         ; form s/reg into characters
        mov     ah,[bx+10]      
        shr     al,1
        mov     RXCP+2,al       ; save 3rd character
        mov     al,[bx+20]
        rcr     ah,1
        rcr     al,1
        shr     ah,1
        mov     RXCP+1,ah       ; save 2nd character
        rcr     al,1
        shr     al,1
        call    RCV4            ; test 1st character for error
        mov     al,RXCP+1
        call    RCV4            ; test 2nd character
        mov     al,RXCP+2       
        call    RCV4            ; test 3rd character

        mov     al,byte ptr TEMP1+1  ; test error flag
        or      al,al           ; SET FLAGS
        jnz     LS00            ; loop back if any error in block
        inc     byte ptr COUNT  ; if good block, count
        mov     ah,COUNT
        cmp     ah,GOOD_CNT     ; test if enough good samples in a row
        jne     LS01            ; if not yet 5, keep counting

        xor     bx,bx           ; if good ...
        mov     CLOCK,bx        ;  ... set clock in step
        call    RBLK            ; receive another block at this time
        mov     al,byte ptr TEMP1+1      ; test any error in this block
        or      al,al           ; SET FLAGS
        jnz     LSN0            ; if error, go back, keep trying

	mov	rx_ascii_case,al ; Set rx case for upper
	mov	RSHFT,al	; reset receive case to LTRS

        mov     TXBUF,al        ; to prevent spurious repeat-suppress
	mov	TXBUF+1,al
	mov	TXBUF+2,al

LS02:   mov     al,byte ptr TEMP1+1      ; test for error
        or      al,al           ; SET FLAGS
        jnz     LS03            ; skip if any error in this block

        mov     NOCY,al         ; else clear timeout count
LS03:   mov     al, byte ptr TEMP1       ; test if any RQ in this block
        or      al,al           ; if so, skip repeat suppression
        jnz     LS06        

        mov     bx,3            ; set count for repeat-supression
        xor     ah,ah           ; clear ah
LS04:   mov     al,[bx+RXBUF-1] ; get character from block
        cmp     al,[bx+TXBUF-1] ; compare with last one
        je      LS05            ; skip if same
        inc     ah
        mov     [bx+TXBUF-1],al ; .. and update last character

LS05:   dec     bx              ; count thru block
        jnz     LS04
        or      ah,ah           ; test if block was repeat
        jz      LS10            ; skip if was repeat
        jmp     SHORT LS07

LS06:   mov     al,66h          ; (RQ code)
        cmp     al,RXBUF        ; test if 1st character is RQ
        je      LS07            ; if so, skip out
        cmp     al,RXBUF+1
        jne     LS06A

        mov     al,9Fh          
        mov     RXBUF+1,al
        jmp     SHORT LS07

LS06A:  cmp     al,RXBUF+2
        jne     LS07            ; if not skip out
        mov     al,84h          ; if this block is 2nd synchblock, replace
        mov     RXBUF+2,al
			      	; Check for closedown sequence
LS07:   mov     al,RXBUF        ; get 1st character in block
        cmp     al,0Fh          ; test if this block is all alphas
        jne     LS07A
        mov     bx, word ptr RXBUF+1  ; RXBUF+1, RXBUF+2 order does'nt matter
        cmp     bx,0F0Fh
        jne     LS07A   
	jmp	SHORT LS12	; skip to (standby) mode if it is

LS07A:	mov     al,RXBUF	; If there is an error, replace char by
	or	al,al		;               a placeholder "_"
	jnz	LS07a
	mov	al,SPEC_SYM	; print special placeholder symbol		
LS07a:  call    OUTC            ; output the characters in this block
        mov     cl,ah           ; save TFC/IDLE status if 1st character
	mov     al,byte ptr RXBUF+1      
	or	al,al	
	jnz	LS07b
	mov	al,SPEC_SYM		
LS07b:  call    OUTC            ; output 2nd character
        mov     al,byte ptr RXBUF+2
	or	al,al	
	jnz	LS07c
	mov	al,SPEC_SYM		
LS07c:  call    OUTC            ; output 3rd character
        mov     ah,cl           ; recover status
LS07C:  mov     al,byte ptr TEMP1+1      ; test if error is in this block
        or      al,al           ; SET FLAGS
        jz      LS08    
        xor     ah,ah           ; if error, reset to error status
        jmp     SHORT LS09

LS08:   mov     al,byte ptr TEMP1        ; test for any RQ in this block
        or      al,al           ; SET FLAGS
        jz      LS09 
        mov     ah,01           ; if RQ, reset to RQ status

LS09:   call    UPDTE           ; output the status from ah
LS10:   mov     ah,TRQFG        ; see if set by +? or user
        or      ah,ah           ; skip out if yes, go resynch
	jne	LS12
        mov     al,NOCY
        cmp     al,4            ; test for continuous errors (32)
        je      LS12            ; if so, skip out and resync
	cmp	byte ptr cs:QFLG,0 ; ready to quit?
	jz	LS10a
	jmp	STBY		; yes, go back to standby
LS10a:	call    RBLK            ; else receive another block
        jmp     LS02      	; and go process it
LS12:   jmp     LISN

;-----------------------------------------------------------------------------
;      SUBROUTINE TO BUMP THE CLOCK
;----------------------------------------------------------------------------
	EVEN
TICK:   mov     bx,CLOCK        ; get CLOCK
        inc     bx              ; add 1ms
        cmp     byte ptr BFLG,0 ; test if mode A or B
        jnz     TIC1            ; skip if mode B
        cmp     bx,450          ; else test end of 450 ms cycle
        jmp     SHORT TIC2
	EVEN
TIC1:   cmp     bx,70           ; test for end of 70ms mode B cycle
TIC2:   jne     TIC3            ; if not end of cycle then skip
        inc     byte ptr NOCY   ; if end of cycle, bump TIMEOUT
        xor     bx,bx           ; and reset to start of cycle
TIC3:   mov     CLOCK,bx        ; and save CLOCK
        ret

;----------------------------------------------------------------------------
;                     INTERRUPT DRIVEN AMTOR MODULE                         
;                VECTORS HERE FROM USER PROGRAM EVERY MS                    
;
; Some notes on how bit phasing works:
; Bit phasing can only work on received data, i.e. slave mode A,
; and receiving mode B or mode L.
; 
; During ARQ master operation, only CTRX and BKRX will be adjusted.
; It is up to the ARQ slave to perform bit phasing.
; From experience, and PC's clocks not being too accurate, it has been
; found that MASTER and SLAVE status need to switch around in rx-tx and
; tx-rx routines. This way whether we are master or slave, we will be able
; to bit phase. This sure helps, however some  TNC's are not as forgiving
; and we get RQ'ed when sending blocks. We still have not found a good
; solution to this problem.
;
; Each character has seven bits and each of the seven bits is 10ms
; in long. Thus it takes ten timer ticks to advance over a
; character bit - 70 ticks to advance over a character and 210 ticks
; to advance over a block of three characters. 
; At each of these timer ticks (when phasing is enabled, i.e. expecting
; received data) the input data is inspected for a new transition. 
; If the transition is too far away from the expected transition (the 
; end of the ten ms interval, which we want to be middle of the bit 
; time), we want to adjust the clock a bit such that the transition will 
; occur at a better place.
; By weighting the time of samplings relative to the start of the
; bit time, (variable PHASE) starting with -5 and increasing
; by 1 till 5, the higher weighted values will amount to larger corrections
; being made near the start and ends of the bit time sampling.
; A cumulative tally is kept over the received characters to keep
; track of how things are going. If the drift is showing a ever
; larger accumulative positive number, the transition occurs too
; close to the end of the bit sampling interval and the clock is held back
; a little so that more signal is allowed to pass by before the next sample.
; If the accumulated error is growing negative, it means that the transition
; occurs too close to the start of the bit sampling time and the clock
; is advanced slightly to move the transition closer to the middle of the 
; bit sampling time.
; The magnitude of the accumulated error when transitions occur is off 
; course relative. For this implenetation the values +/- LIM is used to 
; trigger the clock adjustments.
; It is reasoned that in cases where noise triggers a transition, there
; should be an equal number of noise transitions on either side of the 
; bit sampling centre, i.e. and equal amount added and subtracted into
; the accumulated error, so things should even out.
;
; The routines and variables involved in the bit phasing are:
; "RCV"     set PHASE to -5,
; "INTRP"   increments PHASE, forms the accum. error ACPH and
;           also does the adjustments to the clock.
;  MASTER    controls whether you have master or slave phasing.
;  BFLAG    controls whether bit phasing is on or off.
;
;----------------------------------------------------------------------------
	EVEN
	PUBLIC _INTRP
_INTRP:  
	cmp	byte ptr cs:INT_BUSY,0  ; since we are not re-entrant
	jz	norm_int	     ; check whether were busy

	push	ax		; save ax
	sti			; re-enable interrupts
	jmp	int_ret

norm_int: 
	mov	byte ptr cs:INT_BUSY,1 ; say were busy now to lock out
				; further interrupts
	sti			; Arm interrupts again

	push    ax              ; save registers
        push    bx
        push    cx              
        push    dx              
        push    ds

	mov    ax,cs		; let ds point to cs for data access
	mov    ds,ax
     
        mov     ax,sp           ; save user stack pointer
        mov     USRSP_OFF,ax
        mov     ax,ss
        mov     USRSP_SEG,ax

        mov     ax,AMTSP        ; retrieve AMTOR stack pointer
        mov     sp,ax
        mov     ax,cs           ; SS
        mov     ss,ax

        call    TICK            ; bump the clock

int_t0:	mov	al,byte ptr HIST	; maintain bit history
	mov	byte ptr HIST_1,al	; -1 ms bit

        ; read data from serial port into evm_in
	mov	dx,cs:com_port		; get data byte
	in	al,dx	
	mov	cs:evm_in,al	; most recent data value from EVM
	mov	ah,al
	mov	al,00		; now it is a 16-bit signed value in ax

;------Process recovered data from DSP----------------------------

; Get signal output signal level, separate mark/space
; and update appropiate peak levels
;

	or	ax,ax
	jns	PSSa_a
; ------------------ Its a space signal
	neg	ax		; convert into positive number
	cmp	ax,cs:AD_space_data  ; new peak value?
	jmp	SHORT PSS0_1	;TMP
	jns	SHORT PSS0_1
	mov	ax,cs:AD_space_data
	sub	ax,DECAY	; decay factor
PSS0_1:	mov	cs:AD_space_data,ax
	mov	al,080h
	jmp	SHORT PSSa_out

; ------------------Its a mark signal
PSSa_a:	
	cmp	ax,cs:AD_mark_data  ; new peak value?
	jmp	SHORT PSS0_3	;TMP
	jns	SHORT PSS0_3
	mov	ax,cs:AD_mark_data
	sub	ax,DECAY	    ; decay factor
PSS0_3:	mov	cs:AD_mark_data,ax
	mov	al,0

PSSa_out:
	xor	al,byte ptr cs:MASKIN ; set bit polarity
	mov	cs:data_bit,al
;---------------------------------------------------------------

        ; TODO fix this
        and     al,080h         ; mask to get to data 0/1 bit
	mov	HIST,al

        cmp     byte ptr RAFG,0 ; test if bit phasing active
        je      int8            ; skip bitphasing if not active

        inc     word ptr PHASE  ; its active, bump bitphase

        cmp     al,LINP         ; is it the same as 1ms ago ?
        je      int8            ; skip if yes (no transition)
        mov     LINP,al         ; save new RX data level
                                ; proceed only if there was a transition
        mov     ax,word ptr PHASE ; get bitphase of this transition
        add     ax,word ptr ACPH  ; sum into phase error total  (16 bits)
	mov     ACPH,ax

;;; Master or slave phasing ??
	cmp	byte ptr ORIG,1
        je     slv

; MASTER - attempt to lock on received clocks
; Process bit phasing separately for master and slave
;
        cmp     ax,LIM          ; phase error > 0, test if phase error > LIM
	js      int5            ; skip if less than LIM
	sub     word ptr ACPH,ADJ  ; phase error > LIM, subtract ADJ
	mov	bx,CTRX         ; master, move CTRX late
	inc	bx		;         by 1ms
        jmp     SHORT int6	; and go put it back

int5:   cmp	ax,-LIM		; test if total phase error is less than -LIM
        jns	int8            ; skip if not     
	add     word ptr ACPH,ADJ ; else add ADJ
        mov	bx,CTRX		; move CTRX early ..
	cmp	bx,0		; don't adjust for past zero
	je	int6
  	dec	bx		; .. by	1 ms ..
int6:	mov	CTRX,bx		; and put it back
	mov	BKRX,bx		; .. same for BKRX	
int8:   jmp     SHORT TPR

; SLAVE - note that BKRX is NEVER modified during slave .
; Slave bit phasing (also used for mode B and L)
;

slv:    cmp     ax,LIMS         ; phase error > 0, test if phase error > LIM
	js      slint5          ; skip if less than LIM
	sub     word ptr ACPH,ADJS  ; phase error > LIM, subtract ADJ
        mov	bx,CLOCK	; if slave ..
        dec     bx              ;    .. then put clock back 1ms
        mov	CLOCK,bx
        jmp	SHORT slint8

slint5: cmp	ax,-LIMS	; test if total phase error is less than -LIM
        jns	slint8          ; skip if not     
	add     word ptr ACPH,ADJS ; else add ADJ
        call	TICK		; if slave, put clock forward by 1ms
slint8:

; Transmit data handler
TPR:
				; AMTOR
	mov	al,TXDTA        ; get TX data character from user
	or	al,al		; SET UP FLAGS
	jnz	SHORT TPR1
    	jmp	TPT		; exit if no data from user

; For Amtor transmissions stuff data in small transmit buffer
TPR1:	mov	bx,KIPTR
	call	NXTK		; find next free location in TX buffer		
	cmp	bx,KOPTR	; test if TX buffer is full
	je	TPT		; exit if it is (cant accept data from user)
	and     al,01Fh		; mask TXDATA to ITA2 bits only
    	mov	[bx],al		; put in next TX buffer location
	mov	KIPTR,bx	; save this new location
	mov	byte ptr TXDTA,0 ; acknowledge to user that data was taken

TPT:	cmp     byte ptr RXDTA,0 ; see if user has taken last character
        jne     TPTa            ; exit if not (maybe rx buffer was full) 
        mov     bx,POPTR        ; test is no data in buffer
        cmp     bx,PIPTR
        je      TPT0            ; and exit if its empty
        call    NXTP            ; move to next char to give to user
        mov     al,[bx]         ; get it
        mov     POPTR,bx        ; save its location for next
        mov     RXDTA,al        ; Save character in RXDTA
;
; This section for servicing buffer input and output traffic
;				; As a time saving measure,
TPTa:   mov     al,RXDTA     	;   check if a character is available
        or      al,al           ; else skip if none
	jz	TPT0
	call	EXERX		; process data received through AMTOR
				; As a time-saving measure,
TPT0:	mov	bx,tx_take_from_ptr ; is there anything to tx ?
	cmp	bx,tx_store_at_ptr
	je	TPT1		; skip out if empty
	call	EXETX		; process data to be transmitted by AMTOR

TPT1:

;----------------------------------------------------------------------------
; Control Commands 							    
;                                                                           
; CTRL Q 01=FORCE BREAK                                                     
;      R 02=CLOSE DOWN QRT                                                  
;      S 03=ARQ (MODE A)                                                    
;      T 04=FEC (MODE B)                                                    
;      U 05=LISTEN                                                          
;      V 06=STBY                                                            
;      W 07=TEST (outputs 1kHz signal)
;----------------------------------------------------------------------------
; Implementation of the control call is a little awkward, could be
; done using a call proc_table[si]
; but will work out ok like this too.
; NOTE: not all "CALL's" return, however those that does MUST drop
;       out to EXIT.
;----------------------------------------------------------------------------
CMD:    mov     al,CNTRL        ; get command byte from user
	or	al,al		; SET UP FLAGS
        jz      EXIT            ; exit if no command
        mov     byte ptr CNTRL,0  ; else acknowledge that command has been executed
        cmp     al,01h
        je      CMD01   
        cmp     al,02h
        je      CMD02   
        cmp     al,03h
        je      CMD03   
        cmp     al,04h
        je      CMD04   
        cmp     al,05h
        je      CMD05   
        cmp     al,06h
        je      CMD06
        cmp     al,07h
        je      CMD07
        jmp     SHORT EXIT   
	EVEN
CMD01:  call    TXMIT           ; set TX REQUEST flag
        jmp     SHORT EXIT
	EVEN
CMD02:  call    QRT             ; set QRT REQUEST flag
        jmp     SHORT EXIT
	EVEN
CMD03:  call    CALLA           ; Initiate ARQ
        jmp     SHORT EXIT
	EVEN
CMD04:  call    CALLB           ; Initiate FEC
        jmp     SHORT EXIT
	EVEN
CMD05:  call    LISN            ; Initiate ARQ listen
        jmp     SHORT EXIT
	EVEN
CMD06:  call    STBY            ; Force STBY
        jmp     SHORT EXIT
	EVEN
CMD07:  call    TESTR           ; Enter the test procedure
        jmp     SHORT EXIT

; THIS RTS IS EFFECTIVELY THE RTS TO THE "DROP" SUBROUTINE.
EXIT:   mov     ax,AMTAX        ; all done, recover AMTOR program
        mov     bx,AMTBX     
	mov	cx,AMTCX
        ret                     ; and return to AMTOR program

; SUBROUTINE TO MOVE IDENT OR CALL TO SELCAL BLOCKS
MVSEL:  mov     TEMP2,bx
        mov     al,[bx]         ; get 1st character of 4 char group
        call    STRAN           ; translate ID to AMTOR code
        mov     SLCAL,al        ; put in SELCAL 1st char.
        mov     al,[bx+1]       ; ditto 2nd char.
        call    STRAN
        mov     SLCAL+2,al      ; .. into 1st block 3rd char.
        mov     al,[bx+2]       ; .. 3rd char
        call    STRAN
        mov     SLCAL+3,al      ; .. into 2nd block 1st position
        mov     al,[bx+3]
        call    STRAN
        mov     SLCAL+4,al      ; .. 4th into 2nd block 2nd position
        mov     al,66h          ; put RQ's in it ...
        mov     SLCAL+1,al      ; 1st block 2nd position
        mov     SLCAL+5,al      ; 1st block 2nd position
        ret

; SUBROUTINE TO TRANSLATE SELCAL CHARACTER TO AMTOR
STRAN:  add     al,080h         ; offset to TX table
        call    RCV4            ; translate to AMTOR
        mov     bx,TEMP2        ; recover IDENT/CALL address
        ret

; THIS "SUBROUTINE" ACTUALLY RETURNS CONTROL TO THE USER WHILE
; WAITING FOR THE NEXT INTERRUPT FROM THE 1KHZ TIMER
	EVEN
DROP:   
	mov     AMTAX,ax        ; save amtor program register
        mov     AMTBX,bx
        mov     AMTCX,cx

        mov     ax,sp           ; Save AMTOR stack pointer
        mov     AMTSP,ax

        mov     ax,USRSP_OFF    ; Restore user stack pointer
        mov     sp,ax
        mov     ax,USRSP_SEG
        mov     ss,ax

	mov	byte ptr INT_BUSY,0   ; re-enable interrupt entry

        pop     ds              ; restore registers
        pop     dx
        pop     cx
        pop     bx

int_ret: 

        ;put value in evm_out to serial port
	push	dx
	mov	al,cs:evm_out
	mov	dx,cs:com_port		; get data byte
        out	dx,al
	pop	dx

	mov     al,20h
        out     020H,al          ; send EOI and return
	cmp	cs:irq,08	 ; if IRQ >=8 set EOI to second pic too
	js	int_out
      	out     0A0H,al          ; send EOI and return
int_out:
	pop     ax
	
        iret                   
            
 
;----------------------------------------------------------------------------
; SET TX REQUEST FLAG
TXMIT:  mov     ah,1
        mov     TRQFG,ah
        ret

; SET QRT REQUEST FLAG
QRT:    mov     ah,1
        mov     QFLG,ah
        ret

;----------------------------------------------------------------------------
; System calibration routine
;----------------------------------------------------------------------------
TESTR:  lea     bx,AM_STACK 	; reset AMTOR stack pointer
        mov     sp,bx           ; SP
        mov     ax,cs           ; SS
        mov     ss,ax

TSTPA:  mov     al,055h         ; alternate ones and zeroes
TSTPB:  mov     tst_pat,al      ;
TSTP0:  mov     STSTS,al
        mov     bx,7            ; seven bits in pattern
TSTP00: shr     al,1            ; shift into carry
        call	TXBIT
        call    DROP            ; and wait one ms bit time
        dec     bx              ; count down
        jz      TSTP3           ; skip out if finished
        jmp     TSTP00          ; and go back and do it
        
TSTP3:  mov     al,tst_pat
        cmp     al,055h
        jne     TSTPA 
        mov     al,02Ah
        jmp     TSTPB

;---------------------------------------------------------------------------
; Process data received through AMTOR
; Data characters are stored in a circular buffer
;---------------------------------------------------------------------------
	EVEN
EXERX:	
	mov	bx,rx_store_at_ptr 	    ; is there space in rx buffer ?
	cmp	bx,offset RXBUFR+RXBUFSIZ-1 ; check wrap round case
	jne 	ex00
	cmp	rx_take_from_ptr,offset RXBUFR	; handle wrap around case
	jne 	ex01			; ok there is space left
	jmp	SHORT ex05   		; exit if no space left
ex00:	inc	bx			; normal case
	cmp	bx,rx_take_from_ptr
	je	ex05			; exit if no space available

ex01:	call 	_get_rxdta		; get data if available
        or      al,al                   ; set flags
	jz	ex05			; exit if no data available
	mov	bx,rx_store_at_ptr	; else save it
	mov	byte ptr [bx],al	
	inc	bx			; update buffer pointer
	cmp	bx,offset RXBUFR+RXBUFSIZ
	jne	ex02
	lea	bx,RXBUFR
ex02:	mov	rx_store_at_ptr,bx

ex05:	ret
;---------------------------------------------------------------------------
; Process serial data received from host.
; Data characters are stored in 960 deep circular buffer
; Routine assumes that user has checked that data is available
;---------------------------------------------------------------------------
	EVEN
EXETX:					; do lookahead
	mov	al,byte ptr[bx]		; get character
	xor	ah,ah                   ; and prepare to
        mov     param,ax                ; output via output_txdta
	call	_output_txdta
	or	ax,ax	  		; set flags
	jnz	etx02			; try again 
	inc	bx			; update pointers
	cmp	bx,offset TXBUFR+960		
	jne	etx01
	lea	bx,TXBUFR
etx01: 	mov	tx_take_from_ptr,bx
etx02:	ret 

AMTOR	endp
;----------------------------------------------------------------------------
; Routine to convert ASCII to ITA2 (BAUDOT)
;----------------------------------------------------------------------------
	EVEN
translate_AB  proc  near
        push    bx
        push    cx
        lea     bx,ITASC        ; base of translate table
	push	ax
	xor	ah,ah
	and	al,07Fh
        add     bx,ax           ; compute index	into ASCII to ITA2 table
	pop	ax
	mov     al,cs:[bx]
        and     al,01Fh	 	; mask off extra bits
        pop     cx
        pop     bx
        ret
translate_AB  endp

;----------------------------------------------------------------------------
; Routine to load SELCAL and IDENTIFICATION
;----------------------------------------------------------------------------
_set_SELCAL PROC near
        lea     di,SELCAL               ; point to destination (SELCAL)
        mov     al,byte ptr param+1     ; get next character
        call    translate_AB            ; translate to BAUDOT
        mov     cs:[di],al              ; store the character
        inc     di
        mov     al,byte ptr param  
        call    translate_AB       
        mov     cs:[di],al         
        inc     di
        mov     al,byte ptr param2+1
        call    translate_AB            
        mov     cs:[di],al              
        inc     di
        mov     al,byte ptr param2     
        call    translate_AB            
        mov     cs:[di],al              
        ret
_set_SELCAL ENDP

_set_IDENT PROC near
        lea     di,IDENT                ; point to destination (IDENT)
        mov     al,byte ptr param+1     ; get next character
        call    translate_AB            ; translate to BAUDOT
        mov     cs:[di],al              ; store the character
        inc     di
        mov     al,byte ptr param  
        call    translate_AB       
        mov     cs:[di],al         
        inc     di
        mov     al,byte ptr param2+1
        call    translate_AB            
        mov     cs:[di],al              
        inc     di
        mov     al,byte ptr param2     
        call    translate_AB            
        mov     cs:[di],al              
        ret
_set_IDENT ENDP

;----------------------------------------------------------------------------
; Routine to retrieve received ITA2 character (RXDTA) from AMTOR
; ASCII characters are returned to caller, if available.
;----------------------------------------------------------------------------
	EVEN
_get_rxdta PROC near

        mov     al,cs:RXDTA     	; check if a character is available
        or      al,al           	; exit if its zero
	jnz	ERX0
        jmp     ERX03   
        
ERX0:   mov     byte ptr cs:RXDTA, 0  	; clear to acknowledge

	cmp	al,SPEC_SYM		; special placeholder used for
	jne	ERX00			; mode L
	mov	al,'_'			; replace by special mark '_'
	jmp	SHORT ERX03

ERX00:  cmp	byte ptr RAW_CODE,0	; Raw mode enabled ?
	jz	ERX00A
	jmp	SHORT ERX03		; exit with ITA2 code in al

ERX00A:	and     al,01Fh         	; mask out extra bits

        xor     ah,ah
        cmp     al,01Fh         	; LTRS ?
        je      ERX01
        cmp     al,01Bh         	; FIGS ?
        jne     ERX02
        
        mov     ah,32           	; its FIGS, reset FIGS shift 
ERX01:  mov     cs:RSHFT,ah     	; if LTRS ah=0, if FIGS ah=32
        xor     al,al           	; clear as caller should not try
        jmp     SHORT ERX03     	; to print it

ERX02:	add	al,byte ptr cs:RSHFT   	; add LTR or FIG table offset 
                                       	; to character index  	
	xor	ah,ah			; extend to 16 bits
	lea     bx,cs:EXTAB     	; Table address
	add	bx,ax
        mov     al,byte ptr cs:[bx]     ; Get ASCII character
	cmp	al,0
	jz	ERX03			; ignore undefs
	and	al,07Fh			; mask for 7-bit characters

	cmp	byte ptr case_flag,0 	; upper/lower case enabled ?
	jnz	ERX02A			; skip over if yes

	cmp	al,0 			; when U/L is disabled, let ASCII 
	jnz	ERX03			; nulls pass thru, else 
	mov	al,080h
	jz	ERX03			; skip if not enabled

ERX02A:					; handle upper/lowercase translations
	cmp	al,0			; is it a null ?
	je	ERX_sw_case 		; skip out it is

	cmp	rx_ascii_case,0		; is this a upper case character ?
	je  	ERX_case_reset		; skip out if it is

					; dont force lower case on anything
	cmp	al,'A'			; below the letter 'A'
	js	SHORT	ERX_case_reset

	or	al,020h			; for lower case, force lower case bit
	jmp	SHORT	ERX_case_reset

ERX_sw_case: xor byte ptr rx_ascii_case,01h ; switch case
	or	al,080h			; pass thru ASCII nulls
	jmp	SHORT ERX03

ERX_case_reset:		 		; reset case to upper case
	cmp	al,07h		     	; if not (letter or ascii 21h-3F)
	je	ERX03
	cmp	al,3Fh 			; do not reset case on "?"
	je	ERX03
	cmp	al,021h
	jl	ERX03			; al < 21h
	cmp	al,03Fh
	jg	ERX03		      	; al > 3F
	mov	byte ptr rx_ascii_case,0   ; reset to upper case
	
ERX03:  xor     ah,ah           ; extend to 16 bits

        ret
_get_rxdta ENDP

;----------------------------------------------------------------------------
; Routine to force a ltrs shift
;----------------------------------------------------------------------------
_force_ltrs PROC near
        mov     cs:RSHFT,0      	; force LTRS
	mov	cs:rx_ascii_case,0	; and uppercase
        ret
_force_ltrs ENDP

;----------------------------------------------------------------------------
; Routine to send ITA2 data to AMTOR
; On return, nonzero indicates that character was not accepted
;----------------------------------------------------------------------------
	EVEN
_output_txdta PROC near
	mov	al,07Fh			; assume busy state
	cmp	byte ptr cs:TX_STATE,0 	; Bypass when in receive mode
	je	TETX07
        mov     al,cs:TXDTA             ; is AMTOR module ready to TX ?      
        or      al,al
 	je	ETX0
TETX07: jmp     ETX07                   ; busy, pass by this time

; Amtor/Baudot - some look-ahead processing follows.

ETX0:   mov     ax,cs                   ; point ds to cs for easy data
        mov     ds,ax                   ;     accessing

        mov     ax,param                ; AX = character to be sent
	and	al,07Fh			; remove parity bit

	cmp	al,00h			; check for ASCII null
	jne	ETX00A			; skip out if not

	mov	al,080h			; if ASCII null, send a null 
        mov     cs:TXDTA,al          	; and exit 
	xor	al,al			; acknowledge char was taken 
        jmp     ETX07          

ETX00A: mov     ah,al                   ; get character into ah
	
        mov     al,01Fh                 ; lets say its LTRS
        cmp     ah,0Dh                  ;   -CR-
        jne     ETX01
        mov     byte ptr cs:TSHFT,0     ; a CR clears shift                
                                   ; check if char lies between "!" and "A" 
ETX01:  cmp     ah,21h                  ; i.e. ITA2 (BAUDOT) FIGS shift
        jc      ETX03                   ; this is a control character
        cmp     ah,41h
        jnc     ETX02                   ; yes,  A >char >!
        mov     al,01Bh                 ; else, this is in LTR shift case
ETX02:  cmp     al,cs:TSHFT             ; does this imply a change of shift ?
        je      ETX03                   ; NO
        mov     cs:TSHFT,al             ; Yes, set remember the new shift
        or      al,080h                 ; send the new shift, exit with
        mov     cs:TXDTA,al             ; the busy status so user can try
        jmp     SHORT ETX07             ; again and next time send the
                                        ; actual character

ETX03:  mov     al,ah                   ; get the character to be sent in al
	and	al,07Fh			; 7-bt characters
	cmp	byte ptr case_flag,0 	; upper/lower case enabled ?
	jz	ETX_trans		; skip if not enabled

					; handle upper/lowercase translations
	cmp	al,61h			; is this a lower case character ?
	jc  	ETX_upper		; al < 'a' (uppercase)
					; handle lower case characters
	cmp	byte ptr tx_ascii_case,1 ; do we need to change case ?
	je	ETX_case_reset		; skip if its not
	mov	byte ptr tx_ascii_case,1 ; change case now
	mov	al,0h			; and send a null
        or      al,080h                 ; send the new shift, exit with
        mov     cs:TXDTA,al             ; the busy status so user can try
        jmp     SHORT ETX07             ; again and next time send the
                                        ; actual character

ETX_upper: 
	cmp	al,041h			; separate  special symbols
	jc	ETX_case_reset 		; from true upper case 
					; True upper case characters
	cmp	byte ptr tx_ascii_case,0	
       	je	ETX_case_reset		; skip if its not
	mov	byte ptr tx_ascii_case,0
	mov	al,0h			; send a null 
        or      al,080h                 ; send the new shift, exit with
        mov     cs:TXDTA,al             ; the busy status so user can try
        jmp     SHORT ETX07             ; again and next time send the
                                        ; actual character

ETX_case_reset:		 		; reset case to upper case
	cmp	al,07h		     	; if not (letter or ascii 21h-3F)
	je	ETX_trans
	cmp	al,021h
	jc	ETX_trans
	cmp	al,03Fh
	jnc	ETX_trans
	mov	byte ptr tx_ascii_case,0
	
ETX_trans:   				; translate and send character
        call    translate_AB
        or      al,al                   ; set flags
        je      ETX07                   ; ignore illegal characters
ETX06:	and	al,01Fh			; work with 5 bit codes only
	or      al,080h                 ; send the character
        mov     cs:TXDTA,al
        xor     al,al                   ; acknowledge that char was taken
ETX07:  xor     ah,ah	 		; extend to 16 bit return value

        ret
_output_txdta ENDP

;----------------------------------------------------------------------------
; Function dispatcher
; Function code in AH
;       AH = 0 Initialize
;          = 1 Send character
;          = 2 Receive character
;          = 3 Return status
;          = 4 Inquiry to signature
;          = 5 Drop RTS
;          = 6 Raise RTS
;          = 7 Send break
;          = 8 Non destructive Read
;          = all other codes ignored
;
; For compatibility the following codes are maintained
; These service calls are easier, direct access to PCTOR
;               
;          = 101  __t_speedup
;          = 102
;          = 103  _set_control
;          = 104  _set_txdelay
;          = 105  _set_txmask
;          = 106  _set_rxmask
;          = 107  _get_status
;          = 108  _get_debug
;          = 109  _get_phase_err
;          = 110  _get_debug2
;          = 111  _get_rxdta
;          = 112  _output_txdta
;          = 113  _force_ltrs
;          = 114  _set_SELCAL
;          = 115  _set_IDENT
;          = 116  _get_cases
;          = 117  _close_port
;          = 118
;          = 119  _flush_rxbuffer
;          = 120  _case_translate
;	   = 121  _set_reliability
;	   = 122  _set_raw_mode
;	   = 123  _get_mark_data
;	   = 124  _get_space_data
;	   = 125 
;	   = 126    
;	   = 127  _memARQ  
;	   = 128  _get_RQ_count
;	   = 129  _get_ER1_count
;	   = 130  _get_ER2_count
;	   = 131  _clear_counts
;	   = 132  _set_cd_delay
;	   = 133  _get_RTS

;	   = 202
;	   = 203
;	   = 204
;----------------------------------------------------------------------------
	EVEN
	PUBLIC _new_io_vec
_new_io_vec   proc far

	push	bp			; to access variables
	mov	bp,sp
	mov	ax,CARG2		; get second arg
        mov     cs:param,ax

	mov	ax,CARG3		; get third arg
        mov     cs:param2,ax   

	mov	ax,CARG1		; get first arg

;
; Save registers else C will find its working
; environment all messed up when we return.
;
	PUSHF			; save flags as well
	push	bx
        push    cx              
        push    dx              
        push    si              
        push    di              
        push    ds

        push    cs              ; make data segment the same as
        pop     ds              ; the code seg

;
; Search for function code in jumptable.
; The table is terminated by NULLs.
;
        mov     si,offset jmptab
nextfunc:
        cmp     byte ptr cs:[si],0FFh
        je      nojmp
        cmp     ah,cs:[si]
        je      startjmp
        add     si,3
        jmp     SHORT nextfunc
startjmp:
        add     si,1
        cmp     word ptr cs:[si],0FFFFh
        je      nojmp
;
; Do the jump  to the selected function.
;
        call    word ptr cs:[si]

nojmp:
;
; Return form INT
; Restore all registers except AX are left alone as it is
; used to pass the return data.
;
        pop     ds             
        pop     di             
        pop     si             
        pop     dx             
        pop     cx
	pop	bx
	POPF			; restore flags	as well
        pop     bp              

	ret
_new_io_vec  ENDP
;----------------------------------------------------------------------------
;  Interrupt serivice call jump table                          
;----------------------------------------------------------------------------
	EVEN
jmptab:
	db     0
	dw      _Init  		; initialize
        db      1
        dw      _Send_char
        db      2
        dw      _Receive_char
        db      3
        dw      _Status
        db      4
        dw      _Inquiry
        db      5
        dw      _Ignore
        db      6
        dw      _Ignore
        db      7    
        dw      _Ignore
        db      8    
        dw      _Non_distr_rec
        db      0Ah    
        dw      _Ignore
        db      0Bh    
        dw      _Ignore

        db      101    
        dw      __t_speedup
;;;     db      102  			; Not used anymore         
;;;     dw
	db      103
        dw      _set_control
        db      104
        dw      _set_txdelay
        db      105 
        dw      _set_txmask
        db      106
        dw      _set_rxmask
        db      107
        dw      _get_status
        db      108
        dw      _get_phase_err
        db      109
        dw      _get_debug1
        db      110
        dw      _get_BKRX
        db      111
        dw      _get_rxdta
        db      112
        dw      _output_txdta
        db      113
        dw      _force_ltrs
        db      114
        dw      _set_SELCAL
        db      115
        dw      _set_IDENT
        db      116
        dw      _get_cases
        db      117
        dw      _close_port
        db      118
	dw	_flush_rxbuffer
	db	120
	dw	_set_UL_case
	db	121
	dw	_set_reliability
	db	122
	dw	_set_raw_mode

	db	123
	dw	_get_mark_data
	db	124
	dw	_get_space_data

	db	127			; was 124 ???
	dw	_mem_ARQ
	db	128
	dw	_get_RQ_count
	db	129
	dw	_get_ER1_count
	db	130
	dw	_get_ER2_count
	db	131
	dw	_clear_counts
	db	132
	dw	_set_cd_delay
	db	133
	dw	_get_RTS
	db	134
	dw	_get_MS
 	

	db	200    			; debugging hooks
	dw	_get_debug2
 
        db      0FFh                    ; Table end
        dw      0FFFFh
;----------------------------------------------------------------------------
_Ignore: xor    ax,ax
        ret
;----------------------------------------------------------------------------
; Function code 0 - Inititalize
;----------------------------------------------------------------------------
_Init PROC near                   	
        
        cli                             ; stop interrupts
                                        ; Return code returned on init
        mov     al,030h                 ; modem status: DSR,CTS
        mov     ah,060h                 ; line  status: tx_hold_reg=1,
                                        ;        tx_shft_reg_empty=1,
                                        ;             rcv data rdy=0
	sti                             ; re-enable interrupts 
        ret
_Init ENDP
;----------------------------------------------------------------------------
; Function code 1 - Send character
; AH = 1, AL = character
; Return AH = line status, AL = modem status
;----------------------------------------------------------------------------
; This routine can be in one of four states
; State 1: Control A - awaiting 1st selcal char                   VALUE:01 
;                    - awaiting 2nd selcal char                         02
;                    - awaiting 3rd selcal char                         03
;                    - awaiting 4th selcal char                         04
; State 2: ESC       - awaiting 1st (COMMAND) char in esc seq.    VALUE:10
;                    - awaiting 1st id char following ESC               11
;                    - awaiting 2nd id char following ESC               12
;                    - awaiting 3rd id char following ESC               13
;                    - awaiting 4th id char following ESC               14
;                                                                      
; State 3: Normal    - ready to process and interpret characters  VALUE:00
;                      either as data or controls
;----------------------------------------------------------------------------
	EVEN
_Send_char:
                                        ; Some control commands
                                        ; should be acted upon immediatly

        cmp     al,03h                  ; Break in
        je      j03
        cmp     al,04h                  ; QRT
        je      j04
        cmp     al,06h                  ; Listen mode
        je      j06
        cmp     al,05h                  ; *return status byte*
        jne     send000
        jmp     cntrl05

j03:    jmp     cntrl03
j04:    jmp     cntrl04
j06:    jmp     cntrl06

send000: cmp     byte ptr systate,0     ; Regular input mode ?
        je      send00

        cmp     byte ptr systate,04     ; STATE <= 4 ?
        jg      send_esc                ; skip if not
        jmp     sendSLCL                ; process selcal codes
                                        ; before activating ARQ CALL
send_esc: cmp     byte ptr systate,01Fh ; STATE <= 1F ?
        jg      send00                  ; skip if not
        jmp     sendESC                 ; process escape sequences

; Regular input mode                 
;
send00: cmp     al,01Bh                 ; ESC ?
        jne     send00a                 ; skip if not
        mov     byte ptr systate,10h    ; else ready for 1st char after esc
                                        ; Put signals in orderly state
        ; Turn off TX
	mov	al,EVM_RX
	mov	cs:evm_out,al

        mov     byte ptr CNTRL,06h      ; go to idle mode

	cli	 ; START OF CRITICAL REGION ******************************			
	lea	bx,RXBUFR		; Initialize rx buffer
	mov	rx_store_at_ptr,bx
	mov	rx_take_from_ptr,bx
	sti	 ; END OF CRITICAL REGION   *****************************
        jmp     SHORT send06

send00a: cmp     al,01                  ; Initiate ARQ call
        je      cntrl01
        cmp     al,02h                  ; Initiate FEC
        je      cntrl02
        cmp     al,06h                  ; Listen mode
        je      cntrl06
        cmp     al,07h                  ; Send bell 
        je      cntrl07
        cmp     al,018h                 ; Flush buffers
        je      cntrl18
        cmp     al,07Fh                 ; Force LTRS
	jne	send01
        jmp     cntrl7F
                                        ; else put character in buffer
 					; Check the space situation
send01:	cli	; CRITICAL REGION #####################################
	mov	bx,tx_store_at_ptr	; store character
	mov	byte ptr [bx],al	
	inc	bx			; update buffer pointer
	cmp	bx,offset TXBUFR+960
	jne	send02
	lea	bx,TXBUFR
send02:	mov	tx_store_at_ptr,bx

send05:	sti	; END OF CRITICAL REGION ##########################
send06:	call	_Status
	ret

;
; Handle control commands 01 thru 18,7F
;
cntrl01: mov    byte ptr systate,01     ; Await four selcal characters        
        jmp     SHORT send06       

cntrl02: mov    byte ptr CNTRL,04h      ; Initiate FEC call
        mov    byte ptr systate,0       ; normal input mode now        
        jmp     SHORT send06       

cntrl03: 
	mov	al,STSTS		; ignore when NOT ARQ-RX
	and	al,18h
	cmp	al,10h			; ARQ=1, RX=0
	jne	cntrl03a	
	mov    byte ptr CNTRL,01h       ; Force break
cntrl03a: mov    byte ptr systate,0     ; normal input mode now        
        jmp     SHORT send06       

cntrl04: mov    byte ptr CNTRL,02h      ; QRT
        mov    byte ptr systate,0       ; normal input mode now       
        jmp     SHORT send06       

cntrl05:                                ; Get status byte
        mov     al,STSTS
        and     al,01Fh                 ; clear bits 6,7 
                                        ; since only ARQ and FEC available
        xor     al,010h
        mov     bx,rx_take_from_ptr     ; is there data waiting in
        cmp     bx,rx_store_at_ptr      ;   input buffer ?
        je      cntrl5a                 ; skip if nothing
        or      al,040h                 ; show data waiting (bit 6)
cntrl5a: or      al,080h                ; status byte always has bit 7 set
        mov     status_byte,al          ; save in status byte
        jmp     SHORT send06       
cntrl06: mov    byte ptr CNTRL,05h      ; ARQ listen mode
        mov    byte ptr systate,0       ; normal input mode now #############       
        jmp     send06       
cntrl07: mov     al,07                  ; send BELL
        jmp     send01       
cntrl18:                                ; Flush send buffer
	cli	 ; START OF CRITICAL REGION ******************************			
	lea	bx,TXBUFR		; Initialize 960 byte tx buffer
	mov	tx_store_at_ptr,bx
	mov	tx_take_from_ptr,bx
	sti	 ; END OF CRITICAL REGION   *****************************
        jmp     send06       
cntrl7F: call   _force_ltrs             ; Force a LTR shift on receive
        jmp     send06 

; Build four letter selcal code
;
sendSLCL: cmp     byte ptr systate,01   ; 1st Selcal character
        je      send11    
        cmp     byte ptr systate,02     ; 2nd Selcal character
        je      send12    
        cmp     byte ptr systate,03     ; 3rd Selcal character
        je      send13    
        jmp     SHORT send14            ; 4th Selcal character
send11: mov     byte ptr temp_selcal,al
        inc     byte ptr systate
        jmp     send06
send12: mov     byte ptr temp_selcal+1,al
        inc     byte ptr systate
        jmp     send06
send13: mov     byte ptr temp_selcal+2,al
        inc     byte ptr systate
        jmp     send06
send14: mov     byte ptr temp_selcal+3,al
        mov     byte ptr systate,0      ; reset to normal input mode

        mov     al,temp_selcal          ; move selcal into
        mov     byte ptr param+1,al     ; parameter area
        mov     al,temp_selcal+1
        mov     byte ptr param,al
        mov     al,temp_selcal+2
        mov     byte ptr param2+1,al
        mov     al,temp_selcal+3
        mov     byte ptr param2,al

        call    _set_SELCAL

        mov    byte ptr CNTRL,03h       ; initiate ARQ call
        jmp     send06
send15: 

;
; Build up escape sequence
sendESC:
        cmp     byte ptr systate,010h   ; 1st char after ESC
 	jne	seA
        jmp     sesc1
seA:    cmp     byte ptr systate,011h   ; 1st id char 
	jne	seB
        jmp     sesc2
seB:    cmp     byte ptr systate,012h   ; 2nd id char
	jne	seC
        jmp     sesc3
seC:    cmp     byte ptr systate,013h   ; 3rd id char
	jne	seD
        jmp     sesc4
seD:    cmp     byte ptr systate,014h   ; 4th id char
	jne	seE
        jmp     sesc5
seE:    cmp     byte ptr systate,015h   ; param for L (echo)
	jne	seF
        jmp     sesc6
seF:    mov     byte ptr systate,010h   ; ?? default 
        jmp     send06

sesc1:  cmp     al,'A'                  ; go to STBY
        jne     sesc1a
        mov     byte ptr systate,0
        mov     byte ptr CNTRL,06h       
        jmp     send06
sesc1a: cmp     al,'I'                  ; enter our own id
        jne     sesc1b
        mov     byte ptr systate,011h
        jmp     send06
sesc1b: cmp     al,'L'                  ; turn echo on or off
        jne     sesc1c                  ; L0 = off, L1 = on
        mov     byte ptr systate,015h
        jmp     send06

sesc1c:  cmp     al,'Q'                  ; Send the AMT-1 signature
        jne     sesc1d
	cli	 ; START OF CRITICAL REGION ******************************			
	lea	bx,AMT_1		; AMT-1 signature
	mov	dx,bx
sesA:	mov	bx,dx
	mov	al,[bx]			; get character
	inc	bx
	mov	dx,bx
	cmp	al,0			; skip out if last one
	jz	sesZ	
	mov	bx,rx_store_at_ptr 	    ; is there space in rx buffer ?
	cmp	bx,offset RXBUFR+RXBUFSIZ-1 ; check wrap round case
	jne 	sesB
	cmp	rx_take_from_ptr,offset RXBUFR	; handle wrap around case
	jne 	sesC			; ok there is space left
	jmp	SHORT sesZ   		; exit if no space left
sesB:	inc	bx			; normal case
	cmp	bx,rx_take_from_ptr
	je	sesZ			; exit if no space available
sesC:  	mov	bx,rx_store_at_ptr	; else save it
	mov	byte ptr [bx],al	
	inc	bx			; update buffer pointer
	cmp	bx,offset RXBUFR+RXBUFSIZ
	jne	sesD
	lea	bx,RXBUFR
sesD:	mov	rx_store_at_ptr,bx
	jmp	SHORT sesA  		; do next character
sesZ:	sti	 ; END OF CRITICAL REGION   *****************************
	mov     byte ptr systate,010h	; still in escape state                  
        jmp     send06

sesc1d: mov     byte ptr systate,010h   ; we ignore these, still looking
        jmp     send06                  ; for a valid ESC control char
        
sesc2:  mov     byte ptr temp_selcal,al ; 1st id char
        inc     byte ptr systate
        jmp     send06
sesc3:  mov     byte ptr temp_selcal+1,al ; 2nd id char
        inc     byte ptr systate
        jmp     send06
sesc4:  mov     byte ptr temp_selcal+2,al ; 3rd id char
        inc     byte ptr systate
        jmp     send06
sesc5:  mov     byte ptr temp_selcal+3,al ; 4th id char
        mov     byte ptr systate,010h     ; revert to looking
                                          ; for next escape chars

        mov     al,temp_selcal          ; move id into
        mov     byte ptr param+1,al     ; parameter area
        mov     al,temp_selcal+1
        mov     byte ptr param,al
        mov     al,temp_selcal+2
        mov     byte ptr param2+1,al
        mov     al,temp_selcal+3
        mov     byte ptr param2,al

        call    _set_IDENT
        jmp     send06

sesc6:  cmp     al,'0'                  ; echo off ?
        jne     sesc6a                  ; skip if not
        mov     byte ptr ECHO,0         ; turn echo off
        jmp     SHORT sesc6d
sesc6a: cmp     al,'1'                  ; echo on ?
        jne      sesc6d                 ; ignore invalid values
sesc6b: mov     byte ptr ECHO,1         ; turn echo on
sesc6d: mov     byte ptr systate,010h                  
        jmp     send06                                           

;----------------------------------------------------------------------------
; Function code 2 - Receive character
; Return AH = composite LSR, AL = Character
;----------------------------------------------------------------------------
	EVEN
_Receive_char:
        call    _Status                 ; to set AH

        cmp     status_byte,0           ; pending status byte request ?
        jz      rec0                    ; skip if none
        mov     al, status_byte         ; pick up status byte and return
        mov     byte ptr status_byte,0  ; acknoledge we have it was read
        jmp     SHORT rec_ex

rec0:   cli	 ; START OF CRITICAL REGION ******************************			
	mov	bx,rx_take_from_ptr
	mov	al,byte ptr [bx]	; get the character
	and	al,07Fh			; use 7-bit codes
        cmp     bx,rx_store_at_ptr      ; dont point past end of buffer
        je      rec_nxt
	inc	bx			; update pointer
	cmp	bx,offset RXBUFR+RXBUFSIZ	; test for wraparound
	jne	rec_nxt
	lea	bx,RXBUFR
rec_nxt: mov	rx_take_from_ptr,bx
	sti	 ; END OF CRITICAL REGION   *****************************
rec_ex: ret
;----------------------------------------------------------------------------
; Function code 3 - Get COM Status
; Return AH bit 0 = 1 data waiting
;             5,6 = 1 tx buffer ready to accept data
;----------------------------------------------------------------------------
	EVEN
_Status:   
;
; Set up the line status part
;
        cmp     byte ptr status_byte,0
        jnz     stat00
	cli	; CRITICAL REGION ***************************************
        mov     bx,rx_take_from_ptr ; check if a character is available
        cmp     bx,rx_store_at_ptr
	sti	; END OF CRITICAL REGION ********************************
        je      stat0           
stat00: mov     ah,01h          ; show rx data waiting
        jmp     SHORT stat1
stat0:  xor     ah,ah           ; show rx buffer is empty

stat1:      			; is there tx buffer space available ?      
	cli	; CRITICAL REGION ***************************************
        mov     bx,tx_store_at_ptr      ; buffer space available ?
        cmp     bx,offset TXBUFR+959    ; check for wrap round case
	jne     stat02
	cmp	tx_take_from_ptr,offset TXBUFR
	je  	stat04		        ; no more space
	jmp	SHORT stat03	        ; space ok

stat02:	inc 	bx		        ; normal case
	cmp 	bx,tx_take_from_ptr
	je	stat04
stat03:	or	ah,060h		        ; show space available
				        ; Set up modem status part
stat04:	mov     al,030h                 ; simulate DSR, CTS set
	sti	; END OF CRITICAL REGION ********************************
        ret
;----------------------------------------------------------------------------
; Function code 4 - Inquiry
; For compatibility with MMBIOS we return AH=AAh, AL55h
;----------------------------------------------------------------------------
	EVEN
_Inquiry: mov   ah,0AAh
        mov     al,055h
        ret
;----------------------------------------------------------------------------
; Function code 8 - Non destructive Receive character
;                   no updating of pointers or flags
; Return AH = composite LSR, AL = Character
;----------------------------------------------------------------------------
	EVEN
_Non_distr_rec:
        call    _Status                 ; set AH

        cmp     status_byte,0           ; pending status byte request ?
        jz      nrec0                   ; skip if none
        mov     al, status_byte         ; pick up status byte and return
        jmp     SHORT nrec_ex

nrec0:  mov	bx,rx_take_from_ptr
	mov	al,byte ptr [bx]	; get the character
	and	al,07Fh			; 7-bit codes
nrec_ex:ret

;----------------------------------------------------------------------------
; Function code 0 - Inititalize
;  _t_speedup()       
; Chains into timer interrupt (INT08)
;----------------------------------------------------------------------------
__t_speedup PROC near                   ; registers saved in int_handler
        
        cli                             ; stop interrupts

        lea     bx,cs:AM_STACK
        mov     cs:AMTSP,bx
        mov     al,06h                  ; start off in STBY mode
        mov     cs:CNTRL,al
        mov     al,30                   ; default TXDELAY is 30ms
        mov     cs:DELAY,al
                                        ; set up some more data items
        xor     ax,ax                   ; clear ax       
	mov	cs:INT_BUSY,al		; clear interrupt busy flag                                  
        mov     cs:TXDTA,al             ; clear TXDATA
	mov	cs:RXDTA,al		;
	mov	cs:TSHFT,al		; reset for LTR shift
	mov	cs:RSHFT,al	      	;
	mov	cs:SHIFT,al	       	;
        mov     cs:systate,al           ; clear input state
        mov     cs:status_byte,al       ; clear status byte
        mov     cs:MASKOUT,al           ; Reset masks
        mov     cs:MASKIN,080H		; 
	mov	cs:case_flag,al		; disable upper/lower case
	mov	cs:RELY,al		; disable higher reliability option
	mov	cs:RAW_CODE,al		; disable raw data mode
	mov	cs:AD_mark_data,ax	; A/D data value 
	mov	cs:AD_space_data,ax	; A/D data value 
	mov	cs:TUN_IND,al		; Disable A/D convertor
	mov	cs:Mem_flg,al		; turn memory ARQ-in-progress off
	mov	cs:MARQ,al		; memory ARQ disabled

	mov	word ptr cs:CD_DELAY,5	; to help with master rx recovery

	mov	word ptr cs:RQ_COUNT,ax	; RQ count
	mov	word ptr cs:ER1_COUNT,ax ; ERROR counts
	mov	word ptr cs:ER2_COUNT,ax ; MARQ error fixup counts

    	mov     cs:ECHO,0               ; set ECHO off (!=0)

	lea	bx,RXBUFR		; Initialize rx buffer
	mov	rx_store_at_ptr,bx
	mov	rx_take_from_ptr,bx
	
	lea	bx,TXBUFR		; Initialize 960 byte tx buffer
	mov	tx_store_at_ptr,bx
	mov	tx_take_from_ptr,bx


	mov	dx,com_port1		; disable any interrupts 
	mov	al,00h			; from occurring
	out	dx,al	
	jmp	$+2	 		; brief delay
	jmp	$+2

	sti                             ; re-enable interrupts 
        ret
__t_speedup ENDP
;----------------------------------------------------------------------------
; Restore all interrupts
_close_port PROC near
;-------------------------Reset keyboard interrupts-------------------------
	MOV	DX,CS:OLDKEY
        MOV    	DS,CS:OLDKEYS
        MOV    	AX,2516H     		; reset 16H
        INT    	21H

        MOV    DX,cs:OLD1BO
        MOV    DS,cs:OLD1BS
        MOV    AX,251BH     		; set or reset 1bH
        INT    21H

	mov	word ptr cs:installmsg,0  ; break the signiture	
					  ; forcing a reload of	driver TSR
        ret
_close_port ENDP

;-------------------------AMTOR CONTROL STUFF--------------------------------
; Routine to set AMTOR control mode
;----------------------------------------------------------------------------
	EVEN
_set_control PROC near
        mov     ax,param                ; control code
        mov     byte ptr cs:CNTRL,al 
        ret
_set_control ENDP

;----------------------------------------------------------------------------
; Routine to set transmitter turn-on delay
;----------------------------------------------------------------------------
_set_txdelay PROC near
        mov     ax,param                ; delay
        mov     DELAY,al 
        ret
_set_txdelay ENDP

;----------------------------------------------------------------------------
; Routine to set control delay (5ms min), i.e. the period following
; reception of the master's block. Hleps with rx recovery problems
;----------------------------------------------------------------------------
	EVEN
_set_cd_delay PROC near
        mov     ax,param                ; control code
        mov     word ptr cs:CD_DELAY,ax 
        ret
_set_cd_delay ENDP

;----------------------------------------------------------------------------
; Routine to set output interface mask bits
;----------------------------------------------------------------------------
_set_txmask PROC near
        mov     ax,param                ; mask
        and     al,03h
        mov     byte ptr cs:MASKOUT,al 
        ret
_set_txmask ENDP


_set_rxmask PROC near
        mov     ax,param                ; mask
        and     al,080h
        mov     MASKIN,al 
        ret
_set_rxmask ENDP

;----------------------------------------------------------------------------
; Routine to return AMTOR status
;----------------------------------------------------------------------------
	EVEN
_get_status PROC near

        mov     al,STSTS
;       and     al,01Fh                 ; clear bits 6,7 
                                        ; since only ARQ and FEC available
        xor     al,010h                 ; invert bit
        xor	ah,ah	   
        ret
_get_status ENDP
	EVEN
_get_BKRX PROC near                     ; return BKRX
	mov     ax,cs:BKRX     
        ret
_get_BKRX ENDP
	EVEN
_get_phase_err PROC near                ; return accumulated phase error
        mov   ax,cs:ACPH  
        ret
_get_phase_err ENDP
	EVEN
_get_cases PROC near			; return status of Ascii cases
	mov     ah,cs:rx_ascii_case    
	mov     al,cs:tx_ascii_case    
        ret
_get_cases ENDP

	EVEN
_get_mark_data PROC near
	mov	ax, cs:AD_mark_data
        ret
_get_mark_data ENDP
	EVEN
_get_space_data PROC near
	mov	ax, cs:AD_space_data
        ret
_get_space_data ENDP

;
; Read status of RTS
;
	EVEN
_get_RTS PROC near 		; RTS status
	mov     ah,0     
        mov     dx,com_port4    ; modem control register
        in      al,dx           ; get present state
        xor     al,cs:MASKOUT  	; get proper sense
	and 	al,02h		; isolate RTS bit (RTS controls transmit)
        ret
_get_RTS ENDP

;
; Read MASTER/SLAVE
;
	EVEN
_get_MS PROC near 		; MASTER/SLAVE status
	mov     ah,0  
	mov	al,byte ptr cs:TX_STATE   
        ret
_get_MS ENDP

;
; Debugging hooks
;
	EVEN
_get_debug1 PROC near 			; CTRX
	mov     ax,cs:CTRX     
        ret
_get_debug1 ENDP

_get_debug2 PROC near			; return CS
	mov     ax,cs     
        ret
_get_debug2 ENDP

;----------------------------------------------------------------------------
; Routine to clear rx buffer.
;----------------------------------------------------------------------------
_flush_rxbuffer	 PROC near
	cli	 ; START OF CRITICAL REGION ******************************			
	lea	bx,RXBUFR		; Initialize rx buffer
	mov	rx_store_at_ptr,bx
	mov	rx_take_from_ptr,bx
	mov	byte ptr RXDTA,0	; clear RXDATA
	sti	 ; END OF CRITICAL REGION   *****************************
	ret
_flush_rxbuffer ENDP

;----------------------------------------------------------------------------
; Routine to set the upper/lower case flag.
;----------------------------------------------------------------------------
_set_UL_case	 PROC near
	mov	ax,param		; get user parameter
	mov	cs:case_flag,al
	mov	byte ptr cs:rx_ascii_case,0  ; Reset to Upper case
	mov	byte ptr cs:tx_ascii_case,0
	ret
_set_UL_case ENDP

;----------------------------------------------------------------------------
; Routine to set higher reliabilty option.
;----------------------------------------------------------------------------
_set_reliability PROC near
	mov	ax,param		; get user parameter
	mov	cs:RELY,al
	ret
_set_reliability ENDP

;----------------------------------------------------------------------------
; Routine to set/clear raw data mode
;----------------------------------------------------------------------------
_set_raw_mode PROC near
	mov	ax,param		; get user parameter
	mov	cs:RAW_CODE,al
	ret
_set_raw_mode ENDP

;----------------------------------------------------------------------------
; Turn memory ARQ on/off
;----------------------------------------------------------------------------
_mem_ARQ PROC near
	mov	ax,param		; get user parameter
	mov	byte ptr cs:MARQ,al
	ret	
_mem_ARQ ENDP

;----------------------------------------------------------------------------
; Routine to determine the processor type
; May be called directly from C program.
; This code snippet by .... (CCS INK BBS)
;----------------------------------------------------------------------------
	PUBLIC _processor_type
_processor_type	proc far
;
;	Returns the processor type in the AX
;

	pushf			; Save FLAG registers
	xor	ax,ax		; Clear AX and push onto the stack.
	push	ax
	popf			;  Pop a zero into FLAGs register
	pushf			;  Attempt to set bit 12-15 to a zero
	pop	ax		;  Recover FLAG word
	and	ax, 0f000h	;  If Bits 12-15 are then the processor
	cmp	ax, 0f000h	;  is an 8018x or an 808x
	jz	is_0_1
 
	mov	ax, 07000h	;  Try to set FLAG bits 12-14 (NT, IOPL)
	push 	ax
	popf			;  put 07000H into flags
	pushf
	pop	ax

	and	ax,07000h	;  if bits 12-14 are cleared then the
	jz	is_80286	;	processor is an 286

is_80386:			;	Else it is a 386 
	mov	ax, 386		;	return 386 in AX
	jmp	p_done
is_80286:
	mov	ax, 286		;  return 286 in AX
	jmp	p_done
is_0_1:				;  It is a 8086 or a 80186
	push 	cx		;  Save CX the only other register used
	mov	ax, 0ffffh	;  Set AX to all 1s
	mov	cl, 33		;  will shift it 33 times if it is an
				;  808x or 1 time if it is an 8018x
	shl	ax, cl		;  if we shift 33 times all bits are
	jnz	is_80186	;  zero. If any bits are on it's an 18x
is_8086:			;  Else we have an 8086	or V20/V30
	pop	cx		;  restore cx
	xor	al,al		;  Sets ZF
	mov	al,40h		;  64 decimal
	mul	al		;  64**2 > 255
	jz	is_v30		;  V20/V30 MUL leaves ZF untouched
				;  8086 ZF==OF after MUL
	mov	ax,86		;  Return 86 in AX
	jmp	p_done
is_v30:
	mov	ax,30		;  Return 30 in AX
	jmp	p_done	

is_80186:
	mov 	ax, 186		;  Return 186 in AX
	pop	cx

p_done:
	mov	word ptr cs:processor_type,ax ; save the type of processor
	popf			; Recover original FLAG register
	ret

_processor_type	endp

;----------------------------------------------------------------------------
; Routines for returning error count statistics
	EVEN
_get_RQ_count PROC near
	mov     ax,cs:RQ_COUNT     
        ret
_get_RQ_count ENDP

_get_ER1_count PROC near
	mov     ax,cs:ER1_COUNT     
        ret
_get_ER1_count ENDP

_get_ER2_count PROC near
	mov     ax,cs:ER2_COUNT     
        ret
_get_ER2_count ENDP

_clear_counts PROC near
	xor	ax,ax
	mov     cs:RQ_COUNT,ax     
	mov     cs:ER1_COUNT,ax     
	mov     cs:ER2_COUNT,ax     
        ret
_clear_counts ENDP


;----------------------------------------------------------------------------;
; Routine to set COM port						     ;
;----------------------------------------------------------------------------;
	EVEN
_set_com_port   proc far
	PUBLIC	_set_com_port

	push	bp			; to access variables
	mov	bp,sp
	mov	ax,CARG1		; get argument
	mov	cs:com_port,ax		; COM
	inc	ax
	mov	cs:com_port1,ax		; COM+1
	inc	ax
	mov	cs:com_port2,ax		; COM+2
	inc	ax
	mov	cs:com_port3,ax		; COM+3
	inc	ax
	mov	cs:com_port4,ax		; COM+4
	inc	ax
	mov	cs:com_port5,ax		; COM+5
	inc	ax
	mov	cs:com_port6,ax		; COM+6
        
        mov     ax,CARG2		; IRQ
        mov     cs:irq,ax

	pop	bp
	ret
_set_com_port   endp


;----------------------------------------------------------------------------
; Keyboard break handlers
; From AL Steven's book. DOS 5: Developer's guide
;----------------------------------------------------------------------------
;storage for old 1BH vector
OLD1BO	DW     ?
OLD1BS  DW     ?
; storage for old int 9 vector
OLDKEY  DW     ?
OLDKEYS DW     ?

BRK	PROC   	FAR
;;        PUSH   	DS
;;        PUSH   	AX
;;        PUSH   	BX


;;        POP    	BX
;;        POP    	AX
;;        POP    	DS
        IRET                ; return -- don't terminate program
BRK     ENDP

; new keyboard (INT 16H) handler
; Here we catch functions 0,1 (and 10H, 11H which are the
; same functions except they are aware of the extended
; keyboard keys (like F11 and F12)
; We never allow a program (including DOS) to see a
; ^C (or ALT-numeric pad 3) or ^2

NEWKEY	PROC   	FAR
        CMP    	AH,10H
        JZ     	PKEY
        CMP    	AH,11H
        JZ     	PKEY2
        CMP    	AH,1
        JZ     	PKEY2
        OR     	AH,AH
; only worry about functions 0 and 1 10H & 11H
        JZ     	PKEY
; continue with old handler (its IRET will return to caller)
        JMP    	DWORD PTR CS:[OLDKEY]

; Handle functions 0,10H
PKEY:   PUSH   	AX
        JMP    	SHORT PK1
; call break routine when ^C detected
PKEY1:  INT    	1BH
PK1:    POP    	AX
        PUSH   	AX
        PUSHF  	             ; simulate old INT 16
        CALL   	DWORD PTR CS:[OLDKEY]
        CMP    	AL,3
        JZ     	PKEY1        ; don't return ^C
        CMP    	AX,0300H     ; or ^2
        JZ     	PKEY1
        ADD    	SP,2         ; remove ax off stack
        IRET

; Handle functions 1,11H
PKEY2:  PUSH   	AX
        JMP    	SHORT PK3
; Call break when ^C detected
PKEY3:  INT    	1BH
PK3:    POP    	AX
        PUSH   	AX
        PUSHF               ; simulate old INT 16
        CALL   	DWORD PTR CS:[OLDKEY]
; no key waiting -- report to caller
        JZ     	PKEY6
        CMP    	AL,3
        JZ     	PKEY4        ; don't return ^C
        CMP    	AX,0300H     ; or ^2
        JNZ    	PKEY5
PKEY4:
; "eat" waiting key (don't care about fn 10H return)
        XOR    	AH,AH
        PUSHF
        CALL   	DWORD PTR CS:[OLDKEY]
        JMP    	PKEY3

PKEY5:
        ADD    	SP,2         ; remove ax off stack
        CMP    	AX,0300H     ; clear zero flag
        RET    	2            ; return with flags

PKEY6:
        ADD    	SP,2         ; remove ax off stack
        CMP    	AX,AX        ; set zero flag
        RET    	2            ; return with flags


NEWKEY  ENDP
;----------------------------------------------------------------------------
; This initialization routine for library usage
	PUBLIC _tor_init
_tor_init proc    far

;
; Save registers else C will find its working
; environment all messed up when we return.
;
	PUSHF			; save flags as well
        push    bx              
        push    cx              
        push    dx              
        push    si              
        push    di              
        push    bp
        push    ds

        mov	bx,cs              ; make data segment the same as
        mov     ds,bx              ; the code seg
;***************************************************************
;*                                                             *
;* File: breaks.asm                                            *
;*                                                             *
;* Example of capturing ^C and preventing DOS's ^C display.    *
;*                                                             *
; From Al Stevens, DOS 5: Developer's guide		       *
;***************************************************************

	MOV    	AX,351BH     		; get INT 1BH vector
        INT    	21H
        MOV    	cs:OLD1BO,BX 		; save it
        MOV    	BX,ES
        MOV    	cs:OLD1BS,BX
        MOV    	AX,3516H     		; and INT 16 too
        INT    	21H
        MOV    	CS:OLDKEY,BX 		; save it
        MOV    	BX,ES
        MOV    	CS:OLDKEYS,BX
        MOV    	DX,OFFSET cs:NEWKEY
        MOV    	AX,2516H
        INT    	21H          		; set INT 16
        MOV    	DX,OFFSET cs:BRK
        MOV    	AX,251BH     		; set or reset 1bH
        INT    	21H

;
; Restore all registers except AX are left alone as it is
; used to pass the return data.
;
        pop     ds             
        pop     bp              
        pop     di             
        pop     si             
        pop     dx             
        pop     cx
        pop     bx
	POPF			; restore flags	as well

	ret
_tor_init   endp
;-----------------------------------------------------------------------------

cseg    ends
       END 

