//----------------------------------------------------------------------------
//    PSATOR - Library of routines for amateur radio AMTOR/PACTOR
//    Copyright (C) 1993-1995 Johan B. Forrer, KC7WW
//
//    This is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// Amtor for the PSA PC sound cards
// History:
// Version	0.01  Derived from PCTOR V3.06a
//		0.02-05 Timing derived from PSA card
//            		Built-in modem control
//		0.06/7  User-defined IRQ/PSSBASE
//		        Speaker/line in levels
//		0.08    TX and RX MASKS		        
//              0.09    Use "Mike in" instead of "Line/Aux in"
//                      New timing parameters 8000 - 16.67 MHz
//                                           10000 - 20.00 MHz parts
//                      3 wait states for Soundport in Monitor
//	                Improvements to tuning indicators
//                      Flash RxR invert as display goblin
//		0.9A    Smaller tuning bars
//		        Selectable signal sources
//		0.10 	AFSK output included
//		0.11    50ns timing accuracy restored
//		0.12    Use dithered DSP clock=0
//	        0.13    Auto card detect, ESC614/5
//----------------------------------------------------------------------------
//#define SHAREWARE 0                     // For shareware versions
#define VERS ".13"
#define SERIAL_NUMBER "  S/N 0000"

#include <stdio.h>
#include <dos.h>
#include <ctype.h> 
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <alloc.h>
#include <string.h>
#include <time.h>
#include <bios.h> 
#include <errno.h>
#include <process.h>   
#include <math.h>

unsigned int _stklen=32000;

#include "amt56def.hpp"                   // defines for extended keyboard
#include "amt56bio.hpp"
#include "amt56glb.hpp"

int directvideo=DEFAULT_VIDEO;

main(int argc, char **argv)
{
// List of prototypes to be called
//
extern void store_keyboard_input(Circ_buffer *, char);
int activate_window( int ); 
char command_mode(Circ_buffer *, Circ_buffer *);
void display_cursor();
void send_wru(Circ_buffer *);
void read_setup_data(FILE *); 
void update_status_line(Circ_buffer *, Circ_buffer *); 
void enter_selcal(void); 
void read_directory(void); 
void buffers_menu(Circ_buffer *);  	// show & select text buffer
void files_menu(Circ_buffer *, Circ_buffer *);	// enter filename for transmission
void loop_menu(Circ_buffer *, Circ_buffer *);	// loop on transmission
int harderror_handler(int,int,int,int); // Harderror handler
void show_help(void);			// Show help
void speedup_AMT(void);			// Hook intvec for AMTOR/BAUDOT/ASCII
extern void check_AMT_BD_ASCII_input(void);  // Check input from AMTOR/BD/ASCII
extern void check_AMT_BD_ASCII_buffs(void);  // Check output from buffers	
extern char check_AMT_BD_ASCII_keyboard(void);	// Check output from keyboard


int  output_status, i;
char  ch, time_string[80],*aptr, cmd_line[20];

FILE *setup;
char  cnf_path[30], *env_ptr;

	directvideo=0;		// This ensure BIOS compatibility
				// until the config file is read

// Process command line parameters
//              Video mode
//                 /B	(BW80)
//                 /C	(C80)
//                 /M	(MONO)
// 
//              Window boundary
//                 /Wxx
//
//              Receive-only
//                 /R
//         
//              Color schemes
//                /SC  -- display characters
//                /SE  -- echoed display characters
//                /SB  -- display background
//                /SM  -- message characters
//                /SA  -- announciator characters

	if(argc==1) {
		window_boundary=DEFAULT_BOUNDARY;
		textmode(C80);
		normvideo();
		}

	for(i=1;i<argc;i++) {
		aptr=argv[i];
		if(*aptr!='/') {
			printf("USAGE: TOR[2] [/B|/C|/M] [/Wxx] [/SC|/SE|/SB|/SM|/SA]\n");
			printf("B=black/white, C=color, M=monochrome, W=split screen lines\n");
			printf("COLORS: SC=Display , SE=Echoed , SB=Background, SM=Message, SA=Announciator\n");
			return 0;
			}
		switch (*(aptr+1)) {
			case 'B': textmode(BW80);
				  c_BGCOLOR=BLACK;
				  c_CHCOLOR=WHITE;
				  c_YELLOW=c_CYAN=WHITE;
				  lowvideo();
				  break;
			case 'C': textmode(C80);
				  c_BGCOLOR=BLUE;
				  c_CHCOLOR=LIGHTGRAY;
				  c_YELLOW=YELLOW;
				  c_CYAN=CYAN;
				  normvideo();
				  break;
			case 'M': textmode(MONO);
				  c_BGCOLOR=BLACK;
				  c_CHCOLOR=WHITE;
				  c_YELLOW=c_CYAN=WHITE;
				  lowvideo();
				  break;
			case 'W': sscanf(aptr+2, "%d", &window_boundary);
				  if((window_boundary<4)||(window_boundary>25))
						window_boundary=DEFAULT_BOUNDARY;
				  break;
			case 'R': Receive_only=1;
				  break;
			case 'S': switch (*(aptr+2)) {
					case 'C': sscanf(aptr+3, "%d", &c_CHCOLOR);
						  break;
					case 'E': sscanf(aptr+3, "%d", &c_ECHOCOLOR);
						  break;
					case 'B': sscanf(aptr+3, "%d", &c_BGCOLOR);
						  break;
					case 'M': sscanf(aptr+3, "%d", &c_CYAN);
						  break;
					case 'A': sscanf(aptr+3, "%d", &c_YELLOW);
						  break;
					default:  printf("Color attribute not recognized <%c>\n", *(aptr+2));
					          return 0;
					}
				   break;
			default:
				printf("USAGE: TOR[2] [/B|/C|/M] [/Wxx] [SC|SE|SB|SM|SA]\n");
				printf("B=black/white, C=color, M=monochrome, W=split screen lines\n");
				printf("COLORS: SC=Display , SE=Echoed , SB=Background, SM=Message, SA=Announciator\n");
				return 0;
			}
		}

//
// Open all files before progressing to far into startup
// 	1) .CNF
//

	strcpy(cnf_path,"");
	env_ptr=getenv("AMTOR56");
	if(env_ptr==NULL) strcpy(cnf_path,".");
	else strcpy(cnf_path,env_ptr);

   	if ((setup=fopen(strcat(cnf_path,"\\amtor56.cnf"), "r"))==NULL)
        	{
        	cprintf("\r\nNo setup file (AMTOR56.CNF) supplied\r\n");
        	cprintf("all parameters set to default values\r\n");
		clreol();
        	}

   	proc_type=processor_type();  	// Find out processor class

	tor_init();			// Hook keyboard interrupt to 
					// intercept ^C

	textbackground(c_BGCOLOR); // Default screen colors
	textcolor(c_CHCOLOR);

#ifdef SHAREWARE	      	//-*******************************************
void shareware_banner(void);
        shareware_banner();
#endif				//-*******************************************

// Initialize the screen
	current_window = 1;
	_setcursortype( _NOCURSOR );
	clrscr();
	gotoxy(1,2); cprintf("ͺ AMTOR56 (c)");
	cprintf("%s",version); 
	cprintf(" ");


	x_ind[0]=x_ind[1]=x_ind[2]=x_ind[3]=x_ind[4]=x_ind[5]=1;
	y_ind[0]=y_ind[1]=y_ind[2]=y_ind[3]=y_ind[4]=y_ind[6]=1;

	entered_space=0;
	Raw_mode=0;
	Mode=STBY;
	qrt_flag=0;
	echo_char=0x7F; prev_echo=0x7F; L_selcal[9]='z';

	tzset();	// set up global time from "tz" env. parameter

// Set up default data, then check for setup file
//     and update according to data supplied there
//
	strcpy(ident,"..../------");
	set_IDENT(ident);
	strcpy(selcal,"..../------");
	set_SELCAL(selcal);
	strcpy(wru_reply,"\r\nQRA ----- .... +?");
	strcpy(logfilename,"AMT56LOG");
	strcpy(Log_program,"");
	strcpy(Packet_program,"");
	strcpy(scrollback,"scroll.tmp");

	current_window = activate_window(2);

	read_setup_data(setup);


// Set up COM port
	switch(COM_prt) {
		case 1:
			combase = 0x3F8;
			irq = 4;
			break;
		case 2:
			combase = 0x2F8;
			irq = 3;
			break;
		case 3:
			combase = 0x3E8;
			irq = 4;
			break;
		case 4:
			combase = 0x2E8;
			irq = 3;
			break;
		}

	set_com_port(combase, irq);

// Read data directory and extract filenames with "tor" extensions
	read_directory();

// Open the log file (note BINARY mode to avoid CR/LF problems)
	if((logfile=fopen(logfilename,"ab"))==NULL) {
		printf("Could not open the Log File: <%s>", logfilename);
		printf("Please hit any key to continue using default\n");
		getch();
		strcpy(logfilename,"AMT56LOG");
		if((logfile=fopen(logfilename,"ab"))==NULL) {
			printf("Could not open the default file: <%s> either", logfilename);
			printf("Using a write-protected disk ?");
			return(0);
			}
		}
//
// Open the scrollback
//
	if((scrollfile=fopen(scrollback,"w+"))==NULL) {
		printf("Could not open the Scrollback File: <%s>\n", scrollback);
		printf("Please hit any key to continue using default\n");
		getch();
		strcpy(scrollback,"scroll.tmp");
		if((scrollfile=fopen(scrollback,"w+"))==NULL) {
			printf("Could not open the default file: <%s> either", logfilename);
			printf("Using a write-protected disk ?");
			return(0);
			}
		}
	
//
// Initialize scrolling function
//
	current=0;
	scroll_ptr=scroll_buffer;	
	fgetpos(scrollfile,scroll_ptr++);

	time(&previous_time);
	time(&previous_sw_time);
	previous_biostime=biostime(0,0L);		
	if(Log_option)
		{
		time(&bintime);	// get time in seconds since 00:00:00 UTC,1/1/70  
		if(Time_Type!=GMT) curtime=localtime(&bintime); // convert to local time	
		else curtime=gmtime(&bintime); 	// UTC
		sprintf(time_string,"---->Started at: %s", asctime(curtime));
		fprintf(logfile,"\r\n\r\n\r\n");
		i=0;
		while (time_string[i]!='\0') {
			if(!((time_string[i]=='\r')||(time_string[i]=='\n')))
				fprintf(logfile,"%c",time_string[i]);				
			i++;
			}
		fprintf(logfile,"\r\n");
		}

//
// Clean up startup displays
//
	delay(1500);
	current_window = activate_window(2);  
	gotoxy(1,1);   
        clrscr(); 
	gotoxy(1,window_boundary-4);
        display_cursor();
	current_window = activate_window(3);
	gotoxy(1,1);
        clrscr(); 
        display_cursor();


//
// Initialize system timer for operation on AMTOR/ASCII/BAUDOT
//
	speedup_AMT();

	harderr(harderror_handler);	// Intercept hard errors


	// Set the mode and baud rate of the COM port
	outp(combase+3, 0x83);  // LCR register
	outp(combase+0, (12L*9600L/(long)combaud) & 0xff);
	outp(combase+1,	(12L*9600L/(long)combaud) >> 8);
	outp(combase+3, 0x03);  // LCR register: 8, n, 1

	// Enable the COM port interrupts
	outp(combase+1, 0x01);


//---------------------------------------------------------------------------		
//=========================MAIN PROGRAM LOOP=================================
//---------------------------------------------------------------------------		

   for (;;) {
// Update status lines
	update_status_line(&Keyboard_buffer, &Special_buffer);  
	sys_stat = get_status();	      // get status data
	RQ_count=get_RQ();
	ER1_count=get_ER1();
	ER2_count=get_ER2();

	check_AMT_BD_ASCII_input();
	check_AMT_BD_ASCII_buffs();
	ch=check_AMT_BD_ASCII_keyboard();
	if(ch==(char)0xFF) break;  	// exit via ALT-X
	if(ch==0x1B) if(command_mode(&Keyboard_buffer,	  // Command menu ?
              		&Special_buffer)==0x1B) break; 
      } /* end for */

      close_down();  		// Closedown BIOS, ports etc

#ifdef SHAREWARE
	shareware_banner();
#endif
	normvideo();
        window(1,1,80,25); clrscr();

	if(Log_option)
		{
		time(&bintime);	  	// final time stamp
		if(Time_Type!=GMT) curtime=localtime(&bintime); 	
		else curtime=gmtime(&bintime); 	// UTC
		sprintf(time_string,"---->Closed at: %s", asctime(curtime));	
		fprintf(logfile,"\r\n");
		i=0;
		while (time_string[i]!='\0') {
			if(!((time_string[i]=='\r')||(time_string[i]=='\n')))
				fprintf(logfile,"%c",time_string[i]);				
			i++;
			}
		if(Stats) fprintf(logfile," R:%d E:%d/%d ",
				RQ_count, ER1_count, ER2_count);
		fprintf(logfile,"\r\n");
		}

	fcloseall(); 		// close all files

#ifdef SHAREWARE
 	printf("\nThanks for using PCTOR\n"); 	
#endif
	return 0;

}
//--------------------------------------------------------------------------
//
// define three windows:
// window 1 = status
// window 2 = output from AMTOR
// window 3 = input from user keyboard
// window 4 = keyboard buffer status
// window 5 = tuning indicator
//
int activate_window( int wind_no)
{
   x_ind[current_window-1] = wherex();   // save current windos's location
   y_ind[current_window-1] = wherey();
   switch( wind_no)                      // switch to the new window      
      {
      case 1:   window( 1,1,80,1);
                break;
      case 2:   window( 1,4,80,window_boundary-1);
                break;
      case 3:   window( 1,window_boundary+1,80,25);  // 24??
                break;
      case 4:   window( 1,window_boundary,80,window_boundary);
                break;
      case 5:   window( 1,2,80,3);
                break;
      case 6:   window( 1,2,40,3);
                break;
      }
// retrieve next location
   gotoxy( x_ind[wind_no-1], y_ind[wind_no-1]);   
   return wind_no;
} 
//--------------------------------------------------------------------------
void display_cursor()
{
int x, y;

        x=wherex();y=wherey();
	cprintf("");     // show cursor
//	cprintf("_");
        gotoxy(x,y);
}
//--------------------------------------------------------------------------
//
// Pop-up window for command dialogs
//
char command_mode(Circ_buffer *kb_ptr, 
		Circ_buffer *sp_ptr)
{

int bufsize, left=10,right=71,top=2,bottom=24, delay_input;
char *tbuffer, ch, c, in_str[80];
int i, x, y;
unsigned int Msb, Lsb;
int baudrate_input;
	
	save_xy(&x,&y);				      // save current window
   	_setcursortype( _NORMALCURSOR );	      // co-ordinates
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf("ͻ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf("                                                            ");
        cprintf(" X -> Main display | Q -> quit prgm ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();
        for(;;)
            {
            cprintf( "Ready > ");
            ch = getche();
            cprintf("\r\n");clreol();

// Check for breaking out of endless loop
	    if(ch==(char)0x1B) ch=(char)'X'; 	// change ESC into an 'X'
            if((ch=='X')||(ch=='x'))  break;  
            if((ch=='Q')||(ch=='q')) 
		{
		ch=(char)0x1B;	     // change so it looks like ESC
		break;
		}   
 
           switch (ch)
               {

               case 'a':  //  Toggle memory ARQ mode
               case 'A': cprintf( "Toggle memory ARQ mode (hard error correction): ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old MARQ option was: ");
			 if(Mem_ARQ) cprintf("ON"); else cprintf("OFF");
                         if(Mem_ARQ==0) Mem_ARQ=1;
                         else           Mem_ARQ=0;
                         cprintf( " now: ");
			 if(Mem_ARQ) cprintf("ON\r\n"); else cprintf("OFF\r\n");
                         clreol();
			 set_mem_ARQ(Mem_ARQ);
                         break;

               case 'b':  // Load new sound output level
               case 'B': cprintf( "Adjust Speaker level (using +/-)(m=mute) ");
		    for(;;) {
                         in_str[0]=getch();
 			 cprintf("\r\n");clreol();
			 delay_input=Sound_out_level;
			 if(in_str[0]=='+') delay_input=Sound_out_level+1;
			 if(in_str[0]=='-') delay_input=Sound_out_level-1;
			 if(!((in_str[0]=='+')||(in_str[0]=='-')||(in_str[0]=='m'))) break;
                         if ((delay_input < 0)||(delay_input > 0x3F))
                           {
                           cprintf( "Range exceeded - Level unchanged\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf( "Old value was: %d",Sound_out_level);
                           Sound_out_level = delay_input;
			   if(in_str[0]=='m')  // Mute ?
			   	Adjust_Sound_output_level(Sound_out_level|0x80);
			   else
			   	Adjust_Sound_output_level(Sound_out_level);
                           cprintf( " now: %d",Sound_out_level);
                           clreol();
                           }
                    	 }
                         break;

               case 'e':  // Load new sound output level
               case 'E': cprintf( "Adjust Line in level (using +/-) ");
		    for(;;) {
                         in_str[0]=getch();
 			 cprintf("\r\n");clreol();
			 delay_input=Sound_in_level;
			 if(in_str[0]=='+') delay_input=Sound_in_level+1;
			 if(in_str[0]=='-') delay_input=Sound_in_level-1;
			 if(!((in_str[0]=='+')||(in_str[0]=='-'))) break;
                         if ((delay_input < 0)||(delay_input > 0xF))
                           {
                           cprintf( "Range exceeded - Level unchanged\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf( "Old value was: %d",Sound_in_level);
                           Sound_in_level = delay_input;
			   Adjust_Sound_input_level(Sound_in_level);
                           cprintf( " now: %d",Sound_in_level);
                           clreol();
                           }
                    	 }
                         break;

               case 'i':  // OWN IDENTFICATION
               case 'I': cprintf( "New ID (selcal/call): ");
                         strcpy(in_str, "");
                         in_str[0]=80;
                         cgets(in_str);
                         cprintf("\r\n");clreol();
			 if(in_str[1]==0)
				{
                           	cprintf("ID is: %s",ident);
                           	cprintf("\r\n");clreol();
				break;
				}
			 for(i=2;i<in_str[1]+2;i++) in_str[i]=toupper(in_str[i]);
                         if (in_str[1] < 4)
                           {
                           cprintf( "Four characters required\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf("ID was: %s",ident);
                           strncpy(ident, &in_str[2], in_str[1]);
                           ident[in_str[1]]='\0';     
                           set_IDENT( ident );
                           cprintf( " now: %s",ident);
                           cprintf("\r\n");
                           clreol();
                           }
                         break;

               case 'c':  //  change upper/lower case
               case 'C': cprintf( "Toggle upper/lower case: ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old U/L was: "); 
			 if(UL_case) cprintf("ON"); else cprintf("OFF");
                         if(UL_case==0) UL_case=1;
                         else UL_case =0;
			 set_UL_case(UL_case);
                         cprintf( " now: ");
			 if(UL_case) cprintf("ON\r\n"); else cprintf("OFF\r\n");
                         clreol();
                         break;

               case 'd':  // TX-DELAY VALUES
               case 'D': cprintf("Change Tx delay (TD) and Control delay (CD):\r\n"); 
			 cprintf( "Enter new TD in ms: ");
                         strcpy(in_str, "");
                         in_str[0]=80;
                         cgets(in_str);
 			 cprintf("\r\n");clreol();
			 if(in_str[1]==0) delay_input=tx_delay;
			 else sscanf(&in_str[2],"%d",&delay_input);
                         if ((delay_input <= 0)||(delay_input > MAX_DELAY))
                           {
                           cprintf( "Invalid TD\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf( "Old TD was: %dms",tx_delay);
                           tx_delay = delay_input;
                           set_txdelay( tx_delay );
                           cprintf( " now: %dms",tx_delay);
                           cprintf("\r\n");
                           clreol();
                           }

               		 cprintf( "Enter new CD in ms: ");
                         strcpy(in_str, "");
                         in_str[0]=80;
                         cgets(in_str);
 			 cprintf("\r\n");clreol();
			 if(in_str[1]==0) delay_input=cd_delay;
			 else sscanf(&in_str[2],"%d",&delay_input);
                         if ((delay_input <= 0)||(delay_input > MAX_DELAY))
                           {
                           cprintf( "Invalid control delay\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf( "Old CD was: %dms",cd_delay);
                           cd_delay = delay_input;
                           set_cd_delay( cd_delay );
                           cprintf( " now: %dms",cd_delay);
                           cprintf("\r\n");
                           clreol();
                           }

                         break;
	       case '?':
               case 'h':  // Show help
               case 'H': 
        		cprintf("\r\nA - Toggle Memory ARQ option             ");
        		cprintf("\r\nB - Adjust Speaker level                 ");
        		cprintf("\r\nC - Toggle Upper/Lower case option       ");
        		cprintf("\r\nD - Set tx/control delay (milliseconds)  ");
        		cprintf("\r\nE - Adjust Line in level                 ");
        		cprintf("\r\nH - This help display                    ");
        		cprintf("\r\nI - Enter station Identification Selcal  ");
        		cprintf("\r\nK - Toggle Word/Character keyboard mode  ");
        		cprintf("\r\nL - Toggle capture to disk option        ");
        		cprintf("\r\nR - Toggle Reliability option            ");
        		cprintf("\r\nS - Toggle Statistics option             ");
        		cprintf("\r\nT - Adjust master clock timing           ");
        		cprintf("\r\nU - Toggle Unshift-On-Space option       ");
        		cprintf("\r\nV - Toggle direct video I/O option       ");
        		cprintf("\r\nW - Change 'WRU' text                    ");
        		cprintf("\r\nY - Start the 500 Hz signal generator\r\n");
			clreol();
	       		break;

               case 'k':  //  Word/Character Keyboard mode
               case 'K': cprintf( "Toggle Keyboard Word/Character mode: ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old mode was: ");
			 if(Word_Mode) cprintf("Word"); else cprintf("Character");
                         if(Word_Mode==0) Word_Mode=1;
                         else             Word_Mode=0;
                         cprintf( " now: ");
			 if(Word_Mode) cprintf("Word\r\n"); else cprintf("Character\r\n");
                         clreol();
// Require some cleanup after a mode change
                        kb_ptr->Flush_buf();   
                        sp_ptr->Flush_buf();
			entered_space=0;  		// reset word mode
			set_UL_case(UL_case);		// also reset the cases

                         break;

               case 'l':  //  Logging to a file
               case 'L': cprintf( "Toggle Receive capture (logging): ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old Log option was: ");
			 if(Log_option) cprintf("ON"); else cprintf("OFF");
                         if(Log_option==0) Log_option=1;
                         else              Log_option=0;
                         cprintf( " now: ");
			 if(Log_option) cprintf("ON\r\n"); else cprintf("OFF\r\n");
                         clreol();
                         break;

               case 'r':  //  Toggle reliability option
               case 'R': cprintf( "Toggle reliability option: ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Reliability option was: ");
			 if(Hi_rely) cprintf("ON"); else cprintf("OFF");
                         if(Hi_rely==0) Hi_rely=1;
                         else           Hi_rely=0;
			 set_reliability(Hi_rely);
                         cprintf( " now: ");
			 if(Hi_rely) cprintf("ON\r\n"); else cprintf("OFF\r\n");
                         clreol();
                         break;

               case 's':  //  Toggle statistics
               case 'S': cprintf( "Toggle statistics: ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Stats option was: ");
			 if(Stats) cprintf("ON"); else cprintf("OFF");
                         if(Stats==0) Stats=1;
                         else         Stats=0;
                         cprintf( " now: ");
			 if(Stats) cprintf("ON\r\n"); else cprintf("OFF\r\n");
			 clear_counts(); // Clear the counts as well
                         clreol();
                         break;

               case 't':  // Load new timing sequence
               case 'T': cprintf( "Bump TIMING value (using +/-) ");
                         in_str[0]=getch();
 			 cprintf("\r\n");clreol();
			 delay_input=timing_value;
			 if(in_str[0]=='+') delay_input=timing_value+1;
			 if(in_str[0]=='-') delay_input=timing_value-1;
                         if ((delay_input < 15000)||(delay_input > 22000))
                           {
                           cprintf( "Range exceeded - Timing unchanged\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf( "Old value was: %d",timing_value);
                           timing_value = delay_input;
   			   set_timing(timing_value);
                           cprintf( " now: %d",timing_value);
                           cprintf("\r\n");
                           clreol();
                           }
                         break;

               case 'u':  //  Unshift on space
               case 'U': cprintf( "Toggle UOS: ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old UOS was: ");
			 if(UOS) cprintf("ON"); else cprintf("OFF");
                         if(UOS==0) UOS=1;
                         else       UOS=0;
                         cprintf( " now: ");
			 if(UOS) cprintf("ON\r\n"); else cprintf("OFF\r\n");
                         clreol();
                         break;

               case 'v':  //  change direct/BIOS video
               case 'V': cprintf( "Toggle video mode (DIRECT/BIOS): ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old video mode was: "); 
			 if(directvideo) cprintf("DIRECT"); else cprintf("BIOS");
                         if(directvideo==0) directvideo=1;
                         else directvideo=0;
                         cprintf( " now: ");
			 if(directvideo) cprintf("DIRECT\r\n"); else cprintf("BIOS\r\n");
                         clreol();
                         break;

               case 'w':  // CHANGE WRU RESPONSE
               case 'W': cprintf( "Enter new WRU text: ");
                         strcpy(in_str, "");
                         in_str[0]=80;
                         cgets(in_str);
                         cprintf("\r\n");clreol();
			 if(in_str[1]==0)
				{
                           	cprintf("WRU is: %s\r\n",&wru_reply[2]);
                           	cprintf("\r\n");clreol();
				break;
				}
			 for(i=2;i<in_str[1]+2;i++) in_str[i]=toupper(in_str[i]);
                         if (in_str[1] > 40)
                           {
                           cprintf( "WRU message too long\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf("WRU was: %s\r\n",&wru_reply[2]);
                           clreol();
                           strcpy(wru_reply,"\r\n");
                           strncat(wru_reply, &in_str[2], in_str[1]);
                           strcat(wru_reply," +?");
                           cprintf("now: %s",&wru_reply[2]);
                           cprintf("\r\n");clreol();
                           }
                         break;

               case 'y':  // RUN A 500Hz TIMING TEST
               case 'Y': set_control(0x07); 
                        cprintf( "\r\n500 Hz test started\r\n");
			cprintf( "When done, please hit any key to continue\r\n");
            		ch = getche();
 			cprintf("\r");clreol();
			set_control(0x06);		// FORCE STBY
                        break; 

                default: cprintf("??\r\n"); clreol();
               }
            }


// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);

   	_setcursortype( _NOCURSOR );
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
	gotoxy(x,y);
        return ch;
}

//--------------------------------------------------------------------------
//
// Procedure for handling WRU reply
//
void send_wru(Circ_buffer *buf_ptr)
{
char *ptr;
        ptr=wru_reply;
        while (*ptr!='\0') 
                {       
                buf_ptr->Store_buf(*ptr++);
                }
        set_control(0x01);            // Grab the link now
}

//---------------------------------------------------------------------------
//
// Procedure obtains various setup info from user-supplied file
//
void read_setup_data(FILE *setup)
{
int i, start_op;
char setup_line[80];

   gotoxy(1,3);

   if (setup==NULL)
        {
        cprintf("\r\nNo setup file (AMTOR56.CNF) supplied\r\n");
        cprintf("all parameters set to default values\r\n");
	clreol();
        }
   else
        {
	clreol();
        for(;;)
                {
                fgets(setup_line, 80, setup);
                if(feof(setup)) break;
		strupr(setup_line);
                cprintf("%s\r",setup_line);                  // Show setup
		clreol();

                if(!strncmp(setup_line, "ID:",3))            // id
			if(strlen(setup_line)>ID_LENGTH+3) {
			     	cprintf("ID TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(ident,"..../------");
			     	}
                        else {
				strncpy(ident,&setup_line[3],strlen(setup_line)-3);
                          	ident[strlen(setup_line)-4]='\0';
                          	set_IDENT(ident);
			  	}

                if(!strncmp(setup_line, "SELCAL:",7))        // selcal
			if(strlen(setup_line)>SELCAL_LENGTH+7) {
			     	cprintf("SELCAL TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(selcal,"..../------");
			     	}
                        else {
                        	strncpy(selcal,&setup_line[7],strlen(setup_line)-7);
                        	selcal[strlen(setup_line)-8]='\0';
                        	}

                if(!strncmp(setup_line, "WRU:",4))           // wru
			if(strlen(setup_line)>WRU_LENGTH+4) {
			     	cprintf("WRU TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(wru_reply,"\r\nQRA ----- .... +?");
			     	}
                        else {
				strcpy(wru_reply,"\r\n");
                        	strncat(wru_reply,&setup_line[4],
                                strlen(setup_line)-5);       // -4-1 incl. \r
                        	strcat(wru_reply," +?\0");
                        	}

                if(!strncmp(setup_line, "TXDELAY:",8)) {      // txdelay
                        sscanf(&setup_line[8],"%d",&tx_delay);
                        if ((tx_delay < 0)||(tx_delay > MAX_DELAY)) {
			     	cprintf("BAD TXDELAY - IGNORED: OK?\r\n");
			     	getch();
				tx_delay=DEFAULT_TXDELAY;
			     	}
                        else	set_txdelay(tx_delay);
                        }

                if(!strncmp(setup_line, "CDDELAY:",8)) {      // CD delay
                        sscanf(&setup_line[8],"%d",&cd_delay);
                        if ((cd_delay <= 0)||(cd_delay > MAX_DELAY)) {
			     	cprintf("BAD CD DELAY - IGNORED: OK?\r\n");
			     	getch();
				cd_delay=DEFAULT_CDDELAY;
			     	}
                        else	set_cd_delay(cd_delay);
                        }

                if(!strncmp(setup_line, "TIMING:",7)) {       // clock timing
                        sscanf(&setup_line[7],"%d",&timing_value);
			if ((timing_value < 15000)||(timing_value > 22000)) {
			     	cprintf("BAD TIMING VALUE - IGNORED: OK?\r\n");
			     	getch();
				timing_value=DEFAULT_TIMING;
			     	}
                        }

                if(!strncmp(setup_line, "UOS:",4)) {       // UOS
                        sscanf(&setup_line[4],"%d",&UOS);
                        }

                if(!strncmp(setup_line, "CASE:",5)) {       // Upper/Lower case
                        sscanf(&setup_line[5],"%d",&UL_case);
                        }

                if(!strncmp(setup_line, "FCRLF:",6)) {       // Files CR/LF
                        sscanf(&setup_line[6],"%d",&Files_CRLF);
                        }

                if(!strncmp(setup_line, "LOG:",4)) {       // Logging option
                        sscanf(&setup_line[4],"%d",&Log_option);
                        }

                if(!strncmp(setup_line, "LOGFILE:",8))    // Logfile name
			if(strlen(setup_line)>LOGFILENAME_LENGTH+8) {
			     	cprintf("LOG FILENAME TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(logfilename,"TORLOG");
			     	}
                        else {
                        	strncpy(logfilename,&setup_line[8],strlen(setup_line)-8);
                        	logfilename[strlen(setup_line)-9]='\0';
                        	}

                if(!strncmp(setup_line, "SCROLLFILE:",11))    // Scrollback file name
			if(strlen(setup_line)>LOGFILENAME_LENGTH+8) {
			     	cprintf("SCROLLBACK FILENAME TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(scrollback,"SCROLL.TMP");
			     	}
                        else {
                        	strncpy(scrollback,&setup_line[11],strlen(setup_line)-8);
                        	scrollback[strlen(setup_line)-12]='\0';
                        	}

                if(!strncmp(setup_line, "REL:",4)) {   // Reliability
                        sscanf(&setup_line[4],"%d",&Hi_rely);
                        }

                if(!strncmp(setup_line, "VID:",4)) {       // Direct Video option
                        sscanf(&setup_line[4],"%d",&directvideo);
                        }

                if(!strncmp(setup_line, "REFR:",5))  {      // Refresh rate
                        sscanf(&setup_line[5],"%Ld",&Refresh_rate);
                        }

                if(!strncmp(setup_line, "TIME:",5))  {      // GMT=0 or LOCAL=1
                        sscanf(&setup_line[5],"%d",&Time_Type);
                        }

                if(!strncmp(setup_line, "WORD:",5)) {     // Word/character mode
                        sscanf(&setup_line[5],"%d",&Word_Mode);
                        }

                if(!strncmp(setup_line, "MARQ:",5)) {       // Memory ARQ
                        sscanf(&setup_line[5],"%d",&Mem_ARQ);
			if(!((Mem_ARQ==0)||(Mem_ARQ==1))) 
				Mem_ARQ=DEFAULT_MEM_ARQ;
			set_mem_ARQ(Mem_ARQ);
                        }

                if(!strncmp(setup_line, "STATS:",6)) {      // Statistics
                        sscanf(&setup_line[6],"%d",&Stats);
			if(!((Stats==0)||(Stats==1))) 
				Stats=DEFAULT_STATS;
                        }

                if(!strncmp(setup_line, "COM:",4)) {           // COM port
                        sscanf(&setup_line[4],"%d",&COM_prt);  // COM1/2/3/4
                        if (!((COM_prt>0)&&(COM_prt<=MAX_COM))) 
				COM_prt=DEFAULT_COM_PORT;
			}   

                if(!strncmp(setup_line, "OLVL:",5)) {       // Speaker level
                        sscanf(&setup_line[5],"%d",&Sound_out_level);
			if ((Sound_out_level < 0)||(Sound_out_level > 63)) {
			     	cprintf("BAD SPEAKER LEVEL - IGNORED: OK?\r\n");
			     	getch();
				Sound_out_level=DEFAULT_OLVL;
			     	}
                        }

                if(!strncmp(setup_line, "ILVL:",5)) {       // Line in level
                        sscanf(&setup_line[5],"%d",&Sound_in_level);
			if ((Sound_in_level < 0)||(Sound_in_level > 15)) {
			     	cprintf("BAD LINE IN LEVEL - IGNORED: OK?\r\n");
			     	getch();
				Sound_in_level=DEFAULT_ILVL;
			     	}
                        }

                if(!strncmp(setup_line, "SOURCE:",7)) {       // Input source
                        sscanf(&setup_line[7],"%d",&Source);
			if (!((Source==0)||(Source==1)||(Source==2))) {
			     	cprintf("BAD INPUT SOURCE - IGNORED: OK?\r\n");
			     	getch();
				Source=0;
			     	}
			switch (Source) {
				case 0: Source=0x00;
					break;
				case 1: Source=0x40;
					break;
				case 2: Source=0x80;
					break;
				}
			}

                if(!strncmp(setup_line, "AMTORMDM:",9))    // Modem
                        sscanf(&setup_line[9],"%d",&Amtor_Mdm);

                if(!strncmp(setup_line, "IRQ:",4)) {   // interrupt
                        sscanf(&setup_line[4],"%d",&irq);
			switch(irq) {
				case 3:
				case 5:
				case 7:
				case 9:
				case 10:
				case 11:
				case 12:
					break;
				default:
			     		cprintf("IRQ NOT 3, 5, 7, 9, 10, 11, 12 - SET TO 7: OK?\r\n");
			     		getch();
					irq=DEFAULT_IRQ;
				}
			}

                if(!strncmp(setup_line, "MASKIN:",7)) {   // MASKIN
                        sscanf(&setup_line[7],"%d",&Mask_in);
			if(Mask_in) Mask_in=0x80;
			set_rxmask(Mask_in);
			}

                if(!strncmp(setup_line, "MASKOUT:",8)) {   // MASKOUT
                        sscanf(&setup_line[8],"%d",&Mask_out);
			if(Mask_out) Mask_out=Mask_out&0x03;
			set_txmask(Mask_out);
			}
                }

        
   	if (proc_type==86) cprintf("Processor class: 86/88\r\n");
	else if (proc_type==30) cprintf("Processor class: V20/V30\r\n");
   		else if (proc_type==386) cprintf("Processor class: 386/486\r\n");
	     		else cprintf("Processor class: %d\r\n", proc_type);

        cprintf("Initialization completed\r\n");
	clreol();
        fclose(setup);
        }
}
//---------------------------------------------------------------------------
// Routine to update status window
//
void update_status_line(Circ_buffer *kb_ptr, Circ_buffer *sp_ptr)
{
long biostime_now;
int mark_tuning, space_tuning, i, j;
#define TUN 65
static int mark, space, cases;
char tst_char[2];
char upper_row[]="m¿m";
char lower_row[]="ss";

	biostime_now=biostime(0,0L);

// Tuning indicator display (line 2)
	current_window=activate_window(5);
	get_tuning_data( &mark_tuning, &space_tuning);

	// Scale properly.
	mark_tuning /= 12;
	space_tuning /= 12;
	if(mark_tuning>=10)  mark_tuning=10;
	if(space_tuning>=10) space_tuning=10;

////////if((biostime_now-previous_biostime)>=Refresh_rate) {
		mark=(mark+mark_tuning)>>1;    // Average mark with prev.
		space=(space+space_tuning)>>1; // Average space with prev.


		j=0;
		for(i=TUN; i<TUN+12-mark; i++) {
			gotoxy(i,1); cprintf("%c", upper_row[j]);
			j++;
			}
		j=0;
		for(i=TUN; i<TUN+12-space; i++) {
			gotoxy(i,2); cprintf("%c", lower_row[j]);
			j++;
			}

// >=10  - RED
//   8,9 - YELLOW
//
		if(mark>=10) {
			textattr(0x7F&(RED+(c_BGCOLOR<<4)));
			gotoxy(TUN+12-mark,1); 
			cprintf("");
			}
		else mark++;

		for(i=mark-1; i>=0 ;i--)
			if((mark>7)&&(i>=7)) {
				gotoxy(TUN+12-i,1); 
				textattr(0x7F&(YELLOW+(c_BGCOLOR<<4)));
				cprintf("");
				}
	    		else 
				if((mark>4)&&(i>=4)) {
					gotoxy(TUN+12-i,1); 
					textattr(0x7F&(LIGHTBLUE+(c_BGCOLOR<<4)));
					cprintf("");
					}
				else {
					gotoxy(TUN+12-i,1); 
					textattr(0x7F&(LIGHTGRAY+(c_BGCOLOR<<4)));
					cprintf("");
					}

		textcolor(c_CHCOLOR);
		textbackground(c_BGCOLOR);
		cprintf("%c%c", upper_row[13], upper_row[14]);

		if(space>=10) {
			textattr(0x7F&(RED+(c_BGCOLOR<<4)));
			gotoxy(TUN+12-space,2); 
			cprintf("");
			}
		else space++;

		for(i=space-1; i>=0 ;i--)
			if((space>7)&&(i>=7)) {
				gotoxy(TUN+12-i,2); 
				textattr(0x7F&(YELLOW+(c_BGCOLOR<<4)));
				cprintf("");
				}
			else
				if((space>4)&&(i>=4)) {
					gotoxy(TUN+12-i,2); 
					textattr(0x7F&(LIGHTBLUE+(c_BGCOLOR<<4)));
					cprintf("");
					}
				else {
					gotoxy(TUN+12-i,2); 
					textattr(0x7F&(LIGHTGRAY+(c_BGCOLOR<<4)));
					cprintf("");
					}

		textcolor(c_CHCOLOR);
		textbackground(c_BGCOLOR);
		cprintf("%c%c", lower_row[13], lower_row[14]);
/////////	} 

// Show goblins on second line, input buffer size on second status line

	if((biostime_now-previous_biostime)>Refresh_rate) {
		current_window=activate_window(5);
		gotoxy(50,1);
		if(Log_option) cprintf("");	  
		if(Hi_rely)    cprintf("");
		if(UOS) cprintf("");
		if(Mem_ARQ) cprintf("");

		if(Mask_toggle) {
			textattr(0x80|(c_CHCOLOR+(c_BGCOLOR<<4)));   // flash
			cprintf("i");
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR);
			}

		cases = get_cases();
		if(UL_case==0)    cprintf("");
		if(UL_case) 
			if (cases&0xFF00) cprintf("");
			else        	  cprintf("");
		if(UL_case) 
			if (cases&0xFF)   cprintf("");
			else           	  cprintf("");

		current_window=activate_window(4);
		gotoxy(1,1);
		if(Log_option) {
			textattr(0x7F&(c_CYAN+(c_BGCOLOR<<4)));
			cprintf("LOG=>%s", logfilename);
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR);
			}

		if(strlen(Log_program)) {
			textattr(0x7F&(c_CYAN+(c_BGCOLOR<<4)));
			cprintf("LOGGER=>%sPACKET=>%s", Log_program, Packet_program);
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR);
			}

		if(Printer) {
			textattr(0x7F&(c_CYAN+(c_BGCOLOR<<4)));
   			cprintf("PRINTING"); 
			}

/// ***DEBUG	cprintf("M=%d S=%d", mark_tuning, space_tuning); 

		if(Stats) {
			textattr(0x7F&(c_CYAN+(c_BGCOLOR<<4)));
   			if(Mem_ARQ) {
   				cprintf("RQ:%d ER:%d",RQ_count,ER1_count); 
				cprintf(" MER:%d",ER2_count);
				} 
   			else    cprintf("RQ:%d ER:%d",RQ_count,ER1_count); 
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR);
			}
		
/* Cleanup after text displays, this produces less flickering */
		cprintf("");

/* Display buffer counts in flashing yellow when there is data there */

		textattr(0x7F&(c_CYAN+(c_BGCOLOR<<4)));
		gotoxy(66,1); cprintf("BUFRS: "); 
		gotoxy(73,1); 
		if(kb_ptr->Present_size()) {
			textattr(0x80|(c_YELLOW+(c_BGCOLOR<<4)));   // flash
			cprintf("%d",kb_ptr->Present_size());
			}
		else	cprintf("%d",kb_ptr->Present_size());

		textattr(0x7F&(c_CYAN+(c_BGCOLOR<<4)));
		cprintf("");
		if(sp_ptr->Present_size()) {
			textattr(0x80|(c_YELLOW+(c_BGCOLOR<<4)));   // flash
			cprintf("%d",sp_ptr->Present_size());
			}
		else	cprintf("%d",sp_ptr->Present_size());

		textcolor(c_CHCOLOR);
		textbackground(c_BGCOLOR);
		gettext(wherex()+1,wherey(),wherex()+1,wherey(),tst_char);
		if(tst_char[0]!=' ') while(wherex()<80) cprintf(" ");

// Show status line
		display_status_data();
		previous_biostime=biostime_now;
		}
}

//---------------------------------------------------------------------------
// Routine to update status window
//
void display_status_data()
{
int stat,i;
time_t tnow;
struct tm *tmnow;
char time_str[13], *st_ptr;

      sys_stat = get_status();	      		// get status data

// Update status line
	current_window = activate_window(1);

	textcolor(c_CYAN);

	gotoxy(1,1); 

      	if((sys_stat&0x07)==6) cprintf("STBY:           ");
      	else	{
		if((sys_stat&0x10)==0) cprintf(" ARQ: ");
		else                   cprintf(" FEC: ");
		}

      	switch (sys_stat&0x07)
		{
		case 3: cprintf(" ");
			textattr(0x7F&(c_YELLOW+(BLACK<<4)));
			cprintf("IDL");
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("      "); 
			break;
		case 2:	cprintf("   ");
			textattr(0x7F&(c_YELLOW+(BLACK<<4)));
			cprintf("TFC");
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("    "); 
			break;
		case 0:
			cprintf("     ");
			textattr(c_YELLOW+(BLACK<<4));
			cprintf("ERR");
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("  ");
			break;
		case 7:
			cprintf("     ");
			textattr(c_YELLOW+(BLACK<<4));
			cprintf("ERR");
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("  ");
			break;
		case 1:
			cprintf("       ");
			textattr(c_YELLOW+(BLACK<<4));
			cprintf("RQR");
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("");
			break;
		case 4:
			cprintf(" ");
			textattr(0x80|(c_YELLOW+(BLACK<<4)));   // flash
			cprintf("OVER");
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("     ");
			break;
		case 5:
			cprintf(" "); 
			textattr(0x80|(c_YELLOW+(BLACK<<4)));   // flash
			cprintf("PHASE"); 
			textcolor(c_CYAN);
			textbackground(c_BGCOLOR); 
			cprintf("    "); 
			break;
		}

	time(&tnow);
	if(Time_Type==GMT) {
		tmnow=gmtime(&tnow);
		time_str[9]='U';time_str[10]='T';time_str[11]='C';
		time_str[12]='\0';
		}	
	else	{
		tmnow=localtime(&tnow);
		strncpy(&time_str[9],tzname[0],3);
		time_str[12]='\0';
		}	

	st_ptr=asctime(tmnow);		// fill in the time
	while(*st_ptr++!=':');		// look for start of first field
	st_ptr--;st_ptr--;st_ptr--;
	for(i=0;i<8;i++) time_str[i]=*st_ptr++;
	time_str[8]=' ';

	gotoxy(18,1);
        stat=get_BKRX(); 
	cprintf("%sTD:%2d CD:%2d PD:%3d",selcal,tx_delay,cd_delay,stat);
	for(i=wherex();i<57;i++) cprintf(" "); // clean up after display
        gotoxy(57,1);
	cprintf("%s",time_str);

	textcolor(c_CHCOLOR);
	
	if(get_RTS()) textbackground(c_YELLOW);	// TX=YELLOW
	gotoxy(69,1);cprintf( "");        // Show bit phasing status
      	stat=get_phase_err();
      	if(stat>BITSLIP)  stat= BITSLIP;
      	if(stat<-BITSLIP) stat=-BITSLIP;
      	gotoxy(74+stat*5/BITSLIP,1);
	cprintf(" ");
	textbackground(c_BGCOLOR); 
}

//--------------------------------------------------------------------------
//
// Pop-up window for entering selcals
//
void enter_selcal()
{

int bufsize, left=20,right=60,top=11,bottom=15;
char *tbuffer, ch, c, in_str[80];
int i, x, y;

	save_xy(&x,&y);
   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(" ARQ SELCAL/CALLSIGN ENTRY ͻ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();

        for(;;)
                {
                cprintf( "%s > ",selcal);
//
// This my own cgets, because cgets does not handle ESC correctly
//
		i=2;
		while(i<80){
			in_str[1]=i-2;
			ch=getch();
			if(ch=='\r') {
				in_str[i]='\0';
				break;
				}
			if(ch==0x1B) {
				in_str[1]=0;
				break;
				}
			if(ch==BackSpace)
				if(i>2) {
					i--;
					cprintf("\b \b");
					}
			if(ch!=BackSpace) {
				in_str[i++]=ch;
				cprintf("%c",ch);
				}
			}

		if(i==80) in_str[i]='\0';
                cprintf("\r\n");clreol();
	        if(in_str[1]==0) break;
	        for(i=2;i<in_str[1]+2;i++) in_str[i]=toupper(in_str[i]);
                if(in_str[1]<4)
                        {
                        cprintf( "Four characters required\r\n");
                        clreol();
                        }
                else 
                        {   
                        strncpy( selcal, &in_str[2], in_str[1]);
                        selcal[in_str[1]]='\0';
                        set_SELCAL(selcal);
                        break;
                        }
                }

// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	gotoxy(x,y);
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
   	_setcursortype( _NOCURSOR );
}

//--------------------------------------------------------------------------
//
// Pop-up window for entering buffer numbers
//
void buffers_menu(Circ_buffer *buf_ptr)
{
int bufsize, left=20, right=60, top=5, bottom=20;
char *tbuffer, ch, c, in_str[80], *text_ptr;
char input_line[100];
int i, k, no, flag, x, y;

	save_xy(&x,&y);
   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(" BUFFER DIRECTORY MENU ͻ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();

        if(number_of_files==0) cprintf("No buffers available\r\n");
        else
                {
                for(k=0;k<number_of_files;k++)  
			{
			cprintf("(%d) %s", k+1, buffer_string[k]);
			cprintf(" size %d\r\n", buffer_string_size[k]);
			}
                }
        
        flag=0;
        while(flag==0)
                {
                cprintf("Enter buffer number > ");
//
// This my own cgets, because cgets does not handle ESC correctly
//
		i=2;
		while(i<80){
			in_str[1]=i-2;
			ch=getch();
			if(ch=='\r') {
				in_str[i]='\0';
				break;
				}
			if(ch==0x1B) {
				in_str[1]=0;
				break;
				}
			if(ch==BackSpace)
				if(i>2) {
					i--;
					cprintf("\b \b");
					}
			if(ch!=BackSpace) {
				in_str[i++]=ch;
				cprintf("%c",ch);
				}
			}

		if(i==80) in_str[i]='\0';
 	        cprintf("\r\n");clreol();
	        if(in_str[1]==0) flag=1;
	        else 
                  {
                  sscanf(&in_str[2],"%d",&no);
		  no=no-1;
                  if(no>=number_of_files) cprintf("That buffer is not available\r\n");
                  else    {
  		          buffer_load(no+1, buf_ptr);
                          flag=1;
                          }
                  }
                }

// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
   	_setcursortype( _NOCURSOR );

	while(kbhit()) ch=getch(); // flush extra chars
}

//--------------------------------------------------------------------------
//
// Pop-up window for transferring of file data
// Note that binary read is used because the buffered reads, i.e.
// fgetc and getc strips out the CR characters
//
void files_menu(Circ_buffer *buf_ptr, Circ_buffer *kb_ptr)
{
int bufsize, left=5, right=75, top=5, bottom=20;
char *tbuffer, c, ch, chh, in_str[80];
char input_line[100];
int i, k, flag, buf_stat, x2, x3, y2, y3, loc_x, loc_y;
FILE *f_ptr;
time_t bintime, time_now;

	while(kbhit()) ch=getch(); // flush extra chars

// Save cursur locations for screen 2 and 3 
	current_window = activate_window(2);  
	save_xy(&x2,&y2);
	current_window = activate_window(3);  
	save_xy(&x3,&y3);

   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(
" FILE TRANSMISSION MENU ͻ");

        for (k=1;k<13;k++)
		cprintf(
"                                                                     ");

        cprintf(
"ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();

	flag=0;
        while(flag==0)
                {
                cprintf("Enter file name to be loaded > ");

//
// This my own cgets, because cgets does not handle ESC correctly
//
		i=2;
		while(i<80){
			in_str[1]=i-2;
			ch=getch();
			if(ch=='\r') {
				in_str[i]='\0';
				break;
				}
			if(ch==0x1B) {
				in_str[1]=0;
				break;
				}
			if(ch==BackSpace)
				if(i>2) {
					i--;
					cprintf("\b \b");
					}
			if(ch!=BackSpace) {
				in_str[i++]=ch;
				cprintf("%c",ch);
				}
			}

		if(i==80) in_str[i]='\0';
 	        cprintf("\r\n");clreol();
	        if(in_str[1]==0) flag=1;
	        else 
                  {
                  if((f_ptr=fopen(&in_str[2],"rb"))==NULL)
			 cprintf("Cannot open <%s>\r\n", &in_str[2]);
                  else  {
			cprintf("\r\nReady to transmit <%s>", &in_str[2]); 
			textattr(0x80|(c_YELLOW+(WHITE<<4)));   // flash
			cprintf("\r\n--Hit any key to stop this transmission--\r\n");
			textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
// Priming read
                        fread((void *)&c,sizeof(char),1,f_ptr);
                        while (!feof(f_ptr))
                          	{
//
// Save current location in file transmission window,
//  update the status line, and restore file transmission menu
// 
				save_xy(&loc_x,&loc_y);
				textcolor(c_CHCOLOR);
				textbackground(c_BGCOLOR); 
   				_setcursortype( _NOCURSOR );
				update_status_line( kb_ptr, buf_ptr);
        			window(left+1,top+1,right-1,bottom-2);
				gotoxy(loc_x,loc_y);
				textattr((BLACK+(WHITE<<4))&0x7F);

				if (UL_case==0)	c=toupper(c);

				extend_ascii(buf_ptr,c);
				chh=buf_ptr->Read_buf();
				
				buf_stat=1;		// Send 1/2 character
				while(buf_stat)
					{
					if(kbhit())	// abort when user
						{	// hits any key
						break;
						}

					switch (Files_CRLF) {
					   case 1:   // NL implies CR/LF
                                		if(chh=='\n') buf_stat=output_txdta((int)'\r');
						else if(chh=='\r') buf_stat=0;
						     else          buf_stat=output_txdta((int)chh);
						break; 
					   case 0:   // Pass everything
						buf_stat=output_txdta((int)chh);
						break;
					   }
					}
							// Send 2/2 character
   				if((chh=buf_ptr->Read_buf())!=(char)0xFF)
				  {
				  buf_stat=1;
				  while(buf_stat)
					{
					if(kbhit())	// abort when user
						{	// hits any key
						break;
						}

					switch (Files_CRLF) {
					   case 1:   // NL implies CR/LF
                                		if(chh=='\n') buf_stat=output_txdta((int)'\r');
						else if (chh=='\r') buf_stat=0;
						     else           buf_stat=output_txdta((int)chh);
						break; 
					   case 0:   // Pass everything
						buf_stat=output_txdta((int)chh);
						break;
					   }
					}
				  }

				if(kbhit())  break;		// user want out ?

				if(c!='\n') cprintf("%c", c);	// display the character
				if((Log_option)&&(c!='\n')) 
					fprintf(logfile,"%c", c); // and logfile
				if(c=='\r') 
					{
					cprintf("\n");
					fprintf(logfile,"\n");
					clreol();
					}
				if(wherex()==(right-left-2))	// Line overflow ?
					{
					cprintf("\r\n");
					clreol();
					}
					
                             				// get next character
                        	fread((void *)&c,sizeof(char),1,f_ptr);
//
// Some characters that need to be ignored
//
				if(c==(char)CTL_z) break; // Cntrl_Z
				if(c==(char)TAB) c=' '; // TAB
                                }

                        fclose(f_ptr);
			flag=1;
		       	}  // end got EOF during file reading
		  } // end else got valid filename
		}  // end while

// Restore screen

	if(in_str[1]!=0) 
		{
		cprintf("\r\nTransmission completed");
		time(&bintime);  
		while(1) {    // delay 2 seconds
			time(&time_now);
			if(difftime(time_now, bintime)>2) break;
			}
		}

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
   	_setcursortype( _NOCURSOR );

	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters

	while (kbhit()) getch();	// flush keyboard buffer

// Restore cursors
	current_window = activate_window(2);  
	gotoxy(x2,y2);			// leave cursor as before interrupted
	current_window = activate_window(3);  
	gotoxy(x3,y3);			// leave cursor as before interrupted
}

//--------------------------------------------------------------------------
//
// Pop-up window for continuous looping a file data
// Note that binary read is used because the buffered reads, i.e.
// fgetc and getc strips out the CR characters
//
void loop_menu(Circ_buffer *buf_ptr, Circ_buffer *kb_ptr)
{
int bufsize, left=5, right=75, top=5, bottom=20;
char *tbuffer, c, ch, chh, in_str[80];
char input_line[100];
int i, k, flag, looping, times, buf_stat, x2, x3, y2, y3, loc_x, loc_y;
FILE *f_ptr;
time_t bintime, time_now;

	while(kbhit()) ch=getch(); // flush extra chars

// Save cursur locations for screen 2 and 3 
	current_window = activate_window(2);  
	save_xy(&x2,&y2);
	current_window = activate_window(3);  
	save_xy(&x3,&y3);

   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(
" CONTINUOUS FILE TRANSMISSION MENU ͻ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();


	flag=0;looping=0;times=1;
        while(flag==0)
                {
		if(!looping) cprintf("Enter file name to be loaded > ");

//
// This my own cgets, because cgets does not handle ESC correctly
//
		if(!looping) {
			i=2;
			while(i<80){
				in_str[1]=i-2;
				ch=getch();
				if(ch=='\r') {
					in_str[i]='\0';
					break;
					}
				if(ch==0x1B) {
					in_str[1]=0;
					break;
					}
				if(ch==BackSpace)
					if(i>2) {
						i--;
						cprintf("\b \b");
						}
				if(ch!=BackSpace) {
					in_str[i++]=ch;
					cprintf("%c",ch);
					}
				}
			}

		if(i==80) in_str[i]='\0';
 	        cprintf("\r\n");clreol();
	        if(in_str[1]==0) flag=1;
	        else 
                  {
                  if((f_ptr=fopen(&in_str[2],"rb"))==NULL)
			 cprintf("Cannot open <%s>\r\n", &in_str[2]);
                  else  {
			if(!looping) {
//
// Switch TX ON
//
				set_control(0x04);
				Mode=FEC;

				cprintf("\r\nReady to transmit <%s>", &in_str[2]); 
				textattr(0x80|(c_YELLOW+(WHITE<<4)));   // flash
				cprintf("\r\nCR will stop transmission immediately,\r\n");
				cprintf("any other key will wait until end of file.\r\n");
				textattr((BLACK+(WHITE<<4))&0x7F);
				looping=1;
				}
			else {
				cprintf("\r\nRetransmitting <%s> %d", &in_str[2], times); 
				textattr(0x80|(c_YELLOW+(WHITE<<4)));   // flash
				cprintf("\r\nCR will stop transmission immediately,\r\n");
				cprintf("any other key will wait until end of file.\r\n");
				textattr((BLACK+(WHITE<<4))&0x7F);
				}
// Priming read
                        fread((void *)&c,sizeof(char),1,f_ptr);
                        while (!feof(f_ptr))
                          	{
//
// Save current location in file transmission window,
//  update the status line, and restore file transmission menu
// 
				save_xy(&loc_x,&loc_y);
				textcolor(c_CHCOLOR);
				textbackground(c_BGCOLOR); 
   				_setcursortype( _NOCURSOR );
				update_status_line( kb_ptr, buf_ptr);
        			window(left+1,top+1,right-1,bottom-2);
				gotoxy(loc_x,loc_y);
				textattr((BLACK+(WHITE<<4))&0x7F);

				if (UL_case==0)	c=toupper(c);

				extend_ascii(buf_ptr,c);
				chh=buf_ptr->Read_buf();
				
				buf_stat=1;		// Send 1/2 character
				while(buf_stat)
					{
					switch (Files_CRLF) {
					   case 1:   // NL implies CR/LF
                                		if(chh=='\n') buf_stat=output_txdta((int)'\r');
						else if(chh=='\r') buf_stat=0;
						     else          buf_stat=output_txdta((int)chh);
						break; 
					   case 0:   // Pass everything
						buf_stat=output_txdta((int)chh);
						break;
					   }
					}
							// Send 2/2 character
   				if((chh=buf_ptr->Read_buf())!=(char)0xFF)
				  {
				  buf_stat=1;
				  while(buf_stat)
					{
					switch (Files_CRLF) {
					   case 1:   // NL implies CR/LF
                                		if(chh=='\n') buf_stat=output_txdta((int)'\r');
						else if (chh=='\r') buf_stat=0;
						     else           buf_stat=output_txdta((int)chh);
						break; 
					   case 0:   // Pass everything
						buf_stat=output_txdta((int)chh);
						break;
					   }
					}
				  }

				if(c!='\n') cprintf("%c", c);	// display the character
				if((Log_option)&&(c!='\n')) 
					fprintf(logfile,"%c", c); // and logfile
				if(c=='\r') 
					{
					cprintf("\n");
					fprintf(logfile,"\n");
					clreol();
					}
				if(wherex()==(right-left-2))	// Line overflow ?
					{
					cprintf("\r\n");
					clreol();
					}
					
                             				// get next character
                        	fread((void *)&c,sizeof(char),1,f_ptr);
//
// Some characters that need to be ignored
//
				if(c==(char)CTL_z) break; // Cntrl_Z
				if(c==(char)TAB) c=' '; // TAB
//
// Does user want out ?
// A CR exit the user immediately, anything else will wait till EOF
//
				if(kbhit()) {
					ch=getch();
					if(ch=='\r') {
						flag=1;
						break;
						}
					ungetch(ch);
					}
                                }

                        fclose(f_ptr);

			if(kbhit()) {	 //  terminating after EOF ??
				ch=getch(); 	// flush char
				flag=1;
				}

			times++; // looping counter
		       	}  // end got EOF during file reading

		  } // end else got valid filename
		}  // end while


//
// Switch TX OFF
//
	set_control(0x02); 	// set Quit flag
       	Mode=STBY;

// Restore screen

	if(in_str[1]!=0) 
		{
		cprintf("\r\nTransmission completed");
		time(&bintime);  
		while(1) {    // delay 2 seconds
			time(&time_now);
			if(difftime(time_now, bintime)>2) break;
			}
		}

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
   	_setcursortype( _NOCURSOR );

	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	flush_rxbuffer();		// flush echoed characters
	
	while (kbhit()) getch();	// flush keyboard buffer

// Restore cursors
	current_window = activate_window(2);  
	gotoxy(x2,y2);			// leave cursor as before interrupted
	current_window = activate_window(3);  
	gotoxy(x3,y3);			// leave cursor as before interrupted
}


//--------------------------------------------------------------------------
// Routine to load buffer from ALT-x key
//
void buffer_load(int no, Circ_buffer *buf_ptr)
{
char c,*text_ptr;

// Check for receive only
	if(Receive_only) return;

	no=no-1;
	if((no>=number_of_files)||(no<0))
		{
		cprintf(" \r\n"); clreol();
		cprintf("That buffer is not available\r\n");
		clreol(); display_cursor(); 
		return;
		}
        else	{
		text_ptr=buffer_string_ptr[no];
                buf_ptr->Store_buf('\r');
		buf_ptr->Store_buf('\n');
		entered_space+=2;

	       	current_window = activate_window(3);  
		textcolor(c_CHCOLOR);
		textbackground(c_BGCOLOR); 

		cprintf(" ");			// blank old cursor
		cprintf("\r\n");
		clreol();
		c=(*text_ptr++);  		// Priming character ????????
		while (c!='\0')
                	{
//
// Display buffer characters on screen
//
			if(c=='\r') {
				cprintf(" \r"); // blank out old cursor
				entered_space++;
				}
			else {
				if(c=='\n') {
					cprintf("\n");
					clreol();
					entered_space++;
					}
				else cprintf("%c",c);
				}

			if(c==' ') entered_space++;
//
// Special embedded control characters
//  switches TX ON
//
			if(c=='') {
				set_control(0x04);
				Mode=FEC;
		      		text_ptr++;
				}
//
//  switches TX OFF
//
			if(c=='') {
				set_control(0x02); 	// set Quit flag
                        	Mode=STBY;		// (All modes)
			   	text_ptr++;
				}
//
// Normal characters
//				
			if (UL_case==0)	c=toupper(c);

//  Extend & save in buffer
			extend_ascii(buf_ptr, c); 

                        c=(*text_ptr++);	 // get next character
                        }
		display_cursor();
		}
}

//--------------------------------------------------------------------------
// Routine to read directory and find file with the "tor" extension
// Note that binary read is used because the buffered reads, i.e.
// fgetc and getc strips out the CR characters
//
void read_directory()
{
FILE *file_ptr, *text_file_ptr;
char c, input_line[100], *text_ptr;
int i,j,k;
char  cnf_path[30], *env_ptr, commnd[30];

	entered_space=0;		// reset word mode

   	strcpy(cnf_path,"");  		// Get ENV variable
   	env_ptr=getenv("PCTOR");
   	if(env_ptr==NULL) strcpy(cnf_path,".");
   	else strcpy(cnf_path,env_ptr);
	strcat(cnf_path,"\\");

        gotoxy(1,3);
        cprintf("Reading current directory for data files (.tor)\r\n");

	strcpy(commnd,"dir ");
	strcat(commnd,cnf_path);
	strcat(commnd,"*.tor /o:n> tor$.tmp");
        system(commnd);
        if((file_ptr=fopen("tor$.tmp", "r")) == NULL)
                terminate_prog("Cannot open temporary list file tor$.tmp\n");
        i=0;
        for(;;)
                {
                fgets(input_line, 99, file_ptr);
		strupr(input_line);	// make line all upper case
                if(feof(file_ptr)) break;

// Extract text files with the "TOR" extension
// Save their names and file sizes

                if(strstr(&input_line[8], " TOR ")!=NULL) 
                        {
                        strcpy(buffer_string[i], "");
                        input_line[8]='.';     //  "." to indicate extension
                        k=0;
                        for(j=0;j<12;j++)      // copy over whole of filename
			   {
			   if(input_line[j]!=' ') buffer_string[i][k++]=
                                                               input_line[j];
			   }
                        if(buffer_string[i][k-1]=='.')buffer_string[i][k-1]='\0';
                        else buffer_string[i][k]='\0';
                  	sscanf(&input_line[13],"%d",&buffer_string_size[i]);

// Allocate buffer space for the text

			if ((text_ptr=(char *)malloc(buffer_string_size[i]+1))==NULL)
				{
				cprintf("Not enough room for all message buffers!\r\n");
				break;
				}
			else
				{
			        buffer_string_ptr[i]=text_ptr;

// open text message and load into memory for later retrival
				strcpy(commnd,"");
				strcat(commnd,buffer_string[i]);
				if((text_file_ptr=fopen(commnd,"rb")) == NULL)
                        		{
                        		cprintf("Cannot open <%s>\r\n", buffer_string[i]);
                        		}
                       		else
                          		{
                          		fread((void *)&c,sizeof(char),1,text_file_ptr);
                          		while (!feof(text_file_ptr))
                                		{
                                		if(UL_case==0) c=toupper(c);

						switch (Files_CRLF) {
					   	case 1:   // NL implies CR/LF
                                		   if(c=='\n')  
		                			{
		              				*text_ptr++ = '\r';
		              				*text_ptr++ = '\n';
		              				}
						   else if (c!='\r') *text_ptr++=c;
						   break; 
					        case 0:   // Pass everything
						   *text_ptr++=c;
						   break;
					        }

                          		        fread((void *)&c,sizeof(char),1,text_file_ptr);
						if(c==(char)CTL_z) break; // Cntrl_Z
						if(c==(char)TAB) c=' '; // TAB
                                		}
					*text_ptr='\0';
                          		fclose(text_file_ptr);
                        		i++;
					}
				}
                        }
                }
        
        fclose(file_ptr);
        system("erase tor$.tmp");

        number_of_files=i;
        if(i==0) 
		{
		cprintf("No TOR files found\r\n");
		clreol();
		}
}

//--------------------------------------------------------------------------
//
// Dos shell
//
void dosshell_cmd()
{

int bufsize, left=1,right=80,top=1,bottom=24;
char *tbuffer;
int i, x, y;
	
	save_xy(&x,&y);				      // save current window
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");


   	_setcursortype( _NORMALCURSOR );
	normvideo();
        window(1,1,80,25);
        clrscr();   
	printf("PCTOR is now suspended  - Low level decoder remains active.\n");
	printf("When done, please type EXIT to re-activate PCTOR.\n");

				    // Call command.com as specified
     	system(getenv("COMSPEC"));  // in the COMSPEC env parameter. 

// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);

	_setcursortype( _NOCURSOR );
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
	gotoxy(x,y);
}
//--------------------------------------------------------------------------
//
// Dos shell
//
void logshell_cmd()
{

int bufsize, left=1,right=80,top=1,bottom=24;
char *tbuffer;
int i, x, y;
	
	save_xy(&x,&y);				      // save current window
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");


   	_setcursortype( _NORMALCURSOR );
	normvideo();
        window(1,1,80,25);
        clrscr();   
     	system(Log_program);

// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);

	_setcursortype( _NOCURSOR );
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
	gotoxy(x,y);
}
//---------------------------------------------------------------------------
// Routine to save present x-y coordinates
void save_xy(int *x,int *y)
{
	*x=wherex();
	*y=wherey();
}

//---------------------------------------------------------------------------
// Catastrophic failure error handler function
//
void terminate_prog(char *ptr)
{
	close_down();	      		// Re-instate interrupts
	normvideo();
        window(1,1,80,25); clrscr();
	cprintf("\r\nAMTOR56 detected a serious error - further operation is terminated");
	cprintf("\r\nThe following is a short description of the problem:");
	cprintf("\r\n%s", ptr); 
	cprintf("\r\nIf its a memory allocation problem, you might want to");
	cprintf("\r\nreduce the number or size of your message buffers and");
	cprintf("\r\ntry again");
	exit(1);			// exit to DOS

}
//---------------------------------------------------------------------------
// Critical error handler
//
#define WRITE_PROT      0
#define DRIVE_NOT_READY 2
#define PRINT_ERR	9

#define HARDERR_ABORT   2
#define SOFTERR_RETN    -1

int harderror_handler(int errorcode, int deverror,
                      int bpval, int sival)
{
unsigned far *devhdr;
char dletter;
int bufsize, left=5,right=75,top=5,bottom=12;
char ch, *tbuffer;
int x2, x3, y2, y3;


// Save cursur locations for screen 2 and 3 
	current_window = activate_window(2);  
	save_xy(&x2,&y2);
	current_window = activate_window(3);  
	save_xy(&x3,&y3);

   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((c_YELLOW+(WHITE<<4))&0x7F);   // Yellow letters on white backgnd
        cprintf(
"   SYTEM ERROR HANDLER  ͻ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();


 	// Pointer to device hdr
	switch (errorcode&0xFF)
		{
		case DRIVE_NOT_READY:
			dletter='A'+(deverror&0xFF);  // find out which drive
			cprintf("\r\nDrive %c is not ready - data may be lost", dletter);
			cprintf("\r\nPlease fix - enter (c) to continue: ");
			while(getch()!='c');

        		if(!puttext(left,top,right,bottom,tbuffer))
                		terminate_prog("Puttext - restoring text");
        		free(tbuffer);
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR); 
   			_setcursortype( _NOCURSOR );
			current_window = activate_window(2);  
			gotoxy(x2,y2);		// leave cursor as before interrupted
			current_window = activate_window(3);  
			gotoxy(x3,y3);		// leave cursor as before interrupted

			hardretn(SOFTERR_RETN);

		case WRITE_PROT:
			dletter='A'+(deverror&0xFF);  // find out which drive
			cprintf("\r\Write protection on %c - data may be lost", dletter);
			cprintf("\r\nPlease fix - enter (c) to continue: ");
			while(getch()!='c');

        		if(!puttext(left,top,right,bottom,tbuffer))
                		terminate_prog("Puttext - restoring text");
        		free(tbuffer);
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR); 
   			_setcursortype( _NOCURSOR );
			current_window = activate_window(2);  
			gotoxy(x2,y2);		// leave cursor as before interrupted
			current_window = activate_window(3);  
			gotoxy(x3,y3);		// leave cursor as before interrupted
			while(kbhit()) getch(); // Flush keyboard 
			hardretn(SOFTERR_RETN);

		case PRINT_ERR:	    			// PRINTER
			cprintf("\r\Printer ERROR: R(etry) or A(bort) printing");
			while(1) {
				ch=getch();		
				if((ch=='R')||(ch=='A')||
				  (ch=='r')||(ch=='a')) break;
				}
			switch(ch) {
				case 'a':
				case 'A':
					Printer=0;

        				if(!puttext(left,top,right,bottom,tbuffer))
                				terminate_prog("Puttext - restoring text");
        				free(tbuffer);
					textcolor(c_CHCOLOR);
					textbackground(c_BGCOLOR); 
   					_setcursortype( _NOCURSOR );
					current_window = activate_window(2);  
					gotoxy(x2,y2); 	// leave cursor as before interrupted
					current_window = activate_window(3);  
					gotoxy(x3,y3);	// leave cursor as before interrupted
					while(kbhit()) getch(); // Flush keyboard 
					hardretn(SOFTERR_RETN);
				default:
        				if(!puttext(left,top,right,bottom,tbuffer))
                				terminate_prog("Puttext - restoring text");

        				free(tbuffer);
					textcolor(c_CHCOLOR);
					textbackground(c_BGCOLOR); 
   					_setcursortype( _NOCURSOR );
					current_window = activate_window(2);  
					gotoxy(x2,y2); 	// leave cursor as before interrupted
					current_window = activate_window(3);  
					gotoxy(x3,y3); 	// leave cursor as before interrupted
					while(kbhit()) getch(); // Flush keyboard 
					hardretn(SOFTERR_RETN);
				}

		default:  // Trap further printer errors (offline)

			  if((deverror&0x1000)&&(deverror&0x0100)) {
					cprintf("\r\Printer ERROR: Check if printer is ONLINE - Printing disabled\r\n");
					cprintf("Enter any key to resume\r\n");
					getch();		

					Printer=0;  //switch printer off

        				if(!puttext(left,top,right,bottom,tbuffer))
                				terminate_prog("Puttext - restoring text");
        				free(tbuffer);
					textcolor(c_CHCOLOR);
					textbackground(c_BGCOLOR); 
   					_setcursortype( _NOCURSOR );
					current_window = activate_window(2);  
					gotoxy(x2,y2); 	// leave cursor as before interrupted
					current_window = activate_window(3);  
					gotoxy(x3,y3);	// leave cursor as before interrupted
					while(kbhit()) getch(); // Flush keyboard 
					hardretn(SOFTERR_RETN);
					}

			devhdr = (unsigned *)MK_FP((unsigned)bpval, 
				                   (unsigned)sival);   
			cprintf("\r\nUnknown Critical error: %d\r\n", 
					(int)errorcode&0xFF);
			switch (errorcode&0xFF)	 {
			  case 0: 
				cprintf("Attempted to write to a write protected disk\r\n");
				break;
			  case 1:
				cprintf("Unknown unit (source of error not known)\r\n");
				break;
			  case 2:
				cprintf("Drive not ready\r\n");
				break;
			  case 3:
				cprintf("Unknown command\r\n");
				break;
			  case 4:
				cprintf("CRC indicates error in data\r\n");
				break;
			  case 5:
				cprintf("Length of <drive request structure> is bad\r\n");
				break;
			  case 6:
				cprintf("Seek error\r\n");
				break;
			  case 7:
				cprintf("Unknown media type\r\n");
				break;
			  case 8:
				cprintf("Sector not found\r\n");
				break;
			  case 9:
				cprintf("Printer out of paper\r\n");
				break;
			  case 10:
				cprintf("Write fault\r\n");
				break;
			  case 11:
				cprintf("Read fault\r\n");
				break;
			  case 12:
				cprintf("General failure\r\n");
				break;
			  }

			  if(!deverror&0x1000) 
				cprintf("It's a disk error on %d\r\n",(int)deverror&0xFF);
			  else
				{
				cprintf("Error in character device - ");
				if(!deverror&0x2000) cprintf("Ignore response not allowed\r\n");	
				if(!deverror&0x1000) cprintf("Retry response not allowed\r\n");	
				if(!deverror&0x0800) cprintf("Abort response not allowed\r\n");	
				if( deverror&0x0600==0x0000) cprintf("Occured in MSDOS\r\n");	
				if( deverror&0x0600==0x0200) cprintf("Occurred in FAT\r\n");	
				if( deverror&0x0600==0x0400) cprintf("Occurred in directory\r\n");	
				if( deverror&0x0600==0x0600) cprintf("Occurred in data area\r\n");	
				if( deverror&0x0100) cprintf("Write error\r\n");
				else                 cprintf("Read error\r\n");
				}	

			  if(!*(devhdr+4)&0x8000) cprintf("bad memory image of FAT\r\n");
			  else {
				if(*devhdr==1) cprintf("Error in std input\r\n");
				if(*devhdr==2) cprintf("Error in std output\r\n");
				if(*devhdr==4) cprintf("Error in null device\r\n");
				if(*devhdr==8) cprintf("Error clock device\r\n");
				}
			cprintf("THIS ERROR IS SERIOUS - PLEASE TERMINATE PCTOR IMMEDIATELY\r\n");
			cprintf("Hit any key to resume\r\n");
			getch();		

        		if(!puttext(left,top,right,bottom,tbuffer))
                				terminate_prog("Puttext - restoring text");
        		free(tbuffer);
			textcolor(c_CHCOLOR);
			textbackground(c_BGCOLOR); 
   			_setcursortype( _NOCURSOR );
			current_window = activate_window(2);  
			gotoxy(x2,y2); 	// leave cursor as before interrupted
			current_window = activate_window(3);  
			gotoxy(x3,y3);	// leave cursor as before interrupted

			while(kbhit()) getch(); // Flush keyboard 
			hardretn(SOFTERR_RETN);
		}
	return(0);
}


//---------------------------------------------------------------------------
void show_help()
{
int bufsize, left=5,right=75,top=1,bottom=25;
char *tbuffer, ch;
int x,y;

	save_xy(&x,&y);
   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for help screen

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(
" HELP WITH SYSTEM KEYS ͻ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
" Hit any key to return to main display ͼ");

        window(left+1,top+1,right-1,bottom-2);
        clrscr();
	cprintf("\r\n");
        cprintf("\r\nF1 - CALL FEC     F2  - CALL ARQ         F1  - TRANSMIT");
        cprintf("\r\nF3 - FORCE OVER   F4  - QRT              F4  - RECEIVE");
        cprintf("\r\nF6 - ARQ LISTEN   F10 - FORCE STBY       F10 - RESET");
        cprintf("\r\nF5 - Timestamp");
        cprintf("\r\nF7 - CLR Kbrd     SHFT.F7 - CLR Rcv (buffers/screen)               ");
        cprintf("\r\nF8 - DOSSHELL");
        cprintf("\r\nALT-B Buffer send");
        cprintf("\r\nALT-C Enter remote selcal and start ARQ");
        cprintf("\r\nALT-F File send");
        cprintf("\r\nALT-I Toggle receive polarity");
        cprintf("\r\nALT-O Toggle Printer ON/OFF");
        cprintf("\r\nALT-R Enter remote selcal without ARQ");
        cprintf("\r\nALT-S Send file continuously");
        cprintf("\r\nALT-T Enter UTC time/date into keyboard buffer                     ");
        cprintf("\r\nALT-W Send 'WRU' to remote (ITA2 Figs-D)                           ");
        cprintf("\r\nALT-1 - ALT-9 Send user buffer                                     ");
        cprintf("\r\nEnd - send +?     Ins - send Callsigns                             ");
        cprintf("\r\nDel - Force LTRS  Ctl_Hm - Dump ITA2 in hex \r\n");
	clreol();

	ch=getch();        	// Flush character
	if(ch==0) getch(); 	// if yes, then flush it	

        clrscr();
	cprintf("\r\n   - DISPLAY SYMBOLS-");
	cprintf("\r\n");
        cprintf("\r\n                           ");
	cprintf("\r\n   - logging              ");     
        cprintf("\r\n   - reliability          ");
        cprintf("\r\n   - UOS                  ");
        cprintf("\r\n   - MARQ                 ");
        cprintf("\r\n  - Rx/Tx cases          ");
	clreol();

	ch=getch();      // Flush character
	if(ch==0) getch();

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
   	_setcursortype( _NOCURSOR );

	gotoxy(x,y);			// leave cursor as before interrupted

}

//---------------------------------------------------------------------------
// Hook interrupt for 1000 Hz interrupt rate
//
void speedup_AMT()
{
unsigned int mask;

	asm {CLI};			// prevent any interrupts
					// interrupts will be turned on
					// in -_t_speedup() 


	if (irq<8) {
		oldfunc=getvect(irq+0x08);  // Hook interrupt on PC side
		setvect(irq+0x08,INTRP);
 					// deal with the first 8259
		old_IMR=inp(0x21);	// get IMR
		mask=old_IMR; 		// get the starting interupt mask
		mask&=~(0x0001<<irq); 	// clear the appropriate mask bit
		outp(0x0021,mask); 	// write out the new mask
		}
	else {
		oldfunc=getvect(irq+0x68);  // Hook IRQ on PC side
		setvect(irq+0x68,INTRP);
 					// deal with the second 8259
		old_IMR=inp(0x00a1); 	// get the starting interupt mask
		mask=old_IMR; 		// get the starting interupt mask
		mask&=~(0x0001<<(irq-8)); // clear the appropriate mask bit
		outp(0x00a1,mask);
		}

	_t_speedup();

// These items have been initialized but undone during _t_speedup
	set_IDENT(ident);
	set_SELCAL(selcal);
        set_UL_case(UL_case);
	set_txdelay(tx_delay);
	set_cd_delay(cd_delay);
	set_reliability(Hi_rely);
	set_timing(timing_value);		// Setup timing value
	set_mem_ARQ(Mem_ARQ);			// Setup memory ARQ
	Adjust_Sound_output_level(Sound_out_level);
	Adjust_Sound_input_level(Sound_in_level);
	set_rxmask(Mask_in);
	set_txmask(Mask_out);
}
//---------------------------------------------------------------------------
// When done, remove our interrupt handler
// do this while the interrupts are turned off
void close_down()
{
int i;

	asm {CLI};			// prevent any interrupts 

	//ackInt();			// Clear any pending interrupts

	if (irq<8) { 			// deal with the first 8259
		outp(0x21,old_IMR);
		setvect(irq+0x08, oldfunc);	// Restore INT
		}
	else {
		outp(0x00A1,old_IMR);	// Restore IMR
		setvect(irq+0x68, oldfunc);	// Restore INT
		}

	asm {STI};			// enable interrupts 

	for(i=0;i<20;i++) set_control(0x06);	// Reset interface

   	_setcursortype( _NORMALCURSOR );

	close_port();
}
//--------------------------------------------------------------------------
//
// Shell out to Pactor
//
void pactor_program()
{

int bufsize, left=1,right=80,top=1,bottom=25;
char *tbuffer, timing[5], logopt[5];
int i, x, y;
	
	save_xy(&x,&y);				      // save current window
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");


   	_setcursortype( _NORMALCURSOR );
	normvideo();
        window(1,1,80,25);
        clrscr();

	close_down();

// Close logging file
	if(Log_option) {
		 time(&bintime);   
	  	if(Time_Type!=GMT) curtime=localtime(&bintime); 
	  	else curtime=gmtime(&bintime); 	// UTC
	  	fprintf(logfile,"\r\n---->Shell to DOS at: %s", asctime(curtime));	
		fclose(logfile);   		// Close all files
	       	}


     	sprintf(timing,"%d",timing_value);
     	sprintf(logopt,"%d",Log_option);
	if((spawnl(P_WAIT, 
		"PACTOR.exe",
		"PACTOR.exe",
		timing,
		&ident[5],
		&selcal[5],
		logopt,
		logfilename,
		NULL))!=0) switch (errno) {
			case ENOENT: printf("PC-PacTOR: NO such file or directory\r\n");
			     	     break;
			case ENOMEM: printf("PC-PacTOR: Not enough memory\r\n");
			     	     break;
			case E2BIG:  printf("PC-PacTOR: Bad argument list\r\n");
			     	     break;
			case ENOEXEC:printf("PC-PacTOR: EXEC format error\r\n");
			     	     break;
		}

// Reopen logging
	if(Log_option) {
		if((logfile=fopen(logfilename,"a"))==NULL)
                terminate_prog("Fopen - cannot re-open log file");
		time(&bintime);   
		if(Time_Type!=GMT) curtime=localtime(&bintime); 
		else curtime=gmtime(&bintime); 	// UTC
		fprintf(logfile,"\r\n---->Return from DOS at: %s", asctime(curtime));	
		}

	speedup_AMT();			// Restart TORBIOS

// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);

	_setcursortype( _NOCURSOR );
	textcolor(c_CHCOLOR);
	textbackground(c_BGCOLOR); 
	gotoxy(x,y);
}

//---------------------------------------------------------------------------
// This adjusts the level of the mike in
void Adjust_Sound_input_level(int level)
{
}
//---------------------------------------------------------------------------
// This adjusts the level of the speaker level
void Adjust_Sound_output_level(int level)
{
}
//---------------------------------------------------------------------------
#ifdef SHAREWARE

void shareware_banner()
{

// Put up frame
        window(1,1,80,25);
        clrscr();   

// Fill in the text
        gotoxy(26,10); cputs("  AMTOR56  VERSION");
	               cputs( version );
        gotoxy(22,11); cputs("(c) J.B.Forrer (KC7WW) 1991,1992,1993");
        gotoxy(26,12); cputs("  26553 Priceview Drive    ");
        gotoxy(26,13); cputs("  Monroe, OR 97456, USA    ");
        gotoxy(26,16); cputs("UNREGISTERED COPY ");
                       cputs( serial  );
        gotoxy(26,24); cputs("Hit ESC key to continue    ");
        while(getch()!=0x1B);
}
#endif



//  Member functions for Circ_buffer class

char Circ_buffer::Read_buf()	  	// Read buffer
{
char ch;

	if (read_from_ptr == store_at_ptr) return (char)0xFF;
	
	ch = *read_from_ptr++;
	if (read_from_ptr>=(buffr+buf_size)) 
	 	{
	 	read_from_ptr=buffr;
	 	}
	present_size--;
	return ch;
}

char Circ_buffer::Peek_buf()	  	// Peek into buffer
{
char ch;

	if (read_from_ptr == store_at_ptr) return (char)0xFF;   // Empty ?
	
	ch = *read_from_ptr;
	return ch;
}


int Circ_buffer::Store_buf( char ch)	// Write buffer
{
// Dont accept the character if buffer overrun will occur
        if ((store_at_ptr==buffr+buf_size-1)&&(read_from_ptr==buffr))
                return 1;
        if (store_at_ptr==read_from_ptr-1)
                return 1;
	*store_at_ptr++ = ch;
	if (store_at_ptr>=(buffr+buf_size)) store_at_ptr=buffr;
	present_size++;
	return 0;
}

unsigned char Circ_buffer::Back_space()	// Remove most recent character
{
unsigned char ret;

        if (store_at_ptr==read_from_ptr)   // anything left ??
                return (char)0xFF;         // no
	if(store_at_ptr==buffr) ret=buffr[buf_size];
	else ret=(*(store_at_ptr-1)); 	   // return present value
                                           // handle possible wrap
                                           // arounds
	if (store_at_ptr==buffr) store_at_ptr=buffr+buf_size;
	else store_at_ptr--;
	present_size--;
	return ret;
}

void Circ_buffer::Flush_buf()           // Flush buffer
{	
	   store_at_ptr = read_from_ptr = buffr;
	   present_size=0;
}

int Circ_buffer::Present_size()           // Retrieve buffer size
{	
	   return present_size;
}

