; $Header: E:/HAM/EVM/RCS/firmacs.asm 1.4 1997/02/23 20:21:50 dbraun Exp $
; $Log: firmacs.asm $
; Revision 1.4  1997/02/23 20:21:50  dbraun
; Added macro for IIR 1st order high-pass filter
;
; Revision 1.3  1997/01/31 09:30:06  dbraun
; Added one more macro definition.
;
; Revision 1.2  1997/01/28 15:56:12  dbraun
; Cleaned up a little
;
; Revision 1.1  1997/01/28 15:48:30  dbraun
; Initial revision
;
;

; Macros to generate various kinds of filter FIR coefficients.
; Gathered together from various 56002 programs.
 

; These are intended to be local to this file...
PI_ = 3.14159265358979323846
EX_ = 2.718281745911


;******************************************************************
; This should have a more descriptive name...

BasicBandpassFilter	macro	Length,LowFreq,UppFreq

sigma = @cvf(Length)/2.0/3.0
time = -@cvf(Length/2)+0.5
count = 0
        dup Length
Window = @pow(EX_,-time*time/(2.0*sigma*sigma))
Filter = (@sin(2.0*PI_*time*UppFreq)-@sin(2.0*PI_*time*LowFreq))/(PI_*time)
time = time+1.0
        dc  Window*Filter
count = count+1
        endm

	endm


;******************************************************************
; A different windowing function

CosineWindowBandpassFilter	macro	Length,LowFreq,UppFreq

time = -@cvf(Length/2)+0.5
count = 0
	dup Length
Window = 0.75*@cos((time/@cvf(Length))*PI_)+0.25*@cos(3*(time/@cvf(Length))*PI_)
Filter = (@sin(2.0*PI_*time*UppFreq)-@sin(2.0*PI_*time*LowFreq))/(PI_*time)
time = time+1.0
	dc Window*Filter
count = count+1
	endm

	endm




w0_      equ    0.35875 ;coeff. for Blackman-Harris 4 term
w1_      equ    0.48829 ;minimum sidelobe window (copied from Motorola BBS)
w2_      equ    0.14128
w3_	equ 	0.01168


;******************************************************************
BlackmanHarrisFilterI	macro	Length,LowFreq,UppFreq

time = -@cvf(Length/2)+0.5
count = 0
        dup Length
angle = PI_*time/@cvf(Length/2)
Window = w0_+w1_*@cos(angle)+w2_*@cos(2.0*angle)+w3_*@cos(3.0*angle)
Filter = (@sin(2.0*PI_*time*UppFreq)-@sin(2.0*PI_*time*LowFreq))/(PI_*time)
time = time+1.0
        dc  Window*Filter
count = count+1
        endm

        endm


;******************************************************************
BlackmanHarrisFilterQ	macro	Length,LowFreq,UppFreq

time = -@cvf(Length/2)+0.5
count = 0
        dup Length
angle = PI_*time/@cvf(Length/2)
Window = w0_+w1_*@cos(angle)+w2_*@cos(2.0*angle)+w3_*@cos(3.0*angle)
Filter = (-@cos(2.0*PI_*time*UppFreq)+@cos(2.0*PI_*time*LowFreq))/(PI_*time)
time = time+1.0
        dc  Window*Filter
count = count+1
        endm

	endm



; first order high-pass IIR filter
; input in "acc" (a or b), output in same register
; "store" is an X: memory location.
; frq is -3dB point (fc/fs)

hpass	macro	acc,frq,store
c1	set	-1.0/@cos(6.28319*frq)*(1.0-@sin(6.28319*frq))
	move		    acc,x0
	move		    #>(1.0-c1)/2.0,y0
	mpyr	x0,y0,acc   x:<store,y0
	move		    acc,x:<store
	sub	y0,acc	    x:<store+1,x0
	move		    #>c1,y0
	macr	-x0,y0,acc
	move		    acc,x:<store+1
	endm

