	title "AMTOR56/PACTOR56 modem written by N1OWU with some ideas from KC7WW"
	page    132,60

;
; $Header: E:/HAM/EVM/RCS/tor2.asm 1.4 1997/07/31 10:21:56 dbraun Exp dbraun $
;
; $Log: tor2.asm $
; Revision 1.4  1997/07/31 10:21:56  dbraun
; Added explanitory comments.
;
; Revision 1.3  1997/07/27 11:07:27  dbraun
; Cleaned up amd made the "official" amtor56/pactor56 modem.
;
;
; Revision 1.1  1997/02/20 17:08:28  dbraun
; Initial revision
;
;


; This is a modem for the AMTOR56 and PACTOR56 programs, which
; are modified versions of PC_PACTOR and PSATOR, oringinally written by
; Johan Forrer, KC7WW.  The basic theory of operation is this:
;
; The demodulator makes a soft mark/space decision 1000 times per second,
; and sends a data byte (-128 to 127) out the serial port 1000 times
; per second at the standard 19200 baud rate.  Also 1000 times per second
; it expects to receive a command byte from the host program that tells
; it whether to transmit a mark or a space tone, or not to transmit.
; The advantage of this protocol is that the "standard" EVM serial
; port connections can be used, and no rewiring is necessary.
;
; Note that the host program's timing is derived from the rate of
; the received byte stream.
;
; The command "opcodes" sent from the host program are as follows:
;
; xxxxx110    Transmit a mark
; xxxxx101    Transmit a space
; xxxxx0xx    Do not transmit
;
; Note that some more opcodes should be implemented to switch
; between 100 and 200 baud demodulators, and to reset the modem.
; The assembly language modules of amtor56 and pactor56 have the 
; definitions of the opcodes that these programs send.
;
; There also ought to be a watchdog timer that turns of the transmitter
; if an opcode has not been received for several milliseconds (e.g
; if the host program crashed).


	NOLIST
	include 'leonid.asm'
	LIST

	opt	mu


;******************************************************************************

UpLED   macro mode      ;UP line (a red LED connected)
        b\mode #1,X:$FFE4
        endm

DownLED macro mode      ;DOWN line (a red LED connected)
        b\mode #2,X:$FFE4
        endm

YellowLED macro mode    ;CAT line (a yellow LED connected)
        b\mode #3,X:$FFE4
        endm

DCDLed  macro mode      ; DCD (Green) LED clr/set/chg
        b\mode #13,X:$FFE4
        endm

;******************************************************************************

; Important constants
sampleFreq equ	16000.0
decimationRatio	equ 2
decimatedFreq	equ sampleFreq/decimationRatio

markFreq   equ  2125.0
shift	   equ	170.0

spaceFreq  equ	markFreq+shift
centerFreq equ  markFreq+(shift/2.0)

inputFilterSize   equ     48    ;Number of taps in input bandpass filter.

channelFilterSize equ	  80	;Number of taps in channel filters. Ideally,
				;the length of the filter should be the length
				;of one bit.  Since this is a multi-speed
				;modem, we will make it the longest bit length
				;(8000/80 = 100mS = 100BPS).  For the 200BPS
				;filters, half of the coefficients should
				;be zeros.

buflen	equ	8

PTTRightChannel equ 1

; Various parameters for the DCD algorithm.  Note that these are only
; used for turning the LED on and off.  They do not affect the data
; decoding at all...

; Amount of mS to turn off DCD when we see a false peak
DCDpeakLimit equ	30

; Any peak below this magnitude is considered a false (non-data) peak.
DCDpeakThreshold equ	0.4

; amount of mS that may pass between zero crossings before DCD is turned off.
DCDzeroLimit  equ	64


      if PTTRightChannel
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #4,X:$FFE4
        endm
      else
PTT     macro mode      ;PTT line: clr/set/chg
        b\mode #0,X:$FFE4
        endm
      endif

;******************************************************************************
;************ MAIN PROGRAM ****************************************************

	org     P:user_code
main

        ori #%00000100,omr      ;enable the sine ROM table
                                ;this is for the IQ routine

;---------------------------------------------------------------------------
; Initialize pointer registers, r2, r7
; registers r0, r1, r4, and r5 are used as temporaries
;---------------------------------------------------------------------------

        move    #buffer+2,r7  ; r7 & m7 point to desired sample buffer
        move    #buflen*4-1,m7

        move    #buffer,r2
        move    #buflen*4-1,m2

;
; Program headphone,lineout enabled, speaker off, attn=0
;
	ctrlcd	1,r2,buflen,MIC,3.0,3.0,LINEO|HEADP,0.0,0.0
        opencd  sampleFreq/1000.0,HPF


;--------------- Main Loop -------------------------------------------------
main_loop 
	.loop	#8		; one data byte per 8*2 samples

	.loop	#decimationRatio	; decimation ratio is 2

	waitblk r2,buflen,1		; wait for one sample

	move	X:<inputFilterPtr,r0
        move    #inputFilterSize-1,m0

	move    X:(r2)+,a 	        ; Load the left channel input.
	move	a,X:LEFT_sig

	move    X:(r2),a     		; Load the right channel input.
	move	a,X:RIGHT_sig

	move	a,X:(r0)+		; save in inputFilter buffer
	move	r0,X:<inputFilterPtr

	; Transmitter part

	move X:<TxCarFreq,a   ;update the instant carrier frequency
	move X:<TxCarPhase,x0 ;get the carrier phase
	add x0,a              ;and advance it
	move a1,X:<TxCarPhase
	jsr <IQ               ;compute the sin/cos (modifies a,b,x,y,r0,m0,n0)

	; Mute if not transmitting
	PTT tst
	.if <cc>
	  clr a
  	.endi

	; TODO: LPF it

	asr a		      ; scale it
	move a,Y:(r2)+        ; Put it to the left channel TX output.
	asr a		      ; reduce its volume some more
	asr a
	move a,Y:(r2)	      ; Put to right channel audio monitor output.

;----------------------------------------------------------------------------
; Output RX signal for monitoring
;----------------------------------------------------------------------------

	move    Y:(r2),x0	        ; Get current (right channel) TX sig
	move	X:RIGHT_sig,a		; Get RX signal
	add	x0,a			; mix them
	move    a,Y:(r2)+	        ; Put value back in right channel tx.

	move	(r2)+			; R2 gets incremented 4 times

	.endl   ; #DecimationRatio

;---------------------------------------------------------------------------
; Perform input bandpass + I/Q filtering
; Input sample rate = 16000 SPS 
; Output sample rate = 8000 SPS  (since we are decimating)
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
;  Do I and Q channel filtering together in one loop
;---------------------------------------------------------------------------
        move    #inputFilterSize-1,m0
	move	X:<inputFilterPtr,r0	; R0 points to data

        move    #>-1,m4     		; linear addressing
        move    X:<curInputFilterI,r4  	; R4 points to I filter coefficients
        move    #>-1,m5     		; linear addressing
        move    X:<curInputFilterQ,r5  	; R5 points to Q filter coefficients
	nop

        clr     a                Y:(r4)+,y0     ; priming registers
        clr     b    X:(r0)+,x0  Y:(r5)+,y1
        .loop	#inputFilterSize-1			; allow interrupts
	mac     x0,y0,a              Y:(r4)+,y0
	mac     x0,y1,b  X:(r0)+,x0  Y:(r5)+,y1
	.endl
        macr    x0,y0,a				; final rounding
        macr    x0,y1,b				; I/Q values are in a/b

        move    #channelFilterSize-1,m0
	move	X:<channelFilterIPtr,r0
	nop
	move	a,X:(r0)+          ;save a in I buffer
	move	r0,X:<channelFilterIPtr

        move    #channelFilterSize-1,m0
	move	X:<channelFilterQPtr,r0
	nop
	move	b,X:(r0)+          ;save b in Q buffer
	move	r0,X:<channelFilterQPtr

	.endl  ; #8

;---------------------------------------------------------------------------
; Perform discriminator channel filtering
; Input sample rate = 8000 SPS   Output rate = 1000 SPS
; We do a total of 4 filtrations: mark I, mark Q, space I, space Q
;---------------------------------------------------------------------------

	;--------------------------------------------
	; Do I/Q filtering of mark channel
	;--------------------------------------------
        move    #channelFilterSize-1,m0
        move    #channelFilterSize-1,m1
        move    #>-1,m4     		; linear addressing

	move	X:<channelFilterIPtr,r0	; R0 points to I data
	move	X:<channelFilterQPtr,r1	; R1 points to Q data
        move    X:<curMarkFilter,r4     ; R4 points to mark filter coefficients
        nop

        clr     a    X:(r0)+,x0 
        clr     b    X:(r1)+,x1  Y:(r4)+,y0
        .loop	#channelFilterSize-1		; allow interrupts
        mac     x0,y0,a    X:(r0)+,x0 
        mac     x1,y0,b    X:(r1)+,x1  Y:(r4)+,y0
	.endl
        macr    x0,y0,a			; final rounding
        macr    x1,y0,b			; I/Q mark values are in a/b

        ; Now we have a filtered I/Q value in a/b
        ; Get the magnitude of the I/Q data: sqrt(I*I+Q*Q)

        jsr	IQmag
	move	b,X:<mark_sig


	;--------------------------------------------
	; Do I/Q filtering of space channel
	;--------------------------------------------
        move    #channelFilterSize-1,m0
        move    #channelFilterSize-1,m1
        move    #>-1,m4     		; linear addressing

	move	X:<channelFilterIPtr,r0
	move	X:<channelFilterQPtr,r1
        move    X:<curSpaceFilter,r4   ; R4 points to space filter coefficients
        nop

        clr     a    X:(r0)+,x0 
        clr     b    X:(r1)+,x1  Y:(r4)+,y0
        .loop	#channelFilterSize-1		; allow interrupts
        mac     x0,y0,a    X:(r0)+,x0 
        mac     x1,y0,b    X:(r1)+,x1  Y:(r4)+,y0
	.endl
        macr    x0,y0,a			; final rounding
        macr    x1,y0,b			; I/Q space values are in a/b

        ; Now we have a filtered I/Q value in a/b
        ; Get the magnitude of the I/Q data: sqrt(I*I+Q*Q)

        jsr	IQmag
	move	b,X:<space_sig

;---------------------------------------------------------------------------
; Now use the mark and space magnitudes to derive a data value
;---------------------------------------------------------------------------

	; The demodulated data value will be: 
	;  (|M| - |S|) / (|M| + |S|)
	; 
	; This gives us a value that is independent of the input signal level

	move	X:<mark_sig,x0
	move	X:<space_sig,a
	add	x0,a
	.if <eq>
	  ; zero total amplitude, so avoid doing 0/0 and the data value is 0
	  clr a
  	.else
  	  move a,x0    ; sum goes in x0 to be divisor

  	  move X:<space_sig,a    ; get the difference
	  move X:<mark_sig,x1
	  sub x1,a
	  move a,x1  ; save difference
	  abs a	     ; abs value of difference goes in a to be dividend

	  ; do an unsigned division

	  and #$FE,CCR  ; clear carry
	  rep #$18
	    div x0,a
	  bclr #23,a0      ; now the (positive) quotient is in a0

	  move a0,a
	  btst #23,x1   ; test sign bit of original unscaled difference
	  .if <cs>
	    neg a	; if set, negate the scaled difference
    	  .endi
  	.endi

	move	a,X:<data_val

	btst #0,X:<spy
	.if <cs>
	  jsr <SpyA
  	.endi

        ; Set/reset LEDs based on mark-space decision
        ; If the signal is sufficiently high, turn on the Mark LED.
        ; If it is sufficiently low, turn on the Space LED, otherwise
        ; leave them both off.

	UpLED clr
        DownLED clr
        move	#>0.4,x0
        cmp	x0,a
        .if <gt>
          UpLED	set
  	.else
	  move	#>-0.4,x0
	  cmp	x0,a
	  .if <lt>
	    DownLED set
	  .endi
  	.endi

	move	a,X:<data_val

	; The data will be in the range from -0.99999 to 0.99999.
	; Scale it to fit in mid 8 bits of a1.
	rep	#8
	asr	a


	; Further scale it to the lowest 8 bits of a1
	rep	#8
	asr	a


	; Put byte (-128..127) of data to serial port
	move	a1,x0

	btst #0,X:<spy
	.if <cc>
	  putc 
	.endi


	; Produce a DCD signal for the DCD LED.
	; We make two tests.

	; First test: based on peaks in data values

	; Decide if previous data value represents a peak, i.e. 
	; a local max or min value.
	move	X:<data_val,a
	move	X:<prev_data_val,b
	sub	b,a		; A has D(n)-D(n-1)
	move	a,x0
	move	X:<prev_data_diff,x1	; X1 has D(n-1)-D(n-2)
	mpy	x0,x1,a			; a negative product implies a peak

	; Update variables for next time
	move	x0,X:<prev_data_diff 

	.if <mi>	; test result of multiplication
	  ; If the previous data point is a peak, see if its magnitude is
	  ; below a threshold.  If so, there is no data present.
	  move	#>DCDpeakThreshold,x0
	  cmpm	x0,b
	  .if <lt>
	    ;Mark the test as failed for a while
	    bset #0,X:<DCDflag
	    move #>DCDpeakLimit,x0
	    move x0,X:<DCDpeakCount
	  .endi
        .endi

	; See if the DCD has been off long enough
	clr a
	move X:<DCDpeakCount,a0
	tst a
	.if <eq>
	  bclr #0,X:<DCDflag  ; mark the test as passed
	.else
	  dec a
	  move a0,X:<DCDpeakCount
	.endi

	; Second DCD condition:  Turn off DCD if it's been
	; too long since a zero crossing.  An AMTOR signal
	; has bit transitions at least every six bits.
	; This test keeps a CW carrier on the mark
	; or space frequency from turning on the DCD.

	move	X:<data_val,x0
	move	X:<prev_data_val,x1
	mpy	x0,x1,a
	.if	<pl>
	  clr a
	  move X:<DCDzeroCount,a0
	  ; No zero crossing. Decrement the watchdog counter.
	  tst a
	  .if <ne>
	    dec a
	  .else
	    ; Already timed out, mark the test as failed
	    bset #1,X:<DCDflag
	  .endi
	.else
	  ; There is a zero crossing, so restart the watchdog timer,
	  ; and mark the test as passed.
	  move	#>DCDzeroLimit,a0
	  bclr #1,X:<DCDflag
	.endi
	move a0,X:<DCDzeroCount

	move	X:<DCDflag,a
	tst	a
	.if <eq>
	  DCDLed set
	.else
	  DCDLed clr
	.endi

	; Update variables for next time
	move	X:<data_val,x0
	move	x0,X:<prev_data_val

	; Process any incoming opcode bytes from the serial port

	btst #0,X:<spy
	.if <cc>
	  lookc	0	; flush any input

	.if <cc>	; there is a opcode byte

	  ; Test for things to do

	  btst	#2,x0	; TX/RX bit
	  .if <cs>
	    PTT	set

	    btst #1,x0	; TX-mark bit
	    .if <cs>
	      move X:<curMarkFreq,x0
	      move x0,X:<TxCarFreq
	    .endi

	    btst #0,x0	; TX-space bit
	    .if <cs>
	      move X:<curSpaceFreq,x0
	      move x0,X:<TxCarFreq
	    .endi

	  .else
	    PTT	clr
	  .endi


	.endi

	.endi

	jmp	main_loop


PI      equ     3.14159265358979323846

;this routine computes a cosine/sine pair using the sine ROM
;with a second order (linear+quadrature) approximation between table points
IQ                              ;x0 = angle ( -1 = -PI, +1 = +PI)
        move #>$80,x1   ;shift out 8 most significant bits
        mpy x0,x1,a  #>$FF,x0
        move x0,m0
        and x0,a     #>$100,x0
        or x0,a      #<$40,n0
        move a1,r0      ;put the 8 most significant bits into r0 with offset = $100
        move a0,y0      ;save the remaining bits in y0
        jclr #23,y0,SinTable_lev2
          move (r0)+
SinTable_lev2
        move Y:(r0+n0),x0       ;x0 = coarse cosine
        move Y:(r0),x1          ;x1 = coarse sine
        mpyr x1,y0,a  #PI/256.0,y1
        tfr x0,a  a,x1
        macr -x1,y1,a           ;a = fine cosine
        mpyr x0,y0,b  Y:(r0),x1
        tfr x1,b  b,x1
        macr x1,y1,b  #PI*PI/2.0/65536.0,y1  ;b = fine sine
        mpyr y0,y0,a  a,x0
        move a,y0
        mpyr y0,y1,a
        tfr x0,a  a,y1
        macr -x0,y1,a  b,x1     ;a = super fine cosine
        macr -x1,y1,b           ;b = super fine sine
        rts                     ;x,y are modified
                                ;r0,m0,n0 are modified
                                ;maximum error is about 0.7E-6
                                ;execution time 4+64+4 clock cycles
                                ;including "jsr <IQ_lev2" and "rts"


; IQmag calculates the magnitude of a complex signal in a/b:
;     sqrt(I*I+Q*Q)

IQmag
	move	a,x0
	mpy	x0,x0,a
	move	b,x0
	mac	x0,x0,a  ; A = (I*I) + (Q*Q)  w/ 48-bit precision
	move    a0,y0
	move    a1,y1
	jsr	sqrt3    ; square root in B
	rts


; Full Precision Square Root by Polynomial Approximation. 
; Motorola DSP Operation, 6501 Wm. Cannon Drive W., Austin, Tx., 78735.
; Last Update 09 Feb 87   Version 1.0
;
; Full 23 bit precision square root routine using
; a successive approximation technique.
;
; y  = double precision (48 bit) positive input number
; b  = 24 bit output root
;
; a  = temporary storage
; x0 = guess
; x1 = bit being tested
; y1:y0 = input number
 
sqrt3
        clr     b    #<$40,x0           ; init root and guess
        move                    x0,x1   ; init bit to test
        do      #23,_endl
                                        ;START OF LOOP
        mpy     -x0,x0,a                ; square and negate the guess
        add     y,a                     ; compare to double precision input
        tge     x0,b                    ; update root if input >= guess
        tfr     x1,a                    ; get bit to test
        asr     a                               ; shift to next bit to test
        add     b,a             a,x1            ; form new guess
        move                    a,x0            ; save new guess
_endl                                    ;END OF LOOP
        rts



SpySync move a10,L:<SpySave     ;output: carry=1 => no spy request
        move a2,X:<SpySave+1    ;carry=0 => spy request !
        move x0,Y:<SpySave+1    ;512 words (jsr <SpyA) should follow
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <SpyCont
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        ori #$01,ccr
        jne <Spy_end
        move #>'P',x0
        putc
        move #>0,x0      ; send a 0 meaning non-complex data
        putc
        move #>512,a
        move a,X:<SpyCount
SpyCont andi #$FE,ccr
        jmp <Spy_end

SpyA    move a10,L:<SpySave
        move a2,X:<SpySave+1
        move x0,Y:<SpySave+1
        move x1,Y:<SpyCount
        move X:<SpyCount,a
        tst a
        jne <Spy_copy

Spy_check
        lookc 0
        jcs <Spy_end
        move #>'S',a
        cmp x0,a
        jne <Spy_end
        move #>'P',x0
        putc
        move #>0,x0      ; send a 0 meaning non-complex data
        putc
        move #>512,a
Spy_copy
        move #>1,x0
        sub x0,a
        move a,X:<SpyCount

        move X:<SpySave,a
        rep #8
          lsr a
        move a1,x0
	putc
        move X:<SpySave,a
        rep #16
          lsr a
        move a1,x0
        putc

Spy_end move L:<SpySave,a10
        move X:<SpySave+1,a2
        move Y:<SpySave+1,x0
        move Y:<SpyCount,x1
        ;rts
        rti		;N1OWU: the RTI instruction restores the SR


;*****************************************************************************
; Data definitions
;-----------------------------------------------------------------------------
        LOMEM X:$0000,Y:$0000,L:$0000
        HIMEM X:$00FF,Y:$00FF,L:$00FF

        org L:user_data

SpySave dc 0,0
SpyCount dc 0
SpyTmp   ds 1

LastL	= *
	org     X:LastL
	org	Y:LastL

	org	X:

spy		dc	0	;runtime-patchable SPY switch
LEFT_sig        ds      1       ; storage for Left Signal
RIGHT_sig       ds      1       ; storage for Left Signal
mark_sig 	ds	1
space_sig	ds	1
data_val	ds	1	; demodulated data value (-.999 to .999)
prev_data_val   ds	1	; previous one
prev_data_diff  ds	1	; difference between previous one and one before that
DCDpeakCount	dc	0	; amount of data points before DCD is turned on
DCDzeroCount	dc	0	; amount of data points before DCD is turned on
DCDflag		dc	0     ; DCD status bits


; Pointers to the sets of filter taps currently
; in use.  If we get a command to change modems
; (e.g. from 100 baud amtor to 200 baud pactor),
; we reset these pointers.

curInputFilterI dc	amtorInputFilterI
curInputFilterQ dc	amtorInputFilterQ
curMarkFilter	dc	amtorMarkFilter
curSpaceFilter	dc	amtorSpaceFilter

; These point to the current locations of the filters' data buffers.

inputFilterPtr	dc	inputFilterBuf
channelFilterIPtr dc	channelFilterIBuf
channelFilterQPtr dc	channelFilterQBuf

; The transmitter's NCO data.

TxCarFreq	dc	2.0*markFreq/sampleFreq
TxCarPhase	dc	0

; These will be updated if a "change modem" command specifies
; a modem with different frequencies
curMarkFreq	dc	2.0*markFreq/sampleFreq
curSpaceFreq	dc	2.0*spaceFreq/sampleFreq

; Data buffers for the filters.

inputFilterBuf	dsm	inputFilterSize	; storage for input data  (circular buffer)

;-----------------------------------------------------------------------------
; Coefficient storage in RAM
;-----------------------------------------------------------------------------
	org	Y:

	include 'firmacs.asm'	; various filter generation macros




fl = (centerFreq-(2*shift))/sampleFreq
fh = (centerFreq+(2*shift))/sampleFreq
amtorInputFilterI    BlackmanHarrisFilterI	inputFilterSize,fl,fh
amtorInputFilterQ    BlackmanHarrisFilterQ	inputFilterSize,fl,fh

;----------------------------------------------------------------------------
        LOMEM X:$0200,Y:$0200,L:$0200
        HIMEM X:$37FF,Y:$37FF,L:$37FF

	org	L:$0200

buffer	dsm	buflen*4	;CODEC buffer

LastL	= *
	org     X:LastL
	org	Y:LastL

	org	X:

channelFilterIBuf  dsm	channelFilterSize ; storage for channel filters (circ buffer)
channelFilterQBuf  dsm	channelFilterSize ; storage for channel filters (circ buffer)


	org	Y:

;-----------------------------------------------------------------------------
; A macro to generate "matched" mark/space filters. 
;-----------------------------------------------------------------------------

MarkSpaceFilter	macro	Length,Freq

time = -@cvf(Length/2)+0.5
count = 0
        dup Length
Filter = @sin(2.0*PI_*time*Freq)*2/Length
time = time+1.0
        dc  Filter
count = count+1
        endm

	endm



; 
; f = markFreq/decimatedFreq
; amtorMarkFilter		MarkSpaceFilter		channelFilterSize,f
; 
; f = spaceFreq/decimatedFreq
; amtorSpaceFilter	MarkSpaceFilter		channelFilterSize,f
; 

; In practice, the above matched filters are very fussy to tune, since 
; their bandwidth is rather narrow.  The filter coeffs included below are
; empirically designed to give more of a discriminator-like output,
; where the effects of mistuning are more apparent as a DC skew in the
; output data level.

amtorMarkFilter
	include	'mark100.asm'

amtorSpaceFilter
	include	'space100.asm'


    end
