//    PSATOR - Library of routines for amateur radio AMTOR/PACTOR
//    Copyright (C) 1993-1995 Johan B. Forrer, KC7WW
//
//    This is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//---------------------------------------------------------------------------
//                            PC-PACTOR
//                     (c) Johan Forrer 1992,1993
//
// History:	Beta 0.02 Derived from 0.20
//		     0.05 Status line is updated in "Files" menu
//		     0.06 Full transmit/receive 
//		Alpha 0.01.1 More accurate timing (1/20000)
//		     	     Better slave phasing etc..
//		      0.01.2 Smallish improvements to slave phasing
//		      0.01.3 track ideal_CTRX via 8-stage average buffer
//                           new DSP CLOCK adjustment logic
//                    0.01.4 Compute Tom's RxT tracking parameter       
// 		      0.01.5 Use RxT for phasing corrections
//                           Full MARQ
//                           Echo sent data
//		      0.01.6 Allow for adaptive modem control
//		Beta  0.01   Minor changes to Master level 1 call
//                           Reset DSP on startup
//			     ON RQ blocks, it kept sending new data
//		      0.2 ??
//                    0.3    Add MASKOUT to invert RTS
//		      0.5    ESC615 & fix memory leak in BIOS
//		      0.6    Initial hangup due to uninitialized biostime
//---------------------------------------------------------------------------

#include <stdio.h>
#include <dos.h>
#include <ctype.h> 
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <alloc.h>
#include <string.h>
#include <time.h>
#include <bios.h>    

#include "pac56def.hpp"                   // various defines 


extern "C" void _p_speedup(void);
extern "C" void _p_slowdown(void);
extern "C" void set_control(int);
extern "C" void set_txdelay(int);
extern "C" void set_biostiming(int, int);
extern "C" void	set_USPEED(int);
extern "C" void	set_UREPEAT(int);
extern "C" void	set_MASKOUT(int);
extern "C" int get_rxdta(void);
extern "C" int get_echodta(void);
extern "C" int send_txdta(int);
extern "C" int get_phase_err(void);
extern "C" int get_phase_avg(void);
extern "C" int get_buf_stat(void);
extern "C" void set_calleeID(char *);
extern "C" void set_ourID(char *);
extern "C" int get_status(void);
extern "C" int get_mode(void);
extern "C" int get_sync(void);
extern "C" int get_speed(void);
extern "C" int get_RXsense(void);
extern "C" int get_TXsense(void);
extern "C" int get_err_cnt(void);
extern "C" int get_RQ_cnt(void);
extern "C" int get_CS(void);
extern "C" int get_CTRX(void);
extern "C" int get_CTTX(void);
extern "C" void set_com_port(int, int);
extern "C" int	get_mark_data(void);
extern "C" int	get_space_data(void);
extern "C" int  get_RxT(void);
extern "C" int  get_marq_cnt(void);
extern "C" int	flush_bios_buffers(void);

extern "C" void far interrupt INTRP(...);  // Prototype for new interrupt serv.

//#define SHAREWARE 0                     // For shareware versions
#define VERS " 0.05"
#define SERIAL_NUMBER "  S/N 0000"

#define BUFR_SIZE	1000
#define EOT		0x04

unsigned int _stklen=27000;

//  --------Some "GLOBALS"---------
//
int directvideo=DEFAULT_VIDEO;
int timing_value=DEFAULT_TIMING;
#define DEF_TXDELAY 25;
int tx_delay=DEF_TXDELAY;
int U_Speed=DEFAULT_USPEED;
int U_Repeat=DEFAULT_UREPEAT;
long Refresh_rate=DEFAULT_REFRESH;
int Baudrate;
int Log_option;
int Files_CRLF=DEFAULT_FILECRLF;
int Time_Type=DEFAULT_TIME_TYPE;
int Word_Mode=DEFAULT_WORD_MODE;
int COM_prt=DEFAULT_COM_PORT;
int irq=DEFAULT_IRQ;
int combase=DEFAULT_COMBASE;
int combaud=19200;
int Sound_out_level=DEFAULT_OLVL;
int Sound_in_level=DEFAULT_ILVL;
int Source=0;
int Pactor_Mdm=DEFAULT_PACTR_MDM;
int Dither=0;				// Dither=0-no,=1-yes
int Mask_out=0;				// Leave output mask unchanged

char logfilename[40], scrollback[20];

char version[]=VERS, serial[]=SERIAL_NUMBER;
int activate_window(int);
int x_ind[10], y_ind[10], current_window; 
int window_boundary=20;
int keybuf_size, keybuf_size_old;
int Mode, Mode_old;
char keyboard_buffer[BUFR_SIZE], *store_ptr, *read_ptr;
char callee_ID[9], our_ID[9];
int Data_Mode=0;			// ascii/huffman
int link_stat;				// Monitor - idle/tfc

long previous_biostime;			// controls refresh rate

//---------------HUFFMAN COMPRESSION----------------------------------------
typedef struct tree_node {		 /* The building blocks for the Huffman tree */
	unsigned int symbol;
	struct tree_node *child_0;
	struct tree_node *child_1;
	} NODE;

typedef struct code_lookup {	 /* The building blocks for the code lookup  */
	unsigned int code;
	int code_bits;
	char bit_array[16];
	} CODE;

NODE *root_node, *current_node;
CODE code_data[128];
//---------------HUFFMAN COMPRESSION----------------------------------------


void save_xy(int *, int *);		// save window x-y coordinates
void terminate_prog(char *);		// Catastrophic error handler
char command_mode(void);
void flush_keyboard_buffer(void);
void read_setup_data(FILE *); 
void display_cursor(void);
void store_character(char);
void set_timing(int, int);
void get_tuning_data(int*, int*);
void Adjust_Sound_output_level(int);
void Adjust_Sound_input_level(int);

FILE *logfile;
FILE *setup;
struct tm *curtime;

unsigned char old_IMR;
void interrupt (*oldfunc)(...);		// for storage of old int vector

main()
{

char ch, chh, *ptr;
int i, j, k, ch_data, stat, prev_count, blk_siz;
time_t bintime, previous_time;

void update_status_line(void); 
void files_menu(void);			// Send files window
void enter_selcal(void);		// Entering callsigns
void enter_identity(void);
char check_PACTOR_keyboard(void);	// Checking keyboard
void pactor_speedup();
void close_down();

#ifdef SHAREWARE	      	//-*******************************************
int bufsize, left,right,top,bottom;	// Save screen
char *tbuffer;
int x,y;
#endif				//-*******************************************


//---------------HUFFMAN COMPRESSION----------------------------------------
char inp_code[200];			/* Input buffer size */
char  time_string[80];
unsigned int symb;

/* Function prototypes */
void build_tree(void);
void build_code_array(void);
int test_bits(char);

unsigned int mask;  // For setting com port interrupt vector

//--------------------------------------------------------------------------
	build_code_array();		/* Build the lookup table */

	build_tree();			/* Build the Huffman tree */
//--------------------------------------------------------------------------

//
// Open the configuration file
//
   	if ((setup=fopen("pactor56.cnf", "r"))==NULL)
        	{
        	cprintf("\r\nNo setup file (PACTOR56.CNF) supplied\r\n");
        	cprintf("all parameters set to default values\r\n");
		clreol();
        	}

// Initialize the screen
	textbackground(DEFAULT_BGCOLOR); // Default screen colors
	textcolor(DEFAULT_CHCOLOR);

#ifdef SHAREWARE	      	//-*******************************************
void shareware_banner(void);
        shareware_banner();
#endif				//-*******************************************

	clrscr();
	gotoxy(1,2); cprintf("ͺ EVM-pacTOR (c) KC7WW");
	cprintf("%s",version); 
	cprintf(" ");
//	gotoxy(1,window_boundary); cprintf(
//"");
//
// Initialize some variables
//
	x_ind[0]=x_ind[1]=x_ind[2]=x_ind[3]=x_ind[4]=x_ind[5]=x_ind[6]=1;
	y_ind[0]=y_ind[1]=y_ind[2]=x_ind[3]=x_ind[4]=x_ind[5]=x_ind[6]=1;
	link_stat=0;
	time(&previous_time);
	current_window = 1;
	flush_keyboard_buffer();
	Mode=0;Mode_old=100;
	strcpy(callee_ID,"--------");
	set_calleeID(callee_ID);
	strcpy(our_ID,   "--------");
	set_ourID(our_ID);
	_setcursortype( _NOCURSOR );
	if(get_speed()==0) Baudrate=100;
	else               Baudrate=200;
	strcpy(logfilename,"PACTOR56.LOG");

// Load configuration file
	read_setup_data(setup);

// Set up COM port
	switch(COM_prt) {
		case 1:
			combase = 0x3F8;
			irq = 4;
			break;
		case 2:
			combase = 0x2F8;
			irq = 3;
			break;
		case 3:
			combase = 0x3E8;
			irq = 4;
			break;
		case 4:
			combase = 0x2E8;
			irq = 3;
			break;
		}

	set_com_port(combase, irq);

// Open the log file (note BINARY mode to avoid CR/LF problems)
	if((logfile=fopen(logfilename,"ab"))==NULL) {
		printf("Could not open the Log File: <%s>", logfilename);
		printf("Please hit any key to continue using default\n");
		getch();
		strcpy(logfilename,"PACTOR56.LOG");
		if((logfile=fopen(logfilename,"ab"))==NULL) {
			printf("Could not open the default file: <%s> either", logfilename);
			printf("Using a write-protected disk ?");
			return(0);
			}
		}

// Enter start-up time stamp
	if(Log_option)
		{
		time(&bintime);	// get time in seconds since 00:00:00 UTC,1/1/70  
		if(Time_Type!=GMT) curtime=localtime(&bintime); // convert to local time	
		else curtime=gmtime(&bintime); 	// UTC
		sprintf(time_string,"---->Started at: %s", asctime(curtime));
		fprintf(logfile,"\r\n\r\n\r\n");
		i=0;
		while (time_string[i]!='\0') {
			if(!((time_string[i]=='\r')||(time_string[i]=='\n')))
				fprintf(logfile,"%c",time_string[i]);				
			i++;
			}
		fprintf(logfile,"\r\n");
		}

//
// Clean up startup displays
//
	delay(1500);

	current_window = activate_window(6);  
        clrscr(); 
	textbackground(LIGHTGRAY);
	textcolor(LIGHTBLUE);
	gotoxy(1,1); cprintf("UNPROTO CONNECT BREAKIN  QRT     ?????  MONITOR  CLEAR  DISC     ?????  STANDBY");
	textbackground(DEFAULT_BGCOLOR); // Restore default screen colors
	textcolor(DEFAULT_CHCOLOR);

	current_window = activate_window(2);  
	gotoxy(1,1);   
        clrscr(); 
	gotoxy(1,window_boundary-4);
        display_cursor();
	current_window = activate_window(3);
	gotoxy(1,1);
        clrscr(); 
        display_cursor();
	
	pactor_speedup();

	// Set up the COM port interrupt handler

	if (irq<8) {
		oldfunc=getvect(irq+0x08);  // Hook interrupt on PC side
		setvect(irq+0x08,INTRP);
 					// deal with the first 8259
		old_IMR=inp(0x21);	// get IMR
		mask=old_IMR; 		// get the starting interupt mask
		mask&=~(0x0001<<irq); 	// clear the appropriate mask bit
		outp(0x0021,mask); 	// write out the new mask
		}
	else {
		oldfunc=getvect(irq+0x68);  // Hook IRQ on PC side
		setvect(irq+0x68,INTRP);
 					// deal with the second 8259
		old_IMR=inp(0x00a1); 	// get the starting interupt mask
		mask=old_IMR; 		// get the starting interupt mask
		mask&=~(0x0001<<(irq-8)); // clear the appropriate mask bit
		outp(0x00a1,mask);
		}

	// Set the mode and baud rate of the COM port
	outp(combase+3, 0x83);  // LCR register
	outp(combase+0, (12L*9600L/(long)combaud) & 0xff);
	outp(combase+1,	(12L*9600L/(long)combaud) >> 8);
	outp(combase+3, 0x03);  // LCR register: 8, n, 1

	// Enable the COM port interrupts
	outp(combase+1, 0x01);

        current_window = activate_window(2);
	cprintf("");     	// show cursor upper window

        current_window = activate_window(3);
	cprintf("");     	// show cursor lower window
	
	previous_biostime=biostime(0,0L);  // Initialize

	for(;;)	{

		Data_Mode=get_mode(); /* mode and block count */
      		update_status_line();

//
// Check if there is any special embedded characters
//
		ch=*read_ptr;	       // Check for "OVER"
		if((keybuf_size>=1)&&(ch==(char)0x04)) {
			read_ptr++;
			keybuf_size--;
			if(read_ptr==&keyboard_buffer[BUFR_SIZE])
				read_ptr=keyboard_buffer;
			set_control(0x29);
			}
			
//
// If the link will accept data, load them for sending them.
//  
		if((keybuf_size>=1)&&(get_buf_stat()==0))
			{
         		current_window = activate_window(2);
			ch=*read_ptr++;
			keybuf_size--;
			if(read_ptr==&keyboard_buffer[BUFR_SIZE])
				read_ptr=keyboard_buffer;

			send_txdta((int)ch);
			}


// Check if there is data waiting in the echo buffer
//     presumeably all ASCII ---		

		if(((ch_data=get_echodta())&0xF00)!=0)  {
         		current_window = activate_window(2);
			cprintf("\b \b"); // wipe old cursor
			ch=(char)0x7F&ch_data;

			if(ch=='\r') {// handle CR/LF
				cprintf("\r\n");
				clreol();
				if(Log_option) fprintf(logfile,"\r");
				}
			if(ch>=0x20) {
				cprintf("%c", ch); /* ASCII */
				if(Log_option) fprintf(logfile,"%c",ch);
				}
        		if(wherex()>=80) {	 // handle wrap round
				cprintf("\r\n");
				clreol();
				if(Log_option) fprintf(logfile,"\r");
				}
			cprintf(""); // show cursor
			}

// Check if there is data waiting in the receive buffer

		time(&bintime);
// have link_status time out?
		if(difftime(bintime, previous_time)>(double)1.5) {
			link_stat=0;
			previous_time=bintime;
			}
		if(((ch_data=get_rxdta())&0xF00)!=0)  {
			previous_time=bintime;	 // reset status timeout
			if(get_speed()==0) {
				Baudrate=100;
				blk_siz=8;
				}
			else  {
				Baudrate=200;
				blk_siz=20;
				}
         		current_window = activate_window(2);
			current_node=root_node;	 /* For Huffman */
			Data_Mode=get_mode(); /* mode and block count */
			if((Data_Mode&0x03)!=prev_count) {
			    if(!(((Data_Mode&0x03)==0)&&(prev_count==3)))
			      if((Data_Mode&0x03)!=prev_count+1) {
				cprintf("\b \b"); // wipe old cursor
			    	cprintf(" ");
				if(Log_option)fprintf(logfile," ");
				}
			    prev_count=Data_Mode&0x03;
			    link_stat=2; // Idle
			    k=0;  // Huffman character count
			    for(i=0;i<blk_siz;i++) { // Process a block
				if(ch_data==0) break;  // quit on short block
				ch=(char)ch_data&0xFF;
				if((Data_Mode&0x0C)==0) {
//     --- ASCII ---		
					cprintf("\b \b"); // wipe old cursor
					ch=0x7F&ch;

					if(ch=='\r') {// handle CR/LF
						cprintf("\r\n");
						clreol();
						if(Log_option)
						   fprintf(logfile,"\r");
						}
					if(ch>=0x20) {
						cprintf("%c", ch); /* ASCII */
						if(Log_option)
						   fprintf(logfile,"%c",ch);
						link_stat=1;  // TFC
						}
        				if(wherex()>=80) {	 // handle wrap round
						cprintf("\r\n");
						clreol();
						if(Log_option)
							fprintf(logfile,"\r");
						}
					cprintf(""); // show cursor
					}
/* For Huffman code, traverse the tree from the root down until 
 * all code bits exhausted
 */
				else {
//   --- HUFFMAN ---
					for(j=0;j<8;j++) {
					   if((symb=test_bits(ch))!=0xFFFF) {
						cprintf("\b \b");
						chh=(char)0x7F&symb;

						k++;

						if(chh=='\r') // handle CR/LF
							{
							cprintf("\r\n");
							clreol();
							if(Log_option)
						   		fprintf(logfile,"\r");
							}
						if(chh>=0x20) {
							cprintf("%c",chh);
							link_stat=1;  // TFC
							if(Log_option)
						   		fprintf(logfile,"%c",chh);
							}
						current_node=root_node;	  
        					if(wherex()>=80)	 // handle wrap round
							{
							cprintf("\r\n");
							clreol();
							if(Log_option)
								fprintf(logfile,"\r");
							}
						cprintf(""); // show cursor
						}
					   ch=ch>>1;
					   }
					}
				ch_data=get_rxdta(); 	// next character
			   	}
			    }
                       }
                         

		if(check_PACTOR_keyboard()==(char)0xFF) break;	// Handle keyboard input


		} // End of endless loop

	close_down();

#ifdef SHAREWARE
	shareware_banner();
#endif

   	_setcursortype( _NORMALCURSOR );
	normvideo();
        window(1,1,80,25); clrscr();

	if(Log_option)
		{
		time(&bintime);	  	// final time stamp
		if(Time_Type!=GMT) curtime=localtime(&bintime); 	
		else curtime=gmtime(&bintime); 	// UTC
		sprintf(time_string,"---->Closed at: %s", asctime(curtime));	
		fprintf(logfile,"\r\n");
		i=0;
		while (time_string[i]!='\0') {
			if(!((time_string[i]=='\r')||(time_string[i]=='\n')))
				fprintf(logfile,"%c",time_string[i]);				
			i++;
			}
		fprintf(logfile,"\r\n");
		}

#ifdef SHAREWARE
 	printf("\nThanks for using EVM-pacTOR\n"); 	
#endif

 	return 0;
}
//--------------------------------------------------------------------------
//
// define three windows:
// window 1 = status
// window 2 = output from AMTOR
// window 3 = input from user keyboard
// window 5 = tuning indicator
// window 6 = Function key labels
//
int activate_window(int wind_no)
{
   x_ind[current_window-1] = wherex();   // save current windos's location
   y_ind[current_window-1] = wherey();
   switch( wind_no)                      // switch to the new window      
      {
      case 1:   window( 1,1,80,1);
                break;
      case 2:   window( 1,3,80,window_boundary-1);
                break;
      case 3:   window( 1,window_boundary+1,80,24);
                break;
      case 4:   window( 1,window_boundary,80,window_boundary);
                break;
      case 5:   window( 1,2,80,3);
                break;
      case 6:   window( 1,25,80,25);
                break;
      }
// retrieve next location
   gotoxy( x_ind[wind_no-1], y_ind[wind_no-1]);   
   return wind_no;
} 
//---------------------------------------------------------------------------
// Routine to update status window
//
void update_status_line()
{
int stat, i, j, mark_tuning, space_tuning;
time_t tnow;
struct tm *tmnow;
char time_str[13], *st_ptr;

long biostime_now;

#define TUN 65
static int mark, space;
static char upper_row[]="m¿m";
static char lower_row[]="ss";
static int bar_colors[] = {
	LIGHTGRAY, LIGHTGRAY, LIGHTGRAY, LIGHTGRAY,
	LIGHTBLUE, LIGHTBLUE, LIGHTBLUE, LIGHTBLUE,
	YELLOW, YELLOW, RED, RED, RED, RED };


	biostime_now=biostime(0,0L);
	if((biostime_now-previous_biostime)>Refresh_rate) {
		previous_biostime=biostime_now;
		current_window = activate_window(1);
		Mode=get_status();
		gotoxy(1,1);    cprintf("                ");
		gotoxy(1,1);
		switch (Mode)
			{
			case 0: cprintf("STANDBY");
				break;
			case 0x10: 
				textattr(0x80|(WHITE+(RED<<4)));   // flash
				cprintf("TX-PHASE");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x11: 
				textattr(0x80|(WHITE+(RED<<4)));   // flash
				cprintf("TX-TFC");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x12: 
				textattr(0x80|(WHITE+(RED<<4)));   // flash
				cprintf("TX-IDLE");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x13: 
				textattr(0x80|(WHITE+(RED<<4)));   // flash
				cprintf("TX-RQ");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x14: 
				textattr(0x80|(WHITE+(RED<<4)));   // flash
				cprintf("TX-UNPROT");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x20: 
				textattr(0x80|(WHITE+(GREEN<<4)));   // flash
				cprintf("RX-TFC");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x21: 
				textattr(0x80|(WHITE+(GREEN<<4)));   // flash
				cprintf("RX-IDLE");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x22: 
				textattr(0x80|(WHITE+(GREEN<<4)));   // flash
				cprintf("RX-ERR");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x23: 
				textattr(0x80|(WHITE+(GREEN<<4)));   // flash
				cprintf("RX-SYNC");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x24: 
				textattr(0x80|(WHITE+(GREEN<<4)));   // flash
				cprintf("RX-RQ");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x30: 
				textattr(0x80|(YELLOW+(RED<<4)));   // flash
				cprintf("OVER");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x31: 
				textattr(0x80|(BLACK+(WHITE<<4)));   // flash
				cprintf("MONITOR ");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				switch(link_stat) {
				    case 0:
					textattr(BLACK+(WHITE<<4));  
					cprintf("SCAN  ");
					textbackground(DEFAULT_BGCOLOR); // Default screen colors
					textcolor(DEFAULT_CHCOLOR);
					break;

				    case 1:
					textattr(BLACK+(WHITE<<4));  
					cprintf("RX-TFC");
					textbackground(DEFAULT_BGCOLOR); // Default screen colors
					textcolor(DEFAULT_CHCOLOR);
					break;

				    case 2:
					textattr(BLACK+(WHITE<<4));
					cprintf("RX-IDLE");
					textbackground(DEFAULT_BGCOLOR); // Default screen colors
					textcolor(DEFAULT_CHCOLOR);
					break;
				    }
				break;
			case 0x32: 
				textattr(0x80|(BLACK+(WHITE<<4)));   // flash
				cprintf("MON-IDLE");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			case 0x33: 
				textattr(0x80|(BLACK+(WHITE<<4)));   // flash
				cprintf("MON-CRC");
				textbackground(DEFAULT_BGCOLOR); // Default screen colors
				textcolor(DEFAULT_CHCOLOR);
				break;
			}
		
// Display RxT
		stat=get_RxT();	// Timing factor
		gotoxy(40,1);
		cprintf("RxT:%d ",stat);

		time(&tnow);
		if(Time_Type==GMT) {
			tmnow=gmtime(&tnow);
			time_str[9]='U';time_str[10]='T';time_str[11]='C';
			time_str[12]='\0';
			}	
		else	{
			tmnow=localtime(&tnow);
			strncpy(&time_str[9],tzname[0],3);
			time_str[12]='\0';
			}	

		st_ptr=asctime(tmnow);		// fill in the time
		while(*st_ptr++!=':');		// look for start of first field
		st_ptr--;st_ptr--;st_ptr--;
		for(i=0;i<8;i++) time_str[i]=*st_ptr++;
		time_str[8]=' ';
        	gotoxy(57,1);
		cprintf("%s",time_str);

		gotoxy(17,1);
	
		cprintf("REM:%s", callee_ID);
		gotoxy(69,1);cprintf( "-----0+++++");        // Show bit phasing status
     		stat=get_phase_err();
     		if(stat>5)  stat = 5;
     		if(stat<-5) stat=(-5);
     		gotoxy(74+stat,1);
		textattr((WHITE+(RED<<4))); cprintf(" ");
     		stat=get_phase_avg();
     		if(stat>5)  stat = 5;
     		if(stat<-5) stat=(-5);
     		gotoxy(74+stat,1);
		textattr((WHITE+(GREEN<<4))); cprintf(" ");
		textbackground(DEFAULT_BGCOLOR);
		textcolor(DEFAULT_CHCOLOR);

// Tuning indicator display (line 2)
		current_window=activate_window(5);

		get_tuning_data( &mark_tuning, &space_tuning);
		mark_tuning /= 12;
		space_tuning /= 12;
		if(mark_tuning>=11)  mark_tuning=11;
		if(space_tuning>=11) space_tuning=11;

		mark=(mark+mark_tuning)/2;    // Average mark with prev.
		space=(space+space_tuning)/2; // Average space with prev.


#if 1
		i = 13;
		for(j=0; j<15; j++) {
			gotoxy(TUN+j,1);
			if (mark >= i && i >= 1) {
				// This slot needs to be filled in
				// Pick the right color
				textcolor(bar_colors[i]);
				textbackground(DEFAULT_BGCOLOR);
				putch('');

			} else {
				// Just put the backround character here
				textcolor(DEFAULT_CHCOLOR);
				textbackground(DEFAULT_BGCOLOR);
				putch(upper_row[j]);
			}
			i--;
		}

		i = 13;
		for(j=0; j<15; j++) {
			gotoxy(TUN+j,2);
			if (space >= i && i >= 1) {
				// This slot needs to be filled in
				// Pick the right color
				textcolor(bar_colors[i]);
				textbackground(DEFAULT_BGCOLOR);
				putch('');

			} else {
				// Just put the backround character here
				textcolor(DEFAULT_CHCOLOR);
				textbackground(DEFAULT_BGCOLOR);
				putch(lower_row[j]);
			}
			i--;
		}

#else

		j=0;
		for(i=TUN; i<TUN+12-mark; i++) {
			gotoxy(i,1); cprintf("%c", upper_row[j]);
			j++;
			}
		j=0;
		for(i=TUN; i<TUN+12-space; i++) {
			gotoxy(i,2); cprintf("%c", lower_row[j]);
			j++;
			}
// >=10  - RED
//   8,9 - YELLOW
//
		if(mark>=10) {
			textattr(0x7F&(RED+(DEFAULT_BGCOLOR<<4)));
			gotoxy(TUN+12-mark,1); 
			cprintf("");
			}
		else mark++;
		for(i=mark-1; i>=0 ;i--)
			if((mark>7)&&(i>=7)) {
				gotoxy(TUN+12-i,1); 
				textattr(0x7F&(YELLOW+(DEFAULT_BGCOLOR<<4)));
				cprintf("");
				}
			else 
				if((mark>4)&&(i>=4)) {
					gotoxy(TUN+12-i,1); 
					textattr(0x7F&(LIGHTBLUE+(DEFAULT_BGCOLOR<<4)));
					cprintf("");
					}
				else {
					gotoxy(TUN+12-i,1); 
					textattr(0x7F&(LIGHTGRAY+(DEFAULT_BGCOLOR<<4)));
					cprintf("");
					}

		textcolor(DEFAULT_CHCOLOR);
		textbackground(DEFAULT_BGCOLOR);
		cprintf("%c%c", upper_row[13], upper_row[14]);

		if(space>=10) {
			textattr(0x7F&(RED+(DEFAULT_BGCOLOR<<4)));
			gotoxy(TUN+12-space,2); 
			cprintf("");
			}
		else space++;
		for(i=space-1; i>=0 ;i--)
			if((space>7)&&(i>=7)) {
				gotoxy(TUN+12-i,2); 
				textattr(0x7F&(YELLOW+(DEFAULT_BGCOLOR<<4)));
				cprintf("");
				}
			else
				if((space>4)&&(i>=4)) {
					gotoxy(TUN+12-i,2); 
					textattr(0x7F&(LIGHTBLUE+(DEFAULT_BGCOLOR<<4)));
					cprintf("");
					}
				else {
					gotoxy(TUN+12-i,2); 
					textattr(0x7F&(LIGHTGRAY+(DEFAULT_BGCOLOR<<4)));
					cprintf("");
					}

		textcolor(DEFAULT_CHCOLOR);
		textbackground(DEFAULT_BGCOLOR);
		cprintf("%c%c", lower_row[13], lower_row[14]);

#endif


// Update lower status line
// Compression status

		current_window=activate_window(4);
		gotoxy(1,1);
		textattr(0x7F&(CYAN+(DEFAULT_BGCOLOR<<4)));
		if(get_speed()==0)  cprintf("LO ");
		else                cprintf("HI ");
		stat=get_RXsense();
		if(stat==0)	cprintf("R");
		else            cprintf("R"); 
		stat=get_TXsense();
		if(stat==0)	cprintf("T ");
		else            cprintf("T "); 
		if((Data_Mode&0x0C)==0) cprintf("ASCI");
		if((Data_Mode&0x0C)!=0) cprintf("HUFF");
		cprintf(" Ct:%d", Data_Mode&0x03);
		stat=get_sync();
		cprintf(" Hd:%2X",stat);
		stat=get_err_cnt();
		cprintf(" ERR:%d",stat);
		stat=get_RQ_cnt();
		cprintf(" RQ:%d",stat);
		stat=get_CS();
		switch(stat) {
			case 0:
				cprintf(" T-O ");
				break;
			case 0x4D5:
				cprintf(" CS1 ");
				break;
			case 0xAB2:
				cprintf(" CS2 ");
				break;
			case 0x34B:
				cprintf(" CS3 ");
				break;
			case 0xD2C:
				cprintf(" CS4 ");
				break;
			default: 
				cprintf(" CS? ");
				break;
			}

		stat=get_CTRX();
		if((stat+120+5)>=1250) {
			textattr(0x80|(WHITE+(RED<<4)));   // flash
			cprintf("CTRX:%d ",stat);
			textbackground(DEFAULT_BGCOLOR); // Default screen colors
			textcolor(CYAN);
			}
		else	cprintf("CTRX:%d ",stat);
		stat=get_CTTX(); 
		if((stat+120+5)>=1250) {
			textattr(0x80|(WHITE+(RED<<4)));   // flash
			cprintf("CTTX:%d ",stat);
			textbackground(DEFAULT_BGCOLOR); // Default screen colors
			textcolor(CYAN);
			}
		else		cprintf("CTTX:%d ",stat);

// Number of MARQ blocks recovered
		stat=get_marq_cnt();
		cprintf("MARQ:%d ",stat);

//		while(wherex()<69) cprintf(""); // clean up till buffer stat
		}


	if(keybuf_size!=keybuf_size_old)
		{
		current_window=activate_window(4);
		keybuf_size_old=keybuf_size;
		textbackground(DEFAULT_BGCOLOR);
		textcolor(CYAN);
		gotoxy(69,1); cprintf("BFRS:");
		if(keybuf_size)	{
			textattr(0x80|(YELLOW+(DEFAULT_BGCOLOR<<4))); // flash
			cprintf("%d",keybuf_size);
			}
		else	cprintf("%d",keybuf_size);
		if(keybuf_size<100) cprintf("  ");
		}

	textbackground(DEFAULT_BGCOLOR); // Default screen colors
	textcolor(DEFAULT_CHCOLOR);
}
//--------------------------------------------------------------------------
//
// Pop-up window for transferring of file data
//
void files_menu()
{
int bufsize, left=5,right=75,top=5,bottom=20;
char *tbuffer, c, ch, in_str[80];
char input_line[100];
int i, flag, x, y, loc_x, loc_y, ch_data;
FILE *f_ptr;
time_t bintime, time_now;

	save_xy(&x,&y);
   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
		terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(
" FILE TRANSMISSION MENU ͻ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"                                                                     ");
        cprintf(
"ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();

	flag=0;
        while(flag==0)
                {
                cprintf("Enter file name to be loaded > ");
                strcpy(in_str, "");
                in_str[0]=80;
                cgets(in_str);
 	        cprintf("\r\n");clreol();
	        if(in_str[1]==0) flag=1;
	        else 
                  {
                  if((f_ptr=fopen(&in_str[2],"r"))==NULL)
			 cprintf("Cannot open <%s>\r\n", &in_str[2]);
                  else  {
			cprintf("\r\nReady to transmit <%s>", &in_str[2]); 
			cprintf("\r\n--Hit any key to stop this transmission--\r\n");

//
// Wait for pactor bios to accept a character
// then read a character from the file and transfer them to bios.
// Repeat until end of file reached.
//
			c=getc(f_ptr);	 	// priming read
			for(;;)	{

//
// Check if there are any echoed data, if so, print it
//
				if(((ch_data=get_echodta())&0xF00)!=0)  {
					ch=(char)0x7F&ch_data;

					if(ch=='\n') {// handle CR/LF
						cprintf("\r\n");
						clreol();
						if(Log_option) fprintf(logfile,"\r");
						}
					if(ch>=0x20) {
						if(c==TAB) cprintf(" ");
						else       cprintf("%c", ch);
						if(Log_option) fprintf(logfile,"%c",ch);
						}
					if(wherex()==(right-left-2)) {	// Line overflow ?
						cprintf("\r\n");
						clreol();
						if(Log_option) fprintf(logfile,"\r");
						}
					}

//
// Save current location in file transmission window,
//  update the status line, and restore file transmission menu
// 
				save_xy(&loc_x,&loc_y);
				textcolor(DEFAULT_CHCOLOR);
				textbackground(DEFAULT_BGCOLOR); 
   				_setcursortype( _NOCURSOR );
				update_status_line();
        			window(left+1,top+1,right-1,bottom-2);
				gotoxy(loc_x,loc_y);
				textattr((BLACK+(WHITE<<4))&0x7F);

				while(get_buf_stat()!=0)    // Wait till buffer ready
					if (kbhit()) break; // ...
//
// Interpret NL as CR\LF
//
				if(c=='\n') {
					send_txdta((int)'\r');	// send data
					while(get_buf_stat()!=0)    // Wait till buffer ready
						if (kbhit()) break; // ...
					send_txdta((int)'\n');	// send data
					}				
                             	else 	send_txdta((int)c);	// send data

				c=getc(f_ptr);		   // get next character

			   	if(feof(f_ptr)) { // EOF ?
					cprintf("\r\n---EOF reached\r\n");
					clreol();
					break;
					}

			   	if(kbhit()){ // user needs out ?
					cprintf("\r\n---User terminated file transmission\r\n");
					clreol();
					break;
					}

			   	} // end of file reading

			if(kbhit()) {
				getch();
				fclose(f_ptr);
				flush_keyboard_buffer();
				flush_bios_buffers();
				flag=1;
				break;
				}

			while(get_buf_stat()!=0); // Wait till buffer sent

                        fclose(f_ptr);
			flag=1;
		       	}  // end got EOF during file reading
		  } // end else got valid filename
		}  // end while

// Restore screen

	if(in_str[1]!=0) 	// a file was sent (or partially sent)
		{
		cprintf("\r\nTransmission completed");
		time(&bintime);  
		while(1) {    // delay 2 seconds
			time(&time_now);
			if(difftime(time_now, bintime)>2) break;
			}
		}

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	textcolor(DEFAULT_CHCOLOR);
	textbackground(DEFAULT_BGCOLOR); 
   	_setcursortype( _NOCURSOR );

	gotoxy(x,y);			// leave cursor as before interrupted
}
//---------------------------------------------------------------------------
// Routine to save present x-y coordinates
void save_xy(int *x,int *y)
{
	*x=wherex();
	*y=wherey();
}
//--------------------------------------------------------------------------
//
// Pop-up window for entering remote callsigns
//
void enter_selcal()
{

int bufsize, left=20,right=60,top=11,bottom=15;
char *tbuffer, ch, c, in_str[80], *from_ptr, *to_ptr;
int i,x,y;

	save_xy(&x,&y);
   	_setcursortype( _NORMALCURSOR );
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(" REMOTE CALLSIGN ENTRY ͻ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();

        for(;;)
                {
                cprintf( "%s > ",callee_ID);
                strcpy(in_str, "");
                in_str[0]=80;
                cgets(in_str);
                cprintf("\r\n");clreol();
	        if(in_str[1]==0) break;
	        for(i=2;i<in_str[1]+2;i++) in_str[i]=toupper(in_str[i]);
                if(in_str[1]<4)
                        {
                        cprintf( "At least four characters required\r\n");
                        clreol();
                        }
                else 
                        {
			i=0;
			from_ptr=&in_str[2];
			to_ptr=callee_ID;
			for(;;) {   // Copy data as entered
				if(*from_ptr=='\0') break;
				(*to_ptr++)=(*from_ptr++);
				if(i==7) break;
				i++;
				}
			for(;;) {   // pad remainder with spaces
				*to_ptr=' ';
				if(i==7) break;
				to_ptr++;
				i++;
				}
			callee_ID[8]='\0';
                        set_calleeID(callee_ID);
                        break;
                        }
                }

// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);
	gotoxy(x,y);
	textcolor(DEFAULT_CHCOLOR);
	textbackground(DEFAULT_BGCOLOR); 
   	_setcursortype( _NOCURSOR );
	Mode_old=100;	// Force a refresh of the status line	
}
//--------------------------------------------------------------------------
//
// Pop-up window for command dialogs
//
char command_mode()
{

int bufsize, left=20,right=60,top=8,bottom=24, delay_input;
char *tbuffer, ch, c, in_str[80];
int i, x, y;
	
	save_xy(&x,&y);				      // save current window
   	_setcursortype( _NORMALCURSOR );	      // co-ordinates
        bufsize=2*(bottom-top+1)*(right-left+1);      // allocate buffer space
        if((tbuffer=(char *)malloc(bufsize))==NULL)
                terminate_prog("Malloc - allocating buffer");
        if(!gettext(left,top,right,bottom,tbuffer))   // copy existing data
                terminate_prog("Gettext - saving text!");

// define window for menu and display menu

        window(left,top,right,bottom);
        clrscr();
	textattr((BLACK+(WHITE<<4))&0x7F);   // Black letters on white backgnd
        cprintf(" OPTIONS MENU ͻ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf("                                       ");
        cprintf(" X -> Main display | Q -> quit prgm ͼ");
        window(left+1,top+1,right-1,bottom-2);
        clrscr();
        for(;;)
            {
            cprintf( "Ready > ");
            ch = getche();
            cprintf("\r\n");clreol();

// Check for breaking out of endless loop
	    if(ch==(char)0x1B) ch=(char)'X'; 	// change ESC into an 'X'
            if((ch=='X')||(ch=='x'))  break;  
            if((ch=='Q')||(ch=='q')) 
		{
		ch=(char)0x1B;	     // change so it looks like ESC
		break;
		}   
 
           switch (ch)
               {
		case 'd':  // TX-DELAY VALUE
	      	case 'D': cprintf( "Enter new DELAY in ms: ");
			strcpy(in_str, "");
			in_str[0]=80;
			cgets(in_str);
			cprintf("\r\n");clreol();
			if(in_str[1]==0) delay_input=0;
			else sscanf(&in_str[2],"%d",&delay_input);
			if ((delay_input < 0)||(delay_input > MAX_DELAY)) {
				cprintf( "Invalid delay\r\n");
				clreol();
				}
			else {
				cprintf( "Old delay was: %dms",tx_delay);
				if(delay_input==0) delay_input=DEF_TXDELAY;
				tx_delay = delay_input;
				set_txdelay( tx_delay );
				cprintf( " now: %dms",tx_delay);
				cprintf("\r\n");
				clreol();
				}
			break;

	       case '?':
               case 'h':  // Show help
               case 'H': 
        		cprintf("\r\nD - Adjust Tx delay (ms)");
        		cprintf("\r\nH - This help display");
        		cprintf("\r\nI - Change Station Identification");
        		cprintf("\r\nT - Adjust master clock timing");
        		cprintf("\r\nV - Toggle direct video I/O option");
        		cprintf("\r\nU - Set Unproto Speed/Repeat count");
        		cprintf("\r\nY - 500 Hz Timing test\r\n");
			clreol();
	       		break;
              
               case 'i':  // Set station Identity
               case 'I': 
                	cprintf( "Present ID is: %s \r\nEnter  new ID: ",our_ID);
                	strcpy(in_str, "");
                	in_str[0]=80;
                	cgets(in_str);
                	cprintf("\r\n");clreol();
	        	if(in_str[1]==0) break;
	        	for(i=2;i<in_str[1]+2;i++) in_str[i]=toupper(in_str[i]);
                	if(in_str[1]<4)	{
                        	cprintf( "At least four characters required\r\n");
                        	clreol();
                        	}
                	else 	{   
				strcpy(our_ID,"        ");
                        	strncpy(our_ID, &in_str[2], in_str[1]);
                        	our_ID[8]='\0';
                        	set_ourID(our_ID);
				}
                	cprintf( "Changed ID is: %s",our_ID);
                        cprintf("\r\n");
                        clreol();
                        break;

	 	case 't':  // Load new timing
               	case 'T': cprintf( "Bump TIMING value (using +/-) ");
                         in_str[0]=getch();
 			 cprintf("\r\n");clreol();
			 delay_input=timing_value;
			 if(in_str[0]=='+') delay_input=timing_value+1;
			 if(in_str[0]=='-') delay_input=timing_value-1;
                         if ((delay_input < 14000)||(delay_input > 22000))
                           {
                           cprintf( "Range exceeded - Timing unchanged\r\n");
                           clreol();
                           }
                         else
                           {
                           cprintf( "Old value was: %d",timing_value);
                           timing_value = delay_input;
                           cprintf( " now: %d",timing_value);
                           cprintf("\r\n");
                           clreol();

			   cprintf("Dither? (Y/N) ");
                           in_str[0]=getch();
			   if((in_str[0]=='y')||(in_str[0]=='y'))Dither=1;
			   if((in_str[0]=='n')||(in_str[0]=='N'))Dither=0;
 			   cprintf("\r\n");clreol();

                           set_timing(timing_value, Dither);
                           }
                         break;

		case 'u':  // Unproto speed
	      	case 'U': 
			cprintf( "Unproto Speed: %dbd  -   Repeat: %d\r\n",U_Speed, U_Repeat);
			cprintf( "New Unproto Speed (100/200)?: ");
			strcpy(in_str, "");
			in_str[0]=80;
			cgets(in_str);
			cprintf("\r\n");clreol();
			if(in_str[1]==0) delay_input=0;
			else sscanf(&in_str[2],"%d",&delay_input);
			if (!((delay_input==100)||(delay_input == 200)||
				(delay_input==0))) {
				cprintf( "Invalid Speed\r\n");
				clreol();
				}
			else {
				if(delay_input!=0) U_Speed = delay_input;
				if(U_Speed==100) set_USPEED(0);
				if(U_Speed==200) set_USPEED(0xFF);
				cprintf( "Speed now: %dbd",U_Speed);
				cprintf("\r\n");
				clreol();
				}

	      		cprintf( "New Unproto Repeat Count (1/2/3)?: ");
			strcpy(in_str, "");
			in_str[0]=80;
			cgets(in_str);
			cprintf("\r\n");clreol();
			if(in_str[1]==0) delay_input=0;
			else sscanf(&in_str[2],"%d",&delay_input);
			if (!((delay_input==1)||(delay_input == 2)||
			      (delay_input==3)||(delay_input==0))) {
				cprintf( "Invalid Repeat Count\r\n");
				clreol();
				}
			else {
				if(delay_input!=0) U_Repeat = delay_input;
				set_UREPEAT(U_Repeat);
				cprintf( "Repeat Count now: %d",U_Repeat);
				cprintf("\r\n");
				clreol();
				}
			break;

               case 'v':  //  change direct/BIOS video
               case 'V': cprintf( "Toggle video mode (DIRECT/BIOS): ");
 			 cprintf("\r\n");clreol();
                         cprintf( "Old video mode was: "); 
			 directvideo? cprintf("DIRECT") : cprintf("BIOS");
                         if(directvideo==0) directvideo=1;
                         else directvideo=0;
                         cprintf( " now: ");
			 directvideo? cprintf("DIRECT\r\n") : cprintf("BIOS\r\n");
                         clreol();
                         break;

               case 'y':  // RUN A 500Hz TIMING TEST
               case 'Y': set_control( 0x27 ); 
                        cprintf( "\r\n500Hz test started\r\n");
                        cprintf( "Please check output on Td \r\n");
			cprintf( "When finished, hit any key to continue\r\n");
            		ch = getche();
 			cprintf("\r");clreol();
			set_control(0x26);		// FORCE STBY
                        break; 

                default: cprintf("??\r\n"); clreol();
               }
            }


// Restore screen

        if(!puttext(left,top,right,bottom,tbuffer))
                terminate_prog("Puttext - restoring text");

        free(tbuffer);

   	_setcursortype( _NOCURSOR );
	textcolor(DEFAULT_CHCOLOR);
	textbackground(DEFAULT_BGCOLOR); 
	gotoxy(x,y);
        return ch;
}
//---------------------------------------------------------------------------
// Catastrophic failure error handler function
//
void terminate_prog(char *ptr)
{

        window(1,1,80,25); clrscr();
	cprintf("\r\nPACTOR detected a serious error - further operation is terminated");
	cprintf("\r\nThe following is a short description of the problem:");
	cprintf("\r\n%s", ptr); 
	cprintf("\r\nIf its a memory allocation problem, you might want to");
	cprintf("\r\nreduce the number or size of your message buffers and");
	cprintf("\r\ntry again");
   	cprintf( "\r\nPACTOR TERMINATED\n\r");
   	_setcursortype( _NORMALCURSOR );
	normvideo();
	exit(1);			// exit to DOS

}
/*-----------------HUFFMAN COMPRESSION---------------------------------------
 * Open and read encoding file
 * Note that the PACTOR set has only 128 symbols.
 */
void build_code_array()
{
FILE *codefile;
int i, inp_bits;
unsigned int inp_symbol;
char inp_code[16];
 

	if((codefile=fopen("code.dat","r"))==NULL)
		{
		printf("Could not open the code file:");
		exit(1);
		}

	for(i=0;i<128;i++) {					/* Initialize array */
			code_data[i].code=0xFFFF;
			code_data[i].code_bits=0;
			strcpy(code_data[i].bit_array,"");
			}

	for(;;) {								/* load array */
		fscanf(codefile,"%d %d %s", &inp_symbol, &inp_bits, inp_code);
		if(feof(codefile)) break;
		if(inp_symbol<129) {
			code_data[inp_symbol].code=inp_symbol;
			code_data[inp_symbol].code_bits=inp_bits;
			strcpy(code_data[inp_symbol].bit_array,inp_code);
			}
		}
}
/*--------------------------------------------------------------------------
 * Build Huffman tree
 */
void build_tree()
{
int i, j;

/* Create the root node and initialize it */
	if((root_node=(NODE*)malloc(sizeof(NODE)))==NULL) {
		printf("Malloc failed\n");
		exit(1);
		}
	root_node->symbol=0xFFFF;
	root_node->child_0=NULL;
	root_node->child_1=NULL;

/* For each given code, traverse the tree from the root down, 
 * constructing nodes as we go.
 */
	for(i=0;i<128;i++) {			  /* Get the next code element */
		current_node=root_node;	  /* Start at the root         */

		for(j=0;j<code_data[i].code_bits;j++) {	 /* For each code bit */

			switch (code_data[i].bit_array[j])	  {
				case '0':
				if(current_node->child_0==NULL) {
					if((current_node->child_0=(NODE*)malloc(sizeof(NODE)))==NULL) {
						printf("Malloc failed\n");
						exit(1);
						}
					else {
						current_node=current_node->child_0;
						current_node->symbol=0xFFFF;
						current_node->child_0=NULL;
						current_node->child_1=NULL;
						}
					}
				else current_node=current_node->child_0;
				break;
			
			case '1':
				if(current_node->child_1==NULL) {
					if((current_node->child_1=(NODE*)malloc(sizeof(NODE)))==NULL) {
						printf("Malloc failed\n");
						exit(1);
						}
					else {
						current_node=current_node->child_1;
						current_node->symbol=0xFFFF;
						current_node->child_0=NULL;
						current_node->child_1=NULL;
						}
					}
				else current_node=current_node->child_1;
				break;
				}
		}
		current_node->symbol=code_data[i].code;
	}	

}
/*--------------------------------------------------------------------------
 * Decode Huffman-coded bit stream
 */
int test_bits(char bit)
{
int j, err;


		err=0;
		switch (bit&0x01)  {
			case 0:
					if(current_node->child_0==NULL) {
						err=1;
						break;
						}
					current_node=current_node->child_0;
					break;
			
			case 1:
					if(current_node->child_1==NULL) {
						err=1;
						break;
						}
					current_node=current_node->child_1;
					break;
			}

		if(err) return 0xFFFF;
		else    return current_node->symbol;

}
//---------------------------------------------------------------------------
#ifdef SHAREWARE

void shareware_banner()
{

// Put up frame
        window(1,1,80,25);
        clrscr();   

// Fill in the text
        gotoxy(26,10); cputs("PC-PACTOR  VERSION");
	               cputs( version );
        gotoxy(22,11); cputs("(c) J.B.Forrer (KC7WW) 1991-1994");
        gotoxy(26,12); cputs("  26553 Priceview Drive    ");
        gotoxy(26,13); cputs("  Monroe, OR 97456, USA    ");
        gotoxy(26,16); cputs("UNREGISTERED COPY ");
                       cputs( serial  );
        gotoxy(26,24); cputs("Hit ESC key to continue    ");
        while(getch()!=0x1B);
}
#endif
//---------------------------------------------------------------------------
void flush_keyboard_buffer()
{
	read_ptr=store_ptr=keyboard_buffer;
	keybuf_size=0; keybuf_size_old=100;
}
//---------------------------------------------------------------------------
// Check keyboard for input
//---------------------------------------------------------------------------
char check_PACTOR_keyboard()
{
char ch, *ptr;

		if(kbhit()) {
         		current_window = activate_window(3);
			ch=getch();
         		if(ch==0) {                /* 0 code */
            			ch=getch();
            			switch(ch) {
               				case F1:        // F1 - UNPROTO transmission
					set_control(0x24);
	                      			break;

               				case F2:        // F2 - ARQ transmission
					set_control(0x23);
	                      			break;

               				case F3:        // F3 - ARQ breakin
					set_control(0x21);
	                      			break;

					case F4:	// F4 - QRT request
						set_control(0x2A);
						break;

					case F5:	// F5 - not used
						break;

					case F6:	// F6 - MONITOR
						set_control(0x25);
						break;
					case F8:	// F8 - Not used
						break;
					case F7:	// F7 - USER CLEAR
						current_window = activate_window(3);
                        			gotoxy(1,1);
                        			clrscr(); 
						cprintf(""); 	// show cursor
						flush_keyboard_buffer();
						flush_bios_buffers();
						break;
             				case SHIFT_F7:  // SH-F7 RX CLEAR
						current_window = activate_window(2);
                        			gotoxy(1,1);  
                        			clrscr(); 
						cprintf(""); 	// show cursor
                        			break;

					case F9:	// F9 - Force QRT now!
						set_control(0x22);
						break;
					case F10:	// F10 - Force STBY
						set_control(0x26);
						break;

					case Insert:   // mycall de hiscall
                        			ptr=callee_ID;
                        			for(;;) {
					          if(*ptr==' ')	break;
						  if(*ptr=='\0') break;
						  store_character(*ptr); 
						  ptr++;
						  }
						store_character(' ');
						store_character('D');
						store_character('E');
		        			store_character(' ');
                        			ptr=our_ID;
                        			for(;;) {
					          if(*ptr==' ')	break;
						  if(*ptr=='\0') break;
						  store_character(*ptr); 
						  ptr++;
						  }
		        			store_character(' ');
                        			break;

					case End:	// ARQ over
						store_character((char)0x04);
						break;

                			case ALT_F:
						files_menu();	// Transmit file
                        			break;

                			case ALT_R:             // Enter remote selcal
						enter_selcal();             
                        			break;

					case ALT_X:		// Quick breakout
						ch=0xFF;	//
						break;
						 
  					 }
				}
			else 	{
            			current_window = activate_window(3);
         			if (ch == 0x1B)	{
					ch=command_mode();  // Command mode
              				if(ch==0x1B) ch=0xFF;
					else ch=ch;
					} 
				else store_character(ch);
				}

	 		return ch;
			}        // end kbhit
	 	else return (char)0;
}
//---------------------------------------------------------------------------
//
// Store character in keyboard buffer
//
void store_character(char ch)
{
	if(keybuf_size<BUFR_SIZE) {
		cprintf("\b \b");	 // remove old cursor
		if(ch=='\r') {
			cprintf("\r\n");
			clreol();
			}
		else  cprintf("%c", ch);
        	if(wherex()>=80) {	 // handle wrap round
			cprintf("\r\n");
			clreol();
			}
		cprintf("");     	// show cursor

		*store_ptr++=ch;
		keybuf_size++;
		if(store_ptr==(&keyboard_buffer[BUFR_SIZE]))
			store_ptr=keyboard_buffer;
		}
}

//--------------------------------------------------------------------------
void display_cursor()
{
int x, y;

        x=wherex();y=wherey();
	cprintf("");     // show cursor
//	cprintf("_");
        gotoxy(x,y);
}
//---------------------------------------------------------------------------
//
// Procedure obtains various setup info from user-supplied file
//
void read_setup_data(FILE *setup)
{
int i, start_op;
char setup_line[80];

   gotoxy(1,3);

   if (setup==NULL)
        {
        cprintf("\r\nNo setup file (PACTOR56.CNF) supplied\r\n");
        cprintf("all parameters set to default values\r\n");
	clreol();
        }
   else
        {
	clreol();
        for(;;)
                {
                fgets(setup_line, 80, setup);
                if(feof(setup)) break;
		strupr(setup_line);
                cprintf("%s\r",setup_line);                  // Show setup
		clreol();
                if(!strncmp(setup_line, "ID:",3))            // id
			if(strlen(setup_line)>8+3) {
			     	cprintf("ID TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(our_ID,"........");
			     	}
                        else {
				strncpy(our_ID,&setup_line[3],strlen(setup_line)-3);
                          	our_ID[strlen(setup_line)-4]='\0';
                          	set_ourID(our_ID);
			  	}

                if(!strncmp(setup_line, "TXDELAY:",8)) {      // txdelay
                        sscanf(&setup_line[8],"%d",&tx_delay);
                        if ((tx_delay < 0)||(tx_delay > MAX_DELAY)) {
			     	cprintf("BAD TXDELAY - IGNORED: OK?\r\n");
			     	getch();
				tx_delay=DEFAULT_TXDELAY;
			     	}
                        else	set_txdelay(tx_delay);
                        }

                if(!strncmp(setup_line, "TIMING:",7)) {       // clock timing
                        sscanf(&setup_line[7],"%d",&timing_value);
			if ((timing_value < 14000)||(timing_value > 22000)) {
			     	cprintf("BAD TIMING VALUE - IGNORED: OK?\r\n");
			     	getch();
				timing_value=DEFAULT_TIMING;
			     	}
                        else set_timing(timing_value, Dither);
                        }

                if(!strncmp(setup_line, "DITHER:",7))        // clock dither
                        sscanf(&setup_line[7],"%d",&Dither);

                if(!strncmp(setup_line, "FCRLF:",6))        // Files CR/LF
                        sscanf(&setup_line[6],"%d",&Files_CRLF);

                if(!strncmp(setup_line, "LOG:",4))        // Logging option
                        sscanf(&setup_line[4],"%d",&Log_option);

                if(!strncmp(setup_line, "LOGFILE:",8))    // Logfile name
			if(strlen(setup_line)>LOGFILENAME_LENGTH+8) {
			     	cprintf("LOG FILENAME TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(logfilename,"PACTOR56.LOG");
			     	}
                        else {
                        	strncpy(logfilename,&setup_line[8],strlen(setup_line)-8);
                        	logfilename[strlen(setup_line)-9]='\0';
                        	}

                if(!strncmp(setup_line, "SCROLLFILE:",11))    // Scrollback file name
			if(strlen(setup_line)>LOGFILENAME_LENGTH+8) {
			     	cprintf("SCROLLBACK FILENAME TOO LONG - IGNORED: OK?\r\n");
			     	getch();
				strcpy(scrollback,"PSCROLL.TMP");
			     	}
                        else {
                        	strncpy(scrollback,&setup_line[11],strlen(setup_line)-8);
                        	scrollback[strlen(setup_line)-12]='\0';
                        	}

                if(!strncmp(setup_line, "VID:",4))        // Direct Video option
                        sscanf(&setup_line[4],"%d",&directvideo);

                if(!strncmp(setup_line, "REFR:",5))        // Refresh rate
                        sscanf(&setup_line[5],"%Ld",&Refresh_rate);

                if(!strncmp(setup_line, "TIME:",5))        // GMT=0 or LOCAL=1
                        sscanf(&setup_line[5],"%d",&Time_Type);

                if(!strncmp(setup_line, "TXINVERT:",8)) {   // TXINVERT
                        sscanf(&setup_line[8],"%d",&Mask_out);
			if(Mask_out) Mask_out=Mask_out&0x01;
			set_MASKOUT(Mask_out);
			}

                if(!strncmp(setup_line, "WORD:",5))      // Word/character mode
                        sscanf(&setup_line[5],"%d",&Word_Mode);

                if(!strncmp(setup_line, "COM:",4)) {           // COM port
                        sscanf(&setup_line[4],"%d",&COM_prt);  // COM1/2/3/4
                        if (!((COM_prt>0)&&(COM_prt<=MAX_COM))) 
				COM_prt=DEFAULT_COM_PORT;
			}   

                if(!strncmp(setup_line, "IRQ:",4)) {   // interrupt
			}

                if(!strncmp(setup_line, "OLVL:",5)) {       // Speaker level
                        sscanf(&setup_line[5],"%d",&Sound_out_level);
			if ((Sound_out_level < 0)||(Sound_out_level > 63)) {
			     	cprintf("BAD SPEAKER LEVEL - IGNORED: OK?\r\n");
			     	getch();
				Sound_out_level=DEFAULT_OLVL;
			     	}
                        }

                if(!strncmp(setup_line, "ILVL:",5)) {       // Line in level
                        sscanf(&setup_line[5],"%d",&Sound_in_level);
			if ((Sound_in_level < 0)||(Sound_in_level > 15)) {
			     	cprintf("BAD LINE IN LEVEL - IGNORED: OK?\r\n");
			     	getch();
				Sound_in_level=DEFAULT_ILVL;
			     	}
                        }

                if(!strncmp(setup_line, "SOURCE:",7)) {       // Input source
                        sscanf(&setup_line[7],"%d",&Source);
			if (!((Source==0)||(Source==1)||(Source==2))) {
			     	cprintf("BAD INPUT SOURCE - IGNORED: OK?\r\n");
			     	getch();
				Source=0;
			     	}
			switch (Source) {
				case 0: Source=0x00;
					break;
				case 1: Source=0x40;
					break;
				case 2: Source=0x80;
					break;
				}
			}

                if(!strncmp(setup_line, "PACTORMDM:",10))    // Modem
                        sscanf(&setup_line[10],"%d",&Pactor_Mdm);

                }

        cprintf("Initialization completed\r\n");
	clreol();
        fclose(setup);
        }
}
//--------------------------------------------------------------------------
void set_timing(int value, int dither)    // Load timing
{
	set_biostiming(value, dither);	// to set pacbios timing value
}
//-----------------------------------------------------------------------
// Pactor speedup
//-----------------------------------------------------------------------
void pactor_speedup()
{
	_p_speedup();	// initialize things for EVM

// These variables are initialized by speedup
	set_timing(timing_value, Dither);
	set_txdelay(tx_delay);
	set_UREPEAT(U_Repeat);
	set_USPEED(U_Speed);
	set_MASKOUT(Mask_out);
	Adjust_Sound_output_level(Sound_out_level);
	Adjust_Sound_input_level(Sound_in_level);
}

//-----------------------------------------------------------------------
// Close down PacTOR and reset vectors etc
//-----------------------------------------------------------------------
void close_down()
{
int i;
	asm {CLI};			// prevent any interrupts 

	// TODO: deal with any pending interrupts

	if (irq<8) { 			// deal with the first 8259
		outp(0x21,old_IMR);
		setvect(irq+0x08, oldfunc);	// Restore INT
		}
	else {
		outp(0x00A1,old_IMR);	// Restore IMR
		setvect(irq+0x68, oldfunc);	// Restore INT
		}

	asm {STI};			// enable interrupts 

	for(i=0;i<20;i++) set_control(0x06);	// Reset interface

}

//---------------------------------------------------------------------------
// Get tuning data from DSP
void get_tuning_data(int *t1, int *t2)  // Get tuning data
{
int sig;
	sig=get_mark_data();
	*t1=(sig>>8);
	sig=get_space_data();
	*t2=(sig>>8);
}
//---------------------------------------------------------------------------
// This adjusts the level of the line in
void Adjust_Sound_input_level(int level)
{
}
//---------------------------------------------------------------------------
// This adjusts the level of the speaker level
void Adjust_Sound_output_level(int level)
{
}
//---------------------------------------------------------------------------
