;----------------------------------------------------------------------------
;    EVMTOR - Library of routines for amateur radio AMTOR/PACTOR
;    Copyright (C) 1993-1995 Johan B. Forrer, KC7WW
;
;    This is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
;----------------------------------------------------------------------------;
;                   PACBIOS FOR EVM SOUND CARDS                              ;
;									     ;
; Copyright (C) J.B.Forrer 1991,1992,1993,1994 - All rights reserved         ;         
;									     ;
;									     ;
; History:	Beta 0.04 - Derived from PC-PacTOR 0.20		             ;
;----------------------------------------------------------------------------;
;      TO DO:
;   ==========
; 1). Memory ARQ.
; 2). Full shutdown implementation. 
;    2.1 recognize our call in reverse ASCII + Huffman.
;    2.2 keep scanning for repeated shutdown calls (a small timeout count).
; 3). Huffmann capability from within PACBIOS (via C subroutine calls).
;									     ;
;======================= CONDITIONALS =======================================;				     ;
;;LIMITED		=1	; for limited edition
;;TEST_WITH_PTC		=1	; for testing with PTC
;;LOW_SPEED		=1	; force low speed only
MARQ			=1	; use memory ARQ
;----------------------------------------------------------------------------;

STKSIZE 	= 2048          ; PACTOR stacksize
RXBUFSIZ 	= 2048		; PACTOR RX buffer size
TXBUFSIZ	= 100		; PACTOR TX buffer size
CHKLEN		= 5		; only check first 5 characters of selcal
MASTER_SYNC	= 955		; The "ideal" block time (middle of 10ms bit)		

;----------------------------------------------------------------------------;
; These paramters control the phase correction rate                          ;
;----------------------------------------------------------------------------;
MAX_ADJUST	= 1		; +/- 1 limits of DSP clock adjustment
ALG3_BACKOFF	= 2
;----------------------------------------------------------------------------;
DECAY		= 50		; Decay factor for tuning display
SYNCCS		= 8		; number of good CS bits for sync to be sure
CS_COMP		= 9		; DEF=8 x 5 CS compensation for slave
GDCS		= 5		; number of good CS bits
CTRX_SLOP	= 10		; 1 bit jitter slop allowed
MAX_ERR		= 30		; terminate after 10 errors
SPEED_THR_TX	= 7		; hi/lowspeed changeover threshold
MAX_RQ		= 30		; terminate after 10 RQ's of our CS's
AVE_RXT		= 5		; compute master's RxT over first n cycles

SYNC0		= 055h		; PTC Synch char (normal sense)
SYNC1		= 0AAh		; PTC Synch char (reverse sense)
IDLE		= 01Eh		; PTC Idle character (00011110)

;----------------------------------------------------------------------------;
; CS1/2 alternate acknowledgements					     ;
; Repeat of last CS means negative acknowledgement (RQ)			     ;
; CS3 is used as the head of a changeover packet			     ;
; CS4 serves as speed change indicator					     ;
; CS1/2 is complimentary (bit reverse ordered)				     ;
; CS3/4 is complimentary (bit reverse ordered)				     ;
;----------------------------------------------------------------------------;
CS1		= 04D5h		; control signals
CS2		= 0AB2h
CS3		= 034Bh
CS4		= 0D2Ch
;----------------------------------------------------------------------------
; Access variables off the stack 
; For "Large models", i.e.    mov	bx,ARG1		; get first arg
; C calling sequence func(ARG1, ARG2, ARG3, ARG4)
;
ARG1	EQU	<[BP+6]>
ARG2	EQU	<[BP+8]>
ARG3	EQU	<[BP+10]>
ARG4	EQU	<[BP+12]>
ARG5	EQU	<[BP+14]>
;----------------------------------------------------------------------------

; EVM modem op-codes:

EVM_NOP		=	08h
EVM_TX_PREV	=	04h
EVM_TX_0	=	05h
EVM_TX_1	=	06h
EVM_RX		=	00h
EVM_100		=	10h
EVM_200		=	11h
EVM_RESET	=	20h



	PUBLIC	__p_speedup,_set_control,_set_txdelay,
	PUBLIC  _get_rxdta,_send_txdta,_get_phase_err,_set_biostiming
	PUBLIC  _set_USPEED, _set_UREPEAT, _set_MASKOUT, _get_CTRX
	PUBLIC  _get_buf_stat, _set_calleeID, _set_ourID, _get_status
 	PUBLIC  _get_mode, _get_speed, _get_sync, _get_RXsense, _get_TXsense
	PUBLIC	_get_err_cnt, _get_RQ_cnt, _get_CS, _get_CTTX,_get_phase_avg
	PUBLIC  _set_com_port, _get_mark_data, _get_space_data, _get_RxT
	PUBLIC	_get_marq_cnt, _get_echodta, _flush_bios_buffers

cseg    segment para public 'CODE'

        assume cs:cseg,ds:cseg

;-------NOTES ON USAGE AND SEARCHING FOR THE SYCH CHARACTER------------------;
;		  (see further description in SMPLE_8)			     ;
;----------------------------------------------------------------------------;
RPTR_L	DW      1 dup (?)       ; Search shift reg ptr valid packet 100 bd
RPTR_H	DW      1 dup (?)       ; Search shift reg ptr valid packet 200 bd
RPTR_1  DW      1 dup (?)       ; Search shift reg ptr callsign	100 bd
RPTR_2  DW      1 dup (?)       ; Search shift reg ptr callsign	200 bd
RPTR	DW      1 dup (?)       ; Search shift reg ptr CS search
RPTR_CTL DW     1 dup (?)       ; Search shift reg ptr CS search
CHPTR_1	DW	1 dup (?)	; search register for 100 baud changeover
CHPTR_2	DW	1 dup (?)	; search register for 200 baud changeover

REGS_8  DB      250 dup (?)     ; Shift reg for callsign search 10x9+5x8=130++
REGS_12 DB	32  dup (?)	; Shift registers for CS search
SHIFTREG_1 DB	250 dup (?)	;  10x12=120++	 (100 baud full packet)
SHIFTREG_2 DB	250 dup (?)	;  5x24=120++	 (200 baud full packet)
CHREG_1 DB	200  dup (?)	;  10x10=100++	 (100 baud C/O  packet)
CHREG_2 DB	200 dup (?)	;  5x21=105++	 (200 baud C/O  packet)
;----------------------------------------------------------------------------;
; The data bits arrives at a constant rate (1ms intervals). 80 bits make up  ;
; one byte.								     ;
;									     ;
; Let the data bits be denoted dij where i=0,7 "the ith group of 10 bits"    ;
;                                  and   j=0,9 "the jth bit within the group";
; This denotes a total of 8x10=80 bits.					     ;
;									     ;
;      d0	   d1	       d2	 d3 	   ... d7		     ;
; [..........][..........][..........][..........] ...	DATA BIT STREAM IN   ;
;                                                       GROUPS OF 10 BITS    ;
;									     ;
; If we re-arrange these bits into eight shift registers, r0 - r9 where	     ;
; each shift register contains eight bits,				     ;
;									     ;
; r00=d00, r01=d10, r02=d20, r03=d30, ..., r07=d70  ---- r0		     ;
; r10=d01, r11=d11, r12=d21, r13=d31, ..., r17=d71  ---- r1		     ;
;									     ;
; r90=d09, r91=d19, r92=d29, r93=d39, ..., r97=d79  ---- r9		     ;
;									     ;
; Then each of the shift registers r0 thru r9 contains an eight bit	     ;
; character shifted 1ms in time between registers. To fill these registers   ;
; we use a pointer pointing to a register - on the new data bit sample we    ;
; shift the bit	into the register pointed to (data bit is in the carry flag, ;
; shift right).	Then the test is performed on the byte that was just updated ;
; looking for the synch character. If none is found, the pointer is bumped   ;
; (runs	in a circle) and the next data bit gets shifted into that shift      ;
; register. This way the search moves along "one 1ms at a time". When synch  ;
; is found, we are most probably right on the leading edge of the bit stream,;
; so the middle	of the first bit of the next byte is samples 5ms from the    ;
; moment synch was found.						     ;
; 									     ;
;----------------------------------------------------------------------------;
; 			Misc. variables. 
;----------------------------------------------------------------------------;
	EVEN
USRSP_OFF 	dw    	1 dup (?)        ; USER SP
USRSP_SEG 	dw    	1 dup (?)
       
AMTAX   	DW      1 dup (?)        ; TEMP ax-REG IN PACTOR PROG
AMTBX   	DW      1 dup (?)        ; TEMP bx-REG "
AMTCX		DW	1 dup (?)	 ; TEMP cx-REG "
AMTDX		DW	1 dup (?)	 ; TEMP dx-REG "
AMTSI		DW	1 dup (?)	 ; TEMP si-REG "
AMTDI		DW	1 dup (?)	 ; TEMP di-REG "

AMTSP   	dw	1 dup (?)        ; PACTOR SP
		db	STKSIZE dup (?)	 ; PACTOR STACK
AM_STACK 	db	1 dup (?) 
        	db      512 dup (?)  		
	EVEN
EVM_timing	dw	1 dup (?)	; EVM_clock timing used for phase corrections
EVM_dither	dw	1 dup (?)	; EVM_clock dither
CLOCK		dw      1 dup (?)       ; Main system clock
AUX_CLOCK	dw      1 dup (?)       ; Main system auxiliary
ideal_RxT    	dw      0
RxT		dw	0 		; Time from our last sent bit to
RxT_cnt		dw	0		; ---first received bit
ideal_CTRX    	dw      0       	; ideal time of start of control receive
DS_CTRX    	dw      0       	; for display purposes
CTTX    	dw      1 dup (?)       ;  -for transmit control-
OLD_CTTX    	dw      1 dup (?)       ;  save previous value
CRC_work	dw	1 dup (?)	; CRC_used in CRC routine
CRC_data	dw	1 dup (?)	; pass character in LSB for CRC
control_out	dw	1 dup (?)	; controls are sent from here
CONTROL		dw	1 dup (?) 	; loaded with CS1,2, or 3
ECONTROL	dw	1 dup (?)	;   ditto used in ARQ master
prev_control	dw	1 dup (?)	; previous control during speedup/down
prev_CS		dw	1 dup (?)	; previous CS to prevent CS4 loop
cycle_cnt	dw	5000		; ticks per time dither

SEND_SENSE	dw	1 dup (?)	; bit sense when sending
phase_err	dw	1 dup (?)	; phase error
old_phase1	dw	1 dup (?)
old_phase2	dw	1 dup (?)
old_phase3	dw	1 dup (?)
old_phase4	dw	1 dup (?)
old_phase5	dw	1 dup (?)
old_phase6	dw	1 dup (?)
old_phase7	dw	1 dup (?)
mean_phase	dw	1 dup (?)
old_CTRX	dw 	8 dup(?)	; for computing ideal_CTRX
old_RxT		dw 	8 dup(?)	; for computing ideal_RxT
rq_cnt		db	1 dup (?)	; RQ count for resending our CS's
cs_cnt1		db	1 dup (?)	; cs security counts
cs_cnt2		db	1 dup (?)	; cs security counts
cs_cnt3		db	1 dup (?)	; cs security counts
cs_cnt4		db	1 dup (?)	; cs security counts
err_cnt		db	1 dup (?)	; error count
blk_cnt		db	1 dup (?)	; packet counter 0-3
bit_sense	db	1 dup (?)	; =0 for normal sense =FF for reverse
rx_bit_sense	db	1 dup (?)	; used during ARQ
tx_bit_sense	db	1 dup (?)	; used for character bit sense
tx_sense	db	1 dup (?)	; for display purposes
rx_sense	db	1 dup (?)	; for display purposes
exp_rx_sense	db	1 dup (?)	; expected rx sense
exp_rx_speed	db	1 dup (?)	; expected rx speed
bit_cnt		db	1 dup (?) 	; bit cntr used in x-mission of bits
rx_byte		db	1 dup (?)	; receive bytes are assembled here
byte_size	db	1 dup (?)	; number of bits per char
initialized     db  	1 dup (?)     	; when initialized=5A
STSTS   	DB      1 dup (?)       ; STATUS INPUT used in TESTR
tst_pat 	DB      1 dup (?)       ; used for testing pattern
sync_byte	db	1 dup (?)	; working sync byte
sync		db	1 dup (?)	; actual sync byte	
EVM_cnt		db	1 dup (?)	; slow changes CLOCK
DSP_cnt		db	1 dup (?)	; slow changes the DSP clock
phasing_flag	db	1 dup (?)	; =1 for doing a DSP clock correction
repeat_CS3	db	0		; =1 when CS3 is repeated
clock_state	db	0 		; for clock dithering
ave_RxT_cnt	db	1 dup (?)	; used initial RxT avg in MASTER
	EVEN
MASKOUT DB      1 dup (?)        	; setting output bit levels


STATUS		db	1 dup (?)	; System status

		db	1 dup (?)
;---------------------------------------------------------------------------;
;  00=Standby								    ;
;  10=Sending-phasing (attempting to link)				    ;
;  11=Sending-linked							    ;
;  12=Sending-idle    (linked)						    ;
;  13=Sending-RQ							    ;
;  14=Sending-unproto							    ;
;  20=Receiving-linked							    ;
;  21=Receiving-idle  (linked)						    ;
;  22=Receiving-CRC							    ;
;  30=CHANGE OVER							    ;
;---------------------------------------------------------------------------;									    ;
	EVEN
irq		dw	1 dup (?)	; COM port's IRQ

evm_in		db	1 dup (?)	; Most recent byte from EVM
evm_out		db	1 dup (?)	; Byte to send to EVM

AD_data		dw  	1 dup(?)	; Data value for MARQ
AD_mark_data	dw  	1 dup(?)	; Peak mark  A/D value
AD_space_data	dw  	1 dup(?)	; Peak space A/D value

MODE		db	1 dup (?)	; 00=ASCII, !00=HUFFMAN
SPEED		db	1 dup (?)	; 00=100 BAUD, FF=200 BAUD
TX_SPEED	db	1 dup (?)	; transmitting speed
RX_SPEED	db	1 dup (?)	; receiving speed
DS_SPEED	db	1 dup (?)	; display speed
u_speed		db	1 dup (?)	; UNPROTO SPEED
u_count		db	1 dup (?)	; UNPROTO repeat count
U_REPEAT	db	1 dup (?)	; UNPROTO reapeat constant
MASTER  	db      1 dup (?)       ; 0=>MASTER, !=0=>SLAVE
QFLG    	db      1 dup (?)       ; Set to request a closedown
BREAKIN		db	1 dup (?)	; Request a breakin
OVER		db	1 dup (?)	; Request an "over"
QRT_request	db	1 dup (?)	; set = 1 when ISS sent QRT bit
QRT_pending	db	1 dup (?)	; Keyboard requests a QRT
QRT_sent	db	1 dup (?)	; This shows we have sent a QRT frame
CNTRL   	db      1 dup (?)       ; Command input
tx_delay   	db      1 dup (?)       ; TX-data delay
callee_ID	db	'????????'	; callsign of destination  (8 bytes)
callee_ID_rev	db	'????????'	; "reversed" order for sending QRT
level		db	'1'		; Level "1" software
our_ID		db	'????????'	; our callsign

	EVEN
data_bit	db	1 dup(?)	; last data bit received
old_modem    	db	0		; used for adaptive speed
;----------------------------------------------------------------------------;
; Data store to handle dynamic allocation of COM port			     ;
;----------------------------------------------------------------------------;
	EVEN
COM     	= 03F8h         ; COM1 Port address default
com_port	dw  COM
com_port1	dw  COM+1
com_port2	dw  COM+2
com_port3	dw  COM+3
com_port4	dw  COM+4
com_port5	dw  COM+5
com_port6	dw  COM+6

;----------------------------------------------------------------------------;
selcal_hdr	db	1 dup (?)	; Selcal header
selcal_msg1	db	8 dup (?)	;	
selcal_msg2	db	8 dup (?)	;	
;----------------------------------------------------------------------------;

msg_hdr		db	1 dup (?)	; Data buffer
msg_buffer	db	20 dup (?)	;	
msg_status	db	0h		; status byte
msg_CRC		db	00h		; 2-byte CRC 
		db	00h		
;----------------------------------------------------------------------------;
out_ptr		dw	1 dup (?)	; pointer into buffer
out_msg_hdr	db	1 dup (?)	; Output data buffer
out_msg		db	20 dup (?)	;	
out_msg_stat	db	0h		; status byte for 200 baud
		db	00h		; 2-byte CRC 
		db	00h		
;----------------------------------------------------------------------------;
block_buffer	db	20 dup (?)	; receive block buffer
		db	1 dup (?)	; block status
block_CRC	db	1 dup (?)	; block CRC
		db	1 dup (?)
		db	1 dup (?)
block_ptr	dw	1 dup (?)	; pointer into block buffer
;----------------------------------------------------------------------------;

rx_store_at_ptr	 DW	1 dup (?)	; Store at pointer
rx_take_from_ptr DW	1 dup (?)	; take from pointer	
RXBUFR           DB    RXBUFSIZ dup (?) ; Data received from PACTOR
;----------------------------------------------------------------------------;

echo_store_at_ptr  DW	1 dup (?)	; Store at pointer
echo_take_from_ptr DW	1 dup (?)	; take from pointer	
ECHOBUFR           DB    RXBUFSIZ dup (?) ; Data echoed from PACTOR
;----------------------------------------------------------------------------;
tx_store_at_ptr	 DW	1 dup (?)	; Store at pointer
tx_take_from_ptr DW	1 dup (?)	; take from pointer	
TXBUFR           DB    TXBUFSIZ dup (?) ; Data to transmitted thru PACTOR

;----------------------------------------------------------------------------;
;                         Memory ARQ buffers:				     ;
;									     ;
; Each bit represents a 8-bit encoded signal level. Each frame is 960 ms    ;
; long, with a sample every ms. For 100 baud, 10 samples = 1 bit, for	     ;
;                                   200 baud,  5 samples = 1 bit.	     ;
;----------------------------------------------------------------------------;
mem_arq_accum	dw     970  dup (?)	; accum bit buffer
marq_cnt	db	1 dup (?)	; count number of MARQ successes
;----------------------------------------------------------------------------;


;================START OF PACBIOS CODE SECTION===============================;
PACBIOS	proc	near
;----------------------------------------------------------------------------;
;            Standby mode - scanning for sync's				     ;
;----------------------------------------------------------------------------;
	EVEN
STBY:

IFDEF LIMITED
	jmp	MON
ENDIF	

	lea     bx,AM_STACK
        mov     sp,bx           	; SP
        mov     ax,cs
        mov     ss,ax           	; SS
        
        lea     bx,REGS_8		; set search register pointer
        mov     RPTR_L,bx		; low baudrate bits
        lea     bx,REGS_8+90		; set search register pointer
        mov     RPTR_H,bx		; high baudrate bits
        mov     byte ptr MASTER,1      	; set for slave
        
	mov	byte ptr EVM_cnt,0	; ready for slave phasing
	mov	byte ptr DSP_cnt,0

        xor     ax,ax                                        
        mov     QFLG,al         	; Clear QRT request
        mov     BREAKIN,al         	; Clear BREAKIN request
        mov     OVER,al         	; Clear "over" request
        mov     QRT_request,al         	; Clear qrt_request (ISS)  flag
        mov     QRT_pending,al         	; Clear qrt_request (Kbrd) flag
        mov     QRT_sent,al         	; Clear qrt sent flag
        mov     repeat_CS3,al         	; CS3 repeat flag
	mov	STATUS,al		; STATUS=STANDBY
	mov	RxT,ax			; clear tracker

	call	zero_phase		; initially no phase error
	
	; TODO: initialize modem port here?

;
; Turn DSP AFSK off
;
	mov	al,EVM_RX
	mov	cs:evm_out,al

;----------------------------------------------------------------------------;
; Keep looking for our selcal to fly by				             ;
;----------------------------------------------------------------------------;
STBY0:
	call	SRCH			; Call sign search
	cmp	byte ptr SPEED,0	; 100 baud ?
	jne	S200
	mov	CLOCK,950
	jmp	SHORT S2
S200:	mov	CLOCK,950
S2:
;--------------------------------------------------------------------------;
;                         -IRS PROCESSING-			           ;
; This section is entered at two entry points:				   ;
; a) after recognizing our SELCAL at STBY0				   ;
; b) RR_renter from the ISS section after a turnaround was processed	   ;
;--------------------------------------------------------------------------;
; Some notes on how timing is calculated in this section:		   ;
; We assume the master starts his transmission of our SELCAL or		   ;
; any block for that master at CLOCK = 0000. We then, in return place	   ;
; our CS at:								   ;
;       CTTX = 960 + 5 x CS_COMP (x5 to use PTC terminology)               ;				   
; typically  = 960 + 25 = 985 (+ 20 = 1005 this is where actual bit lays)  ;
;--------------------------------------------------------------------------;
	xor	ax,ax
	mov	al,tx_delay
	mov	bx,960
	sub	bx,ax
	add	bx,5*CS_COMP		; CS compensation
	mov	CTTX,bx			; this is where we put our CS

	call	reset_MARQ		; ensure we start off empty MARQ

	mov	al,byte ptr cs:bit_sense ; Rx sense as found in call serach
	mov	byte ptr cs:rx_sense,al	; for display purposes

;----------------------------------------------------------------------------;
; The expected receive sense can be interrogated after a packet is received  ;
; whether the packet is valid or timed out, it will reflect correct sense    ;
; It is toggled back and forth in the CS transmit routine.                    ;
;----------------------------------------------------------------------------;
	mov	byte ptr exp_rx_sense,al

	mov	al,byte ptr cs:sync_byte ; byte as from call search
	mov	byte ptr cs:sync,al	; for display puposaes

; Update status showing link was established
	mov	byte ptr STATUS,23h	; Show sync is in progress
					 
	mov	byte ptr blk_cnt,0	; so that first block count = 1
					; first block cnt =1, sync=AAh
	mov	byte ptr err_cnt,0	; Clear errors
	mov	byte ptr rq_cnt,0	; Clear RQ count
	mov	byte ptr marq_cnt,0	; Clear marq counts

; For establishing 200 baud link CS1, else for 100 baud CS4
;===============================================
IFNDEF LOW_SPEED
	mov	CONTROL,CS1		; Assume low speed
ELSE
	mov	CONTROL,CS4		; TEST (FORCE LOW SPEED)
ENDIF
;===============================================

	mov	al,byte ptr SPEED	; found in SRCH routine
	mov	byte ptr RX_SPEED,al
	mov	byte ptr DS_SPEED,al
	cmp	al,00
	je	RR_lsp
	mov	CONTROL,CS4

RR_lsp:
	mov	byte ptr exp_rx_speed,0	; Set expected RX SPEED
	cmp	CONTROL,CS4
	je	RR_l	
	mov	byte ptr exp_rx_speed,0FFh

RR_l:	mov	SEND_SENSE,0 		; start off normal sense

	mov	bx,CTTX
	call	HOLD
	call	TCTL			; send control
	mov	RxT_cnt,0		; initialize RxT counter

RR_renter:
	call	init_ptrs		; initialize pointers    
	jmp	RR_next			; get into cycle, i.e. time=000.

RR_LSN:
	cmp	CLOCK,960		; Are we timing out ?
	js	RR_LS1
	jmp	RR_to			; yes, skip out
RR_LS1:
	call	SAMPLE			; get another bit sample
	cmp	byte ptr QFLG,1        	; test if qrt is set
        je      RR_STBYL           	; exit to STBY if yes
        jmp     SHORT RR_rd_blk	      	; else keep searching

RR_STBYL: jmp	STBY 			; go to STANDBY

;----------------------------------------------------------------------------;
; Look at the rest of the block. Both the right speed and polarity is        ;
; necessary.					                             ;
;----------------------------------------------------------------------------;
RR_rd_blk:
	cmp	byte ptr exp_rx_speed,0
	jne	RR_200

	mov	byte ptr cs:SPEED,00		; 100 baud operation 
	cmp	byte ptr exp_rx_sense,0
	jne	RR_rd1

	mov	byte ptr cs:bit_sense,00	; Normal sense
	call	test_blk_1
	or	al,al
	jnz	RR_LSRRR3a			; Received a good block	lo speed	
	jmp	RR_LSN   			; keep trying
RR_rd1:
	mov	byte ptr cs:bit_sense,0FFh	; Reverse sense
	call	test_blk_1_r
	or	al,al
	jnz	RR_LSRRR3a			; Received a good block	lo speed	
	jmp	RR_LSN   			; keep trying

RR_200:
	mov	byte ptr cs:SPEED,0FFh		; 200 baud operation 
	cmp	byte ptr exp_rx_sense,0
	jne	RR_200_reverse
	mov	byte ptr cs:bit_sense,00	; Normal sense
	call	test_blk_2
	or	al,al
	jnz	RR_LSRRR3a			; Received a good block	hi speed	
	jmp	RR_LSN   			; keep trying

RR_200_reverse:
	mov	byte ptr cs:bit_sense,0FFh	; Reverse sense
	call	test_blk_2_r
	or	al,al
	jnz	RR_LSRRR3a			; Received a good block	hi speed	
	jmp	RR_LSN   			; keep trying

RR_LSRRR3a:
	mov	al,byte ptr SPEED
	mov	byte ptr RX_SPEED,al 
	mov	byte ptr DS_SPEED,al

	jmp	RR_LSRRR3     			; long jump

;----------------------------------------------------------------------------;
; Bad block (CRC has failed or timeout)			                     ;
; If error count is low, start RQ process, i.e.				     ;
; send last CS as negative acknowledgement.				     ;
;									     ;
; We cannot perform phase error correction during this cycle as we dont know ;
; what state the last phase error is in					     ;
;									     ;
; If the error count is too high go back to standby and remain there.	     ;
;----------------------------------------------------------------------------;
RR_to:

;=========================CONDITIONAL
IFDEF MARQ
	call	test_MARQ			; does MARQ help?
	or	al,al
	je	RR_no_good
	inc	byte ptr marq_cnt		; YES !
	jmp	RR_LSRRR3aa
ENDIF
;================================

RR_no_good:
	inc	byte ptr err_cnt  		; bump error count
	cmp	byte ptr err_cnt,MAX_ERR	; too many ?
	js	RR_repeat
	jmp	STBY			

;----------------------------------------------------------------------------;
; If we are experiencing too many errors at high speed, request		     ;
; a speed down - CS4. This CS4 will be interpreted as a RQ by ISS.	     ;
; Do not attempt a speeddown when were at low speed already.		     ;
; Also reset previous control to CS2, so that we expect to send CS1	     ;
; on the next round							     ;
;----------------------------------------------------------------------------;
RR_repeat:
	cmp	byte ptr RX_SPEED,0h		; already at low speed ?
	jz	RR_rpt				; yes, skip out
	cmp	byte ptr err_cnt,SPEED_THR_TX	; error < change-thresh ?
	js	RR_rpt				; yes, skip out
	mov	prev_control,CS2
	mov	CONTROL,CS4			; else send speed down request
	mov	byte ptr exp_rx_speed,0		; now we expect low speed
RR_rpt:	mov	bx,CTTX 
	call	HOLD
	call	TCTL
	mov	RxT_cnt,0		; initialize RxT counter
	
; Update status showing ERROR
	mov	byte ptr STATUS,22h
	jmp	RR_next

;----------------------------------------------------------------------------;
; We received a good block as a result of MARQ 						     ;
; Move the data block into receive buffer regardless whether its	     ;
; a new block or not (the main program sorts that out) 			     ;
; and signal that data is available.					     ;
; However, since we timed out, CLOCK=960 now				     ;
;----------------------------------------------------------------------------;
RR_LSRRR3aa:
	mov	al,byte ptr SPEED
	mov	byte ptr RX_SPEED,al 
	mov	byte ptr DS_SPEED,al
	mov	bx,960
	jmp	SHORT Z0
;----------------------------------------------------------------------------;
; We received a good block 						     ;
; Move the data block into receive buffer regardless whether its	     ;
; a new block or not (the main program sorts that out) 			     ;
; and signal that data is available.					     ;
;----------------------------------------------------------------------------;
RR_LSRRR3:
;----------------------------------------------------------------------------;
; Compute the average RxT, for the first time use the first value            ;
;----------------------------------------------------------------------------;
	mov	bx,950			; XXXXX
	cmp	byte ptr RX_SPEED,00 	; check baudrate
	jz	Z0
	mov	bx,955			; 200
Z0:	
	mov	ax,RxT_cnt
	sub	ax,bx
	cmp	RxT,0
	jnz	Z1
	call	init_RxT		; initialize ideal_RxT
Z1:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update ideal_RxT

	mov	ax,phase_err	
	call	ave_ph_err		; arg. in ax
	call	slave_sync_IRS

	cmp	byte ptr err_cnt,0	; Clear error counter
	jz	RR_good
	mov	byte ptr err_cnt,0
 
;----------------------------------------------------------------------------;
; Update mode byte(Ascii/Huffman:packet count:QRT bit)			     ;
;----------------------------------------------------------------------------;
RR_good:
	mov	al,block_buffer+8 	; get status byte (assume 100 baud)
	cmp	byte ptr RX_SPEED,00 	; check speed
	jz	RR_LSRB4a
	mov	al,block_buffer+20 	; get status byte (200 baud)
	
RR_LSRB4a:
	mov	byte ptr cs:MODE,al

	mov	al,byte ptr cs:bit_sense
	mov	byte ptr cs:rx_sense,al
	mov	al,byte ptr cs:sync_byte
	mov	byte ptr cs:sync,al

;----------------------------------------------------------------------------;
; ISS requesting QRT?							     ;
;----------------------------------------------------------------------------;
	mov	al,byte ptr cs:MODE
	and	al,080h			; bit 7 is QRT bit
	jz	RR_LSRB4b
	mov	byte ptr QRT_request,1	; Yes
	jmp	SHORT RR_LSR_tt		; dont print this QRT block 

;----------------------------------------------------------------------------;
; Move data from internal buffer to output buffer for printing               ;
; Handle both 100 and 200 baud blocks                                        ;
;----------------------------------------------------------------------------;
RR_LSRB4b:
	lea	bx,block_buffer		; Data, move data from block_buffer
	mov	si,bx
	mov	cx,8	    		; assume 8 bytes (100 baud)
	cmp	byte ptr RX_SPEED,00 	; check baudrate
	jz	RR_LSRB4
	mov	cx,20			; else 20 bytes (200 baud)
	
RR_LSRB4:
	mov	al,byte ptr cs:[si]	; move bytes into receive buffer
	call	STORE_RX_BYTE
	inc	si
	loop	RR_LSRB4   

;----------------------------------------------------------------------------;
; Update sense and sync	information for status and control purposes          ;
;----------------------------------------------------------------------------;
RR_LSR_tt:

;----------------------------------------------------------------------------;
; Send acknowledgement: alternate CS1/2					     ;
; but first check whether this is not an RQ, i.e. 			     ;
; (our previous CS was not recvd).					     ;
;----------------------------------------------------------------------------;
	cmp	CONTROL,CS4		; CS4 is an exception as it
	je	RR_new			; the data was to be repackaged
	mov	al,byte ptr cs:MODE	; is this a repeated block ?
	and	al,03h
	cmp	al,byte ptr blk_cnt
	jne	RR_new

;----------------------------------------------------------------------------;
; Same packet counter as before means we received RQ - thus repeat last CS.  ;
;----------------------------------------------------------------------------;
	inc	byte ptr rq_cnt		; update RQ count
	cmp	byte ptr rq_cnt,MAX_RQ	; Too many ?
	js	RR_rq
	jmp	STBY			; too many RQ's in a row - terminate

RR_rq:
	cmp	CONTROL,CS4		; XXXX
	jne	RR_r00
	mov	ax,prev_control		; CS4
	mov	CONTROL,ax

RR_r00:	mov	bx,CTTX			; Get to CS time
	call	HOLD	
	call	TCTL			; repeat last CS
	mov	RxT_cnt,0		; initialize RxT counter

;
; Finally update status showing RQ, and go get next packet
;
	mov	byte ptr STATUS,24h
	jmp	RR_next

;----------------------------------------------------------------------------;
; Here after reception of a new packet.					     ;
;									     ;
; Does the operator have a breakin request or did we receive an over ?	     ;
; Note that a breakin is only acceptable AFTER reception of a good packet.   ;
;----------------------------------------------------------------------------;
RR_new:
	mov	ah,byte ptr cs:MODE
	and	ah,040h			; "over" ?
	jnz	RRR_CLA20a		; yes
	cmp	byte ptr BREAKIN,0	; Breakin ?
	jne	RRR_CLA20		; yes

;----------------------------------------------------------------------------;
; If we were sending CS4, and get a good block at LOW speed,		     ;
; then the CS4 was correctly received. If we get a good			     ;
; block at HIGH speed, then the CS4 was not received and we must	     ;
; do an RQ.								     ;
;----------------------------------------------------------------------------;
RR_new1:
	cmp	CONTROL,CS4
	jne	RR_new2
	cmp	byte ptr RX_SPEED,00
	je	RR_new2
	jmp	RR_rpt	

; Ready to go get next packet
RR_new2:
	mov	byte ptr blk_cnt,al  		; update packet count
	mov	byte ptr rq_cnt,0

	call	reset_MARQ			; ensure we start off empty MARQ

RR_ng:	
	cmp	CONTROL,CS4
	je	RR_g3
	cmp	CONTROL,CS1			; A new block
	jne	RR_g2
	mov	CONTROL,CS2			; CS2
	jmp	SHORT RR_g1

RR_g2:	mov	CONTROL,CS1
	jmp	SHORT RR_g1

RR_g3:	mov	ax,prev_control
	mov	CONTROL,ax

RR_g1:
	mov	bx,CTTX
	call	HOLD	
	call	TCTL
	mov	RxT_cnt,0		; initialize RxT counter

; Update status showing traffic
	mov	byte ptr STATUS,20h

RR_next:
;----------------------------------------------------------------------------;
; If we have received a QRT request, quit right here	                     ;
;----------------------------------------------------------------------------;
	or	byte ptr QRT_request,0
	jz	no_qrt_yet
	jmp	STBY				; DONE!
no_qrt_yet:
;----------------------------------------------------------------------------;
;  Here to collect data for next packet. 				     ;
;  Note this is not the entire packet!					     ;
;----------------------------------------------------------------------------;
	mov	cx,900				; 960  XXXXXX
RR_LSRB5:
	call	SAMPLE				;  get another bit sample
	loop	RR_LSRB5
	jmp	RR_LSN				;  now start checking CRC




;----------------------------------------------------------------------------; 	
;                     -Process breakin request-			             ;
;----------------------------------------------------------------------------; 	
; Fill the output buffer with idles					     ;
;----------------------------------------------------------------------------; 	
RRR_CLA20:
	mov	byte ptr BREAKIN,0		; Show we processed it
RRR_CLA20a:
	mov	byte ptr out_msg_stat,0 	; packet cntr=0

	cmp	byte ptr repeat_CS3,0		; this a repeat??		
	je	R1
	jmp	RR_TO3				; yes, resend last frame
						; else load a new buffer
;----------------------------------------------------------------------------;
; If there is data available, send it,					     ;
; Fill packet (7 bytes at 100 baud, 18 bytes at 200 baud)		     ;
; If no data, pad with idle bytes					     ;
;----------------------------------------------------------------------------;
R1:
	mov	byte ptr out_msg_stat,0 	; packet cntr=0

	mov	cx,18			; Initially fill message with idles
	mov	al,byte ptr IDLE
	lea	bx, out_msg+2
RR_CLA21:
	mov	byte ptr [bx],al
	inc	bx
	loop	RR_CLA21

	cmp	byte ptr TX_SPEED,0FFh
	je	C50B0_hi
	mov	cx,7	  			; 7 bytes
	mov	al,byte ptr out_msg_stat	; set up status byte
	mov	byte ptr out_msg_hdr+9,al	; (included in CRC)
	jmp	SHORT C50B0_ad
C50B0_hi:
	mov	cx,18				; 18 bytes
C50B0_ad:
	lea	bx,out_msg+2			; destination
	mov	out_ptr,bx

C50B0_fill:
	mov	ax,cs:tx_take_from_ptr		; is there data ?
	cmp	ax,cs:tx_store_at_ptr
	jne	C50B0_get_dat			; yes, retrieve it

	mov	al,IDLE				; else pad with IDLE
	jmp	SHORT C50B0_stor

C50B0_get_dat:
	mov	bx,tx_take_from_ptr
	mov	al,byte ptr [bx]		; get the character
	inc	bx				; update pointer
	cmp	bx,offset TXBUFR+TXBUFSIZ 	; test for wraparound
	jne	C50B0_ge_nxt
	lea	bx,TXBUFR
C50B0_ge_nxt: 
	mov	tx_take_from_ptr,bx

	call	ECHO_RX_BYTE		; note ax,bx is saved
C50B0_stor:
	mov	bx,out_ptr
	mov	byte ptr [bx],al
	inc	bx
	mov	out_ptr,bx
	loop	C50B0_fill

;----------------------------------------------------------------------------;
; Prepare for CRC calculation						     ;
;----------------------------------------------------------------------------;
	mov	cx,19 
	cmp	byte ptr RX_SPEED,0		; Check speed
	jne	RR_TO1
	mov	cx,8				; low speed
	mov	al,byte ptr out_msg_stat
	mov	byte ptr out_msg+9,al 		; packet cntr=0

RR_TO1:
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	lea	bx,out_msg+2
RR_blk2_2:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	RR_blk2_2

; Convert the CRC to an equivalent form
	cmp	byte ptr RX_SPEED,0	  	; Check speed
	jne	RR_TO2

	mov	al,byte ptr CRC_work 		; 100 baud
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg+11,al
				      
	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg+10,al
	jmp	RR_TO3

RR_TO2:
	mov	al,byte ptr CRC_work 		; 200 baud
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg_hdr+23,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg_hdr+22,al
;
; Send CS3 + remainder of message
RR_TO3:
	mov	bx,CTTX 		; Get to CS time
	call	HOLD	
RRR_send:
	mov	CONTROL,CS3		; Now handle the breakin
	call	TCTL			; Send CS3 w/o following TXOFF
	
	cmp	byte ptr RX_SPEED,0
	jne	RR_TO5

; Sending at 100 baud
	mov	byte ptr tx_bit_sense,0FFh ; figure out send sense
	cmp	word ptr SEND_SENSE,0h
	je	RRR_CLA2ax
	mov	byte ptr tx_bit_sense,0h

RRR_CLA2ax:
	lea	bx,out_msg+2
	call	TXCH_S			; send filler nibble

	mov	cx,10			; Next 10 bytes is at 100 baud
RRR_CLA3x:	
	call    TXCH_L			; TXCH includes xor with bit sense
	inc	bx			; 100 baud
	loop	RRR_CLA3x
	jmp	RR_TO8

; Sending at 200 baud
RR_TO5:
	mov	byte ptr tx_bit_sense,0FFh ; figure out send sense
	cmp	SEND_SENSE,0h
	je	RRR_CLA2a
	mov	byte ptr tx_bit_sense,0h
RRR_CLA2a:
	mov	cx,21			; Next 21 bytes is at 200 baud
	lea	bx,out_msg+2
RRR_CLA3:	
	call    TXCH_H			; TXCH includes xor with bit sense
	inc	bx			; 200 baud
	loop	RRR_CLA3

RR_TO8:
	call	TXOFF			; turn TX off

	mov	RxT_cnt,0		; reset RxT counter
;----------------------------------------------------------------------------;
; Response to our CS3 block:						     ;
;									     ;
; 100 baud:								     ;
;	CS1 - ACK - continue as ISS		                             ;
;	CS3 - ACK - we are not allowed to continue as ISS, go process this   ;
;                   as in the ISS breakin section                            ;							     ;
;	CS2 - NAK - resend CS3 packet                                        ;    	packet sent at 100 baud      ;
;       T/O -(NAK)- resend CS3 packet					     ;
;									     ;
; 200 baud:								     ;
;	CS1 - ACK - continue as ISS				             ;
;	CS3 - ACK - we are not allowed to continue as ISS, go process this   ;
;                   as in the ISS breakin section                            ;							     ;
;	CS2 - NAK - resend CS3 packet                                        ;    	packet sent at 100 baud      ;
;       T/O -(NAK) -resend CS3 packet					     ;
;       CS4 - NAK - if the CS3 frame is resent too many times, IRS can       ;
;                   request a speedown. Continue in ISS CS4 section	     ;
;----------------------------------------------------------------------------;

	mov	bx,CLOCK    		; set up auxilliary clock
	mov	AUX_CLOCK,bx		; so that we can restore our old time

;----------------------------------------------------------------------------;
; Assuming that our breakin packet is going to be received OK.		     ;
;----------------------------------------------------------------------------;

	xor	bx,bx 			; resets our time frame
	mov	bl,tx_delay
	add	bx,960
	mov	CLOCK,bx

	call	AN_CS 			; get response
;----------------------------------------------------------------------------;
; Received turn-round acknowledgement, set up rest of time frame	     ;
;----------------------------------------------------------------------------;

	cmp	CONTROL,CS1		; Turnround packet accepted ?
	je	RRR_CLA4
	cmp	CONTROL,CS3		; Rebreak ?
	je	RRR_CLA4
	cmp	CONTROL,CS4		; Speedown ?
	je	RRR_CLA4

;----------------------------------------------------------------------------;
; Reject (CS2) or T/O							     ;
; Our CS3 block failed to check out at IRS, resend the data in the CS3 packet;
;----------------------------------------------------------------------------;
	cmp     byte ptr QFLG,1    	; test for manual quit request
	jne	RRR_RJ1		   	; skip out if not
	call	TXOFF			; else, turn TX off
	jmp	STBY		   	; go to STANDBY

RRR_RJ1:
	inc	byte ptr rq_cnt		; update RQ count
	cmp	byte ptr rq_cnt,MAX_RQ	; Too many ?
	js	RRR_RJ1a
;
; If too many RQ's in a row - terminate
	jmp	STBY

RRR_RJ1a:				; not too many retries yet
	mov	bx,AUX_CLOCK		; reset our clock
	mov	CLOCK,bx

	xor	byte ptr tx_bit_sense,0FFh ; ready next transmit sense

	mov	byte ptr repeat_CS3,1	; show thats its going to be a repeat
					; to prevent reload of data buffer
	jmp	RRR_CLA20 		; go resend CS3 packet

;----------------------------------------------------------------------------;
; Continue as ISS now							     ;
;----------------------------------------------------------------------------;
RRR_CLA4:
	mov	ax,RxT_cnt		; compute RxT parameter
	sub	ax,110+GDCS

	cmp	RxT,0
	jnz	Z2
	call	init_RxT		; initialize ideal_RxT
Z2:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update old_RxT array

	mov	ax,CLOCK		; XXXXXX
	sub	ax,110+GDCS 		; Mirroring calc in proc. CLA40	
	mov	ideal_CTRX,ax
	call	init_CTRX		; initialize ideal_CTRX

	mov	byte ptr TX_SPEED,0FFh	 ; assume high speed
	cmp	byte ptr RX_SPEED,0	 ; determine speed
	jne	RRR_CLA4a	
	mov	byte ptr TX_SPEED,0	 ; set low

;
; Ensure sending bit sense is correct
RRR_CLA4a:
	mov	byte ptr tx_bit_sense,0h
	cmp	word ptr SEND_SENSE,0
	je	RRR_0
	mov	byte ptr tx_bit_sense,0FFh
RRR_0:	
	mov	byte ptr STATUS,11h  	; STATUS=sending linked

	mov	byte ptr repeat_CS3,0	; clear for future use
;----------------------------------------------------------------------------;
; At this point, several different paths are possible:                        ;
; CS1 - continue with normal ISS  processing				     ;
; CS3 - continue with ISS rebreak processing				     ;
; CS4 - continue with ISS speedown processing				     ;
;----------------------------------------------------------------------------;
	cmp	CONTROL,CS3		; Rebreak
	jne	RRR_1
	jmp	rebreak_entry
RRR_1:
	cmp	CONTROL,CS4		; Speedown
	jne	RRR_2
	jmp	CR_CS4_OK
RRR_2:					; CS1 normal continuation as ISS
	mov	word ptr CS:ECONTROL,CS1   ; next expect a CS1
	mov	byte ptr out_msg_hdr,SYNC1 ; with header=55h

	jmp	CLA40a
;============================END OF IRS======================================;

;---------------- Some Helper functions -------------------------------------;
;                    Call sign search 					     ;
;----------------------------------------------------------------------------;
SRCH:   mov     cx,960           	; initially fill search registers
SRC0:   call    SMPLE_8
        loop	SRC0
	
SRC1:	call	SMPLE_8			;  get another bit sample
	mov	bx,RPTR_L
	cmp	byte ptr [bx],SYNC0 	; Do we have a SYNCH (normal sense)
	je	SYNC_NORM			
	cmp	byte ptr [bx],SYNC1 	; perhaps a SYNCH (reverse sense ?)
	je	SYNC_REV

SRC2:   mov     ah,QFLG         	; test if qrt is set (during RESYNCH)
	or	ah,ah			; SET UP FLAGS
        jnz     STBYL           	; exit to STBY if yes
        jmp     SHORT SRC1      	; else keep searching

STBYL:	jmp	STBY 			; go to STANDBY
;
; Make sure this is'nt a spurious sync, get another sample
;
SYNC_NORM:  		       
	call	SMPLE_8			;  get another bit sample
	mov	bx,RPTR_L
	cmp	byte ptr [bx],SYNC0 	; Do we have a SYNCH (normal sense)
	je	SYNC_NORM1			
	jmp	SHORT SRC1		; seems spurious, continue searching

SYNC_REV:
	call	SMPLE_8			;  get another bit sample
	mov	bx,RPTR_L
	cmp	byte ptr [bx],SYNC1 	; perhaps a SYNCH (reverse sense ?)
	je	SYNC_REV1
	jmp	SHORT SRC1		; seems spurious, continue searching
;
; Looks like a good sync
;
SYNC_NORM1:  				; SYNC detected
SYNC_REV1:
;----------------------------------------------------------------------------;
; A synch character has been detected					     ;
; analyse remainder of selcal packet.					     ;
;----------------------------------------------------------------------------;
;----------------------------------------------------------------------------;
; Test 200 baud address first						     ;
;----------------------------------------------------------------------------;
rd_first_bk:			
	mov	byte ptr cs:SPEED,0FFh	; 200 baud

	mov	byte ptr cs:bit_sense,00  ; assume normal sense
	mov	bx,RPTR_L
	mov	al,byte ptr cs:[bx]	; sync byte
	mov	sync_byte,al

	mov	bx,RPTR_H
	mov	al,byte ptr cs:[bx]
	cmp	al,our_ID
	jne	RRX_2a
	mov	al,byte ptr cs:[bx+5]
	cmp	al,our_ID+1
	jne	RRX_2a
	mov	al,byte ptr cs:[bx+10]
	cmp	al,our_ID+2
	jne	RRX_2a
	mov	al,byte ptr cs:[bx+15]
	cmp	al,our_ID+3
	jne	RRX_2a
	mov	al,byte ptr cs:[bx+20]
	cmp	al,our_ID+4
	jne	RRX_2a
	jmp	RR_sync
RRX_2a:	
	mov	byte ptr cs:bit_sense,0FFh 	; try inverted sense
	mov	bx,RPTR_L
	mov	al,byte ptr cs:[bx]	; sync byte
	xor	al,bit_sense
	mov	sync_byte,al

	mov	bx,RPTR_H
	mov	al,byte ptr cs:[bx]
	xor	al,bit_sense
	cmp	al,our_ID
	jne	RRX_1
	mov	al,byte ptr cs:[bx+5]
	xor	al,bit_sense
	cmp	al,our_ID+1
	jne	RRX_1
	mov	al,byte ptr cs:[bx+10]
	xor	al,bit_sense
	cmp	al,our_ID+2
	jne	RRX_1
	mov	al,byte ptr cs:[bx+15]
	xor	al,bit_sense
	cmp	al,our_ID+3
	jne	RRX_1
	mov	al,byte ptr cs:[bx+20]
	xor	al,bit_sense
	cmp	al,our_ID+4
	jne	RRX_1
	jmp	RR_sync

; If 200 baud address failed, test 100 baud address
RRX_1:
	mov	byte ptr cs:SPEED,00	; 100 baud

	mov	byte ptr cs:bit_sense,00  ; assume normal sense
	mov	bx,RPTR_L
	mov	al,byte ptr cs:[bx]	; sync byte
	mov	sync_byte,al

	mov	al,byte ptr cs:[bx+10]
	cmp	al,our_ID
	jne	RRXa
	mov	al,byte ptr cs:[bx+20]
	cmp	al,our_ID+1
	jne	RRXa
	mov	al,byte ptr cs:[bx+30]
	cmp	al,our_ID+2
	jne	RRXa
	mov	al,byte ptr cs:[bx+40]
	cmp	al,our_ID+3
	jne	RRXa
	mov	al,byte ptr cs:[bx+50]
	cmp	al,our_ID+4
	jne	RRXa
	jmp	SHORT RR_sync
RRXa:	
	mov	byte ptr cs:bit_sense,0FFh 	; try inverted sense
	mov	bx,RPTR_L
	mov	al,byte ptr cs:[bx]	; sync byte
	xor	al,bit_sense
	mov	sync_byte,al

	mov	bx,RPTR_L
	mov	al,byte ptr cs:[bx+10]
	xor	al,bit_sense
	cmp	al,our_ID
	jne	RRX
	mov	al,byte ptr cs:[bx+20]
	xor	al,bit_sense
	cmp	al,our_ID+1
	jne	RRX
	mov	al,byte ptr cs:[bx+30]
	xor	al,bit_sense
	cmp	al,our_ID+2
	jne	RRX
	mov	al,byte ptr cs:[bx+40]
	xor	al,bit_sense
	cmp	al,our_ID+3
	jne	RRX
	mov	al,byte ptr cs:[bx+50]
	xor	al,bit_sense
	cmp	al,our_ID+4
	jne	RRX
	jmp	SHORT RR_sync

RRX:	jmp	SRC1 			; Keep scanning for selcal packet

RR_sync:				; Exit - selcal packet detected
	ret

;----------------------------------------------------------------------------;
;      Stores data received through PACTOR in receive buffer		     ;
;      Character to be stored in AL					     ;
;----------------------------------------------------------------------------;
STORE_RX_BYTE:
	push	ax
	push	bx
	mov	bx,rx_store_at_ptr 	    ; is there space in rx buffer ?
	cmp	bx,offset RXBUFR+RXBUFSIZ-1 ; check wrap round case
	jne 	SB0
	cmp	rx_take_from_ptr,offset RXBUFR	; handle wrap around case
	jne 	SB1			; ok there is space left
	jmp	SHORT SB5   		; exit if no space left
SB0:	inc	bx			; normal case
	cmp	bx,rx_take_from_ptr
	je	SB5			; exit if no space available

SB1:	mov	bx,rx_store_at_ptr	; else save it
	mov	byte ptr [bx],al	
	inc	bx			; update buffer pointer
	cmp	bx,offset RXBUFR+RXBUFSIZ
	jne	SB2
	lea	bx,RXBUFR
SB2:	mov	rx_store_at_ptr,bx

SB5:	pop	bx
	pop	ax
	ret

;----------------------------------------------------------------------------;
;      Stores data echoed through PACTOR in receive buffer		     ;
;      Character to be echoed in AL					     ;
;----------------------------------------------------------------------------;
ECHO_RX_BYTE:
	push	ax
	push	bx
	mov	bx,echo_store_at_ptr 	    ; is there space in rx buffer ?
	cmp	bx,offset ECHOBUFR+RXBUFSIZ-1 ; check wrap round case
	jne 	ESB0
	cmp	echo_take_from_ptr,offset ECHOBUFR	; handle wrap around case
	jne 	ESB1			; ok there is space left
	jmp	SHORT ESB5   		; exit if no space left
ESB0:	inc	bx			; normal case
	cmp	bx,echo_take_from_ptr
	je	ESB5			; exit if no space available

ESB1:	mov	bx,echo_store_at_ptr	; else save it
	mov	byte ptr [bx],al	
	inc	bx			; update buffer pointer
	cmp	bx,offset ECHOBUFR+RXBUFSIZ
	jne	ESB2
	lea	bx,ECHOBUFR
ESB2:	mov	echo_store_at_ptr,bx

ESB5:	pop	bx
	pop	ax
	ret
;----------------------------------------------------------------------------;
; Transmit in unprotected mode, entered with control code 24		     ;
; Note that either 100 or 200 baud rate could be use and that a variable     ;
; number of repeats may be included as well. No special timing frames are    ;
; used, except that frames follow right up one another.			     ;
;----------------------------------------------------------------------------;
CALLB:  
	lea     bx,AM_STACK 		; reset AMTOR stack pointer
        mov     sp,bx           	; SP
        mov     ax,cs           	; SS
        mov     ss,ax

	mov	byte ptr STATUS,14h  	; STATUS=Sending Unproto
	mov	byte ptr tx_bit_sense,0 ; normal transmit bit sense
;
; Prepare output packets
;
	mov	byte ptr out_msg_stat,0 	; packet cntr=0
	mov	byte ptr out_msg_hdr,SYNC0 	; 055h

	call    TXON            	; turn TX on and wait for TX-delay
;
; If there is data available, send it,
; unless a QRT request is pending

CALB0:	mov	al,U_REPEAT
	mov	byte ptr u_count,al	; Set Unproto repeat count
				
	cmp     byte ptr QFLG,1    	; test if qrt request is set
	je	CALBX1		   	; empty buffer first

	cmp     byte ptr QFLG,0FFh    	; test if final qrt request is set
	je	CALBX2		   	; then go to STBY
;
; Fill packet (8 bytes at 100 baud, 20 bytes at 200 baud)
; If no data, pad with idle bytes
;
CALB0_data:
	cmp	byte ptr u_speed,0FFh
	je	CALB0_hi
	mov	cx,8	  		; 8 bytes
	mov	al,byte ptr out_msg_stat  ; set up status byte
	mov	byte ptr out_msg_hdr+9,al ; (included in CRC)
	jmp	SHORT CALB0_ad
CALB0_hi:
	mov	cx,20			; 20 bytes
CALB0_ad:
	lea	bx,out_msg		; destination
	mov	out_ptr,bx

CALB0_fill:
	mov	ax,cs:tx_take_from_ptr	; is there data ?
	cmp	ax,cs:tx_store_at_ptr
	jne	CALB0_get_dat		; yes, retrieve it

	mov	al,IDLE			; else pad with IDLE
	jmp	SHORT CALB0_stor

;----------------------------------------------------------------------------;
; Before the shutdown, make sure buffer is empty			     ;
;----------------------------------------------------------------------------;
CALBX1:	
	mov	ax,cs:tx_take_from_ptr	; data left?
	cmp	ax,cs:tx_store_at_ptr
	jne	CALB0_data		; yes, keep sending it

	mov	byte ptr QFLG,0FFh	; signal sending of QRT status
	or	byte ptr out_msg_stat,080h ; set QRT bit in status byte
	jmp	CALB0_data		; and send last one

CALBX2:	
	call	TXOFF			; else, turn TX off
	jmp	STBY		   	; go to STANDBY

CALB0_get_dat:
	mov	bx,tx_take_from_ptr
	mov	al,byte ptr [bx]	; get the character
	inc	bx			; update pointer
	cmp	bx,offset TXBUFR+TXBUFSIZ ; test for wraparound
	jne	CALB0_ge_nxt
	lea	bx,TXBUFR
CALB0_ge_nxt: 
	mov	tx_take_from_ptr,bx

	call	ECHO_RX_BYTE		; note ax,bx is saved

CALB0_stor:
	mov	bx,out_ptr
	mov	byte ptr [bx],al
	inc	bx
	mov	out_ptr,bx
	loop	CALB0_fill

CALB0_repeat:
	mov	al,byte ptr u_speed
	mov	byte ptr TX_SPEED,al
	call	TXD_pack

; Update next status byte
CALB2b: mov	al,byte ptr out_msg_stat
	inc 	al
	and	al,03h
	mov	byte ptr out_msg_stat,al

CALB2d:	jmp	CALB0			; ready for next block

;----------------------------------------------------------------------------;
;                         - ISS section	-				     ;
;  Entered entered with control code 24	from control section (user initiated ;
;  or from the IRS section aftera turn round received - entry at CLA40a	     ;
;----------------------------------------------------------------------------;
; A word about how timing is interpreted in this section:                    ;
; CLOCK = 0000 is where we pull in our PTT, wait for TX-DELAY, and the end   ;
; of the last databit from the data block goes out at 960+TX_DELAY 	     ;								     ;
;									     ;
; We see the leading edge of the received CS response at CTRX. This includes ;
; some propagation delays and the IRS's	tx_delay + CS_delay times            ;
;----------------------------------------------------------------------------;
CALLA:  

ifdef LIMITED
	jmp	MON
endif	

	lea     bx,AM_STACK 		; reset AMTOR stack pointer
        mov     sp,bx           	; SP
        mov     ax,cs           	; SS
        mov     ss,ax

	mov	byte ptr EVM_cnt,0	; ready slave sync CLOCK
	mov	byte ptr DSP_cnt,0	; ready slave sync DSP clock

	mov	byte ptr STATUS,10h  	; STATUS=Attempting to link

	mov	byte ptr MASTER,0    	; Master phasing
	mov	byte ptr tx_bit_sense,0 ; normal transmit bit sense
	mov	byte ptr tx_sense,0	; for display
	mov	byte ptr rx_sense,0	; for display
	mov	byte ptr rq_cnt,0	; count of RQ's sent 
	mov	byte ptr err_cnt,0	; Clear errors
	mov	byte ptr marq_cnt,0	; Clear marq counts
        mov     QRT_request,0         	; Clear qrt_request (ISS)  flag
        mov     QRT_pending,0         	; Clear qrt_request (KBrd) flag
        mov     QRT_sent,0         	; Clear qrt sent flag

	mov	byte ptr ave_RxT_cnt,AVE_RXT	; compute initial RxT over 5 cycles

	mov	ax,0
	mov	RxT,ax			; clear tracker
	mov	phase_err,ax		; for display
	call	init_RxT		; initialize ideal_RxT

	call	zero_phase		; initially no phase error
	mov	word ptr prev_CS,0	; just so that last CS not CS4

;----------------------------------------------------------------------------;
; Prepare selcal block for initiating the ARQ call			     ;
;----------------------------------------------------------------------------;
	mov	byte ptr selcal_hdr,SYNC0 ; header = 0x55
	lea	bx, callee_ID		; source address
	push	si
	mov	si,bx
	lea	bx,selcal_msg1		; selcal at 100 baud
	mov	cx,8			; byte count
CLA0:	mov	al,cs:[si]
	mov	cs:[bx],al

;===============================================
IFNDEF LOW_SPEED
	mov	cs:[bx+8],al
ELSE
	mov	al,055h
	mov	cs:[bx+8],al
ENDIF
;===============================================

	inc	si
	inc	bx
	loop	CLA0
	pop	si

;----------------------------------------------------------------------------;
; Ready to send out an ARQ call 					     ;
;----------------------------------------------------------------------------;
	mov	CLOCK,0
	jmp	SHORT CLA1

CLA00:	mov	bx,0			; Entry point also for CS retries,
	call	HOLD			; thus we wait till CLOCK=0
					; this also entry point for T-O's

CLA1:	cmp     byte ptr QFLG,1    	; test if qrt request is set
	jne	CLA1a		   	; skip out if not
CLA1aa:	call	TXOFF			; else, turn TX off
	jmp	STBY		   	; go to STANDBY

CLA1a:	call    TXON            	; turn TX on and wait for TX-delay
	mov	cx,9			; first 9 bytes is at 100 baud
	lea	bx,selcal_hdr
CLA2:	call    TXCH_L			; TXCH includes xor with bit sense
	inc	bx			; 100 baud
	loop	CLA2

	mov	cx,6			; Next 6 bytes is at 200 baud
	lea	bx,selcal_msg2
CLA3:	call    TXCH_H			; TXCH includes xor with bit sense
	inc	bx			; 200 baud
	loop	CLA3

	call	TXOFF			; turn TX off

	mov	RxT_cnt,0		; reset tracker 

	xor	byte ptr tx_bit_sense,0FFh	; ready for next bit sense

;----------------------------------------------------------------------------;
; Get a CS response							     ;
; Note that we are looking for several consequitive samples (#SYNCCS) bits   ;
; of a CS. This is necessary as noise often looks like a valid CS	     ;
;----------------------------------------------------------------------------;
CLA11_0:
	mov 	al,SYNCCS		; reset good CS counts
	mov	byte ptr cs_cnt1,al
	mov	byte ptr cs_cnt4,al

;----------------------------------------------------------------------------;
; We need to flush the input buffer so that we are sure that we are	     ;
; only looking at a new CS, not part of the old data buffer		     ;
;----------------------------------------------------------------------------;
	mov	cx,50			
  	lea     bx,REGS_12      	; set search register pointer
        mov     RPTR_CTL,bx
CL1:	call	SMPLE_12
   	loop	CL1

CLA12:	call	SMPLE_12		;  get another sample
	mov	bx,RPTR_CTL
	mov	ax, word ptr [bx]
	mov	cx,4
	shr	ax,cl			; convert to 12 bits

	mov	cx,CS1			; CS1 normal sense ?
	cmp	ax,cx
	je	CL_CS1
	mov	cx,CS1			; CS1 inverted ?
	xor	cx,0FFFh
	cmp	ax,cx
	je	CL_CS1r

	mov	cx,CS4			; link reply CS4 ?
	cmp	ax,cx
	jne	CLA12yy
	jmp	CL_CS4
CLA12yy:
	mov	cx,CS4			; CS4 inverted ?
	xor	cx,0FFFh
	cmp	ax,cx
	jne	SHORT CLA12xx
	jmp	CL_CS4r

CLA12xx:
	cmp	word ptr CLOCK,1230 	; Timed out ?
	jns	CLA12a
	jmp	SHORT CLA12		; not yet, look some more
CLA12a:	mov	word ptr CONTROL,0	; for status display
	mov 	al,SYNCCS
	mov	byte ptr cs_cnt1,al
	mov	byte ptr cs_cnt4,al
	jmp	CLA00 			; skip out if timed out

;----------------------------------------------------------------------------;
; Received a valid sync							     ;
;----------------------------------------------------------------------------;
CL_CS1:	mov	cx,CLOCK		; save clock
	mov	byte ptr cs:rx_sense,0	; for status display
	mov	byte ptr rx_bit_sense,0	; Received normal bit sense
	mov	byte ptr TX_SPEED,0FFh 	; 100 baud
	mov	word ptr CONTROL,CS1	; for status display
	dec	byte ptr cs_cnt1
	jnz	SHORT CL_CS1xx
	jmp	CLA20a
CL_CS1xx:
	mov 	al,SYNCCS
	mov	byte ptr cs_cnt4,al
	jmp	CLA20_more_a

CL_CS1r:				; reversed sense
	mov	cx,CLOCK		; save clock
	mov	byte ptr cs:rx_sense,0FFh	; for status display
	mov	byte ptr rx_bit_sense,0FFh	; Received normal bit sense
	mov	byte ptr TX_SPEED,0FFh 	; 100 baud
	mov	word ptr CONTROL,CS1	; for status display
	dec	byte ptr cs_cnt1
	jz	CLA20a
	mov 	al,SYNCCS
	mov	byte ptr cs_cnt4,al
	jmp	CLA20_more_a

CL_CS4:
	mov	cx,CLOCK		; save clock
	mov	byte ptr cs:rx_sense,0	; for status display
	mov	byte ptr rx_bit_sense,0	; Received normal bit sense
	mov	byte ptr TX_SPEED,0h	; 200 baud
	mov	word ptr CONTROL,CS4	; for status display
	dec	byte ptr cs_cnt4
	jz	CLA20a
	mov 	al,SYNCCS
	mov	byte ptr cs_cnt1,al
	jmp	SHORT CLA20_more_a

CL_CS4r:
	mov	cx,CLOCK		; save clock
	mov	byte ptr cs:rx_sense,0FFh	; for status display
	mov	byte ptr rx_bit_sense,0FFh	; Received normal bit sense
	mov	byte ptr TX_SPEED,0h	; 200 baud
	mov	word ptr CONTROL,CS4	; for status display
	dec	byte ptr cs_cnt4
	jz	CLA20a
	mov 	al,SYNCCS
	mov	byte ptr cs_cnt1,al
	jmp	SHORT CLA20_more_a

; Scanning for more valid bits to confirm the same CS
;
CLA20_more_a:
	jmp	CLA12			; get another bit of CS

;----------------------------------------------------------------------------;
; Ready to send "system level" message					     ;
; But make sure we have at least two good CS's in a row			     ;
;----------------------------------------------------------------------------;
CLA20a:
	sub	cx,110+SYNCCS 		   ; actual CTRX is 120ms earlier	
	mov	ideal_CTRX,cx		   ; save for tracking bit phasing			
	mov	ax,cx
	call	init_CTRX		   ; initalize ideal_CTRX

;----------------------------------------------------------------------------;
; Compute the average RxT, for the first time use the first value            ;
;----------------------------------------------------------------------------;
	mov	ax,RxT_cnt
	sub	ax,110+SYNCCS

	cmp	RxT,0
	jnz	Z3
	call	init_RxT		; initialize ideal_RxT
Z3:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update old_RxT array

	mov	byte ptr out_msg_stat,1    ; packet  cntr = 1
	mov	byte ptr out_msg_hdr,SYNC0 ; first header = 55H

	mov	cx,20			; Initially fill message with idles
	mov	al,IDLE
	lea	bx, out_msg
CLA21:
	mov	byte ptr [bx],al
	inc	bx
	loop	CLA21

	mov	cx,7	  			; 7+1 bytes in Level message

	cmp	byte ptr TX_SPEED,0FFh
	je	CLA21_hi
	mov	al,byte ptr out_msg_stat	; set up status byte
	mov	byte ptr out_msg_hdr+9,al	; (included in CRC)

CLA21_hi:
	lea	bx,out_msg			; destination
	mov	di,bx

	lea	bx,level			; level message
	mov	si,bx

CLA21_fill:
	mov	al,byte ptr [si]	; get the character
	inc	si			; update pointer

	call	ECHO_RX_BYTE		; note ax,bx is saved

	mov	byte ptr [di],al
	inc	di
	loop	CLA21_fill

	mov	al,0Dh			; end with CR
	call	ECHO_RX_BYTE		; note ax,bx is saved
	mov	byte ptr [di],al	; end with CR

;----------------------------------------------------------------------------;
; Since CS1/CS4 was received at the selcal				     ;
; the expected control is either CS1/CS4 for a repeat or CS2 for	     ;
; to send the next block. 						     ;
;----------------------------------------------------------------------------;
CLA22a:
	mov	word ptr ECONTROL,CS2	; If all is well, expect a CS2
	cmp	TX_SPEED,00
	jne	CLA22b			; For low speed
	mov	word ptr ECONTROL,CS1	; If all is well, expect a CS1
CLA22b:
	mov	byte ptr STATUS,11h  	; STATUS=sending linked

; The main transmit loop
CLA29:
	mov 	al,GDCS
	mov	byte ptr cs_cnt1,al
	mov	byte ptr cs_cnt2,al
	mov	byte ptr cs_cnt3,al
	mov	byte ptr cs_cnt4,al

CLA29b:
	mov	bx,0			; wait till CLOCK=0
	call	HOLD

	mov	al,byte ptr TX_SPEED
	mov	byte ptr DS_SPEED,al	; for status display
	call	TXD_pack		; Send it
	call	TXOFF

	mov	RxT_cnt,0		; reset timing tracker

; Alternate receive bit sense of last CS
	xor	byte ptr rx_bit_sense,0FFh

;----------------------------------------------------------------------------;
; Get a response.							     ;
; Note that we are looking for several consequitive samples (#GDCS) bits     ;
; of a CS. This is necessary as noise often looks like a valid CS.	     ;
; Further note that SYNCCS is more strict than GDCS.                         ;
;----------------------------------------------------------------------------;
	mov	bx,ideal_CTRX
	sub	bx,CTRX_SLOP		; allow for IRS's CS drift
	call	HOLD			; now wait ---

;----------------------------------------------------------------------------;	
; Collect 110ms worth of data - for "priming" the input buffer               ;
;----------------------------------------------------------------------------;	
  	lea     bx,REGS_12
        mov     RPTR_CTL,bx
	mov	cx,110
CLA31:	call	SMPLE_12
	loop	CLA31

CLA32:	call	SMPLE_12		;  get another sample
	mov	bx,RPTR_CTL
	mov	ax, word ptr [bx]
	mov	cx,4
	shr	ax,cl			; convert to 12 bits

;----------------------------------------------------------------------------;
; Look only at data coming in at the right signal sense	to reduce the chance ;
; of an erroneously received CS slipping in.                                 ;
;----------------------------------------------------------------------------;
	cmp	byte ptr rx_bit_sense,0	; Are we expecting normal sense ?
	jne	CLA32_inv		; No, skip out

;-------The normal set -----------------------------
	mov	cx,CS1			; link reply CS1 ?
	cmp	ax,cx
	jne	CLA32p
	jmp	CR_CS1

CLA32p:	mov	cx,CS2			; link reply CS2 ?
	cmp	ax,cx
	jne	CLA32q
	jmp	CR_CS2

CLA32q:	mov	cx,CS3			; link reply CS3 ?
	cmp	ax,cx			; Link turnaround request
	jne	CLA32u
	jmp	CR_CS3xx

CLA32u:	mov	cx,CS4			; link reply CS4 ?
	cmp	ax,cx
	jne	CLA32w
	jmp	CR_CS4

;-------The inverted set -----------------------------
CLA32_inv:
	mov	cx,CS1
	xor	cx,0FFFh
	cmp	ax,cx
	jne	CLA32r
	jmp	CR_CS1_INV

CLA32r:	mov	cx,CS2
	xor	cx,0FFFh
	cmp	ax,cx
	jne	CLA32s
	jmp	CR_CS2_INV

CLA32s:	mov	cx,CS3
	xor	cx,0FFFh
	cmp	ax,cx
	jne	CLA32t
	jmp	CR_CS3_INVxx

CLA32t:	mov	cx,CS4
	xor	cx,0FFFh
	cmp	ax,cx
	jne	CLA32w
	jmp	CR_CS4_INV

; A timeout is the same as not receiving the correct response
;
CLA32w:	mov	bx,ideal_CTRX
	add	bx,120+CTRX_SLOP	; CS length + slop	
	cmp	word ptr CLOCK,bx 	; Timed out ?
	jns	CLA32a
	jmp	CLA32			; not yet, look some more

;----------------------------------------------------------------------------;
; Timed out -								     ;
; We should not attempt slave phasing on timeouts as we have no idea how     ;
; far well drift away.							     ;
;----------------------------------------------------------------------------;
CLA32a:
	mov	word ptr CONTROL,0	; for status display
	cmp     byte ptr QFLG,1    	; test if qrt request is set
	jne	CLA32b		   	; skip out if not
	call	TXOFF			; else, turn TX off
	jmp	STBY		   	; go to STANDBY
CLA32b:	

;-----------------------------------------------------------------------------;
; On time-outs, update the rq_count. If too many RQ's - quit		      ;
;-----------------------------------------------------------------------------;
CLA32c:
	inc	byte ptr rq_cnt
	cmp	byte ptr rq_cnt,MAX_RQ
	js	CLA_CS_RQa
	jmp	STBY
CLA_CS_RQa:
	mov 	al,GDCS
	mov	byte ptr cs_cnt1,al
	mov	byte ptr cs_cnt2,al
	mov	byte ptr cs_cnt3,al
	mov	byte ptr cs_cnt4,al
	jmp	CLA29b			; resend last packet

;----------------------------------------------------------------------------;
;   --- ISS control received--(EXCLUDING CS3 - THAT FOLLOWS THIS CODE)----   ;
; Received CS4								     ;
; If we are sending at high speed, this is a request for speed down	     ;
;                                  as well as reqesting a repeat	     ;
; If we are at low speed, this is a request for speed up as well as	     ;
;                                  an acknowledgement			     ;
;----------------------------------------------------------------------------;
CR_CS4:
	mov	byte ptr cs:rx_sense,0	; for status display
	mov	byte ptr rx_bit_sense,0	; Received normal bit sense
	mov	word ptr CONTROL,CS4	; for status display
	dec	byte ptr cs_cnt4	; good CS ?
	jnz	CR_more_jmp
	jmp	SHORT CR_CS4_OK

CR_more_jmp: 	
	mov	byte ptr cs_cnt1,GDCS
	mov	byte ptr cs_cnt2,GDCS
	mov	byte ptr cs_cnt3,GDCS
	jmp	CLA32

CR_CS4_INV:
	mov	byte ptr cs:rx_sense,0FFh	; for status display
	mov	byte ptr rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS4	; for status display
	dec	byte ptr cs_cnt4	; good CS ?
	jnz	CR_more_jmp

;----------------------------------------------------------------------------;
; Test if there wasnt a speed change in process,			     ;
; i.e. have we received a good CS since ?				     ;
; This is to prevent CS4, speed down, CS4(RQ), speed up, CS4(RQ) ....	     ;
;----------------------------------------------------------------------------;
CR_CS4_OK:
	cmp	word ptr prev_CS,CS4
	jne	CR_CS4_AOKa
	jmp	CR_CS_RQ		; skip out if change is in progress

;----------------------------------------------------------------------------;
; If we received a CS4 with (rq_cnt > SPEED_THR_TX) and ECONTROL=CS2,	     ;
; the toggle ECONTROL as well						     ;
;----------------------------------------------------------------------------;
CR_CS4_AOKa:

	mov	ax,RxT_cnt		; compute timing tracker
	sub	ax,110+GDCS

	cmp	RxT,0
	jnz	Z4
	call	init_RxT		; initialize ideal_RxT
Z4:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update old_RxT array

	mov	ax,CLOCK		; XXXXXXX
	sub	ax,110+GDCS		; actual CTRX is 120ms earlier
	mov	DS_CTRX,ax		; for tuning display

	mov	ax,phase_err
	call	ave_ph_err  		; arg. in ax
	call	slave_sync_ISS

	mov	ax,DS_CTRX
	call	mean_CTRX

	mov	word ptr prev_CS,CS4	; we received CS4
	cmp	rq_cnt,SPEED_THR_TX
	js	CR_CS4t
	cmp	ECONTROL,CS2
	jne	CR_CS4t

	cmp	word ptr CS:ECONTROL,CS1
	jne	CR_tc1
	mov	word ptr CS:ECONTROL,CS2
	jmp	SHORT CR_tc2
CR_tc1:	mov	word ptr CS:ECONTROL,CS1
CR_tc2:

CR_CS4t:
	mov	word ptr CONTROL,CS4	; for status display
	cmp	byte ptr TX_SPEED,0FFh	; presently at Hi speed ?
	jne	CR_CS4_hi		; no, make it high
	mov	byte ptr TX_SPEED,0	; else make it lo
	jmp	CLA43			; else, resend last packet	

CR_CS4_hi: 		
;----------------------------------------------------------------------------;
; For time being, if we get down to 100 baud, keep it there.		     ;
;----------------------------------------------------------------------------;
	mov	byte ptr TX_SPEED,0FFh	; else make it high
	jmp	CLA40

;----------------------------------------------------------------------------;
; Received a CS1							     ;
;----------------------------------------------------------------------------;
CR_CS1:		      			; Did we expect a CS1?
	mov	byte ptr cs:rx_sense,0	; for status display
	mov	byte ptr rx_bit_sense,0	; Received normal bit sense
	dec	byte ptr cs_cnt1	; good CS ?
	jnz	SHORT CR_CS1aaa
	jmp	SHORT CR_CS1a
CR_CS1aaa:
	mov	byte ptr cs_cnt2,GDCS
	mov	byte ptr cs_cnt3,GDCS
	mov	byte ptr cs_cnt4,GDCS
	jmp	CLA32
CR_CS1_INV:
	mov	byte ptr rx_bit_sense,0FFh	; Received inverted bit sense
	mov	byte ptr cs:rx_sense,0FFh	; for status display
	dec	byte ptr cs_cnt1	; good CS ?
	jnz	CR_CS1aaa
CR_CS1a: 
	mov	word ptr prev_CS,0	; update last CS
	mov	word ptr CONTROL,CS1	; for status display
	cmp	word ptr ECONTROL,CS1	; CS1
	je	CR_xx
	jmp	CR_CS_RQ		; if not, resend last packet	
CR_xx:
	jmp	CLA40			; yes

;----------------------------------------------------------------------------;
; Received CS2								     ;
;----------------------------------------------------------------------------;
CR_CS2:
	mov	byte ptr rx_bit_sense,0	; Received normal bit sense
	mov	byte ptr cs:rx_sense,0	; for status display
	dec	byte ptr cs_cnt2	; good CS ?
	jnz	SHORT CR_CS2aaa
	jmp	SHORT CR_CS2a
CR_CS2aaa:
	mov	byte ptr cs_cnt1,GDCS
	mov	byte ptr cs_cnt3,GDCS
	mov	byte ptr cs_cnt4,GDCS
	jmp	CLA32
CR_CS2_INV:
	mov	byte ptr rx_bit_sense,0FFh	; Received inverted bit sense
	mov	byte ptr cs:rx_sense,0FFh	; for status display
	dec	byte ptr cs_cnt2	; good CS ?
	jnz	CR_CS2aaa
CR_CS2a: 
	mov	word ptr prev_CS,0	; update last CS
	mov	word ptr CONTROL,CS2	; for status display
	cmp	word ptr ECONTROL,CS2
	jne	CR_CS_RQ
	jmp	CLA40			; yes

;----------------------------------------------------------------------------;
; We received the wrong CS, meaning an RQ request			     ;
; Update rq_cnt, when too many RQ's - quit				     ;
;----------------------------------------------------------------------------;
; Compute phase error
;----------------------------------------------------------------------------;
CR_CS_RQ:
	mov	ax,RxT_cnt		; compute timing tracker
	sub	ax,110+GDCS

	cmp	RxT,0
	jnz	Z5
	call	init_RxT		; initialize ideal_RxT
Z5:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update ideal_RxT

	mov	ax,CLOCK		; XXXXXXX
	sub	ax,110+GDCS		; actual CTRX is 120ms earlier
	mov	DS_CTRX,ax		; for tuning display

	mov	ax,phase_err
	call	ave_ph_err 		; arg in ax
	call	slave_sync_ISS

	mov	ax,DS_CTRX
	call	mean_CTRX

	inc	byte ptr rq_cnt
	cmp	byte ptr rq_cnt,MAX_RQ
	js	CR_CS_RQa
	jmp	STBY
CR_CS_RQa:
;;;;;	jmp	CLA43			; ????? else, resend last packet	
	jmp	CLA29			; else, resend last packet	

;----------------------------------------------------------------------------;
; Here when we received acknowledgement that last packet was OK, ready for   ;
; sending net new packet.						     ;
; 									     ;
; May also get here after processing failed speeddown CS4's 		     ;
; However, if we get here after sending a QRT frame, this is the sign	     ;
; to quit and go back to standby		                             ;
;----------------------------------------------------------------------------;
CLA40:
	or	byte ptr QRT_sent,0
	jz	not_quitting_yet
	jmp	STBY

;----------------------------------------------------------------------------;
; Update next header byte						     ;
;----------------------------------------------------------------------------;
not_quitting_yet:

	mov	ax,RxT_cnt		; compute timing tracker
	sub	ax,110+GDCS

	cmp	RxT,0
	jnz	Z6
	call	init_RxT		; initialize ideal_RxT
Z6:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update ideal_RxT

	mov	ax,CLOCK		; XXXXXX
	sub	ax,110+GDCS		; actual CTRX is 121ms earlier	
	mov	DS_CTRX,ax		; for tuning display

	mov	ax,phase_err
	call	ave_ph_err		; arg. in ax
	call	slave_sync_ISS

	mov	ax,DS_CTRX 
	call	mean_CTRX

;----------------------------------------------------------------------------;
; The header byte gets changed 55H <-> AAh on every good received block      ;
;----------------------------------------------------------------------------; 
CLA40a:
	cmp	byte ptr out_msg_hdr,SYNC0
	jne	CLA41
	mov	byte ptr out_msg_hdr,SYNC1
	jmp	SHORT CLA42
CLA41:	mov	byte ptr out_msg_hdr,SYNC0

; Update next expected control
CLA42:	cmp	word ptr ECONTROL,CS1
	jne	CLA41a
	mov	word ptr ECONTROL,CS2
	jmp	SHORT CLA42a
CLA41a:	mov	word ptr ECONTROL,CS1

; Update next status byte and clear RQ error counter
;
CLA42a:
	mov	al,byte ptr out_msg_stat
	inc 	al
	and	al,03h
	or	al,byte ptr OVER	; include "over"
	mov	byte ptr out_msg_stat,al

	mov	byte ptr rq_cnt,0	; clear RQ count 
					; (or should it be decrement ??)
CLA43:	
	cmp     byte ptr QFLG,1    	; test if qrt request is set
	jne	CLA43b		   	; skip out if not
	call	TXOFF			; else, turn TX off
	jmp	STBY		   	; go to STANDBY

;---------------------------------------------------------------------------;
; This where we get ready to load data into the output packet               ;
;									    ;
; If the keyboard requested a QRT, move the caller callsign in reversed     ;
; order, padded with idles, and set the QRT bit in the status byte too      ;
; Hoever, first empty the data in the buffer before the QRT.	            ;
;---------------------------------------------------------------------------;
CLA43b:	
	or	byte ptr QRT_pending,0
	je	nothing_pending

	mov	ax,cs:tx_take_from_ptr		; is there data ?
	cmp	ax,cs:tx_store_at_ptr
	jne	nothing_pending			; yes, retrieve it

	mov	al,byte ptr out_msg_stat	; set up status byte
	and	al,03h				; leave packet cntr
	or	al,80h				; set QRT bit
	mov	byte ptr out_msg_stat,al	; set up status byte
	mov	byte ptr out_msg_hdr+9,al	; (included in CRC)

REV_CALL:
	lea	bx,callee_ID_rev 		; source
	mov	si,bx
	lea	bx,out_msg			; destination
	mov	cx,7

REV1:	mov	al,byte ptr [si]		; transfer the call
	mov	byte ptr [bx],al
	inc	bx
	inc	si
	loop	REV1

	mov	byte ptr QRT_sent,1		; show we sent QRT
	jmp	CLA29				; send QRT packet

;----------------------------------------------------------------------------;
; If there is data available, send it,					     ;
; Fill packet (8 bytes at 100 baud, 20 bytes at 200 baud)		     ;
; If no data, pad with idle bytes					     ;
;----------------------------------------------------------------------------;
nothing_pending:
	cmp	byte ptr TX_SPEED,0FFh
	je	CLA50B0_hi
	mov	cx,8	  			; 8 bytes
	mov	al,byte ptr out_msg_stat	; set up status byte
	mov	byte ptr out_msg_hdr+9,al	; (included in CRC)
	jmp	SHORT CLA50B0_ad
CLA50B0_hi:
	mov	cx,20				; 20 bytes
CLA50B0_ad:
	lea	bx,out_msg			; destination
	mov	out_ptr,bx

CLA50B0_fill:
	mov	ax,cs:tx_take_from_ptr	; is there data ?
	cmp	ax,cs:tx_store_at_ptr
	jne	CLA50B0_get_dat		; yes, retrieve it

	mov	al,IDLE			; else pad with IDLE
	jmp	SHORT CLA50B0_stor

CLA50B0_get_dat:
	mov	bx,tx_take_from_ptr
	mov	al,byte ptr [bx]	; get the character
	inc	bx			; update pointer
	cmp	bx,offset TXBUFR+TXBUFSIZ ; test for wraparound
	jne	CLA50B0_ge_nxt
	lea	bx,TXBUFR
CLA50B0_ge_nxt: 
	mov	tx_take_from_ptr,bx
	call	ECHO_RX_BYTE		; note ax,bx is saved

CLA50B0_stor:
	mov	bx,out_ptr
	mov	byte ptr [bx],al
	inc	bx
	mov	out_ptr,bx
	loop	CLA50B0_fill

	jmp	CLA29			; send new packet

;----------------------------------------------------------------------------;
;                       Reply to CS3 logic				     ;
;----------------------------------------------------------------------------;
; 		data is checking out OK            - CS1	             ;
;		we want an immediate rebreak       - CS3		     ;
;               CRC failed                         - CS2		     ;
;          (200 baud) too many retries, speedown   - CS4 		     ;
;----------------------------------------------------------------------------;
; CS3 normal sense                                                           ;
;----------------------------------------------------------------------------;
CR_CS3xx: 
	mov	word ptr prev_CS,0	; update CS
	mov	byte ptr cs:rx_sense,0	; for status display
	mov	rx_bit_sense,0		; Received normal bit sense
	mov	word ptr CONTROL,CS3	; for status display
	dec	byte ptr cs_cnt3	; good CS3 ?
	jz	CR_CS3yy
	mov	byte ptr cs_cnt1,GDCS
	mov	byte ptr cs_cnt2,GDCS
	mov	byte ptr cs_cnt4,GDCS
	jmp	CLA32

;----------------------------------------------------------------------------;
; CS3 inverse sense                                                          ;
;----------------------------------------------------------------------------;
CR_CS3_INVxx:
	mov	word ptr prev_CS,0	; update CS
	mov	byte ptr cs:rx_sense,0FFh ; for status display
	mov	rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS3	; for status display
	dec	byte ptr cs_cnt3	; good CS3 ?
	jnz	CR_CS3_INVzz
	jmp	CR_CS3yy
CR_CS3_INVzz:
	mov	byte ptr cs_cnt1,GDCS
	mov	byte ptr cs_cnt2,GDCS
	mov	byte ptr cs_cnt4,GDCS
	jmp	CLA32
;----------------------------------------------------------------------------;

; Compute phase error
CR_CS3yy:

	mov	ax,RxT_cnt		; update tracker
	sub	ax,110+GDCS		; actual CTRX is 120ms earlier	

	cmp	RxT,0
	jnz	Z7
	call	init_RxT		; initialize ideal_RxT
Z7:	mov	RxT,ax
	sub	ax,ideal_RxT
	mov	phase_err,ax		; for display
	mov	ax,RxT
	call	mean_RxT		; update ideal_RxT

	mov	ax,CLOCK
	sub	ax,110+GDCS		; actual CTRX is 120ms earlier	
	mov	DS_CTRX,ax		; for tuning display
;;	sub	ax,ideal_CTRX
;;	mov	phase_err,ax		; for status display

	mov	ax,phase_err
	call	ave_ph_err		; arg. in ax
	call	slave_sync_ISS 

	mov	ax,DS_CTRX
	call	mean_CTRX

;----------------------------------------------------------------------------;
; Set up new time frame and other working parameters			     ;
; CLOCK gets set so that incoming block starts at reference=0		     ;
;----------------------------------------------------------------------------;
rebreak_entry:
	mov	bx,CTTX	   	; save CTTX for later use
	mov	OLD_CTTX,bx

	mov	bx,CLOCK	; save CLOCK for when CS2 needs to be sent
	mov	AUX_CLOCK,bx
	
	mov	bx,110+GDCS	; to middle of data bit
	mov	CLOCK,bx
;----------------------------------------------------------------------------;
; An estimate for CTTX. This is based on the fact that RxT is a constant     ;
; thus CTTX should be put one CS-time, minus RxT before the end of the new   ;
; frame. Tx-delay must then be compensated                                   ;
;----------------------------------------------------------------------------;
	xor	ax,ax		; this is an estimate for CTRX
	mov	al,tx_delay	; TX-delay
	mov	bx,1250		; from end of frame
	sub	bx,ideal_RxT	; one RxT time
	sub	bx,120		; CS length
	sub	bx,ax
	mov	CTTX,bx		; update CTTX

; Update status showing changeover in progress
	mov	byte ptr STATUS,30h
;
;Process the remainder of the turn-round packet
;
	cmp	byte ptr TX_SPEED,0FFh		; 200 Baud ?
	jne	D_00a
	jmp	D_00

;----------------------------------------------------------------------------;
;         Decoding for 100 baud				                     ;
;----------------------------------------------------------------------------;
D_00a:
	lea     bx,CHREG_1		; set search register pointer
        mov     CHPTR_1,bx

D_1aa:	
	cmp	CLOCK,1000		; timed out ?
	jns	D_1zz			; yes, skip out

	call	CHSAMP1			; get another bit sample
	cmp	byte ptr QFLG,1        	; test if qrt is set
	jne	D_2a			;
D_STBY_jmpx:
        jmp     D_STBY_jmp           	; exit to STBY if yes

;----------------------------------------------------------------------------;
; 100 baud - Not timed out yet, so keep searching			     ;
;----------------------------------------------------------------------------;
D_2a:
	mov	byte ptr cs:SPEED,0h		; 100 baud operation 
	cmp	byte ptr rx_bit_sense,0
	jne	D_2ra

	mov	byte ptr cs:bit_sense,00	; Normal sense
	call	ch_blk1
	or	al,al
	jnz	D_3a				; Received a good CRC
	jmp	D_1aa  				; keep trying

D_2ra:
	mov	byte ptr cs:bit_sense,0FFh	; Reverse sense
	call	ch_blk1_r
	or	al,al
	jnz	D_3a				; Received a good CRC
	jmp	D_1aa  				; keep trying

;----------------------------------------------------------------------------;
; 100 baud - Good CRC!							     ;
; Transfer data to receive buffer and acknowledge with CS1		     ;
;           and proceed to IRS section.					     ;
;----------------------------------------------------------------------------;
D_3a:
	lea	bx,block_buffer		; Data, move data from block_buffer
	mov	si,bx
	mov	cx,7	    		; 7 bytes (100 baud C/O packet)

D_5a:	mov	al,byte ptr cs:[si]	; move bytes into receive buffer
	call	STORE_RX_BYTE
	inc	si
	loop	D_5a   
	
;----------------------------------------------------------------------------;
; 100 baud - Update mode byte(Ascii/Huffman:packet count:QRT bit)	     ;
;----------------------------------------------------------------------------;
	mov	al,block_buffer+7 	; get status byte 
	mov	byte ptr cs:MODE,al

	jmp	D_6 			; send general acknowledgement

;----------------------------------------------------------------------------;
; 100 baud: Timed out or CRC failed, send CS2				     ;
;----------------------------------------------------------------------------;
D_1zz:
	inc	byte ptr err_cnt  	; bump error count
	cmp	byte ptr err_cnt,MAX_ERR ; too many ?
	js	D_xx
	jmp	STBY			
D_xx:
	cmp	byte ptr QFLG,1		; manual abort ?
	jne	D_xy
	jmp	STBY
D_xy:

	mov	CONTROL,CS2		; NAK to request resend of CS3 packet
;;;	mov	prev_control,CS3	; XXXX

	mov	SEND_SENSE,0		; Prepare CTL sense
	cmp	tx_bit_sense,0
	je	D_1zza
	mov	SEND_SENSE,0FFFh	; CNTL is 12 bits wide

D_1zza:
	mov	bx,CTTX			; send the control signal now
	call	HOLD	
	call	TCTL			; this updates SEND_SENSE
	mov	RxT_cnt,0		; resets tracking counter

	mov	bx,AUX_CLOCK		; reset CLOCK to ISS frame
	mov	CLOCK,bx

	mov	bx,OLD_CTTX		; reset CTTX
	mov	CTTX,bx

	xor	tx_bit_sense,0FFh	; ready for next send sense

;----------------------------------------------------------------------------;
; 		Wait for next CS3 to be sent again				     ;
;----------------------------------------------------------------------------;
CS3_repeat:
	mov	al,GDCS			; make sure we have at least n bits
	mov	cs_cnt3,al

	mov	bx,ideal_CTRX		; wait till CTRX time
	sub	bx,CTRX_SLOP		; allow for IRS's CS drift
	call	HOLD			; now wait ---

;----------------------------------------------------------------------------;	
; Collect 120ms worth of data - for "priming" the input buffer               ;
;----------------------------------------------------------------------------;	
  	lea     bx,REGS_12
        mov     RPTR_CTL,bx
	mov	cx,120
CL2:	call	SMPLE_12
	loop	CL2

rep2:	call	SMPLE_12		;  get another sample
	mov	bx,RPTR_CTL
	mov	ax, word ptr [bx]
	mov	cx,4
	shr	ax,cl			; convert to 12 bits

	mov	cx,CS3
	cmp	ax,cx
	jne	rep2a
	jmp	SHORT rep_CS3
rep2a:	xor	cx,0FFFh
	cmp	ax,cx
	jne	rep2b
	jmp	SHORT rep_CS3_INV

rep2b:	mov	bx,ideal_CTRX
	add	bx,120+CTRX_SLOP	; CS length + slop	
	cmp	word ptr CLOCK,bx 	; Timed out ?
	jns	rep3

	mov	al,GDCS			; no, search some more
	mov	cs_cnt3,al
	jmp	SHORT rep2		; not yet, look some more

;----------------------------------------------------------------------------;
; T/o - send CS2 this time. The ISS will thus be forced to resend his 	     ;
; CS3 packet again. Keep waiting for this new CS3.                           ;
;----------------------------------------------------------------------------;
rep3:	mov	word ptr CONTROL,0	; Timed out. For status display
	jmp	D_1zz			; send CS2 and get anoter CS3 block

;----------------------------------------------------------------------------;
; Received CS3 - normal sense							     ;
;----------------------------------------------------------------------------;
rep_CS3: 
	mov	word ptr prev_CS,0	; update CS
	mov	byte ptr cs:rx_sense,0	; for status display
	mov	rx_bit_sense,0		; Received normal bit sense
	mov	word ptr CONTROL,CS3	; for status display
	dec	cs_cnt3			; good CS?
	jz	rep_exit
	jmp	SHORT rep2

;----------------------------------------------------------------------------;
; Received CS3 - inverse sense							     ;
;----------------------------------------------------------------------------;
rep_CS3_INV: 
	mov	word ptr prev_CS,0	; update CS
	mov	byte ptr cs:rx_sense,0FFh ; for status display
	mov	rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS3	; for status display
	dec	byte ptr cs_cnt3	; good CS3 ?
	jz	rep_exit
	jmp	rep2

rep_exit:
	jmp	rebreak_entry		; Good CS3, go to IRS CS3 processing

;----------------------------------------------------------------------------;
;         Decoding for 200 baud						     ;
;----------------------------------------------------------------------------;
D_00:	
	lea     bx,CHREG_2		; set search register pointer
        mov     CHPTR_2,bx

D_1:	
	cmp	CLOCK,1000		; timed out ?
	jns    	 D_1a			; yes, skip out

	call	CHSAMP2			;  get another bit sample
	cmp	byte ptr QFLG,1        	; test if qrt is set
        je      D_STBY_jmp           	; exit to STBY if yes
        jmp     SHORT D_2	      	; else keep searching

D_STBY_jmp: 
	jmp	STBY 			; go to STANDBY
	
;----------------------------------------------------------------------------;
; 200 baud - Not timed out yet, so keep searching			     ;
;----------------------------------------------------------------------------;
D_2:
	mov	byte ptr cs:SPEED,0FFh		; 200 baud operation 
	cmp	byte ptr rx_bit_sense,00
	jne	D_2r
	mov	byte ptr cs:bit_sense,00	; Normal sense
	call	ch_blk2
	or	al,al
	jnz	D_3				; Received a good block	hi speed	
     	jmp	D_1
D_2r:
	mov	byte ptr cs:bit_sense,0FFh	; Reverse sense
	call	ch_blk2_r
	or	al,al
	jnz	D_3				; Received a good block	hi speed	
	jmp	D_1   				; keep trying

;----------------------------------------------------------------------------;
; 200 baud - Good CRC!							     ;
; Transfer data to receive buffer and acknolege with CS1		     ;
;           and proceed to IRS section.					     ;
;----------------------------------------------------------------------------;
D_3:
	lea	bx,block_buffer		; Data, move data from block_buffer
	mov	si,bx
	mov	cx,18	    		; 18 bytes (200 baud C/O packet)

D_5:	mov	al,byte ptr cs:[si]	; move bytes into receive buffer
	call	STORE_RX_BYTE
	inc	si
	loop	D_5   
	
;----------------------------------------------------------------------------;
; 200 baud - Update mode byte(Ascii/Huffman:packet count:QRT bit)            ;
;----------------------------------------------------------------------------;
	mov	al,block_buffer+18 	; get status byte 
	mov	byte ptr cs:MODE,al

	jmp	D_6

;----------------------------------------------------------------------------;
; 200 baud - Timed out or bad CRC send up to four CS2 (NAK)		     ;
; if still failing, CS4 (NAK and request for speeeddown)                     ;
;           and proceed to IRS section.					     ;
;----------------------------------------------------------------------------;
D_1a:
	cmp	byte ptr err_cnt,4	; try at least four times (PTC)
	js	D_1az
	jmp	D_1zz			; send CS2
D_1az:
	mov	byte ptr STATUS,23h	; Show we are IRS now
				  	
	mov	byte ptr blk_cnt,3	; first block cnt =0, sync=55h
	mov	al, byte ptr TX_SPEED
	mov	byte ptr RX_SPEED,al
	mov	byte ptr DS_SPEED,al

	mov	byte ptr OVER,0		; clear over request now

;-------------------------------------

	mov	CONTROL,CS4		; NAK C/O block	for speeddown
					; prev_control must be setup here
	mov	prev_control,CS2  	; (CS1) so that we expect a C1 next.

	mov	byte ptr TX_SPEED,0	 ; ensure we have the right speed
	
	mov	SEND_SENSE,0		; Prepare CTL sense
	cmp	tx_bit_sense,0
	je	D_6aa
	mov	SEND_SENSE,0FFFh	; CNTL is 12 bits wide
	jmp	D_6aa

;----------------------------------------------------------------------------;
; General block acknowledgement (100/200)				     ;
;----------------------------------------------------------------------------;
D_6:
	mov	byte ptr STATUS,23h	; Show we are IRS now
				  	
	mov	byte ptr blk_cnt,3	; 
					; first block cnt =0, sync=55h
	mov	byte ptr err_cnt,0	; Clear errors
	mov	byte ptr rq_cnt,0	; Clear RQ count

	mov	al, byte ptr TX_SPEED
	mov	byte ptr RX_SPEED,al
	mov	byte ptr DS_SPEED,al

	mov	byte ptr OVER,0		; clear over request now

;-------------------------------------

	mov	CONTROL,CS1		; All OK new block

	mov	SEND_SENSE,0		; Prepare CTL sense
	cmp	tx_bit_sense,0
	je	D_6aa
	mov	SEND_SENSE,0FFFh	; CNTL is 12 bits wide

;----------------------------------------------------------------------------;
; This is where the actual Breakin CS is sent				     ;
;----------------------------------------------------------------------------;
D_6aa:
	mov	al,byte ptr rx_bit_sense ; Set up the expected rx sense
	mov	byte ptr exp_rx_sense,al ; (its toggled in TCTL)

	mov	al, byte ptr TX_SPEED	 ; ensure we have the right speed
	mov	byte ptr exp_rx_speed,al

	mov	bx,CTTX			; send the control signal now
	call	HOLD	
	call	TCTL			; this updates SEND_SENSE
	mov	RxT_cnt,0		; initialize RxT counter

; Update status showing traffic
	mov	byte ptr STATUS,20h

	call	reset_MARQ		; ensure we start off empty MARQ

	jmp	RR_renter		;  now start checking CRC
;=======================END OF ISS SECTION=================================;

;-------------------SOME HELPER FUNCTIONS----------------------------------;
; Routine to calculate CRC for outgoing messages			   ;
; and transmit packet.							   ;
; Over 9 bytes for 100 baud, 21 bytes for 200 baud			   ;
;--------------------------------------------------------------------------;
TXD_pack:
	cmp	byte ptr TX_SPEED,0FFh
	je	TXD_big
	mov	cx,9
	jmp	SHORT TXD_in
TXD_big:
	mov	cx,21
TXD_in:
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	lea	bx,out_msg		; Over 9 bytes
TXDSRB2_r:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	TXDSRB2_r

; Convert the CRC to an equivalent form
; Note CRC locations are different for 100 and 200 baud

	cmp 	byte ptr TX_SPEED,0FFh
	je	TXD_deep

	mov	al,byte ptr CRC_work 		; 100 baud
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg_hdr+11,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg_hdr+10,al

	lea	bx,out_msg_hdr		; Ready to send message

;
; prepare status display bytes
;
	mov	al,byte ptr out_msg_hdr
	mov	byte ptr sync,al
	mov	al,byte ptr out_msg_stat
	mov	byte ptr MODE,al

; Send the message now
	cmp	CONTROL,CS3
	je	TXD1aa
	call	TXON				
TXD1aa:	mov	cx,12			
TXD1a:	call    TXCH_L			; TXCH includes xor with bit sense
	inc	bx			; 100 baud
	loop	TXD1a
	jmp	TXD2_updte

TXD_deep:
	mov	al,byte ptr CRC_work 		; 200 baud
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg_hdr+23,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr out_msg_hdr+22,al
	lea	bx,out_msg_hdr		; Ready to send message

;
; prepare status display bytes
;
	mov	al,byte ptr out_msg_hdr	; for status display
	mov	byte ptr sync,al
	mov	al,byte ptr out_msg_stat ; for status display
	mov	byte ptr MODE,al

; Send the message now				
	cmp	CONTROL,CS3
	je	TXD1bb
	call	TXON				
TXD1bb:	mov	cx,24			
TXD1b:	call    TXCH_H			; TXCH includes xor with bit sense
	inc	bx			; 200 baud
	loop	TXD1b

; Update next bit sense
TXD2_updte:
	xor	byte ptr tx_bit_sense,0FFh ; invert bits
	ret

;----------------------------------------------------------------------------;
; This routine to initialize accumulated memory ARQ buffer		     ;
;----------------------------------------------------------------------------;
reset_MARQ:
	lea	bx,mem_arq_accum	; compute bit index
	mov	cx,960
	xor	ax,ax
R_MARQ:
	mov	[bx],ax
	inc	bx
	inc	bx 			; words!
	loop	R_MARQ
	ret

;---------------------------------------------------------------------------;
; Routine to interpret MARQ buffers					    ;
;---------------------------------------------------------------------------;
test_MARQ:

;----------------------------------------------------------------------------;
; Character extraction depends on speed
;----------------------------------------------------------------------------;
	cmp	byte ptr exp_rx_speed,0
	je	t_M100
	jmp	t_M200

;----------------------------------------------------------------------------;
;       Memory ARQ 100 baud check
;----------------------------------------------------------------------------;
t_M100:
	xor	al,al			; carry on at low speed
	lea	bx,mem_arq_accum
	add	bx,10			; middle of bits 5 * 2
	mov	si,bx
	mov	cx,8
t_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10*2 away	(words!)
	cmp	bx,0
	je	t_M2
	js	t_M2
	or	al,080h			; positive = 1, negative/zero=0
t_M2:	loop	t_M1

;  --- The header byte is first checked for AA/55? --
	cmp	al,SYNC0
	je	t_tb_1
	cmp	al,SYNC1
	je	t_tb_1
	jmp	t_LSRB3			; failed
	
t_tb_1:					; header looks OK	
;---------------------------------------CH1
	xor	al,al
	mov	cx,8			; 8 bit byte
t1_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t1_M2
	js	t1_M2
	or	al,080h			; positive = 1, negative/zero=0
t1_M2:	loop	t1_M1
	mov	block_buffer,al 	; Save character in block buffer

;---------------------------------------CH2
	xor	al,al
	mov	cx,8			; 8 bit byte
t2_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t2_M2
	js	t2_M2
	or	al,080h			; positive = 1, negative/zero=0
t2_M2:	loop	t2_M1
	mov	block_buffer+1,al 	; Save character in block buffer

;---------------------------------------CH3
	xor	al,al
	mov	cx,8			; 8 bit byte
t3_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t3_M2
	js	t3_M2
	or	al,080h			; positive = 1, negative/zero=0
t3_M2:	loop	t3_M1
	mov	block_buffer+2,al 	; Save character in block buffer

;---------------------------------------CH4
	xor	al,al
	mov	cx,8			; 8 bit byte
t4_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t4_M2
	js	t4_M2
	or	al,080h			; positive = 1, negative/zero=0
t4_M2:	loop	t4_M1
	mov	block_buffer+3,al 	; Save character in block buffer

;---------------------------------------CH5
	xor	al,al
	mov	cx,8			; 8 bit byte
t5_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t5_M2
	js	t5_M2
	or	al,080h			; positive = 1, negative/zero=0
t5_M2:	loop	t5_M1
	mov	block_buffer+4,al 	; Save character in block buffer

;---------------------------------------CH6
	xor	al,al
	mov	cx,8			; 8 bit byte
t6_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t6_M2
	js	t6_M2
	or	al,080h			; positive = 1, negative/zero=0
t6_M2:	loop	t6_M1
	mov	block_buffer+5,al 	; Save character in block buffer

;---------------------------------------CH7
	xor	al,al
	mov	cx,8			; 8 bit byte
t7_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t7_M2
	js	t7_M2
	or	al,080h			; positive = 1, negative/zero=0
t7_M2:	loop	t7_M1
	mov	block_buffer+6,al 	; Save character in block buffer

;---------------------------------------CH8
	xor	al,al
	mov	cx,8			; 8 bit byte
t8_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t8_M2
	js	t8_M2
	or	al,080h			; positive = 1, negative/zero=0
t8_M2:	loop	t8_M1
	mov	block_buffer+7,al 	; Save character in block buffer

;---------------------------------------STAT
	xor	al,al
	mov	cx,8			; 8 bit byte
t9_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t9_M2
	js	t9_M2
	or	al,080h			; positive = 1, negative/zero=0
t9_M2:	loop	t9_M1
	mov	block_buffer+8,al 	; Save character in block buffer

;---------------------------------------CRC1
	xor	al,al
	mov	cx,8			; 8 bit byte
t10_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t10_M2
	js	t10_M2
	or	al,080h			; positive = 1, negative/zero=0
t10_M2:	loop	t10_M1
	mov	block_buffer+9,al 	; Save character in block buffer

;---------------------------------------CRC2
	xor	al,al
	mov	cx,8			; 8 bit byte
t11_M1:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,20			; next bit is 10 away
	cmp	bx,0
	je	t11_M2
	js	t11_M2
	or	al,080h			; positive = 1, negative/zero=0
t11_M2:	loop	t11_M1
	mov	block_buffer+10,al 	; Save character in block buffer

;----------------------------------------------------------------------------;
; Compute CRC of 100 baud memory ARQ'd received block (data+status).
;----------------------------------------------------------------------------;
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 9 bytes
	mov	cx,9
t_LSRB2:	
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	t_LSRB2

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+9
	cmp	al, byte ptr CRC_work+1	
	jne	t_LSRB3			; MSB failed

	mov	al,block_buffer+10
	cmp	al, byte ptr CRC_work	
	jne	t_LSRB3			; LSB failed

	mov	al,0FFh			; signal success
	ret

t_LSRB3:	
	xor	al,al			; signal failure
	ret				; Keep searching			

;----------------------------------------------------------------------------;
;       Memory ARQ 200 baud check
;----------------------------------------------------------------------------;
t_M200:
	xor	al,al
	lea	bx,mem_arq_accum
	add	bx,4			; middle of bit 2 * 2
	mov	si,bx
	mov	cx,8
t_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5*2 away
	cmp	bx,0
	je	t_M4
	js	t_M4
	or	al,080h			; positive = 1, negative/zero=0
t_M4:	loop	t_M3

;--- first see if header is OK ---
	cmp	al,SYNC0
	je	t_tb_2
	cmp	al,SYNC1
	je	t_tb_2
	jmp	t_LSRB3_2		; failed

t_tb_2:	
;---------------------------------------CH1
	xor	al,al
	mov	cx,8
t1_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t1_M4
	js	t1_M4
	or	al,080h			; positive = 1, negative/zero=0
t1_M4:	loop	t1_M3
	mov	block_buffer,al 	; Save character in block buffer

;---------------------------------------CH2
	xor	al,al
	mov	cx,8
t2_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t2_M4
	js	t2_M4
	or	al,080h			; positive = 1, negative/zero=0
t2_M4:	loop	t2_M3
	mov	block_buffer+1,al 	; Save character in block buffer

;---------------------------------------CH3
	xor	al,al
	mov	cx,8
t3_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t3_M4
	js	t3_M4
	or	al,080h			; positive = 1, negative/zero=0
t3_M4:	loop	t3_M3
	mov	block_buffer+2,al 	; Save character in block buffer

;---------------------------------------CH4
	xor	al,al
	mov	cx,8
t4_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t4_M4
	js	t4_M4
	or	al,080h			; positive = 1, negative/zero=0
t4_M4:	loop	t4_M3
	mov	block_buffer+3,al 	; Save character in block buffer

;---------------------------------------CH5
	xor	al,al
	mov	cx,8
t5_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t5_M4
	js	t5_M4
	or	al,080h			; positive = 1, negative/zero=0
t5_M4:	loop	t5_M3
	mov	block_buffer+4,al 	; Save character in block buffer

;---------------------------------------CH6
	xor	al,al
	mov	cx,8
t6_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t6_M4
	js	t6_M4
	or	al,080h			; positive = 1, negative/zero=0
t6_M4:	loop	t6_M3
	mov	block_buffer+5,al 	; Save character in block buffer

;---------------------------------------CH7
	xor	al,al
	mov	cx,8
t7_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t7_M4
	js	t7_M4
	or	al,080h			; positive = 1, negative/zero=0
t7_M4:	loop	t7_M3
	mov	block_buffer+6,al 	; Save character in block buffer

;---------------------------------------CH8
	xor	al,al
	mov	cx,8
t8_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t8_M4
	js	t8_M4
	or	al,080h			; positive = 1, negative/zero=0
t8_M4:	loop	t8_M3
	mov	block_buffer+7,al 	; Save character in block buffer

;---------------------------------------CH9
	xor	al,al
	mov	cx,8
t9_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t9_M4
	js	t9_M4
	or	al,080h			; positive = 1, negative/zero=0
t9_M4:	loop	t9_M3
	mov	block_buffer+8,al 	; Save character in block buffer

;---------------------------------------CH10
	xor	al,al
	mov	cx,8
t10_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t10_M4
	js	t10_M4
	or	al,080h			; positive = 1, negative/zero=0
t10_M4:	loop	t10_M3
	mov	block_buffer+9,al 	; Save character in block buffer

;---------------------------------------CH11
	xor	al,al
	mov	cx,8
t11_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t11_M4
	js	t11_M4
	or	al,080h			; positive = 1, negative/zero=0
t11_M4:	loop	t11_M3
	mov	block_buffer+10,al 	; Save character in block buffer

;---------------------------------------CH12
	xor	al,al
	mov	cx,8
t12_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t12_M4
	js	t12_M4
	or	al,080h			; positive = 1, negative/zero=0
t12_M4:	loop	t12_M3
	mov	block_buffer+11,al 	; Save character in block buffer

;---------------------------------------CH13
	xor	al,al
	mov	cx,8
t13_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t13_M4
	js	t13_M4
	or	al,080h			; positive = 1, negative/zero=0
t13_M4:	loop	t13_M3
	mov	block_buffer+12,al 	; Save character in block buffer

;---------------------------------------CH14
	xor	al,al
	mov	cx,8
t14_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t14_M4
	js	t14_M4
	or	al,080h			; positive = 1, negative/zero=0
t14_M4:	loop	t14_M3
	mov	block_buffer+13,al 	; Save character in block buffer

;---------------------------------------CH15
	xor	al,al
	mov	cx,8
t15_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t15_M4
	js	t15_M4
	or	al,080h			; positive = 1, negative/zero=0
t15_M4:	loop	t15_M3
	mov	block_buffer+14,al 	; Save character in block buffer

;---------------------------------------CH16
	xor	al,al
	mov	cx,8
t16_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t16_M4
	js	t16_M4
	or	al,080h			; positive = 1, negative/zero=0
t16_M4:	loop	t16_M3
	mov	block_buffer+15,al 	; Save character in block buffer

;---------------------------------------CH17
	xor	al,al
	mov	cx,8
t17_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t17_M4
	js	t17_M4
	or	al,080h			; positive = 1, negative/zero=0
t17_M4:	loop	t17_M3
	mov	block_buffer+16,al 	; Save character in block buffer

;---------------------------------------CH18
	xor	al,al
	mov	cx,8
t18_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t18_M4
	js	t18_M4
	or	al,080h			; positive = 1, negative/zero=0
t18_M4:	loop	t18_M3
	mov	block_buffer+17,al 	; Save character in block buffer

;---------------------------------------CH19
	xor	al,al
	mov	cx,8
t19_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t19_M4
	js	t19_M4
	or	al,080h			; positive = 1, negative/zero=0
t19_M4:	loop	t19_M3
	mov	block_buffer+18,al 	; Save character in block buffer

;---------------------------------------CH20
	xor	al,al
	mov	cx,8
t20_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t20_M4
	js	t20_M4
	or	al,080h			; positive = 1, negative/zero=0
t20_M4:	loop	t20_M3
	mov	block_buffer+19,al 	; Save character in block buffer

;---------------------------------------Status
	xor	al,al
	mov	cx,8
t21_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t21_M4
	js	t21_M4
	or	al,080h			; positive = 1, negative/zero=0
t21_M4:	loop	t21_M3
	mov	block_buffer+20,al 	; Save character in block buffer

;---------------------------------------CRC1
	xor	al,al
	mov	cx,8
t22_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t22_M4
	js	t22_M4
	or	al,080h			; positive = 1, negative/zero=0
t22_M4:	loop	t22_M3
	mov	block_buffer+21,al 	; Save character in block buffer

;---------------------------------------CRC2
	xor	al,al
	mov	cx,8
t23_M3:	shr	al,1			; ready for next bit
	mov	bx,[si]
	add	si,10			; next bit is 5 away
	cmp	bx,0
	je	t23_M4
	js	t23_M4
	or	al,080h			; positive = 1, negative/zero=0
t23_M4:	loop	t23_M3
	mov	block_buffer+22,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 21 bytes
	mov	cx,21
t_LSRB2_2:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	t_LSRB2_2

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+21
	cmp	al, byte ptr CRC_work+1	
	jne	t_LSRB3_2			; MSB failed

	mov	al,block_buffer+22
	cmp	al, byte ptr CRC_work	
	jne	t_LSRB3_2			; LSB failed

	mov	al,0FFh			; signal success
	ret

t_LSRB3_2:
	xor	al,al			; signal failure
	ret				; Keep searching			

;----------------------------------------------------------------------------;
; This routine to initialize old_CTRX array				     ;
; AX contains value to be initialized to                                     ;
;----------------------------------------------------------------------------;
init_CTRX:
	mov	cx,8
	lea	bx,old_CTRX
init_C: mov	[bx],ax
	inc	bx    			; these are words!
	inc	bx
	loop	init_C
	mov	ideal_CTRX,ax		; initialize this one too
	ret

;----------------------------------------------------------------------------;
; This routine to update ideal_CTRX					     ;
; Average based on 8-wide shift register				     ;
; AX contains value to be initialized to                                     ;
;----------------------------------------------------------------------------;
mean_CTRX:
	push	ax
	sub	ax,ideal_CTRX		; limit changes to +/- 2 ticks
	je	mean_C0	
	jns	mean_C0a
					; CTRX < ideal_CTRX
	add	ax,2			; less/equal to 2 ?
	jns	mean_C0
	pop	ax			; use a value ideal_CTRX-2
	mov	ax,ideal_CTRX
	sub	ax,2
	push	ax
	jmp	mean_C0

mean_C0a:				; CTRX > ideal_CTRX
	sub	ax,2			; greater/equal to 2
	je	mean_C0
	js	mean_C0
	pop	ax			; use a value ideal_CTRX-2
	mov	ax,ideal_CTRX
	add	ax,2
	push	ax

;
; Now we are ready for the shift register's update
;
mean_C0:   				
	mov	cx,7
	lea	bx,old_CTRX+2*7		; word offsets
mean_C1:			 	; update shift register
	mov	ax,[bx-2]
	mov	[bx],ax
	dec	bx    			; these are words!
	dec	bx
	loop	mean_C1
	pop	ax
	mov	[bx],ax

	mov	cx,8			; compute average CTRX
	lea	bx,old_CTRX
	xor	ax,ax
mean_C2:
	add	ax,[bx]
	inc	bx			; these are words!
	inc	bx
	loop	mean_C2
	sar	ax,1
	sar	ax,1
	sar	ax,1
	mov	ideal_CTRX,ax
	ret		

;----------------------------------------------------------------------------;
; This routine to initialize old_RxT array				     ;
; AX contains value to be initialized to                                     ;
;----------------------------------------------------------------------------;
init_RxT:
	mov	cx,8
	lea	bx,old_RxT
init_R: mov	[bx],ax
	inc	bx    			; these are words!
	inc	bx
	loop	init_R
	mov	ideal_RxT,ax		; initialize this one too 
	ret

;----------------------------------------------------------------------------;
; This routine to update ideal_RxT					     ;
; Average based on 8-wide shift register				     ;
; AX contains value to be initialized to                                     ;
;----------------------------------------------------------------------------;
mean_RxT:
	push	ax
;
; Allow RxT to be updated only in slave mode, or only during the
; first ave_RxT_cnt cycles
;
	cmp	byte ptr MASTER,0
	jne	Rx1			; pass slave calls
	cmp	ave_RxT_cnt,0
	je	Rx0
	dec	ave_RxT_cnt
	jmp	SHORT Rx2
Rx0:	pop	ax
	ret
Rx2:
    	mov	bx,ideal_RxT
	pop	ax
	add	ax,bx
	sar	ax,1
	mov	ideal_RxT,ax
	ret

Rx1:
	sub	ax,ideal_RxT		; limit changes to +/- 2 ticks
	je	mean_R0	
	jns	mean_R0a
					; RxT < ideal_RxT
	add	ax,2			; less/equal to 2 ?
	jns	mean_R0
	pop	ax			; use a value ideal_RxT-2
	mov	ax,ideal_RxT
	sub	ax,2
	push	ax
	jmp	mean_R0

mean_R0a:				; RxT > ideal_RxT
	sub	ax,2			; greater/equal to 2
	je	mean_R0
	js	mean_R0
	pop	ax			; use a value ideal_RxT-2
	mov	ax,ideal_RxT
	add	ax,2
	push	ax

;
; Now we are ready for the shift register's update
;
mean_R0:   				
	mov	cx,7
	lea	bx,old_RxT+2*7		; word offsets
mean_R1:			 	; update shift register
	mov	ax,[bx-2]
	mov	[bx],ax
	dec	bx    			; these are words!
	dec	bx
	loop	mean_R1
	pop	ax
	mov	[bx],ax

	mov	cx,8			; compute average RxT
	lea	bx,old_RxT
	xor	ax,ax
mean_R2:
	add	ax,[bx]
	inc	bx			; these are words!
	inc	bx
	loop	mean_R2
	sar	ax,1
	sar	ax,1
	sar	ax,1
	mov	ideal_RxT,ax
	ret		

;----------------------------------------------------------------------------;
;                     Listen (Monitor) mode				     ;
;----------------------------------------------------------------------------;
MON:	lea     bx,AM_STACK
        mov     sp,bx           	; SP
        mov     ax,cs
        mov     ss,ax           	; SS
        
        xor     ax,ax                                        
        mov     QFLG,al         	; Clear QRT request
	mov	byte ptr STATUS,31h	; STATUS=Monitor
	call	zero_phase		; initially no phase error

;
; Turn DSP AFSK off
;
	mov	al,EVM_RX
	mov	cs:evm_out,al

	call	init_ptrs		; initialize scanning pointers
	
LSN:	call	SAMPLE			;  get another bit sample
	cmp	byte ptr QFLG,1        	; test if qrt is set
        je      LS_STBYL           	; exit to STBY if yes
        jmp     SHORT LS_rd_blk	      	; else keep searching

LS_STBYL: jmp	STBY 			; go to STANDBY

;----------------------------------------------------------------------------;
; Look at the rest of the block.					     ;
;----------------------------------------------------------------------------;
LS_rd_blk: 
	mov	byte ptr cs:SPEED,0
	mov	byte ptr cs:bit_sense,00	; Normal sense
	call	test_blk_1
	or	al,al
	jnz	LSRRR3			; Received a good block	lo speed	
	mov	byte ptr cs:bit_sense,0FFh	; Reverse sense
	call	test_blk_1_r
	or	al,al
	jnz	LSRRR3			; Received a good block	lo speed	

	mov	byte ptr cs:SPEED,0FFh
	mov	byte ptr cs:bit_sense,00	; Normal sense
	call	test_blk_2
	or	al,al
	jnz	LSRRR3			; Received a good block	hi speed	
	mov	byte ptr cs:bit_sense,0FFh	; Reverse sense
	call	test_blk_2_r
	or	al,al
	jnz	LSRRR3			; Received a good block	hi speed	

	jmp	LSN   			; keep trying

; We received a good block 
; Move the data block into receive buffer 
; and signal that data is available.
;
LSRRR3:
	mov	ax,CLOCK		; Compute phase difference
	sub	ax,MASTER_SYNC		; from the ideal
	mov	phase_err,ax
	mov	CLOCK,MASTER_SYNC	

	mov	al,byte ptr SPEED	; for status display 
	mov	byte ptr DS_SPEED,al
	lea	bx,block_buffer		; Data, move data from block_buffer
	mov	si,bx
	mov	cx,8	    		; assume 8 bytes (100 baud)
	cmp	byte ptr cs:SPEED,00 	; check baudrate
	jz	LSRB4
	mov	cx,20			; else 20 bytes (200 baud)
	
LSRB4:	mov	al,byte ptr cs:[si]
	call	STORE_RX_BYTE
	inc	si
	dec	cx
	jnz	LSRB4   

; Update mode byte=Ascii/Huffman
	
	mov	al,block_buffer+8 	; get status byte (assume 100 baud)
	cmp	byte ptr cs:SPEED,00 	; check speed
	jz	LSRB4a
	mov	al,block_buffer+20 	; get status byte (200 baud)
	
LSRB4a:	mov	cs:MODE,al

; update sense and sync
	mov	al,byte ptr cs:bit_sense
	mov	byte ptr cs:rx_sense,al
	mov	al,byte ptr cs:sync_byte
	mov	byte ptr cs:sync,al

;  collect next block (10 ticks short)
	mov	cx,MASTER_SYNC		;  number of tics in block 960
				
LSRB5:	call	SAMPLE			;  get another bit sample
	loop	LSRB5

	jmp	LSN

;----------------------------------------------------------------------------;
;               System calibration/test routine				     ;
;----------------------------------------------------------------------------;
TESTR:  lea     bx,AM_STACK 	; reset PACTOR stack pointer
        mov     sp,bx           ; SP
        mov     ax,cs           ; SS
        mov     ss,ax

TSTPA:  mov     al,055h         ; alternate ones and zeroes
TSTPB:  mov     tst_pat,al      ;
TSTP0:  mov     STSTS,al
        mov     bx,7            ; seven bits in pattern
TSTP00: shr     al,1            ; shift into carry
	call	TXBIT
        call    DROP            ; and wait one ms bit time
        dec     bx              ; count down
        jz      TSTP3           ; skip out if finished
        jmp     TSTP00          ; and go back and do it
        
TSTP3:  mov     al,tst_pat
        cmp     al,055h
        jne     TSTPA 
        mov     al,02Ah
        jmp     TSTPB
	ret

;-----MORE HELPERS----------------------------------------------------------;
;  Obtain good CS response
;---------------------------------------------------------------------------;
AN_CS:
	mov	al,GDCS			; make sure we have at least n bits
	mov	cs_cnt1,al
	mov	cs_cnt2,al
	mov	cs_cnt3,al
	mov	cs_cnt4,al

  	lea     bx,REGS_12      	; set search register pointer
	mov     cx,10         		; clear shift reg
	xor	ax,ax
A11:  	mov	word ptr [bx],ax
	inc	bx
	inc	bx
        loop	A11
	
AN_CS0:
;----------------------------------------------------------------------------;	
; Collect 120ms worth of data - for "priming" the input buffer               ;
;----------------------------------------------------------------------------;	
  	lea     bx,REGS_12
        mov     RPTR_CTL,bx
	mov	cx,120
CL3:	call	SMPLE_12
	loop	CL3

A12:	call	SMPLE_12		;  get another sample
	mov	bx,RPTR_CTL
	mov	ax, word ptr [bx]
	mov	cx,4
	shr	ax,cl			; convert to 12 bits

	mov	cx,CS1
	cmp	ax,cx
	jne	A12a
	jmp	AA_CS1
A12a:	xor	cx,0FFFh
	cmp	ax,cx
	jne	A12b
	jmp	AA_CS1_INV

A12b:	mov	cx,CS2
	cmp	ax,cx
	jne	A12c
	jmp	AA_CS2
A12c:	xor	cx,0FFFh
	cmp	ax,cx
	jne	A12d
	jmp	AA_CS2_INV

A12d:	mov	cx,CS3
	cmp	ax,cx
	jne	A12e
	jmp	AA_CS3
A12e:	xor	cx,0FFFh
	cmp	ax,cx
	jne	A12f
	jmp	AA_CS3_INV

A12f:	mov	cx,CS4
	cmp	ax,cx
	jne	A12g
	jmp	AA_CS4
A12g:	xor	cx,0FFFh
	cmp	ax,cx
	jne	A12h
	jmp	AA_CS4_INV

A12h:	cmp	word ptr CLOCK,0 	; Timed out ?
	je	AA12x
	cmp	word ptr CLOCK,1 	; Timed out ?
	je	AA12x

	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt2,al
	mov	cs_cnt3,al
	mov	cs_cnt4,al
	jmp	A12			; not yet, look some more

AA12x:	mov	word ptr CONTROL,0	; Timed out. For status display
	ret 				; skip out if timed out
;---------------------------------------------------------------
AA_CS1: dec	cs_cnt1			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt2,al
	mov	cs_cnt3,al
	mov	cs_cnt4,al
	jz	AA_CS1a
	jmp	A12
AA_CS1a:
	mov	rx_bit_sense,0		; Received normal bit sense
	mov	word ptr CONTROL,CS1	; for status display
	jmp	AA_AGN
AA_CS1_INV: dec	cs_cnt1			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt2,al
	mov	cs_cnt3,al
	mov	cs_cnt4,al
	jz	AA_CS1_INVa
	jmp	A12
AA_CS1_INVa:
	mov	rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS1	; for status display
	jmp	AA_AGN

;---------------------------------------------------------------
AA_CS2: dec	cs_cnt2			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt3,al
	mov	cs_cnt4,al
	jz	A12_CS2a
	jmp	A12
A12_CS2a:
	mov	rx_bit_sense,0		; Received normal bit sense
	mov	word ptr CONTROL,CS2	; for status display
	jmp	AA_AGN
AA_CS2_INV: dec	cs_cnt2			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt3,al
	mov	cs_cnt4,al
	jz	AA_CS2_INVa	
	jmp	A12	
AA_CS2_INVa:	
	mov	rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS2	; for status display
	jmp	AA_AGN

AA_CS3: dec	cs_cnt3			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt2,al
	mov	cs_cnt4,al
	jz	AA_CS3a
	jmp	A12
AA_CS3a:
	mov	rx_bit_sense,0		; Received normal bit sense
	mov	word ptr CONTROL,CS3	; for status display
	jmp	AA_AGN
AA_CS3_INV: dec	cs_cnt3			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt2,al
	mov	cs_cnt4,al
	jz	AA_CS3_INVa
	jmp	A12
AA_CS3_INVa:
	mov	rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS3	; for status display
	jmp	AA_AGN

AA_CS4: dec	cs_cnt4
	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt2,al
	mov	cs_cnt3,al
	jz	AA_CS4a
	jmp	A12
AA_CS4a:
	mov	rx_bit_sense,0		; Received normal bit sense
	mov	word ptr CONTROL,CS4	; for status display
	jmp	AA_AGN
AA_CS4_INV: dec	cs_cnt4			; good CS?
	mov	al,GDCS			; no, search some more
	mov	cs_cnt1,al
	mov	cs_cnt2,al
	mov	cs_cnt3,al
	jz	AA_CS4_INVa
	jmp	A12
AA_CS4_INVa:
	mov	rx_bit_sense,0FFh	; Received inverted bit sense
	mov	word ptr CONTROL,CS4	; for status display

AA_AGN:	ret

;----------------------------------------------------------------------------;
; Routine to initialize scan pointers					     ;
;----------------------------------------------------------------------------;
	EVEN
init_ptrs:
	lea     bx,SHIFTREG_1		; set search register pointer
        mov     RPTR_1,bx
	lea     bx,SHIFTREG_2		; set search register pointer
        mov     RPTR_2,bx
	ret

;----------------------------------------------------------------------------;
; Sampling routine for Listen/IRS mode					     ;
;----------------------------------------------------------------------------;

;----------------------------------------------------------------------------;
; 8 bytes at 100 baud full packets shift register to holds 10 samples/bit    ;
;----------------------------------------------------------------------------;
	EVEN
SAMPLE: 
	call    DROP          		; go to user till next IRQ
					; at this point, a new data
					; value is in evm_in and data_bit

;----------------------------------------------------------------------------;
; Process MARQ array for 100 baud elements				     ;
; CLOCK is used as index, thus only collect data for the interval 0 - 959    ;
;----------------------------------------------------------------------------;
	mov	ax,CLOCK
	cmp	ax,961			; should we collect data for MARQ?
	jns	SA1			; no, skip out

	lea	bx,mem_arq_accum	; yes, compute bit index
	add	bx,ax
	add	bx,ax  			; words!

	mov	ax,cs:AD_data		; get A/D data
	sar	ax,1			; scale it to prevent overflow
	sar	ax,1
	sar	ax,1
	sar	ax,1

	cmp	byte ptr exp_rx_sense,0	 ; if transient data is of
	jne	SA0			 ; reversed sense, use it's
	neg	ax			 ; 2's compliment

SA0:	add	ax,[bx]
	sar	ax,1
	mov	[bx],ax
SA1:
;----------------------------------------------------------------------------;
; 8 bytes at 100 baud full packets. At 100 baud shift register holds 10/bit  ;
;----------------------------------------------------------------------------;
        mov     bx,RPTR_1       	; get current shiftreg pointer
        inc     bx              	; ready for next register
        cmp     bx,offset SHIFTREG_1+10 ; test gone too far ?
        jne     SNXR1
        lea     bx,SHIFTREG_1       	; if so start round again
SNXR1:  mov	RPTR_1,bx

	mov	al,data_bit

        shl     al,1            	; shift into carry 
        rcr     byte ptr [bx+110],1 
        rcr     byte ptr [bx+100],1 
        rcr     byte ptr [bx+90],1 
        rcr     byte ptr [bx+80],1 
        rcr     byte ptr [bx+70],1 
        rcr     byte ptr [bx+60],1 
        rcr     byte ptr [bx+50],1 
        rcr     byte ptr [bx+40],1 
        rcr     byte ptr [bx+30],1 
        rcr     byte ptr [bx+20],1 
        rcr     byte ptr [bx+10],1 
        rcr     byte ptr [bx],1 

;----------------------------------------------------------------------------;
; 20 bytes at 200 baud full packets. At 200 baud shift register holds 5/bit  ;
;----------------------------------------------------------------------------;
        mov     bx,RPTR_2 	; get current shiftreg pointer
        inc     bx              ; ready for next register
        cmp     bx,offset SHIFTREG_2+5 ; test gone too far ?
        jne     SNXR1_2
        lea     bx,SHIFTREG_2    ; if so start round again
SNXR1_2:
	mov	RPTR_2,bx

	mov	al,data_bit

        shl     al,1            ; shift into carry 

        rcr     byte ptr [bx+115],1	; additional bytes for 200 baud 
        rcr     byte ptr [bx+110],1 
        rcr     byte ptr [bx+105],1 
        rcr     byte ptr [bx+100],1 
        rcr     byte ptr [bx+95],1 
        rcr     byte ptr [bx+90],1 
        rcr     byte ptr [bx+85],1 
        rcr     byte ptr [bx+80],1 
        rcr     byte ptr [bx+75],1 
        rcr     byte ptr [bx+70],1 
        rcr     byte ptr [bx+65],1 
        rcr     byte ptr [bx+60],1 

        rcr     byte ptr [bx+55],1	; bytes for 100 baud 
        rcr     byte ptr [bx+50],1 
        rcr     byte ptr [bx+45],1 
        rcr     byte ptr [bx+40],1 
        rcr     byte ptr [bx+35],1 
        rcr     byte ptr [bx+30],1 
        rcr     byte ptr [bx+25],1 
        rcr     byte ptr [bx+20],1 
        rcr     byte ptr [bx+15],1 
        rcr     byte ptr [bx+10],1 
        rcr     byte ptr [bx+5],1 
        rcr     byte ptr [bx],1 

	ret

;---------------------------------------------------------------------------;
; Sampling routine for Changovers					    ;
; Size of shift register to hold 10 samples of entire block		    ;
; 7 bytes + 1 nibble  (unused) + status + CRC1/CRC2 at 100 baud		    ;
;---------------------------------------------------------------------------;
	EVEN
CHSAMP1:
	call    DROP          ; go to user till next IRQ
        mov     bx,CHPTR_1 	; get current shiftreg pointer
        inc     bx              ; ready for next register
        cmp     bx,offset CHREG_1+10 ; test gone too far ?
        jne     CHXR1_2
        lea     bx,CHREG_1    	; if so start round again
CHXR1_2:
	mov	CHPTR_1,bx

	mov	al,data_bit

        shl     al,1            ; shift into carry 

        rcr     byte ptr [bx+90],1 	; CRC2
        rcr     byte ptr [bx+80],1 	; CRC1
        rcr     byte ptr [bx+70],1 	; Status

        rcr     byte ptr [bx+60],1 	; D7
        rcr     byte ptr [bx+50],1 	; D6
        rcr     byte ptr [bx+40],1 	; D5
        rcr     byte ptr [bx+30],1 	; D4
        rcr     byte ptr [bx+20],1 	; D3
        rcr     byte ptr [bx+10],1 	; D2
        rcr     byte ptr [bx],1 	; D1

	ret

;----------------------------------------------------------------------------;
;	CHANGOVER 200 baud						     ;
; Size of shift register to hold 5 samples of entire block		     ;
; 18 bytes + status + CRC1/CRC2 at 200 baud				     ;
;----------------------------------------------------------------------------;
	EVEN
CHSAMP2:
	call    DROP          ; go to user till next IRQ
        mov     bx,CHPTR_2 	; get current shiftreg pointer
        inc     bx              ; ready for next register
        cmp     bx,offset CHREG_2+5 ; test gone too far ?
        jne     CHXR2_2
        lea     bx,CHREG_2    	; if so start round again
CHXR2_2:
	mov	CHPTR_2,bx

	mov	al,data_bit

        shl     al,1            ; shift into carry 

        rcr     byte ptr [bx+100],1 	; CRC2
        rcr     byte ptr [bx+95],1 	; CRC1
        rcr     byte ptr [bx+90],1 	; Status
        rcr     byte ptr [bx+85],1 	; D18
        rcr     byte ptr [bx+80],1 	; D17
        rcr     byte ptr [bx+75],1 	; D16
        rcr     byte ptr [bx+70],1 	; D15
        rcr     byte ptr [bx+65],1 	; D14
        rcr     byte ptr [bx+60],1 	; D13

        rcr     byte ptr [bx+55],1	; D12 
        rcr     byte ptr [bx+50],1 	; D11
        rcr     byte ptr [bx+45],1 	; D10
        rcr     byte ptr [bx+40],1 	; D9
        rcr     byte ptr [bx+35],1 	; D8
        rcr     byte ptr [bx+30],1 	; D7
        rcr     byte ptr [bx+25],1 	; D6
        rcr     byte ptr [bx+20],1 	; D5
        rcr     byte ptr [bx+15],1 	; D4
        rcr     byte ptr [bx+10],1 	; D3
        rcr     byte ptr [bx+5],1 	; D2
        rcr     byte ptr [bx],1 	; D1

	ret

;---------------------------------------------------------------------------;
; Test a block for correct CRC						    ;
; 100 baud full packets							    ;
;---------------------------------------------------------------------------;
test_blk_1:
	mov	bx,RPTR_1

	mov	al,byte ptr [bx]
	mov	byte ptr cs:sync_byte,al
	cmp	al,SYNC0
	je	tb_1
	cmp	al,SYNC1
	je	tb_1
	jmp	LSRB3			; failed
	
tb_1:	mov	al,byte ptr [bx+10]	; Char #1
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #2
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #3
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #4
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #5
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #6
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Char #7
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; Char #8
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; Status byte
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,[bx+100]		; CRC MSByte
	mov	block_buffer+9,al 	; Save character in block buffer

	mov	al,[bx+110]		; CRC LSByte
	mov	block_buffer+10,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 9 bytes
	mov	cx,9
LSRB2:	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	LSRB2

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+9
	cmp	al, byte ptr CRC_work+1	
	jne	LSRB3			; MSB failed

	mov	al,block_buffer+10
	cmp	al, byte ptr CRC_work	
	jne	LSRB3			; LSB failed

	mov	al,0FFh			; signal success
	ret

LSRB3:	xor	al,al			; signal failure
	ret				; Keep searching			

;----------------------REVERSE SENSE-----------------------------------------;
test_blk_1_r:
	mov	bx,RPTR_1

	mov	al,byte ptr [bx]
	xor	al,0FFh
	mov	byte ptr cs:sync_byte,al
	cmp	al,SYNC0
	je	tb_1_r
	cmp	al,SYNC1
	je	tb_1_r
	jmp	LSRB3_r			; failed
	
tb_1_r:	
	mov	al,byte ptr [bx+10]	; Char #1
	xor	al,0FFh
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #2
	xor	al,0FFh
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #3
	xor	al,0FFh
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #4
	xor	al,0FFh
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #5
	xor	al,0FFh
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #6
	xor	al,0FFh
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Char #7
	xor	al,0FFh
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; Char #8
	xor	al,0FFh
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; Status byte
	xor	al,0FFh
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,[bx+100]		; CRC MSByte
	xor	al,0FFh
	mov	block_buffer+9,al 	; Save character in block buffer

	mov	al,[bx+110]		; CRC LSByte
	xor	al,0FFh
	mov	block_buffer+10,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 9 bytes
	mov	cx,9
LSRB2_r:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	LSRB2_r

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+9
	cmp	al, byte ptr CRC_work+1	
	jne	LSRB3_r			; MSB failed

	mov	al,block_buffer+10
	cmp	al, byte ptr CRC_work	
	jne	LSRB3_r			; LSB failed

	mov	al,0FFh			; signal success
	ret

LSRB3_r:
	xor	al,al			; signal failure
	ret				; Keep searching			

;---------------------------------------------------------------------------;
;                    200 baud full packets				    ;
;---------------------------------------------------------------------------;
test_blk_2:
	mov	bx,RPTR_2

	mov	al,byte ptr [bx]
	mov	byte ptr cs:sync_byte,al
	cmp	al,SYNC0
	je	tb_2
	cmp	al,SYNC1
	je	tb_2
	jmp	LSRB3_2			; failed

tb_2:	mov	al,byte ptr [bx+5]	; Char #1
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+10]	; Char #2
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+15]	; Char #3
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #4
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+25]	; Char #5
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #6
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+35]	; Char #7
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #8
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+45]	; Char #9
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #10
	mov	block_buffer+9,al 	; Save character in block buffer

	mov	al,byte ptr [bx+55]	; Char #11
	mov	block_buffer+10,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #12
	mov	block_buffer+11,al 	; Save character in block buffer

	mov	al,byte ptr [bx+65]	; Char #13
	mov	block_buffer+12,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Char #14
	mov	block_buffer+13,al 	; Save character in block buffer

	mov	al,byte ptr [bx+75]	; Char #15
	mov	block_buffer+14,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; Char #16
	mov	block_buffer+15,al 	; Save character in block buffer

	mov	al,byte ptr [bx+85]	; Char #17
	mov	block_buffer+16,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; Char #18
	mov	block_buffer+17,al 	; Save character in block buffer

	mov	al,byte ptr [bx+95]	; Char #19
	mov	block_buffer+18,al 	; Save character in block buffer

	mov	al,byte ptr [bx+100]	; Char #20
	mov	block_buffer+19,al 	; Save character in block buffer

	mov	al,byte ptr [bx+105]	; Status byte
	mov	block_buffer+20,al 	; Save character in block buffer

	mov	al,byte ptr [bx+110] 	; CRC MSByte
	mov	block_buffer+21,al 	; Save character in block buffer

	mov	al,byte ptr [bx+115]	; CRC LSByte
	mov	block_buffer+22,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 21 bytes
	mov	cx,21
LSRB2_2:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	LSRB2_2

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+21
	cmp	al, byte ptr CRC_work+1	
	jne	LSRB3_2			; MSB failed

	mov	al,block_buffer+22
	cmp	al, byte ptr CRC_work	
	jne	LSRB3_2			; LSB failed

	mov	al,0FFh			; signal success
	ret

LSRB3_2:
	xor	al,al			; signal failure
	ret				; Keep searching			
;------------------REVERSE SENSE--------------------------------------------;
test_blk_2_r:
	mov	bx,RPTR_2

	mov	al,byte ptr [bx]
	xor	al,0FFh
	mov	byte ptr cs:sync_byte,al
	cmp	al,SYNC0
	je	tb_2_r
	cmp	al,SYNC1
	je	tb_2_r
	jmp	LSRB3_2_r		; failed

tb_2_r:	
	mov	al,byte ptr [bx+5]	; Char #1
	xor	al,0FFh
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+10]	; Char #2
	xor	al,0FFh
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+15]	; Char #3
	xor	al,0FFh
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #4
	xor	al,0FFh
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+25]	; Char #5
	xor	al,0FFh
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #6
	xor	al,0FFh
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+35]	; Char #7
	xor	al,0FFh
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #8
	xor	al,0FFh
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+45]	; Char #9
	xor	al,0FFh
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #10
	xor	al,0FFh
	mov	block_buffer+9,al 	; Save character in block buffer

	mov	al,byte ptr [bx+55]	; Char #11
	xor	al,0FFh
	mov	block_buffer+10,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #12
	xor	al,0FFh
	mov	block_buffer+11,al 	; Save character in block buffer

	mov	al,byte ptr [bx+65]	; Char #13
	xor	al,0FFh
	mov	block_buffer+12,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Char #14
	xor	al,0FFh
	mov	block_buffer+13,al 	; Save character in block buffer

	mov	al,byte ptr [bx+75]	; Char #15
	xor	al,0FFh
	mov	block_buffer+14,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; Char #16
	xor	al,0FFh
	mov	block_buffer+15,al 	; Save character in block buffer

	mov	al,byte ptr [bx+85]	; Char #17
	xor	al,0FFh
	mov	block_buffer+16,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; Char #18
	xor	al,0FFh
	mov	block_buffer+17,al 	; Save character in block buffer

	mov	al,byte ptr [bx+95]	; Char #19
	xor	al,0FFh
	mov	block_buffer+18,al 	; Save character in block buffer

	mov	al,byte ptr [bx+100]	; Char #20
	xor	al,0FFh
	mov	block_buffer+19,al 	; Save character in block buffer

	mov	al,byte ptr [bx+105]	; Status byte
	xor	al,0FFh
	mov	block_buffer+20,al 	; Save character in block buffer

	mov	al,byte ptr [bx+110] 	; CRC MSByte
	xor	al,0FFh
	mov	block_buffer+21,al 	; Save character in block buffer

	mov	al,byte ptr [bx+115]	; CRC LSByte
	xor	al,0FFh
	mov	block_buffer+22,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 21 bytes
	mov	cx,21
LSRB2_2_r:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	LSRB2_2_r

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+21
	cmp	al, byte ptr CRC_work+1	
	jne	LSRB3_2_r		; MSB failed

	mov	al,block_buffer+22
	cmp	al, byte ptr CRC_work	
	jne	LSRB3_2_r		; LSB failed

	mov	al,0FFh			; signal success
	ret

LSRB3_2_r:
	xor	al,al			; signal failure
	ret				; Keep searching			

;---------------------------------------------------------------------------;
;       Checking change-over message - 100 Baud				    ;
;---------------------------------------------------------------------------;
ch_blk1:
	mov	bx,CHPTR_1

	mov	al,byte ptr [bx]	; Char #1
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+10]	; Char #2
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #3
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #4
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #5
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #6
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #7
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Status byte
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; CRC MSB byte
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; CRC LSB byte
	mov	block_buffer+9,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 8 bytes
	mov	cx,8
ch_blk1_2:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	ch_blk1_2

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+8
	cmp	al, byte ptr CRC_work+1	
	jne	ch_blk1_3			; MSB failed

	mov	al,block_buffer+9
	cmp	al, byte ptr CRC_work	
	jne	ch_blk1_3			; LSB failed

	mov	al,0FFh				; signal success
	ret

ch_blk1_3:
	xor	al,al			; signal failure
	ret				; Keep searching			
;------------------REVERSE SENSE--------------------------------------------;
ch_blk1_r:
	mov	bx,CHPTR_1

	mov	al,byte ptr [bx]	; Char #1
	xor	al,0FFh
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+10]	; Char #2
	xor	al,0FFh
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #3
	xor	al,0FFh
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #4
	xor	al,0FFh
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #5
	xor	al,0FFh
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #6
	xor	al,0FFh
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #7
	xor	al,0FFh
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Status byte
	xor	al,0FFh
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; CRC MSB 
	xor	al,0FFh
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; CRC LSB
	xor	al,0FFh
	mov	block_buffer+9,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 8 bytes
	mov	cx,8

ch_blk1_2_r:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	ch_blk1_2_r

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+8
	cmp	al, byte ptr CRC_work+1	
	jne	ch_blk1_3_r		; MSB failed

	mov	al,block_buffer+9
	cmp	al, byte ptr CRC_work	
	jne	ch_blk1_3_r		; LSB failed

	mov	al,0FFh			; signal success
	ret

ch_blk1_3_r:
	xor	al,al			; signal failure
	ret				; Keep searching			
;---------------------------------------------------------------------------;
;        Checking change over packet - 200 Baud				    ;
;---------------------------------------------------------------------------;
ch_blk2:
	mov	bx,CHPTR_2

	mov	al,byte ptr [bx]	; Char #1
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+5]	; Char #2
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+10]	; Char #3
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+15]	; Char #4
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #5
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+25]	; Char #6
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #7
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+35]	; Char #8
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #9
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,byte ptr [bx+45]	; Char #10
	mov	block_buffer+9,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #11
	mov	block_buffer+10,al 	; Save character in block buffer

	mov	al,byte ptr [bx+55]	; Char #12
	mov	block_buffer+11,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #13
	mov	block_buffer+12,al 	; Save character in block buffer

	mov	al,byte ptr [bx+65]	; Char #14
	mov	block_buffer+13,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Char #15
	mov	block_buffer+14,al 	; Save character in block buffer

	mov	al,byte ptr [bx+75]	; Char #16
	mov	block_buffer+15,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; Char #17
	mov	block_buffer+16,al 	; Save character in block buffer

	mov	al,byte ptr [bx+85]	; Char #18
	mov	block_buffer+17,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; Status byte
	mov	block_buffer+18,al 	; Save character in block buffer

	mov	al,byte ptr [bx+95]	; CRC MSB byte
	mov	block_buffer+19,al 	; Save character in block buffer

	mov	al,byte ptr [bx+100]	; CRC LSB byte
	mov	block_buffer+20,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 19 bytes
	mov	cx,19
ch_blk2_2:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	ch_blk2_2

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+19
	cmp	al, byte ptr CRC_work+1	
	jne	ch_blk2_3			; MSB failed

	mov	al,block_buffer+20
	cmp	al, byte ptr CRC_work	
	jne	ch_blk2_3			; LSB failed

	mov	al,0FFh				; signal success
	ret

ch_blk2_3:
	xor	al,al			; signal failure
	ret				; Keep searching			
;------------------REVERSE SENSE--------------------------------------------;
ch_blk2_r:
	mov	bx,CHPTR_2

	mov	al,byte ptr [bx]	; Char #1
	xor	al,0FFh
	mov	block_buffer,al 	; Save character in block buffer

	mov	al,byte ptr [bx+5]	; Char #2
	xor	al,0FFh
	mov	block_buffer+1,al 	; Save character in block buffer

	mov	al,byte ptr [bx+10]	; Char #3
	xor	al,0FFh
	mov	block_buffer+2,al 	; Save character in block buffer

	mov	al,byte ptr [bx+15]	; Char #4
	xor	al,0FFh
	mov	block_buffer+3,al 	; Save character in block buffer

	mov	al,byte ptr [bx+20]	; Char #5
	xor	al,0FFh
	mov	block_buffer+4,al 	; Save character in block buffer

	mov	al,byte ptr [bx+25]	; Char #6
	xor	al,0FFh
	mov	block_buffer+5,al 	; Save character in block buffer

	mov	al,byte ptr [bx+30]	; Char #7
	xor	al,0FFh
	mov	block_buffer+6,al 	; Save character in block buffer

	mov	al,byte ptr [bx+35]	; Char #8
	xor	al,0FFh
	mov	block_buffer+7,al 	; Save character in block buffer

	mov	al,byte ptr [bx+40]	; Char #9
	xor	al,0FFh
	mov	block_buffer+8,al 	; Save character in block buffer

	mov	al,byte ptr [bx+45]	; Char #10
	xor	al,0FFh
	mov	block_buffer+9,al 	; Save character in block buffer

	mov	al,byte ptr [bx+50]	; Char #11
	xor	al,0FFh
	mov	block_buffer+10,al 	; Save character in block buffer

	mov	al,byte ptr [bx+55]	; Char #12
	xor	al,0FFh
	mov	block_buffer+11,al 	; Save character in block buffer

	mov	al,byte ptr [bx+60]	; Char #13
	xor	al,0FFh
	mov	block_buffer+12,al 	; Save character in block buffer

	mov	al,byte ptr [bx+65]	; Char #14
	xor	al,0FFh
	mov	block_buffer+13,al 	; Save character in block buffer

	mov	al,byte ptr [bx+70]	; Char #15
	xor	al,0FFh
	mov	block_buffer+14,al 	; Save character in block buffer

	mov	al,byte ptr [bx+75]	; Char #16
	xor	al,0FFh
	mov	block_buffer+15,al 	; Save character in block buffer

	mov	al,byte ptr [bx+80]	; Char #17
	xor	al,0FFh
	mov	block_buffer+16,al 	; Save character in block buffer

	mov	al,byte ptr [bx+85]	; Char #18
	xor	al,0FFh
	mov	block_buffer+17,al 	; Save character in block buffer

	mov	al,byte ptr [bx+90]	; Status byte
	xor	al,0FFh
	mov	block_buffer+18,al 	; Save character in block buffer

	mov	al,byte ptr [bx+95]	; CRC MSB 
	xor	al,0FFh
	mov	block_buffer+19,al 	; Save character in block buffer

	mov	al,byte ptr [bx+100]	; CRC LSB
	xor	al,0FFh
	mov	block_buffer+20,al 	; Save character in block buffer

; Compute CRC of received block (data+status).
	
	mov	word ptr CRC_work,0FFFFh ; initial CRC
	
	lea	bx,block_buffer		; Over 19 bytes
	mov	cx,19

ch_blk2_2_r:
	mov	al,byte ptr [bx]
	call	CALC_CRC
	inc	bx
	loop	ch_blk2_2_r

; Convert the CRC to an equivalent form

	mov	al,byte ptr CRC_work
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work,al

	mov	al,byte ptr CRC_work+1
	call	Revers
	xor	al,0FFh
	mov	byte ptr CRC_work+1,al

; Does the CRC's match ?

	mov	al,block_buffer+19
	cmp	al, byte ptr CRC_work+1	
	jne	ch_blk2_3_r		; MSB failed

	mov	al,block_buffer+20
	cmp	al, byte ptr CRC_work	
	jne	ch_blk2_3_r		; LSB failed

	mov	al,0FFh			; signal success
	ret

ch_blk2_3_r:
	xor	al,al			; signal failure
	ret				; Keep searching			

;----------------------------------------------------------------------------;
;         Some notes on how the phasing algorithms works:		     ;
;----------------------------------------------------------------------------;
; This phasing algorithm estimates an error, then slowly		     ;
; adjusts the DSP timer clock. Several parameters controls the phasing	     ;
; process: 								     ;
;									     ;
; 1) Since changes take place rather slowly, limit the magnitude of	     ;
;    changes to the clock to a step or two above and below what we consider  ;
;    an ideal timer setting.						     ;
;									     ;
; 2) It has been found that the actual phase error is very variable and thus ;
;    must be stabilized. A mean value is calculated for phase error and that ;
;    is used to control the timer updating.				     ;
;									     ;
;                     SUMMARY OF PARAMETERS:				     ;
;									     ;
; mean_phase    - this is the aritmetic mean of the phase error over the past;
;                 n cycles (n=2,8 - typically 2 - 4)			     ;
; MAX_ADJUST    - sets the limit on how many ticks can be adjusted up/down   ;
; UPDAT_BACKOFF - limit the frequency of updating			     ;
;    									     ;
;----------------------------------------------------------------------------;
slave_sync_ISS:

	cmp	byte ptr MASTER,0
	jnz	sl_ISS_slave
	ret

sl_ISS_slave:
	jmp	sl_ALG3			; CLOCK forwards adjustment
	ret
;----------------------------------------------------------------------------;
slave_sync_IRS:

	cmp	byte ptr MASTER,0
	jnz	sl_IRS_slave
  	ret

sl_IRS_slave:
	jmp	sl_ALG3		; CLOCK forwards
	ret

;----------------------------------------------------------------------------;
; Algoritm with CLOCK forwards adjustment - with phase zero		     ;
;----------------------------------------------------------------------------;
sl_ALG3:

	cmp	mean_phase,0		; Anything to adjust?
	jnz	SLaa			; yes, go ahead then
	ret				; no, return

SLaa:	cmp	byte ptr EVM_cnt,0	; slow changes to DSP clock
	jz	SLab
	ret

SLab:	mov	byte ptr EVM_cnt,ALG3_BACKOFF

	cmp	mean_phase,0		; What direction?
	js	SLad			; positive or negative adjustment?
	cmp	CLOCK,0
	jnz	SLac
	mov	CLOCK,1249		; backoff
	jmp	SHORT SLac1
SLac:	dec	CLOCK
SLac1:	call	zero_phase	
	ret	

SLad:	inc	CLOCK			; advance
	cmp	CLOCK,1250
	js	SLae
	mov	CLOCK,0
SLae:
	call	zero_phase
SLxx:	ret

;---------------------------------------------------------------------------;
;  Routine to compute average phase error over 8 cycles			    ;
;---------------------------------------------------------------------------;
ave_ph_err:
	push	ax			; these are words!	
	add	ax,old_phase1
	add	ax,old_phase2
	add	ax,old_phase3
	add	ax,old_phase4
	add	ax,old_phase5
	add	ax,old_phase6
	add	ax,old_phase7
	sar	ax,1
	sar	ax,1
	sar	ax,1
	mov	mean_phase,ax
	mov	bx,old_phase6
	mov	old_phase7,bx
	mov	bx,old_phase5
	mov	old_phase6,bx
	mov	bx,old_phase4
	mov	old_phase5,bx
	mov	bx,old_phase3
	mov	old_phase4,bx
	mov	bx,old_phase2
	mov	old_phase3,bx
	mov	bx,old_phase1
	mov	old_phase2,bx
	pop	ax
	mov	old_phase1,ax
	ret
;----------------------------------------------------------------------------;
;   Routine to reset all phase error history				     ;
;----------------------------------------------------------------------------;
zero_phase:
	xor	ax,ax
	mov	phase_err,ax
	mov	old_phase1,ax
	mov	old_phase2,ax
	mov	old_phase3,ax
	mov	old_phase4,ax
	mov	old_phase5,ax
	mov	old_phase6,ax
	mov	old_phase7,ax
	mov	mean_phase,ax
	ret
;----------------------------------------------------------------------------;
;            Routine to set EVM card's timing				     ;
;	     Parameter is in EVM_timing and EVM_dither
;----------------------------------------------------------------------------;
set_EVM_timing:
	; TODO
	ret	
;----------------------------------------------------------------------------;
; Routine samples receive data at next 1ms and shift bit into a 8-bit	     ;
; shift register.							     ;
;									     ;
; Scanning for sync uses ten shift registers called (REGS_8) using pointer   ;
; RPTR. On return from "DROP", i.e. after the next 1ms period, the data      ;
; input line is sampled, and the bit rotated into the shiftregister being    ;
; pointed to. Note that the shift register is only a single byte wide. 	     ;
; After 80 cycles through SMPLE_8, we end up with ten shift registers	     ;
; full of data, each shift register an 8 bit character, and the where	     ;
; the shiftregisters are 1ms apart in time. Synch search proceed at 1ms      ;
; resolution.								     ;
;----------------------------------------------------------------------------;
	EVEN
SMPLE_8: call    DROP          ; go to user till next IRQ

        mov     bx,RPTR_H      	; get current shiftreg pointer
        inc     bx              ; ready for next register
        cmp     bx,offset REGS_8+90+5 ; test gone too far ?
        jne     NXR1a
        lea     bx,REGS_8+90    ; if so start round again
NXR1a:   mov	RPTR_H,bx

	mov	al,data_bit

        shl     al,1            ; shift into carry 

;----------------------------------------------------------------------------;
;                      6 bytes at 200 baud				     ;
;----------------------------------------------------------------------------;
        rcr     byte ptr [bx+25],1 
        rcr     byte ptr [bx+20],1 
        rcr     byte ptr [bx+15],1 
        rcr     byte ptr [bx+10],1 
        rcr     byte ptr [bx+5],1 
        rcr     byte ptr [bx],1
;---------------------------------------------------------------------------;
;                     9 bytes at 100 baud				    ;
;---------------------------------------------------------------------------;
	pushf
        mov     bx,RPTR_L      	; get current shiftreg pointer
        inc     bx              ; ready for next register
        cmp     bx,offset REGS_8+10 ; test gone too far ?
        jne     NXR1
        lea     bx,REGS_8       ; if so start round again
NXR1:   mov	RPTR_L,bx

	popf
        rcr     byte ptr [bx+80],1
        rcr     byte ptr [bx+70],1 
        rcr     byte ptr [bx+60],1 
        rcr     byte ptr [bx+50],1 
        rcr     byte ptr [bx+40],1 
        rcr     byte ptr [bx+30],1 
        rcr     byte ptr [bx+20],1 
        rcr     byte ptr [bx+10],1 
        rcr     byte ptr [bx],1 ; and into sync byte

	ret

;----------------------------------------------------------------------------;
;         12 bit version - 10 ms signalling elements			     ;
;----------------------------------------------------------------------------;
	EVEN
SMPLE_12:  call    DROP         ; go to user till next IRQ
        mov     bx,RPTR_CTL     ; get current shiftreg pointer
        inc     bx              ; ready for next register
	inc	bx
        cmp     bx,offset REGS_12+20 ; test gone too far ?
        jne     NXR12
        lea     bx,REGS_12	; if so start round again
NXR12:  mov	RPTR_CTL,bx

	mov	al,data_bit

        shl     al,1            ; shift into carry 
        rcr     word ptr [bx],1 ; and into shiftregister
	ret

;----------------------------------------------------------------------------;
; Timed Wait routine							     ;
; Delay until 'CLOCK'= bx-REGISTER					     ;
;----------------------------------------------------------------------------;
	EVEN
HOLD1:	call	DROP		; go to user while waiting for next IRQ
HOLD:	cmp	bx,word ptr CLOCK	
	jne	HOLD1		; loop till bx = CLOCK
	ret

;----------------------------------------------------------------------------;
;            Delay for 10ms						     ;
;----------------------------------------------------------------------------;
	EVEN
WTEN:	mov	ah,10	        ; set count of ten
WAITE:	call	DROP		; go to user till next IRQ
	dec	ah	        ; count down
	jne	WAITE		; loop back till done
	ret		
;----------------------------------------------------------------------------;
;           Delay for 5ms						     ;
;----------------------------------------------------------------------------;
	EVEN
WFIVE:	mov	ah,5	        ; set count of five
WAITE_5:
	call	DROP		; go to user till next IRQ
	dec	ah	        ; count down
	jne	WAITE_5		; loop back till done
	ret		
;----------------------------------------------------------------------------;
;          Turn TX ON at the right time.     				     ;
;----------------------------------------------------------------------------;
TXON: 	push    ax
        mov     al,EVM_TX_PREV
        out     dx,al 
        pop     ax               

	mov	ah,tx_delay	; get TX delay constant
	jmp	WAITE		; and wait that long, then return

;----------------------------------------------------------------------------;
; Turn TX off 								     ;
;----------------------------------------------------------------------------;
TXOFF:  push    ax
;
; Turn DSP AFSK off
;
	mov	al,EVM_RX
	mov	cs:evm_out,al

        pop     ax                
	ret


;----------------------------------------------------------------
; Transmit the bit on the carry flag, preserving AX
;----------------------------------------------------------------
TXBIT: push    ax  
        jc      TXCS1           	; is bit 1 or 0
        mov     al,EVM_TX_0
        jmp     SHORT TXCS2     

TXCS1:  mov     al,EVM_TX_1
TXCS2:	xor     al,MASKOUT
	mov	cs:evm_out,al
        pop     ax
	ret


;----------------------------------------------------------------------------;
; Transmit a character at 100 baud 					     ;
; Character is pointed to by (bx) 					     ;
;----------------------------------------------------------------------------;
	EVEN
TXCH_L:	mov	al,byte ptr [bx]  	; get character to send
	xor	al,tx_bit_sense		; set the logic sense
TXC2:	mov	byte ptr bit_cnt,8	; set bit count
TXC3:	shr	al,1			; shift end bit into carry
        call	TXBIT
        call	WTEN			; and wait one bit time
        dec	byte ptr bit_cnt	; has 8th bit been transmitted?
        jnz     TXC3            	; skip if preloaded bit now in carry

	mov	al,tx_bit_sense	; for display
	mov	tx_sense,al
	ret
;----------------------------------------------------------------------------;
; Transmit half a character at 100 baud 				     ;
; Character is pointed to by (bx) 					     ;
;----------------------------------------------------------------------------;
	EVEN
TXCH_S:	mov	al,byte ptr [bx]  	; get character to send
	xor	al,tx_bit_sense		; set the logic sense
TXC2s:	mov	byte ptr bit_cnt,4	; set bit count
TXC3s:	shr	al,1			; shift end bit into carry
	call	TXBIT
        call	WTEN			; and wait one bit time
        dec	byte ptr bit_cnt	; has 8th bit been transmitted?
        jnz     TXC3s            	; skip if preloaded bit now in carry

	mov	al,tx_bit_sense	; for display
	mov	tx_sense,al
	ret

;----------------------------------------------------------------------------;
; Transmit a character at 200 baud 					     ;
; Character is pointed to by (bx) 					     ;
;----------------------------------------------------------------------------;
	EVEN
TXCH_H:	mov	al,byte ptr [bx]  	; get character to send
	xor	al,tx_bit_sense		; set the logic sense
TXC2_H:	mov	byte ptr bit_cnt,8	; set bit count
TXC3_H:	shr	al,1			; shift end bit into carry
        call	TXBIT
        call	WFIVE			; and wait one bit time
        dec	byte ptr bit_cnt	; has 8th bit been transmitted?
        jnz     TXC3_H            	; skip if preloaded bit now in carry

	mov	al,tx_bit_sense	; for display
	mov	tx_sense,al
	ret
;----------------------------------------------------------------------------;
;                 Transmit a control code in ax				     ;
; Control code is XOR'ed with the appropriate send sense.		     ;
;----------------------------------------------------------------------------;
	EVEN
TCTL:					; Here for next new CS 
	xor	byte ptr exp_rx_sense,0FFh ; next expected receive sense

	mov	bx,SEND_SENSE
	or	bx,bx
	je	TCTL0
	mov	SEND_SENSE,0
	jmp	SHORT TCTL0a
TCTL0:
	mov	SEND_SENSE,0FFFh	

TCTL0a:	mov     ax,CONTROL		; get control to be sent
	xor	ax,bx			; set the logic sense
	mov	control_out,ax

	call	TXON			; go wait till time and turn TX on

	mov	byte ptr bit_cnt,12	; set bit count
TCTL3:	shr	word ptr control_out,1	; shift end bit into carry
        call	TXBIT
        call	WTEN			; and wait one bit time
        dec	byte ptr bit_cnt	; has 12th bit been transmitted?
        jnz     TCTL3            	; next bit if not

	cmp	CONTROL,CS3		; if control=CS3, dont turn TX off
	je	TCTS3
	call	TXOFF			; else turn TX off
TCTS3:	
	mov	ax,SEND_SENSE	; for display
	mov	tx_sense,al
	ret

;----------------------------------------------------------------------------;
;      Bit reversing							     ;
;----------------------------------------------------------------------------;
Revers:
	push	cx	   
	mov	cx,8				
	mov	ah,al 
CRC0:	shl	ah,1
	rcr	al,1
	dec	cx
	jnz	CRC0  
	pop	cx
	ret		

;----------------------------------------------------------------------------;
; Calculate CRC using CCITT 16 bit polynomial				     ;
;----------------------------------------------------------------------------;
	EVEN
CALC_CRC:
	push	bx
	push	dx
	push	cx
	
	call	Revers
	xor	ah,ah			; prepare data
		    
	mov	bx,CRC_work		; get "dx"	
	xchg	bh,bl			
	xor	ax,bx			; get "data"		
	mov	dx,ax		 	
					; dx = dx XOR (Lo(dx) SHR 4)
	mov	cx,4 			 
	shr	al,cl	
	xor	dl,al
					; dx = dx XOR ((swap(Lo(dx)) shl 4)
	mov	ah,dl
	mov	cx,4
	shl	ah,cl
	xor	dh,ah
					; dx = dx XOR ((Lo(dx)) shl
	mov	al,dl
	mov	ah,0
	mov	cx,5
	shl	ax,cl
	xor	dx,ax
	mov	CRC_work,dx

	pop	cx
	pop	dx
	pop	bx

	ret
;----------------------------------------------------------------------------;
; Routine to bump the clock						     ;
; entered every 1ms tick.						     ;
;----------------------------------------------------------------------------;
	EVEN
TICK:   mov     bx,CLOCK        ; get CLOCK
        inc     bx              ; add 1ms
        cmp     bx,1250         ; 1250 else test end of 1.25s cycle
	jne     TIC3            ; if not end of cycle then skip
        xor     bx,bx           ; and reset to start of cycle

	cmp	byte ptr DSP_cnt,0	; count down DSP clock backoff
	jz	TT3
	js	TT3
	dec	byte ptr DSP_cnt
TT3:
	cmp	byte ptr EVM_cnt,0	; count down phasing backoff
	jz	TIC3
	js	TIC3
	dec	byte ptr EVM_cnt

TIC3:   mov     CLOCK,bx        ; and save CLOCK

; Do the same with AUX_CLOCK

	mov     bx,AUX_CLOCK    ; get AUX_CLOCK
        inc     bx              ; add 1ms
        cmp     bx,1250         ; 1250 else test end of 1.25s cycle
	jne     TIC3A           ; if not end of cycle then skip
        xor     bx,bx           ; and reset to start of cycle
TIC3A:  mov     AUX_CLOCK,bx    ; and save AUX_CLOCK

	inc	RxT_cnt		; timing tracking paramater
       ret

;----------------------------------------------------------------------------;
;                     INTERRUPT DRIVEN PACTOR MODULE                         ;
;                 Vectors here from user program every 1ms                   ; 
;----------------------------------------------------------------------------;
	EVEN
	PUBLIC _INTRP
_INTRP:  

norm_int: 

	push    ax              ; save registers
	push    bx
	push    cx              
	push    dx              
	push    ds
	push	si
	push	di

	mov    ax,cs		; let ds point to cs for data access
	mov    ds,ax
     
        mov     ax,sp           ; save user stack pointer
        mov     USRSP_OFF,ax
        mov     ax,ss
        mov     USRSP_SEG,ax

        mov     ax,AMTSP        ; retrieve PACTOR stack pointer
        mov     sp,ax
        mov     ax,cs           ; SS
        mov     ss,ax


        call    TICK            ; bump the clock

        ; read data from serial port into evm_in
	mov	dx,cs:com_port		; get data byte
	in	al,dx	
	mov	cs:evm_in,al	; most recent data value from EVM
	mov	ah,al
	mov	al,00


;-----------------Process recovered data from DSP----------------------------;
; Get signal output signal level, separate mark/space			     ;
; and update appropiate peak levels.					     ;
; Save actual data value for MARQ					     ;
;----------------------------------------------------------------------------;
	; TODO: make sure value is in correct range
	mov	cs:AD_data,ax	; for MARQ purposes

	or	ax,ax
	jns	EVMa_a		; TODO make this correct
; ------------------ Its a space signal
	neg	ax		; convert into positive number
	cmp	ax,cs:AD_space_data  ; new peak value?
	jns	SHORT EVM0_1
	mov	ax,cs:AD_space_data
	sub	ax,DECAY	; decay factor
EVM0_1:	mov	cs:AD_space_data,ax
	mov	al,080h
	jmp	SHORT EVMa_out

; ------------------Its a mark signal
EVMa_a:	
	cmp	ax,cs:AD_mark_data  ; new peak value?
	jns	SHORT EVM0_3
	mov	ax,cs:AD_mark_data
	sub	ax,DECAY	    ; decay factor
EVM0_3:	mov	cs:AD_mark_data,ax
	mov	al,0

EVMa_out:
	mov	cs:data_bit,al


;----------------------------------------------------------------------------;
; Coordinate DSP clock corrections to occur only during this instance 	     ;
;----------------------------------------------------------------------------;
	cmp	byte ptr cs:phasing_flag,0
	jz	no_correction
	call	set_EVM_timing		; for ARQ slave phase sync
	mov	byte ptr cs:phasing_flag,0
no_correction:

;----------------------------------------------------------------------------;
; Did the modem change ? If so, setup the new modem too                      ;
;----------------------------------------------------------------------------;
	mov	al,byte ptr old_modem
	cmp	byte ptr DS_SPEED,al
	je	same_m
	
	mov	ah,byte ptr DS_SPEED	; load new modem
	;;mov	cs:evm_out,ah  ; TODO Data to EVM data is code for new modem
	mov	al,byte ptr DS_SPEED
	mov	byte ptr old_modem,al

same_m:
;----------------------------------------------------------------------------;
; Control Commands 							     ;
;                                                                            ;
; CTRL  21=FORCE BREAK                                                       ;
;       22=CLOSE DOWN QRT                                                    ;
;       23=INITIATE ARQ (MASTER)                                             ;       
;       24=UNPROTO SEND                                                      ;
;       25=MONITOR (listen mode)                                             ;             
;       26=STBY                                                              ;
;       27=TEST (outputs 1kHz signal at DTR)        			     ;
;----------------------------------------------------------------------------;
; Implementation of the control call is a little awkward, could be	     ;
; done using a call proc_table[si]					     ;
; but will work out ok like this too.					     ;
; NOTE: not all "CALL's" return, however those that does MUST drop	     ;
;       out to EXIT.							     ;
;----------------------------------------------------------------------------;
CMD:    mov     al,CNTRL        ; get command byte from user
	or	al,al		; SET UP FLAGS
        jnz     CMDa
        jmp      EXIT           ; exit if no command
CMDa:   mov     byte ptr CNTRL,0  ; else acknowledge that command has been executed
        cmp     al,21h
        je      CMD01   
        cmp     al,22h
        je      CMD02   
        cmp     al,23h
        je      CMD03   
        cmp     al,24h
        je      CMD04   
        cmp     al,25h
        je      CMD05   
        cmp     al,26h
        je      CMD06
        cmp     al,27h
        je      CMD07
        cmp     al,28h
        je      CMD08
        cmp     al,29h
        je      CMD09
        cmp     al,2Ah
        je      CMD0A
        jmp     SHORT EXIT   
	EVEN
CMD01:  mov     byte ptr BREAKIN,1
        jmp     SHORT EXIT
	EVEN
CMD02:  
;----------------------------------------------------------------------------;
;     Set QRT request flag						     ;
;----------------------------------------------------------------------------;
	mov     byte ptr QFLG,1
        jmp     SHORT EXIT
	EVEN
CMD03:  call    CALLA           ; Initiate MASTER CALL
        jmp     SHORT EXIT
	EVEN
CMD04:  call    CALLB           ; Initiate UNPROTO
        jmp     SHORT EXIT
	EVEN
CMD05:  call    MON          	; Initiate MONITOR
        jmp     SHORT EXIT
	EVEN
CMD06:	xor	al,al
	mov	byte ptr cs:QRT_request,al
	mov	byte ptr cs:QRT_pending,al
	mov	byte ptr cs:QRT_sent,al
	call    STBY            ; Force STBY
        jmp     SHORT EXIT
	EVEN
CMD07:  call    TESTR           ; Enter the test procedure
        jmp     SHORT EXIT
	EVEN
CMD08:  xor	al,al
	mov	byte ptr cs:QRT_request,al
	mov	byte ptr cs:QRT_pending,al
	mov	byte ptr cs:QRT_sent,al
	call    STBY            ; Enter STBY
        jmp     SHORT EXIT
	EVEN
CMD09:  mov     byte ptr OVER,040h 	; Set "over" request
        jmp     SHORT EXIT
	EVEN
CMD0A:  mov     byte ptr QRT_pending,01 ; Set QRT_pending flag
        jmp     SHORT EXIT

; THIS RTS IS EFFECTIVELY THE RTS TO THE "DROP" SUBROUTINE.

EXIT:   mov     ax,AMTAX        ; all done, recover PacTOR program
        mov     bx,AMTBX     
	mov	cx,AMTCX
	mov	dx,AMTDX
	mov	si,AMTSI
	mov	di,AMTDI
        ret                     ; and return to Pactor program

;----------------------------------------------------------------------------;
; THIS "SUBROUTINE" ACTUALLY RETURNS CONTROL TO THE USER WHILE		     ;
; WAITING FOR THE NEXT 1KHz INTERRUPT FROM THE serial port
;----------------------------------------------------------------------------;
	EVEN
DROP:   
        ;put value in evm_out to serial port
	mov	al,cs:evm_out
	mov	dx,cs:com_port		; get data byte
        out	dx,al

	mov     cs:AMTAX,ax        ; save amtor program register
        mov     cs:AMTBX,bx
        mov     cs:AMTCX,cx
	mov	cs:AMTDX,dx
	mov	cs:AMTSI,si
	mov	cs:AMTDI,di

        mov     ax,sp           ; Save AMTOR stack pointer
        mov     AMTSP,ax

        mov     ax,USRSP_OFF    ; Restore user stack pointer
        mov     sp,ax
        mov     ax,USRSP_SEG
        mov     ss,ax

	pop	di
	pop	si
	pop     ds              	; restore registers
	pop     dx
	pop     cx
	pop     bx

int_ret: 
	mov     al,20h
        out     020H,al          ; send EOI and return
	cmp	cs:irq,08	 ; if IRQ >=8 set EOI to seconf pic too
	js	int_out
      	out     0A0H,al          ; send EOI and return
int_out:

	pop     ax

	sti			; Arm interrupts again

        iret                   
            
PACBIOS endp

;--------------------------SUPPORT AND CONTROL-------------------------------;
;         SINCE THESE ARE CALLED DIRECTLY FROM THE MAIN ROUTINE, 	     ;
;                      THEY MUST BE FAR POINTERS			     ;
;									     ;
;  _p_speedup()       							     ;
;----------------------------------------------------------------------------;
__p_speedup PROC far                   ; registers saved in int_handler
        
        cli                             ; stop interrupts

        lea     bx,cs:AM_STACK
        mov     cs:AMTSP,bx
        mov     byte ptr cs:CNTRL,26h	; start off in STBY mode
        mov     byte ptr cs:tx_delay,40	; default TXDELAY is 40ms
	mov	byte ptr cs:U_REPEAT,1  ; default unproto repeat
                                        ; set up some more data items
        xor     al,al                   ; clear al       
	mov	byte ptr cs:MODE,al	; assume ascii

	mov	byte ptr cs:BREAKIN,al	; breakin
	mov	byte ptr cs:QFLG,al	; quit request
	mov	byte ptr cs:OVER,al	; "over"
	mov	byte ptr cs:QRT_request,al	; ISS QRT request
	mov	byte ptr cs:QRT_pending,al	; Keyboard requests QRT
	mov	byte ptr cs:QRT_sent,al		; IRS sent a QRT - we are ack'ing now

	mov	byte ptr cs:SPEED,al	; assume 100 baud
	mov	byte ptr cs:u_speed,al	; assume 100 baud UNPROTO

        mov     cs:MASKOUT,al           ; Reset output masks
 
	lea	bx,cs:RXBUFR		; Initialize rx buffer
	mov	cs:rx_store_at_ptr,bx	; (flush it)
	mov	cs:rx_take_from_ptr,bx

	lea	bx,cs:ECHOBUFR		; Initialize echo buffer
	mov	cs:echo_store_at_ptr,bx	; (flush it)
	mov	cs:echo_take_from_ptr,bx

	lea	bx,cs:TXBUFR		; Initialize tx buffer
	mov	cs:tx_store_at_ptr,bx	; (flush it)
	mov	cs:tx_take_from_ptr,bx
	
	mov	word ptr cs:AD_mark_data,0
	mov	word ptr cs:AD_space_data,0
	
	mov	al,EVM_RESET
	mov	cs:evm_out,al

        cmp     byte ptr cs:initialized,05Ah    ; have we initialized
        je      t_out                	        ; already ?    
        mov     byte ptr cs:initialized,05Ah
t_out:
	sti                             ; re-enable interrupts 
        ret
__p_speedup ENDP

;-------------------------PACTOR CONTROL STUFF-------------------------------;
;               Routine to set PACTOR control mode		             ;
;----------------------------------------------------------------------------;
_set_control PROC far
	push	bp			; access variables
	mov	bp,sp
        mov     ax,ARG1			; control code
        mov     cs:CNTRL,al
	pop	bp 
        ret
_set_control ENDP
;----------------------------------------------------------------------------;
; Routine to set transmitter turn-on delay				     ;
;----------------------------------------------------------------------------;
_set_txdelay PROC far
	push	bp			; access variables
	mov	bp,sp
        mov     ax,ARG1          	; delay
        mov     cs:tx_delay,al
	pop	bp 
        ret
_set_txdelay ENDP
;----------------------------------------------------------------------------;
; Routine to set EVM card's timing value				     ;
;----------------------------------------------------------------------------;
_set_biostiming PROC far
	push	bp			; access variables
	mov	bp,sp
        mov     ax,ARG1          	; timing value
        mov     cs:EVM_timing,ax
        mov     ax,ARG2          	; dither
        mov     cs:EVM_dither,ax
        mov     byte ptr cs:phasing_flag,1 ; update on next tick
	pop	bp 
        ret
_set_biostiming ENDP
;----------------------------------------------------------------------------;
; Routine to set UNPROTO speed						     ;
;----------------------------------------------------------------------------;
_set_USPEED PROC far
	push	bp			; access variables
	mov	bp,sp
        mov     ax,ARG1          	; speed value
        mov     cs:u_speed,al
	pop	bp 
        ret
_set_USPEED ENDP
;----------------------------------------------------------------------------;
; Routine to set MASKOUT						     ;
;----------------------------------------------------------------------------;
_set_MASKOUT PROC far
	push	bp			; access variables
	mov	bp,sp
        mov     ax,ARG1          	; maskout value
        mov     cs:MASKOUT,al
	pop	bp 
        ret
_set_MASKOUT ENDP
;----------------------------------------------------------------------------;
; Routine to set unproto repeat count					     ;
;----------------------------------------------------------------------------;
_set_UREPEAT PROC far
	push	bp			; access variables
	mov	bp,sp
        mov     ax,ARG1          	; repeat count
        mov     byte ptr cs:U_REPEAT,al
	pop	bp 
        ret
_set_UREPEAT ENDP

;----------------------------------------------------------------------------;
; Routine to retrieve received data					     ;
;----------------------------------------------------------------------------;
_get_rxdta PROC far
	mov	ax,cs:rx_take_from_ptr	; is there data ?
	cmp	ax,cs:rx_store_at_ptr
	jne	ge_dat
	xor	ax,ax			; no, return NULL
	jmp	SHORT ge_ex
	
ge_dat:	cli	 ; START OF CRITICAL REGION ******************************			
	push	bx
	mov	bx,cs:rx_take_from_ptr
	mov	al,byte ptr cs:[bx]	; get the character
	mov	ah,1			; show we have data
	inc	bx			; update pointer
	cmp	bx,offset cs:RXBUFR+RXBUFSIZ	; test for wraparound
	jne	ge_nxt
	lea	bx,cs:RXBUFR
ge_nxt: mov	cs:rx_take_from_ptr,bx
	pop	bx
	sti	 ; END OF CRITICAL REGION   *****************************

ge_ex: ret

_get_rxdta ENDP

;----------------------------------------------------------------------------;
; Routine to retrieve echoed data					     ;
;----------------------------------------------------------------------------;
_get_echodta PROC far
	mov	ax,cs:echo_take_from_ptr ; is there data ?
	cmp	ax,cs:echo_store_at_ptr
	jne	geh_dat
	xor	ax,ax			; no, return NULL
	jmp	SHORT geh_ex
	
geh_dat: 
	cli	 ; START OF CRITICAL REGION ******************************			
	push	bx
	mov	bx,cs:echo_take_from_ptr
	mov	al,byte ptr cs:[bx]	; get the character
	mov	ah,1			; show we have data
	inc	bx			; update pointer
	cmp	bx,offset cs:ECHOBUFR+RXBUFSIZ	; test for wraparound
	jne	geh_nxt
	lea	bx,cs:ECHOBUFR
geh_nxt: mov	cs:echo_take_from_ptr,bx
	pop	bx
	sti	 ; END OF CRITICAL REGION   *****************************

geh_ex: ret

_get_echodta ENDP

;----------------------------------------------------------------------------;
; Routine to store data for transmission				     ;
;----------------------------------------------------------------------------;
_send_txdta PROC far

	push	bp			; get pointer to data
	mov	bp,sp
	mov	ax,ARG1			; get the character

	push	bx

	mov	bx,cs:tx_store_at_ptr 	; is there space in tx buffer ?
	cmp	bx,offset cs:TXBUFR+TXBUFSIZ-1 ; check wrap round case
	jne 	SD0
	cmp	cs:tx_take_from_ptr,offset cs:TXBUFR ; handle wrap around case
	jne 	SD1			; ok there is space left
	jmp	SHORT SD5   		; exit if no space left
SD0:	inc	bx			; normal case
	cmp	bx,cs:tx_take_from_ptr
	je	SD5			; exit if no space available

SD1:
	cli	 ; START OF CRITICAL REGION ******************************			
	mov	bx,cs:tx_store_at_ptr	; and store it
	mov	byte ptr cs:[bx],al	
	inc	bx			; update buffer pointer
	cmp	bx,offset cs:TXBUFR+TXBUFSIZ
	jne	SD2
	lea	bx,cs:TXBUFR
SD2:	mov	cs:tx_store_at_ptr,bx
	sti	 ; END OF CRITICAL REGION   *****************************

SD5:	pop	bx
	pop	bp
 	xor	ax,ax
	ret

_send_txdta ENDP
;----------------------------------------------------------------------------;
; Retrieve message buffer status					     ;
; return = 0 for space available, else = 0xFF				     ;
;----------------------------------------------------------------------------;
_get_buf_stat PROC far
	push	bx
	mov	bx,cs:tx_store_at_ptr 	    ; is there space in tx buffer ?
	cmp	bx,offset cs:TXBUFR+TXBUFSIZ-1 ; check wrap round case
	jne 	GB0
	cmp	cs:tx_take_from_ptr,offset cs:TXBUFR	; handle wrap around case
	jne 	GB1			; ok there is space left
	mov	al,0FFh
	jmp	SHORT GB5   		; exit if no space left
GB0:	inc	bx			; normal case
	cmp	bx,cs:tx_take_from_ptr
	mov	al,0FFh
	je	GB5			; exit if no space available
GB1:	xor	al,al			; space is OK
GB5:	xor	ah,ah  
	pop bx
        ret
_get_buf_stat ENDP

;----------------------------------------------------------------------------;
; Retrieve accumulated phase error					     ;
;----------------------------------------------------------------------------;
_get_phase_err PROC far
        mov   ax,cs:phase_err  
        ret
_get_phase_err ENDP

;----------------------------------------------------------------------------;
; Retrieve average accumulated phase error					     ;
;----------------------------------------------------------------------------;
_get_phase_avg PROC far
        mov   ax,cs:mean_phase  
        ret
_get_phase_avg ENDP

;----------------------------------------------------------------------------;
; Retrieve ascii/huffman mode						     ;
;----------------------------------------------------------------------------;
_get_mode PROC far
        mov   	al,byte ptr cs:MODE
	xor	ah,ah  
        ret
_get_mode ENDP

;----------------------------------------------------------------------------;
; Routine to set COM port						     ;
;----------------------------------------------------------------------------;
	EVEN
_set_com_port   proc far

	push	bp			; to access variables
	mov	bp,sp
	mov	ax,ARG1			; get argument
	mov	cs:com_port,ax		; COM
	inc	ax
	mov	cs:com_port1,ax		; COM+1
	inc	ax
	mov	cs:com_port2,ax		; COM+2
	inc	ax
	mov	cs:com_port3,ax		; COM+3
	inc	ax
	mov	cs:com_port4,ax		; COM+4
	inc	ax
	mov	cs:com_port5,ax		; COM+5
	inc	ax
	mov	cs:com_port6,ax		; COM+6
        
        mov     ax,ARG2			; IRQ
        mov     cs:irq,ax

	pop	bp
	ret
_set_com_port   endp

;----------------------------------------------------------------------------;
; Loads callee callsign							     ;
; Upon the first null, or space character, fill remainder of		     ;
; selcal with 55h							     ;
;----------------------------------------------------------------------------;
_set_calleeID proc far

	push	bp		; get pointer to data
	mov	bp,sp
	push	bx
	push	si
	push	cx
	les	bx,[bp+06]
	mov	si,bx
	lea	bx,cs:callee_ID
	mov	cx,8	 	; move block of 8 bytes
se_ee:	mov	al,es:[si]
	cmp	al,00h		; short string?
	je	se_fill		; else fill with 0x55h
	cmp 	al,' '	
	je	se_fill		; else fill with 0x55h
	mov	cs:[bx],al
	inc	si
	inc	bx
	loop	se_ee
	jmp	SHORT se_exit

se_fill:
	mov	al,55h		; short strings are filled with 55h 
	mov	cs:[bx],al
	inc	bx
	loop	se_fill

se_exit:

;----------------------------------------------------------------------------;
; We also need to build a reversed order string for QRT packets		     ;
;----------------------------------------------------------------------------;
	mov	cx,8			; fill reversed ID with IDLES
	lea	bx,cs:callee_ID_rev	; as used in QRT packets
se1:	mov	byte ptr cs:[bx],IDLE
	inc	bx
	loop	se1

;
; Now find the last callsign character
	mov	cx,8

	lea	bx,cs:callee_ID
	add	bx,7
	mov	si,bx

	lea	bx,cs:callee_ID_rev

se2:					; load our ID in reverse
	mov	al,byte ptr cs:[si]
	cmp	al,55h
	jne	se3
	dec	si
	loop	se2			
	jmp	SHORT se4
se3:	mov	byte ptr cs:[bx],al
	dec	si
	inc	bx
	loop	se2
se4:
	pop	cx
	pop	si
	pop	bx
	pop	bp

	ret
_set_calleeID ENDP

;----------------------------------------------------------------------------;
; Loads our own callsign						     ;
;----------------------------------------------------------------------------;
_set_ourID proc far

	push	bp			; get pointer to data
	mov	bp,sp
	push	bx
	push	si
	push	cx
	les	bx,[bp+06]
	mov	si,bx
	lea	bx,cs:our_ID
	mov	cx,8	 		; move block of 8 bytes
se_ou:	mov	al,es:[si]
	mov	cs:[bx],al
	inc	si
	inc	bx
	loop	se_ou


	pop	cx
	pop	si
	pop	bx
	pop	bp

	ret
_set_ourID ENDP

;----------------------------------------------------------------------------;
; Get status								     ;
;----------------------------------------------------------------------------;
_get_status PROC far
        mov   al,cs:STATUS
	xor	ah,ah  
        ret
_get_status ENDP

;----------------------------------------------------------------------------;
; Get header byte								     ;
;----------------------------------------------------------------------------;
_get_sync PROC far
        mov   al,cs:sync
	xor	ah,ah  
        ret
_get_sync ENDP

;----------------------------------------------------------------------------;
; Get bit sense								     ;
;----------------------------------------------------------------------------;
_get_RXsense PROC far
        mov   al,cs:rx_sense
	xor	ah,ah  
        ret
_get_RXsense ENDP
;----------------------------------------------------------------------------;
_get_TXsense PROC far
        mov   al,cs:tx_sense
	xor	ah,ah  
        ret
_get_TXsense ENDP

;----------------------------------------------------------------------------;
; Get speed (100/200)							     ;
;----------------------------------------------------------------------------;
_get_speed proc far
        mov     al,byte ptr cs:DS_SPEED
	xor	ah,ah
        ret
_get_speed ENDP

;----------------------------------------------------------------------------;
; Get recoved modem signal						     ;
;----------------------------------------------------------------------------;
_get_mark_data proc far
        mov     ax,cs:AD_mark_data
        ret
_get_mark_data ENDP
_get_space_data proc far
        mov     ax,cs:AD_space_data
        ret
_get_space_data ENDP

;---------------------------------------------------------------------------;
; Get error count							    ;
;---------------------------------------------------------------------------;
_get_err_cnt proc far
        mov     al,byte ptr cs:err_cnt
	xor	ah,ah
        ret
_get_err_cnt ENDP

;---------------------------------------------------------------------------;
; Get number of MARQ successes							    ;
;---------------------------------------------------------------------------;
_get_marq_cnt proc far
        mov     al,byte ptr cs:marq_cnt
	xor	ah,ah
        ret
_get_marq_cnt ENDP

;---------------------------------------------------------------------------;
; Get RxT parameter							    ;
;---------------------------------------------------------------------------;
_get_RxT proc far
	mov     ax,cs:ideal_RxT
        ret
_get_RxT ENDP

;---------------------------------------------------------------------------;
; Get RQ count								    ;
;---------------------------------------------------------------------------;
_get_RQ_cnt proc far
        mov     al,byte ptr cs:rq_cnt
	xor	ah,ah
        ret
_get_RQ_cnt ENDP

;---------------------------------------------------------------------------;
; Get CS								    ;
;---------------------------------------------------------------------------;
_get_CS proc far
        mov     ax, word ptr cs:CONTROL
        ret
_get_CS ENDP

;---------------------------------------------------------------------------;
; Get CTRX								    ;
;---------------------------------------------------------------------------;
_get_CTRX proc far
        mov     ax, word ptr cs:DS_CTRX
        ret
_get_CTRX ENDP

;---------------------------------------------------------------------------;
; Get CTTX								    ;
;---------------------------------------------------------------------------;
_get_CTTX proc far
        mov     ax, word ptr cs:CTTX
        ret
_get_CTTX ENDP

;---------------------------------------------------------------------------;
; Flush bios buffers								    ;
;---------------------------------------------------------------------------;
_flush_bios_buffers proc far
	push	bx
	cli	; ========CRITICAL REGION============
	lea	bx,cs:RXBUFR		; Initialize rx buffer
	mov	cs:rx_store_at_ptr,bx	; (flush it)
	mov	cs:rx_take_from_ptr,bx

	lea	bx,cs:ECHOBUFR		; Initialize echo buffer
	mov	cs:echo_store_at_ptr,bx	; (flush it)
	mov	cs:echo_take_from_ptr,bx

	lea	bx,cs:TXBUFR		; Initialize tx buffer
	mov	cs:tx_store_at_ptr,bx	; (flush it)
	mov	cs:tx_take_from_ptr,bx
	sti	; =======END OF CRITICAL REGION ======
	pop	bx
        ret
_flush_bios_buffers ENDP

;---------------------------------------------------------------------------;
cseg   	ends
       	END
